/*
***********************************************************************
LICENSE AGREEMENT
***********************************************************************
 READ THIS LICENSE AGREEMENT CAREFULLY BEFORE USING THIS PRODUCT. BY 
 USING THIS PRODUCT YOU INDICATE YOUR ACCEPTANCE OF THE TERMS OF THE 
 FOLLOWING AGREEMENT. THESE TERMS APPLY TO YOU AND ANY SUBSEQUENT 
 LICENSEE OF THIS PRODUCT.

               License Agreement for RegularPlace 1.0

Copyright (c)  2008 by 
Ashutosh Chakraborty, Anurag Kumar, David Z. Pan
All rights reserved

ATTRIBUTION ASSURANCE LICENSE
Redistribution of the any code, with or without modification, are permitted 
provided that the conditions below are met. These conditions require a 
modest attribution to Ashutosh Chakraborty, Anurag Kumar and Dr. 
David Z. Pan (the "Author").

1. Redistributions of any code, with or without modification (the "Code"), 
must be accompanied by any documentation and, each time the resulting 
executable program or a program dependent thereon is launched, a 
prominent display (e.g., splash screen or banner text) of the Author's 
attribution information, which includes:
  (a) Ashutosh Chakraborty, Anurag Kumar, and Dr. David Z. Pan ("AUTHOR"),
  (b) The University of Texas at Austin ("PROFESSIONAL IDENTIFICATION"), and
  (c) http://www.cerc.utexas.edu/utda/ ("URL").

2. Use is limited to academic research groups only. Users who are interested 
in industry or commercial purposes must notify Author and request 
separate license agreement.  

3. Neither the name nor any trademark of the Author may be used to 
endorse or promote products derived from this software without specific 
prior written permission.

4. Users are entirely responsible, to the exclusion of the Author and 
any other persons, for compliance with (a) regulations set by owners or 
administrators of employed equipment, (b) licensing terms of any other 
software, and (c) local, national, and international regulations regarding 
use, including those regarding import, export, and use of encryption software. 

THIS FREE SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR 
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE AUTHOR OR ANY CONTRIBUTOR BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, EFFECTS OF UNAUTHORIZED OR MALICIOUS NETWORK 
ACCESS; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <iostream>
#include <vector>
#include <string>
#include <map>
#include <fstream>
#include <cfloat>
#include "math.h"

using namespace std;

#include "Util.h"
#include "BBox.h"
#include "Net.h"
#include "Tile.h"
#include "Data.h"
#include "Design.h"

CBBox::CBBox()
    : mx(DBL_MAX), mX(DBL_MIN), my(DBL_MAX), mY(DBL_MIN) {}

CBBox::CBBox(const CNet* aNet)
{
  vector<CNode*> v = aNet->GetCellsInNet();

  assert(v.size() > 0);

  double xMin = DBL_MAX;
  double xMax = DBL_MIN;
  double yMin = DBL_MAX;
  double yMax = DBL_MIN;

  for (int i = 0; i < v.size(); i++)
  {
    double currX = v[i]->X();
    double currY = v[i]->Y();

    if (xMin > currX) xMin = currX;
    if (xMax < currX) xMax = currX;
    if (yMin > currY) yMin = currY;
    if (yMax < currY) yMax = currY;
  }

  assert(xMin <= xMax);
  assert(yMin <= yMax);

  mx = xMin;
  mX = xMax;
  my = yMin;
  mY = yMax;
}

bool CBBox::Contains(CNode *aNode) const
{
  double x = aNode->X();
  double y = aNode->Y();
  return Contains(x, y);
}

bool CBBox::Contains(double x, double y) const
{
  return x >= mx && x <= mX && y >= my && y <= mY;
}

CBBox& CBBox::operator=(const CBBox & that)
{
  // Check for self copy
  if (this == &that)
    return *this;

  // Copy each private member
  mx = that.mx;
  mX = that.mX;
  my = that.my;
  mY = that.mY;

  return *this;
}

double CBBox::HPWL() const
{
  double wl = (mX + mY - mx - my);
  assert(DBL_ZERO <= wl);
  return wl;
}

double CBBox::ShortestDistance(CNode* aNode) const
{
  double xDist = 0.0;
  if (aNode->X() < mx)
    xDist = mx - aNode->X();
  else if (aNode->X() > mX)
    xDist = aNode->X() - mX;

  double yDist = 0.0;
  if (aNode->Y() < my)
    yDist = my - aNode->Y();
  else if (aNode->Y() > mY)
    yDist = aNode->Y() - mY;

  return xDist + yDist;
}

CBBox* CBBox::GetIntersectionBB(const CTile* t) const
{
  return getIntersectionBB(
           t->X(), t->X() + gDesign->mTileWidth,
           t->Y(), t->Y() + gDesign->mTileHeight);
}

CBBox* CBBox::GetIntersectionBB(const CBBox* b) const
{
  return getIntersectionBB(b->mx, b->mX, b->my, b->mY);
}

#define max(a, b) a>b ? a : b
#define min(a, b) a<b ? a : b
CBBox* CBBox::getIntersectionBB(double x, double X, double y, double Y) const
{
  if (mx > X || mX < x || my > Y || mY < y)
    return NULL;

  CBBox *tmpBB = new CBBox(
    max(mx, x),
    min(mX, X),
    max(my, y),
    min(mY, Y)
  );

#ifdef _VERBOSE
  cout << " Intersection of " << mx << "," << my << " " << mX << "," << mY <<
  " and " << x << "," << y << " " << X << "," << Y << " is " << tmpBB->mx <<
  "," << tmpBB->my << " " << tmpBB->mX << "," << tmpBB->mY << endl;
#endif

  return tmpBB;
}
