/*
***********************************************************************
LICENSE AGREEMENT
***********************************************************************
 READ THIS LICENSE AGREEMENT CAREFULLY BEFORE USING THIS PRODUCT. BY 
 USING THIS PRODUCT YOU INDICATE YOUR ACCEPTANCE OF THE TERMS OF THE 
 FOLLOWING AGREEMENT. THESE TERMS APPLY TO YOU AND ANY SUBSEQUENT 
 LICENSEE OF THIS PRODUCT.

               License Agreement for RegularPlace 1.0

Copyright (c)  2008 by 
Ashutosh Chakraborty, Anurag Kumar, David Z. Pan
All rights reserved

ATTRIBUTION ASSURANCE LICENSE
Redistribution of the any code, with or without modification, are permitted 
provided that the conditions below are met. These conditions require a 
modest attribution to Ashutosh Chakraborty, Anurag Kumar and Dr. 
David Z. Pan (the "Author").

1. Redistributions of any code, with or without modification (the "Code"), 
must be accompanied by any documentation and, each time the resulting 
executable program or a program dependent thereon is launched, a 
prominent display (e.g., splash screen or banner text) of the Author's 
attribution information, which includes:
  (a) Ashutosh Chakraborty, Anurag Kumar, and Dr. David Z. Pan ("AUTHOR"),
  (b) The University of Texas at Austin ("PROFESSIONAL IDENTIFICATION"), and
  (c) http://www.cerc.utexas.edu/utda/ ("URL").

2. Use is limited to academic research groups only. Users who are interested 
in industry or commercial purposes must notify Author and request 
separate license agreement.  

3. Neither the name nor any trademark of the Author may be used to 
endorse or promote products derived from this software without specific 
prior written permission.

4. Users are entirely responsible, to the exclusion of the Author and 
any other persons, for compliance with (a) regulations set by owners or 
administrators of employed equipment, (b) licensing terms of any other 
software, and (c) local, national, and international regulations regarding 
use, including those regarding import, export, and use of encryption software. 

THIS FREE SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR 
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE AUTHOR OR ANY CONTRIBUTOR BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, EFFECTS OF UNAUTHORIZED OR MALICIOUS NETWORK 
ACCESS; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

/*
 * =======================================================================
 *
 *       Filename:  BBox.h
 *
 *    Description:  Class representing bounding box of a net
 *
 *        Version:  1.0
 *        Created:  08/24/2008 10:11:11 AM CDT
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ashutosh Chakraborty (), ashutosh@mail.utexas.edu
 *        Company:  University of Texas, Austin
 *
 * =======================================================================
 */

#ifndef _BBOX_HXX
#define _BBOX_HXX

class CNode;
class CNet;
class CTile;

class CBBox
{
  public: // Constructor/destructors
    CBBox();

    CBBox(double axm, double axM, double aym, double ayM)
        : mx(axm), mX(axM), my(aym), mY(ayM) {}

    // Construct BBox for the given net and return it
    CBBox(const CNet *aNet);

  public:  // Copy constructor
    CBBox& operator=(const CBBox& that);

  public: // Set and Get function
    double x() const
    {
      return mx;
    }
    double X() const
    {
      return mX;
    }
    double y() const
    {
      return my;
    }
    double Y() const
    {
      return mY;
    }

  public: // Other functions
    bool Contains(CNode *aNode)           const;
    bool Contains(double x, double y)     const;
    double ShortestDistance(CNode* aNode) const;
    CBBox* GetIntersectionBB(const CTile *)     const;
    CBBox* GetIntersectionBB(const CBBox *)     const;
    double HPWL() const;

    friend std::ostream& operator<<(std::ostream& s, const CBBox& b)
    {
      s << "(" << b.mx << "," << b.my << ") (" <<
      b.mX << "," << b.mY << ")";
      return s;
    }

  private:
    double mx;  // minimum x
    double mX;  // maximum x
    double my;  // minimum y
    double mY;  // maximum y
    CBBox* getIntersectionBB(double, double, double, double) const;
};

#endif
