/*
***********************************************************************
LICENSE AGREEMENT
***********************************************************************
 READ THIS LICENSE AGREEMENT CAREFULLY BEFORE USING THIS PRODUCT. BY 
 USING THIS PRODUCT YOU INDICATE YOUR ACCEPTANCE OF THE TERMS OF THE 
 FOLLOWING AGREEMENT. THESE TERMS APPLY TO YOU AND ANY SUBSEQUENT 
 LICENSEE OF THIS PRODUCT.

               License Agreement for RegularPlace 1.0

Copyright (c)  2008 by 
Ashutosh Chakraborty, Anurag Kumar, David Z. Pan
All rights reserved

ATTRIBUTION ASSURANCE LICENSE
Redistribution of the any code, with or without modification, are permitted 
provided that the conditions below are met. These conditions require a 
modest attribution to Ashutosh Chakraborty, Anurag Kumar and Dr. 
David Z. Pan (the "Author").

1. Redistributions of any code, with or without modification (the "Code"), 
must be accompanied by any documentation and, each time the resulting 
executable program or a program dependent thereon is launched, a 
prominent display (e.g., splash screen or banner text) of the Author's 
attribution information, which includes:
  (a) Ashutosh Chakraborty, Anurag Kumar, and Dr. David Z. Pan ("AUTHOR"),
  (b) The University of Texas at Austin ("PROFESSIONAL IDENTIFICATION"), and
  (c) http://www.cerc.utexas.edu/utda/ ("URL").

2. Use is limited to academic research groups only. Users who are interested 
in industry or commercial purposes must notify Author and request 
separate license agreement.  

3. Neither the name nor any trademark of the Author may be used to 
endorse or promote products derived from this software without specific 
prior written permission.

4. Users are entirely responsible, to the exclusion of the Author and 
any other persons, for compliance with (a) regulations set by owners or 
administrators of employed equipment, (b) licensing terms of any other 
software, and (c) local, national, and international regulations regarding 
use, including those regarding import, export, and use of encryption software. 

THIS FREE SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR 
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE AUTHOR OR ANY CONTRIBUTOR BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, EFFECTS OF UNAUTHORIZED OR MALICIOUS NETWORK 
ACCESS; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <iostream>
#include <vector>
#include <list>
#include <string>
#include <cfloat>
#include <map>
#include <fstream>

#include <math.h>
using namespace std;

#include "Design.h"
#include "Data.h"
#include "BBox.h"
#include "Net.h"

CNode::CNode(string name, int aType, int aClkDomain)
    : mName(name), mX(0), mY(0), mType(aType),
    isFixed(false), mPartitionId(-1)
{
  assert(false);
  makeNetsDirty();
}

void CNode::makeNetsDirty() const
{
//   assert(0 < mIncidentNets.size());
  for (int i = 0; i < mIncidentNets.size(); i++)
    mIncidentNets[i]->Dirty(true);
}

CNode::CNode(string name)
    : mName(name), mX(0.0), mY(0.0), mType(-1),
    mWd(-1), mHt(-1), isFixed(false), mPartitionId(-1),
    mBestBBox(NULL), mDistFromBestBox(0),
    mCurrLocationInHessian(-1), hasNoMovableNeighbor(true),
    mPartnIdX(-1), mPartnIdY(-1)
{
  mNeighbors.clear();
  mWtsNeighbors.clear();
  static int currId;
  mId = currId++;
  makeNetsDirty();
}

CNode::CNode(void)
    : mName(""), mX(0), mY(0), mType(0), isFixed(false), mPartitionId(-1)
{
  assert(false);
}

CNode::~CNode(void)
{
}

void CNode::Initialize(int iType, string cName)
{
  mType = iType;
  mName = cName;
}

static const int gMaxNumNodesInNetForBBoxComputation = 100;
void CNode::ComputeOptimalBBox()
{
  assert(0 < mIncidentNets.size());

  vector <double> allX;
  vector <double> allY;

#if 0
  allX.reserve(mIncidentNets.size() << 1);
  allY.reserve(mIncidentNets.size() << 1);
#endif

  int maxX = static_cast<int>(gDesign->mNumTilesX * (gDesign->mTileWidth + gDesign->mInterTileSpaceX));
  int maxY = static_cast<int>(gDesign->mNumTilesY * (gDesign->mTileHeight + gDesign->mInterTileSpaceY));

  for (int i = 0; i < mIncidentNets.size(); ++i)
  {
    if (mIncidentNets[i]->mNodes.size() > gMaxNumNodesInNetForBBoxComputation)
      continue;
    CBBox *tmp = mIncidentNets[i]->GetBBoxWithoutNode(this);
    allX.push_back(std::min<double>(std::max<double>(tmp->x(), gDesign->mIOSpaceX), maxX - 1));
    allX.push_back(std::min<double>(std::max<double>(tmp->X(), gDesign->mIOSpaceX), maxX - 1));
    allY.push_back(std::min<double>(std::max<double>(tmp->y(), gDesign->mIOSpaceY), maxY - 1));
    allY.push_back(std::min<double>(std::max<double>(tmp->Y(), gDesign->mIOSpaceY), maxY - 1));
    delete tmp;
  }

  for (int i = allX.size() - 1; i >= 0; --i)
  {
    assert(allX[i] >= 0);
    assert(allX[i] < maxX);
    assert(allY[i] >= 0);
    assert(allY[i] < maxY);
  }

  assert(allX.size() == allY.size());
//   assert(allX.size() == (mIncidentNets.size() << 1));

  std::sort(allX.begin(), allX.end());
  std::sort(allY.begin(), allY.end());

  if (2 > allX.size() || 2 > allY.size())
  {
    cout << "No optimal BBox computed for " << mName << endl;
//     cout << "For node " << mName << " something wrong" << endl;
//     abort();
    assert(mBestBBox == NULL);
    mBestBBox = NULL;
    mIsInBestBox = true;
    mDistFromBestBox = -INT_MAX;
    return;
  }

// #define _VERBOSE
#ifdef _VERBOSE
  cout << "X: ";
  for (int i = 0; i < allX.size(); ++i)
    cout << allX[i] << " ";
  cout << endl;

  cout << "Y: ";
  for (int i = 0; i < allY.size(); ++i)
    cout << allY[i] << " ";
  cout << endl;
#endif

  double x = allX[(allX.size() >> 1) - 1];
  double X = allX[allX.size() >> 1];
  double y = allY[(allY.size() >> 1) - 1];
  double Y = allY[allY.size() >> 1];

  if (mBestBBox != NULL)
  {
    delete mBestBBox;
    mBestBBox = NULL;
  }
  mBestBBox = new CBBox(x, X, y, Y);
  ComputeBestBBDistance();
  mIsInBestBox = (fabs(mDistFromBestBox) < DBL_ZERO_ABS);

#ifdef _VERBOSE
  cout << *mBestBBox << endl;
#endif
}

double CNode::GetIncidentWL()
{
  double wl = 0.0;
//   cout << mName << " has " << mIncidentNets.size() << " nets attaced to it" << endl;
  for (int i = 0; i < mIncidentNets.size(); ++i)
  {
    wl += mIncidentNets[i]->GetHPWL();
  }
  return wl;
}

bool cmpNodes(const CNode* a, const CNode* b)
{
//   cout << "Comparing " << a->Name() << " and " << b->Name() << endl;
  if (a->Y() < b->Y())
    return true;
#if 0
  if (a->Y() > b->Y())
    return false;
  if (a->X() < b->X())
    return true;
  if (a->X() > b->X())
    return false;
#endif
  return false;
}

void CNode::SwapLocation(CNode *n)
{
  n->makeNetsDirty();
  this->makeNetsDirty();

  double x = n->mX;
  double y = n->mY;

  n->mX = mX;
  n->mY = mY;

  mX = x;
  mY = y;
}

bool cmpNodeBBDistance(const CNode* a, const CNode* b)
{
//   cout << "Comparing " << *a << " and " << *b << endl;
  // Easy decisions.  Farther is better.
  if (a->DistToBB() > b->DistToBB())
    return true;;
  if (a->DistToBB() < b->DistToBB())
    return false;

#if 0
  // If the distance to the best bounding box is same for two nodes,
  // the one which has less #incident nets is better suited.
  if (a->GetNumIncidentNets() > b->GetNumIncidentNets())
    return false;
  else return true;
#endif

  return false;
}

vector<CNode*> CNode::GetConnectedCellSorted(int &numMovable, int &numFixed)
{
  map <int, bool> tmpMap;

  for (unsigned i = 0; i < mIncidentNets.size(); i++)
  {
    CNet *tmpNet = mIncidentNets[i];
    vector<CNode *> connCells = tmpNet->GetCellsInNet();
    for (unsigned j = 0; j < connCells.size(); j++)
    {
      if (this->ID() == connCells[j]->ID())
        continue;
      if (tmpMap.end() == tmpMap.find(connCells[j]->ID()))
      {
        tmpMap[connCells[j]->ID()] = true;
      }
      else
      {
        // Cell already added in map
      }
    }
  }

  if (tmpMap.size() == 0)
    cout << " Unconnected cell\n";

  numMovable = 0;
  numFixed = 0;

  vector<CNode *> result;

  map<int, bool>::iterator it = tmpMap.begin();
  for (; it != tmpMap.end(); ++it)
  {
    CNode *tmpNode = gDesign->mAllNodes[(*it).first];
    if (true == tmpNode->Fixed())
      continue;
    result.push_back(tmpNode);
  }
  numMovable = result.size();

  it = tmpMap.begin();
  for (; it != tmpMap.end(); ++it)
  {
    CNode *tmpNode = gDesign->mAllNodes[(*it).first];
    if (false == tmpNode->Fixed())
      continue;
    result.push_back(tmpNode);
  }
  numFixed = result.size() - numMovable;

  for (int i = 0; i < numMovable; i++)
    assert(false == result[i]->Fixed());

  for (int i = 0; i < numFixed; i++)
    assert(true == result[numMovable + i]->Fixed());

  return result;
}

void CNode::UniquifyNeighbor()
{
  map <CNode*, double> mapNodeToWts;
  assert(mNeighbors.size() == mWtsNeighbors.size());

  for (int i = 0; i < mNeighbors.size(); i++)
  {
    if (mapNodeToWts.find(mNeighbors[i]) == mapNodeToWts.end())
    {
      mapNodeToWts[mNeighbors[i]] = mWtsNeighbors[i];
    }
    else
    {
      mapNodeToWts[mNeighbors[i]] += mWtsNeighbors[i];
    }
  }

  mNeighbors.clear();
  mWtsNeighbors.clear();

  map <CNode*, double>::iterator it = mapNodeToWts.begin();
  for (; it != mapNodeToWts.end(); it++)
  {
    mNeighbors.push_back(it->first);
    mWtsNeighbors.push_back(it->second);
    assert(it->second > 0);
  }

  assert(mNeighbors.size() == mWtsNeighbors.size());
}
