/*
***********************************************************************
LICENSE AGREEMENT
***********************************************************************
 READ THIS LICENSE AGREEMENT CAREFULLY BEFORE USING THIS PRODUCT. BY 
 USING THIS PRODUCT YOU INDICATE YOUR ACCEPTANCE OF THE TERMS OF THE 
 FOLLOWING AGREEMENT. THESE TERMS APPLY TO YOU AND ANY SUBSEQUENT 
 LICENSEE OF THIS PRODUCT.

               License Agreement for RegularPlace 1.0

Copyright (c)  2008 by 
Ashutosh Chakraborty, Anurag Kumar, David Z. Pan
All rights reserved

ATTRIBUTION ASSURANCE LICENSE
Redistribution of the any code, with or without modification, are permitted 
provided that the conditions below are met. These conditions require a 
modest attribution to Ashutosh Chakraborty, Anurag Kumar and Dr. 
David Z. Pan (the "Author").

1. Redistributions of any code, with or without modification (the "Code"), 
must be accompanied by any documentation and, each time the resulting 
executable program or a program dependent thereon is launched, a 
prominent display (e.g., splash screen or banner text) of the Author's 
attribution information, which includes:
  (a) Ashutosh Chakraborty, Anurag Kumar, and Dr. David Z. Pan ("AUTHOR"),
  (b) The University of Texas at Austin ("PROFESSIONAL IDENTIFICATION"), and
  (c) http://www.cerc.utexas.edu/utda/ ("URL").

2. Use is limited to academic research groups only. Users who are interested 
in industry or commercial purposes must notify Author and request 
separate license agreement.  

3. Neither the name nor any trademark of the Author may be used to 
endorse or promote products derived from this software without specific 
prior written permission.

4. Users are entirely responsible, to the exclusion of the Author and 
any other persons, for compliance with (a) regulations set by owners or 
administrators of employed equipment, (b) licensing terms of any other 
software, and (c) local, national, and international regulations regarding 
use, including those regarding import, export, and use of encryption software. 

THIS FREE SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR 
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE AUTHOR OR ANY CONTRIBUTOR BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, EFFECTS OF UNAUTHORIZED OR MALICIOUS NETWORK 
ACCESS; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

#ifndef __GRID_H__
#define __GRID_H__

#include "Util.h"
#include "BBox.h"
class CNet;
class Cell;

class CNode
{
  public:
    CNode(void);
    CNode(string name, int aType, int aClkDomain);
    CNode(string name);
    ~CNode(void);
    void Initialize(int, string);

  public: // Set and Get Parameters
    double X() const
    {
      return mX;
    }
    double Y() const
    {
      return mY;
    }
    string Name() const
    {
      return mName;
    }
    int    Type() const
    {
      return mType;
    }
    int    Width() const
    {
      return mWd;
    }
    int    Height() const
    {
      return mHt;
    }
    int    Area() const
    {
      return mWd * mHt;
    }
    bool   Fixed() const
    {
      return isFixed;
    }
    void   Width(int aWd)
    {
      mWd = aWd;
    }
    void   Height(int aHt)
    {
      mHt = aHt;
    }
    void   Fixed(bool aIsFixed)
    {
      isFixed = aIsFixed;
    }
    void PutAt(double x, double y)
    {
      this->makeNetsDirty();
      mX = x;
      mY = y;
    }
    void   X(double aX)
    {
      makeNetsDirty();
      mX = aX;
    }
    void   Y(double aY)
    {
      makeNetsDirty();
      mY = aY;
    }
    void   Name(string aName)
    {
      mName = aName;
    }
    void   Type(int aType)
    {
      mType = aType;
    }
    void ComputeBestBB();
    void AddClockDomain(int aClkDomID)
    {
      mClkDom.push_back(aClkDomID);
    }
    void AddNet(CNet *aNet)
    {
      mIncidentNets.push_back(aNet);
    }

    void   ComputeOptimalBBox();
    double GetIncidentWL();

    void   ComputeBestBBDistance()
    {
      assert(mBestBBox != NULL);
      mDistFromBestBox = mBestBBox->ShortestDistance(this);
    }

  public:
    void ID(int aID)
    {
      mId = aID;
    }
    int ID() const
    {
      return mId;
    }

    void PartitionId(int aPartitionId)
    {
      mPartitionId = aPartitionId;
    }
    int PartitionId()
    {
      return mPartitionId;
    }

    int GetNumIncidentNets() const
    {
      return mIncidentNets.size();
    }
    void SwapLocation(CNode* n);
    double DistToBB() const
    {
      return mDistFromBestBox;
    }
    void   DistToBB(double aDist)
    {
      mDistFromBestBox = aDist;
    }
    CBBox* BestBB() const
    {
      return mBestBBox;
    }

  private:
    void makeNetsDirty() const;

  public:  // To allow easy access in Tile.cpp
    vector <int> mClkDom;

  private:
    double mX;
    double mY;

  private:
    vector <CNet*> mIncidentNets;
    string mName;
    int mType;
    int mWd;
    int mHt;
    bool isFixed;
    int mId;

    // Defunct.  Should not use.
    int mPartitionId;

  private:
    CBBox *mBestBBox;
    bool  mIsInBestBox;

  public: // To access from compareNodePtr
    double mDistFromBestBox;

  public:
    Cell *mGlobalCell;

  public:
    bool operator==(const CNode &other) const
    {
      abort();
      return (mId == other.mId);
    }

    friend std::ostream& operator<<(std::ostream& s, const CNode& t)
    {
      s << "Cell " << t.mName << " @ x/y=(" << t.X() << "/" << t.Y() <<
      ") @ dist=" << t.mDistFromBestBox;
      return s;
    }

    vector<CNode*> GetConnectedCellSorted(int &a, int &b);


    // Functions related to neighbour computation.
  public:
    void ClearNeighbors()
    {
      mNeighbors.clear();
      mWtsNeighbors.clear();
    }

    void UniquifyNeighbor();

    void AddNeighbor(CNode *n, double w)
    {
      if (n->Name() == this->Name())
        cout << " failing for " << n->Name() << endl;
      mNeighbors.push_back(n);
      mWtsNeighbors.push_back(w);
      if (false == n->Fixed())
        hasNoMovableNeighbor = false;
    }

    int GetNumNeighbors()
    {
      return mNeighbors.size();
    }
    vector<CNode *>& GetNeighborCells()
    {
      return mNeighbors;
    }
    vector<double>& GetNeighborWts()
    {
      return mWtsNeighbors;
    }

    // Vector of the cells neighboring this cell.
  private:
    vector <CNode*> mNeighbors;
    vector <double> mWtsNeighbors;

  public:
    void SetCurLocInHess(const int a)
    {
      mCurrLocationInHessian = a;
    }
    int  GetCurLocInHess()
    {
      return mCurrLocationInHessian;
    }
    int mCurrLocationInHessian;

  public:
    int  PartitionIDX()
    {
      return mPartnIdX;
    }
    int  PartitionIDY()
    {
      return mPartnIdY;
    }
    void PartitionIDX(const int id)
    {
      mPartnIdX = id;
    }
    void PartitionIDY(const int id)
    {
      mPartnIdY = id;
    }

  private:
    int mPartnIdX;
    int mPartnIdY;

  public:
    bool hasNoMovableNeighbor;
};

#endif
