/************************************************************************
LICENSE AGREEMENT
***********************************************************************
 READ THIS LICENSE AGREEMENT CAREFULLY BEFORE USING THIS PRODUCT. BY 
 USING THIS PRODUCT YOU INDICATE YOUR ACCEPTANCE OF THE TERMS OF THE 
 FOLLOWING AGREEMENT. THESE TERMS APPLY TO YOU AND ANY SUBSEQUENT 
 LICENSEE OF THIS PRODUCT.

               License Agreement for RegularPlace 1.0

Copyright (c)  2008 by 
Ashutosh Chakraborty, Anurag Kumar, David Z. Pan
All rights reserved

ATTRIBUTION ASSURANCE LICENSE
Redistribution of the any code, with or without modification, are permitted 
provided that the conditions below are met. These conditions require a 
modest attribution to Ashutosh Chakraborty, Anurag Kumar and Dr. 
David Z. Pan (the "Author").

1. Redistributions of any code, with or without modification (the "Code"), 
must be accompanied by any documentation and, each time the resulting 
executable program or a program dependent thereon is launched, a 
prominent display (e.g., splash screen or banner text) of the Author's 
attribution information, which includes:
  (a) Ashutosh Chakraborty, Anurag Kumar, and Dr. David Z. Pan ("AUTHOR"),
  (b) The University of Texas at Austin ("PROFESSIONAL IDENTIFICATION"), and
  (c) http://www.cerc.utexas.edu/utda/ ("URL").

2. Use is limited to academic research groups only. Users who are interested 
in industry or commercial purposes must notify Author and request 
separate license agreement.  

3. Neither the name nor any trademark of the Author may be used to 
endorse or promote products derived from this software without specific 
prior written permission.

4. Users are entirely responsible, to the exclusion of the Author and 
any other persons, for compliance with (a) regulations set by owners or 
administrators of employed equipment, (b) licensing terms of any other 
software, and (c) local, national, and international regulations regarding 
use, including those regarding import, export, and use of encryption software. 

THIS FREE SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR 
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE AUTHOR OR ANY CONTRIBUTOR BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, EFFECTS OF UNAUTHORIZED OR MALICIOUS NETWORK 
ACCESS; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <iostream>
#include <vector>
#include <string>
#include <queue>
#include <map>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <cfloat>

using namespace std;

#include <math.h>

#include "Design.h"
#include "BBox.h"
#include "Net.h"
#include "Row.h"
#include "Region.h"
#include "munkres.h"
#include "Tile.h"
#include "MetaPlacer/metaPl.h"

// Is used only for tile creation and manipulation
vector<CTile*> gTilesVec;
extern bool cmpTiles(const CTile *a, const CTile *b);
extern bool cmpNodes(const CNode *a, const CNode *b);
bool cmpNodeBBDistance(const CNode* a, const CNode* b);
void dumpMatrix(Matrix<int>& m);
string makeString(int x);

static const int gWidthOfEcellAndDff = 60;

static const int gWidthTile    = 92;
static const int gHeightTile   = 64;

CDesign* gDesign;

CDesign::CDesign(void)
    : mPlatformTop(NULL), mNumNodes(-1), mNumNets(-1), mNumPins(-1),
    mNumTerms(-1), mWidth(-1), mHeight(-1), mIOSpaceX(-1),
    mIOSpaceY(-1), mCoreWidth(-1), mCoreHeight(-1), mNumGroups(-1),
    mIsShrunk(false), mTopInstName("")
{
  mPlatformTop = NULL;
  gDesign = this;
}

CDesign::~CDesign(void)
{
}

void CDesign::Read(
  string nodesFileName,
  string netFileName,
  string propFileName,
  string platRegionName,
  string platSclName,
  string initPlaceName)
{
  // Read the .nodes file
  readNodesFile(nodesFileName);

  // Read the .nets file
  readNetsFile(netFileName);

  // Read the .props file
  readPropsFile(propFileName);

#if 1
  // Read the platform .region file
  readPlatformRegion(platRegionName);

  // Read the platform .scl file
  readPlatformScl(platSclName);
#endif

  // Read the .pl file associated with benchmark
  ReadPLFile(initPlaceName);
}

void CDesign::dumpPartitionedFiles(string aFileName, int iPartitionId)
{
  cout << "STAT: Enter dumpPartitionedFiles" << endl;
  string aNodeFileName = aFileName + ".nodes";
  string aPlFileName = aFileName + ".pl";
  int iNumberOfNodes = 0, iNumberOfTerminals = 0;
  // Dump Nodes file

  FILE *fNodeOutFile = fopen(aNodeFileName.c_str(), "w");
  if (!fNodeOutFile)
    cout << "ERROR: Could not open Node file for writing\n";
  FILE *fPlOutFile = fopen(aPlFileName.c_str(), "w");
  if (!fPlOutFile)
    cout << "ERROR: Could not open Pl file for writing\n";
  fprintf(fPlOutFile, "UCLA pl 1.0\n");
  fprintf(fPlOutFile, "# Created : Anurag Kumar\n");
  fprintf(fPlOutFile, "# User    : eASIC netlist2nodes\n\n");

  fprintf(fNodeOutFile, "UCLA nodes 1.0\n");
  fprintf(fNodeOutFile, "# Created : Anurag Kumar\n");
  fprintf(fNodeOutFile, "# User    : eASIC netlist2nodes\n\n");

  for (unsigned i = 0; i < mAllNodes.size(); ++i)
  {
    if (iPartitionId == mAllNodes[i]->PartitionId())
    {
      ++iNumberOfNodes;
      if (mAllNodes[i]->Fixed())
      {
        ++iNumberOfTerminals;
      }
    }
  }
  fprintf(fNodeOutFile, "NumNodes : %d\n", iNumberOfNodes);
  fprintf(fNodeOutFile, "NumTerminals : %d\n", iNumberOfTerminals);

  for (unsigned i = 0; i < mAllNodes.size(); ++i)
  {
    if (iPartitionId == mAllNodes[i]->PartitionId())
    {
      fprintf(fPlOutFile, "%s %lf %lf : N\n", mAllNodes[i]->Name().c_str(), mAllNodes[i]->X(), mAllNodes[i]->Y());
      fprintf(fNodeOutFile, "%s %d %d", mAllNodes[i]->Name().c_str(), mAllNodes[i]->Width(), mAllNodes[i]->Height());
      if (mAllNodes[i]->Fixed())
        fprintf(fNodeOutFile, " terminal");
      fprintf(fNodeOutFile, "\n");
    }
  }
  if (fNodeOutFile)
    fclose(fNodeOutFile);
  if (fPlOutFile)
    fclose(fPlOutFile);

  string aNetFileName = aFileName + ".nets";
  FILE *fNetOutFile = fopen(aNetFileName.c_str(), "w");
  fprintf(fNetOutFile, "                                       \n");
  fprintf(fNetOutFile, "                                       \n");
  fprintf(fNetOutFile, "                                       \n");
  fprintf(fNetOutFile, "                                       \n");
  fprintf(fNetOutFile, "                                       \n");
  fprintf(fNetOutFile, "                                       \n");
  fprintf(fNetOutFile, "                                       \n");
  fprintf(fNetOutFile, "                                       \n");
  fprintf(fNetOutFile, "                                       \n");
  int iNetDegree;
  int aNumNets = 0;
  int aNumPins = 0;
  vector <CNode*> vConnectedNodes;
  for (int i = 0; i < mAllNets.size(); ++i)
  {
    iNetDegree = mAllNets[i]->GetNumNodesInPartition(iPartitionId);
    if (iNetDegree > 1)
    {
//    cout << "NetDegree " << iNetDegree << " " << mAllNets[i]->Name() << endl;
      ++aNumNets;
      fprintf(fNetOutFile, "NetDegree : %d %s\n", iNetDegree, mAllNets[i]->Name().c_str());
      vConnectedNodes = mAllNets[i]->GetCellsInNet();
      for (unsigned int j = 0; j < vConnectedNodes.size(); ++j)
      {
        if (vConnectedNodes[j]->PartitionId() == iPartitionId)
        {
          ++aNumPins;
//     cout << vConnectedNodes[j]->Name() << " I : " << vConnectedNodes[j]->X() << " " << vConnectedNodes[j]->Y() << endl;
          fprintf(fNetOutFile, "%s I : %lf %lf\n", vConnectedNodes[j]->Name().c_str(), vConnectedNodes[j]->X(), vConnectedNodes[j]->Y());
        }
      }
    }
  }
  fseek(fNetOutFile, 0, SEEK_SET);
  fprintf(fNetOutFile, "UCLA nets 1.0\n");
  fprintf(fNetOutFile, "# Created : Anurag Kumar\n");
  fprintf(fNetOutFile, "# User    : eASIC design2nets\n\n");
  fprintf(fNetOutFile, "NumNets : %d\nNumPins : %d\n", aNumNets, aNumPins);
  if (fNetOutFile)
    fclose(fNetOutFile);
  cout << "STAT: Exiting dumpPartitionedFiles" << endl;
}

void CDesign::dumpInputForHmetis(string aHmetisInputFileName)
{
  cout << "STAT: Enter dumpInputForHmetis" << endl;

  FILE *pFile = fopen("tempfile1", "w");
  CNet *pCurrentNet;
  CNode *pCurrentNode;
  vector <CNode*> curNodes;
  map <string, int> mMapNameToInt;
  map <int, string> mMapIntToName;
  int k = 1;
  for (int i = 0; i < mAllNets.size(); ++i)
  {
    curNodes = mAllNets[i]->GetCellsInNet();
    for (int j = 0; j < curNodes.size(); ++j)
    {
      if (mMapNameToInt[curNodes[j]->Name()] != 0)
      {
        fprintf(pFile, "%d ", mMapNameToInt[curNodes[j]->Name()]);
      }
      else
      {
        mMapNameToInt[curNodes[j]->Name()] = k;
        mMapIntToName[k] = curNodes[j]->Name();
        fprintf(pFile, "%d ", k);
        ++k;
      }
    }
    fprintf(pFile, "\n");
  }
  assert(mMapIntToName.size() == mMapNameToInt.size());
  cout << mMapNameToInt.size() << ":" << mAllNodes.size() << ":" << mMapIntToName.size() << endl;
//   exit(1);
//   fprintf(pFile,"No of nodes: %d ",k);
  if (pFile)
    fclose(pFile);

  FILE *pTempFile = fopen("tempfile", "w");
  fprintf(pTempFile, "%d %d\n", mAllNets.size(), k);
  if (pTempFile)
    fclose(pTempFile);
  string sCommandString = "cat tempfile tempfile1 > " + aHmetisInputFileName + " ; rm tempfile1 tempfile";
  system(sCommandString.c_str());

  string line;
  vector <string> tok;
  int iPartition;
  int iCountElements = 1;
  std::ifstream in;
  in.open("inputHmetis.hgr.part.4");
  if (!in)
  {
    cout << "EROR: ** Could not open hgr file " <<  endl;
  }

//   itrMapNameToInt = mMapNameToInt.begin();
  // Get number of nets and pins
  while (!in.eof())
  {
    getline(in, line);
    Tokenize(line, tok, "\n\t ");
    if (tok.size() != 1)
    {
      break;
    }
    if (iCountElements > mMapNameToInt.size())
      break;

//     mMapNameToInt[iCountElements];
    iPartition = atoi(tok[0].c_str());
    string sCurNodeName = mMapIntToName[iCountElements];
//  string sCurNodeName = (*itrMapNameToInt).first;
//  cout << iPartition << ":" << sCurNodeName << ":"  << iCountElements << " :" << mMapNameToInt[sCurNodeName] << endl;
//  cout << iPartition << ":" << sCurNodeName << ":" << (*itrMapNameToInt).second << " :" << iCountElements << endl;
    mMapNameToNode[sCurNodeName]->PartitionId(iPartition);
//  itrMapNameToInt++;
    ++iCountElements;
    tok.clear();
  }

  cout << "STAT: Exit dumpInputForHmetis" << endl;

  dumpPartitionedFiles("test", 1);
}

#if 0
void CDesign::dumpInputForHmetis(string aHmetisInputFileName)
{
  cout << "STAT: Enter dumpInputForHmetis" << endl;

  std::ofstream in;
  in.open(aHmetisInputFileName.c_str());

  if (!in)
  {
    cout << "EROR: ** Could not open hmetis output file " <<
    aHmetisInputFileName << endl;
  }

  in << mAllNets.size() << " " << mAllNodes.size() << endl;

  for (unsigned int i = 0; i < mAllNets.size(); i++)
  {
    for (unsigned int j = 0; j < mAllNets[i]->mNodes.size(); j++)
      in << mAllNets[i]->mNodes[j]->Id() << " ";

    in << endl;
  }

  in.close();

  FILE *pFile = fopen("tempfile1", "w");
  CNet *pCurrentNet;
  CNode *pCurrentNode;
  vector <CNode*> curNodes;
  map <string, int> mMapNameToInt;
  map <int, string> mMapIntToName;
//   map <string, int>::iterator itrMapNameToInt;
  int k = 1;
//   for(unsigned i=0; i<mAllNodes.size(); i++)
//   {
//       if(mAllNodes[i]->GetNumIncidentNets() == 0)
//        continue;
//       mMapNameToInt[mAllNodes[i]->Name()] = k;
//    mMapIntToName[k] = mAllNodes[i]->Name();
//    k++;
//   }

  k = 1;
  for (unsigned i = 0; i < mAllNets.size(); i++)
  {
    curNodes = mAllNets[i]->GetConnectedNodes();
    for (unsigned j = 0; j < curNodes.size(); j++)
    {
      if (mMapNameToInt[curNodes[j]->Name()] != 0)
      {
        fprintf(pFile, "%d ", mMapNameToInt[curNodes[j]->Name()]);
      }
      else
      {
        mMapNameToInt[mAllNodes[i]->Name()] = k;
        fprintf(pFile, "%d ", k);
        k++;
      }

//    fprintf(pFile,"%d ",mMapNameToInt[curNodes[j]->Name()]);
    }
    fprintf(pFile, "\n");
  }

//   printf("No of nodes: %d %d %d %d\n",k,mMapIntToName.size(),mMapNameToInt.size(),mAllNodes.size());
//   exit(0);
  if (pFile)
    fclose(pFile);

  FILE *pTempFile = fopen("tempfile", "w");
  fprintf(pTempFile, "%d %d\n", mAllNets.size(), k);
  if (pTempFile)
    fclose(pTempFile);
  string sCommandString = "cat tempfile tempfile1 > " + aHmetisInputFileName + " ; rm tempfile1 tempfile";

  system(sCommandString.c_str());



  string line;
  vector <string> tok;
  int iPartition;
  int iCountElements = 1;
  std::ifstream in;
  in.open("inputHmetis.hgr.part.4");
  if (!in)
  {
    cout << "EROR: ** Could not open hgr file " <<  endl;
  }

//   itrMapNameToInt = mMapNameToInt.begin();
  // Get number of nets and pins
  while (!in.eof())
  {
    getline(in, line);
    Tokenize(line, tok, "\n\t ");
    if (tok.size() != 1)
    {
      break;
    }
//     mMapNameToInt[iCountElements];
    iPartition = atoi(tok[0].c_str());
    string sCurNodeName = mMapIntToName[iCountElements];
//  string sCurNodeName = (*itrMapNameToInt).first;
//  cout << iPartition << ":" << sCurNodeName << ":"  << iCountElements << " :" << mMapNameToInt[sCurNodeName] << endl;
//  cout << iPartition << ":" << sCurNodeName << ":" << (*itrMapNameToInt).second << " :" << iCountElements << endl;
    mMapNameToNode[sCurNodeName]->PartitionId(iPartition);
//  itrMapNameToInt++;
    iCountElements++;
    tok.clear();
  }

  dumpPartitionedFiles("test", 1);
  cout << "STAT: Exit dumpInputForHmetis" << endl;
}
#endif

inline int gMapCellType(string s)
{
  if ("ecell" == s)
    return 0;
  else if ("edff" == s)
    return 1;
  else if ("reg_file" == s || "regfile" == s)
    return 2;
  else if ("bram" == s)
    return 3;
  abort();
}

int CDesign::readPropsFile(string aPropsFileName)
{
  cout << "STAT:  Enter readPropsFile" << endl;

  char cLine[128], cName[100], cValue[100];
  string tmpLine, cNodeName, line;
  char tmpStr[100];
  vector<string> tok;

  // Read the PROPS file
  FILE* pFile = fopen(aPropsFileName.c_str(), "rt");
  while (fgets(cLine, sizeof(cLine), pFile))
  {
    if (strncmp(cLine, "PropClass", sizeof("PropClass") - 1) == 0)
    {
      if (!fgets(cLine, sizeof(cLine), pFile))
        break;

      sscanf(cLine, "Name : %s", &cName);

      // Possible types of cells captured here
      if (0 == strcmp(cName, "type"))
      {
        while (fgets(cLine, sizeof(cLine), pFile))
        {
          if (strncmp(cLine, "EndPropClass",
                      sizeof("EndPropClass") - 1) == 0)
          {
            break;
          }

          sscanf(cLine, "Value : %s", &cValue);

          tmpLine = cValue;
        }
      }

      // Clock Domain names captured here
      else if (0 == strcmp(cName, "clock_domain"))
      {
//         printf ( "clock_domain\n" );

        while (fgets(cLine, sizeof(cLine), pFile))
        {
          if (strncmp(cLine, "EndPropClass",
                      sizeof("EndPropClass") - 1) == 0)
          {
            break;
          }

          sscanf(cLine, "Value : %s", &cValue);

          tmpLine = cValue;
          mMapClockDomainNameToInt[tmpLine] = mMapClockDomainNameToInt.size();
          mClockDomains.push_back(tmpLine);
          assert(mMapClockDomainNameToInt.size() == mClockDomains.size());
        }
      }
    }
    else if (strncmp(cLine, "Node", sizeof("Node") - 1) == 0)
    {
      // Nodes are captured here
      if (!fgets(cLine, sizeof(cLine), pFile))
        break;

      sscanf(cLine, "Name : %s", &tmpStr);

      cNodeName = tmpStr;

      // Check that such a node should exist
      CNode *tmpNode = mMapNameToNode[cNodeName];
      assert(NULL != tmpNode);

      if (!fgets(cLine, sizeof(cLine), pFile))
        break;

      while (strncmp(cLine, "Prop : ", sizeof("Prop : ") - 1) == 0)
      {
        line = cLine;
        Tokenize(line, tok, "\n\t ");

        if ("clock_domain" == tok[2])
        {
          int numDomains = tok.size() - 5;
          // Last token is always "\n" due to reading the data using fgets
          // Skip it by iterating upto numDomains-1
          for (int i = 0; i < numDomains; ++i)
            tmpNode->AddClockDomain(mMapClockDomainNameToInt[tok[5+i]]);
        }
        else if ("type" == tok[2])
        {
          tmpNode->Type(gMapCellType(tok[5]));
        }
        else
        {
          // Should not come here
          assert(false);
        }

        tok.clear();

        if (!fgets(cLine, sizeof(cLine), pFile))
          break;
      }
    }
  }

  for (int i = 0; i < mAllNodes.size(); ++i)
  {
    switch (mAllNodes[i]->Type())
    {
    case 0:
      mECells.push_back(i);
      break;
    case 1:
      mDffs.push_back(i);
      break;
    case 2:
      mRegs.push_back(i);
      break;
    case 3:
      mBrams.push_back(i);
      break;
    case - 1:
      assert(mAllNodes[i]->Fixed());
      mFixeds.push_back(i);
      break;
    default:
      abort();
    }
  }

  cout << "ecell=" << mECells.size() << " edff=" << mDffs.size() <<
  " regs=" << mRegs.size() << " bram=" << mBrams.size() << endl;

  cout << "STAT:  Exiting readPropsFile" << endl;

  return 0;
}


// This function
// READS .nodes file and POPULATES mAllNodes vector with
// pointer to cells and also sets fixed/unfixed type and
// the width and the height of each cell
int CDesign::readNodesFile(string aNodesFileName)
{
  cout << "STAT:  Enter readNodesFile" << endl;
  string line;
  vector <string> tok;

  std::ifstream in;
  in.open(aNodesFileName.c_str());

  if (!in)
    cout << "EROR: ** Could not open nets file " << aNodesFileName << endl;

  // Get number of nets and pins
  while (true)
  {
    tok.clear();
    getline(in, line);
    Tokenize(line, tok, "\n\t ");

    if (0 == tok.size())
      continue;

    if ("NumNodes" == tok[0])
    {
      mNumNodes = atoi(tok[2].c_str());
      cout << "  - Expecting " << mNumNodes << " nodes" << endl;
    }
    else if ("NumTerminals" == tok[0])
    {
      mNumTerms = atoi(tok[2].c_str());
      cout << "  - Expecting " << mNumTerms << " terminals" << endl;
      break;
    }
  }

  mAllNodes.reserve(mNumNodes);

#if 1
  getline(in, line);
#else
  do
  {
    getline(in, line);
  }
  while (line == "");
#endif

  int numTerm = 0;
  int numCell = 0;
  while (!in.eof())
  {
    getline(in, line);

    if (0 == line.length())
      break;

    tok.clear();
    Tokenize(line, tok, "\t ");
    assert(3 == tok.size() || 4 == tok.size());
    CNode* tmpNode = new CNode(tok[0]);
    tmpNode->Width(atoi(tok[1].c_str()));
    tmpNode->Height(atoi(tok[2].c_str()));
    // If the line has 4 parts, the last parts is
    // "terminal" meaning the node is fixed
    tmpNode->Fixed(4 == tok.size());

    if (4 == tok.size())
      ++numTerm;
    else
      ++numCell;

    mMapNameToNode[tok[0]] = tmpNode;
    mAllNodes.push_back(tmpNode);
    tok.clear();
  }

//   cout << "Got term = " << numTerm  << " expecting " << mNumTerms << endl;
  assert(numTerm == mNumTerms);
  assert(numCell == mNumNodes - mNumTerms);
  assert(-1 != mNumTerms);
  assert(-1 != mNumNodes);

  cout << "STAT:  Exiting readNodesFile (" << mAllNodes.size() <<
  ") nodes read" << endl;

  return mAllNodes.size();
}

// Read the .info platform file
int CDesign::readPlatformInfo(string aFileName)
{
  cout << "STAT:  Enter readPlatformInfo" << endl;

  string line;
  vector <string> tok;

  std::ifstream in;
  in.open(aFileName.c_str());

  if (!in)
  {
    cout << "EROR: ** Could not open nets file " << aFileName << endl;
    assert(false);
  }

  // Get platform information
  string sclFileName, regionFileName;

  while (!in.eof())
  {
    getline(in, line);
    Tokenize(line, tok, "\n\t ");

    if ("size" == tok[0])
    {
      assert("w" == tok[1] && "h" == tok[3] && 5 == tok.size());
      mWidth = atoi(tok[4].c_str());
      mHeight = atoi(tok[2].c_str());
    }
    else if ("files" == tok[0])
    {
      assert(":" == tok[1] && 4 == tok.size());
      sclFileName = tok[2];
      regionFileName = tok[3];
    }
    else if ("groups" == tok[0])
    {
      assert(2 == tok.size());
      mNumGroups = atoi(tok[1].c_str());
    }

    tok.clear();
  }

  cout << "PLAT:  Width = " << mWidth << " Height = " << mHeight <<

  " #Groups = " << mNumGroups << endl;
  cout << "WARN: ** FLIPPED WIDTH AND HEIGHT.  Is it still needed?" << endl;

  cout << "STAT:  Exiting readPlatformInfo" << endl;
  return 0;
}

// Read the regions file of Platform
map <string, CRegion*> gMapRegionNameToPtr;
int CDesign::readPlatformRegion(string aFileName)
{
  cout << "STAT:  Enter readPlatformRegion" << endl;
  string line;
  vector <string> tok;
  std::ifstream regionFile;
  regionFile.open(aFileName.c_str());

  if (!regionFile)
  {
    cout << "EROR: ** Could not open region file " << aFileName << endl;
    abort();
  }

  mPlatformTop = NULL;
  while (!regionFile.eof())
  {
    getline(regionFile, line);
    tok.clear();
    Tokenize(line, tok, " \t\n");

    if (0 == tok.size())
      continue;

    if ("PropArea" == tok[0])
    {
      CRegion *tmpReg = new CRegion();

      // Consume all the proparea cluster

      while ("EndPropArea" != tok[0])
      {
        tok.clear();
        getline(regionFile, line);
        Tokenize(line, tok, " \t\n");

        if ("Name" == tok[0])
        {
          tmpReg->Name(tok[2]);
          gMapRegionNameToPtr[tmpReg->Name()] = tmpReg;
        }
        else if ("Width" == tok[0])
          tmpReg->Width(atoi(tok[2].c_str()));
        else if ("Height" == tok[0])
          tmpReg->Height(atoi(tok[2].c_str()));
        else if ("Property" == tok[0] && "type" == tok[2])
          tmpReg->Type(gMapCellType(tok[5]));
        else if ("Property" == tok[0] && "clock_domain" == tok[2])
          tmpReg->NumClkDom(atoi(tok[5].c_str()));
        else if ("PropAreaInst" == tok[0])
        {
          CRegion *tmpChild = gMapRegionNameToPtr[tok[2]];
          assert(NULL != tmpChild);
          double xOffset = atof(tok[4].c_str());
          double yOffset = atof(tok[6].c_str());
#if 0
          if (tmpReg->mYOffset.size() == 1 && "group" == tok[2])
          {
            cout << "WARN:  Second group being read.  Computing inter tile height\n";
            mInterTileSpaceY = yOffset - tmpReg->mYOffset[0] - gHeightTile;
            cout << "    Inter Tile height = " << mTileHeight << endl;
            exit(1);
          }
#endif
          tmpReg->AddRegion(tmpChild, xOffset, yOffset);
        }
      }
    }
    else if ("PropAreaInst" == tok[0])
    {
      // Must be the top level module now
      mPlatformTop = gMapRegionNameToPtr[tok[2]];
      mIOSpaceX    = atoi(tok[4].c_str());
      mIOSpaceY    = atoi(tok[6].c_str());
      mTopInstName = tok[2];
    }
  }
  regionFile.close();
  assert(NULL != mPlatformTop);
  mPlatformTop->Flatten(mIOSpaceX, mIOSpaceY);

  cout << "STAT:  Exiting readPlatformRegion" << endl;
  return 0;
}

/*********************************************/
/*
    Function reads .nets file of given benchmark
       - populates the data structure
       - Outputs Inputhmetis.hgr for Hmetis partitioning
    - Outputs .nets in Standard format
*/
/*********************************************/
int CDesign::readNetsFile(string netsFileName)
{
  cout << "STAT:  Enter readNetsFile" << endl;

  string line;
  vector <string> tok;

  std::ifstream netFile;
  netFile.open(netsFileName.c_str());

  if (!netFile)
    cout << "EROR: ** Could not open nets file " << netsFileName << endl;

  // Get number of nets and pins
  while (true)
  {
    getline(netFile, line);
    tok.clear();
    Tokenize(line, tok);

    if (0 == tok.size())
      continue;

    if ("NumNets" == tok[0])
    {
      mNumNets = atoi(tok[2].c_str());
      cout << "  - Expecting " << mNumNets << " nets" << endl;
    }
    else if ("NumPins" == tok[0])
    {
      mNumPins = atoi(tok[2].c_str());
      cout << "  - Expecting " << mNumPins << " pins" << endl;
      break;
    }
    else if ("NetDegree" == tok[0])
    {
      cout << "WARN: Parsing beyond NumPins detected.  Check" << endl;
      assert(false);
    }
  }

  tok.clear();

  int numNets = 0;
  int numPins = 0;
  mAllNets.clear();
  mAllNets.reserve(mNumNets);

  while (!netFile.eof())
  {
    getline(netFile, line);
    tok.clear();
    Tokenize(line, tok);

    if (0 == tok.size())
      continue;

    if ("NetDegree" == tok[0])
    {
      int degree = atoi(tok[2].c_str());
      if (0 == degree)
      {
      }
      else
      {
        CNet *tmpNet = new CNet(tok[3]);

        for (int i = 0; i < degree; ++i)
        {
          tok.clear();
          getline(netFile, line);
          Tokenize(line, tok, " \n\t");

          CNode *tmpNode = mMapNameToNode[tok[0]];
          assert(NULL != tmpNode);

          tmpNode->X(atof(tok[3].c_str()));
          tmpNode->Y(atof(tok[4].c_str()));

          tmpNet->AddNode(tmpNode);
          tmpNode->AddNet(tmpNet);
          ++numPins;
        }

        mAllNets.push_back(tmpNet);
        ++numNets;
      }
    }
  }

  netFile.close();

  assert(-1 != mNumNets);
  assert(-1 != mNumPins);
  assert((int) mAllNets.size() == mNumNets);
  assert(numNets == mNumNets);
  assert(mNumPins == numPins);

  cout << "STAT:  Exiting readNetsfile (" << mAllNets.size() <<
  ") nets read" << endl;
  return mAllNets.size();
}

/* This function maps the cell information on platform */
int CDesign::MapInfoToGrid()
{
  /*
  vector<SRegions>::iterator itr_m_sRegions;
  vector<pSAreaInstance>::iterator itr_pInstances;
  vector<pSAreaInstance> pInstances;
  pSAreaInstance curInstance;
  for ( itr_m_sRegions = m_sRegions.begin();itr_m_sRegions != m_sRegions.end();itr_m_sRegions++ )
  {
    pInstances = ( *itr_m_sRegions ).pInstances;
    for ( itr_pInstances = pInstances.begin();itr_pInstances != pInstances.end();itr_pInstances++ )
    {
      curInstance = ( *itr_pInstances );
      cout << itr_m_sRegions->cTopName << ":New instance found:" << curInstance->cInstanceName << ":Name:" << curInstance->iStartX << ":Value:" << curInstance->iStartY << endl;

    }
  }

  vector<SRegions> vSRegions;

  for ( itr_m_sRegions = m_sRegions.begin();itr_m_sRegions != m_sRegions.end();itr_m_sRegions++ )
  {
    SRegions curRegion;
    curRegion.cTopName = strdup( itr_m_sRegions->cTopName );
    vector<pSAreaInstance> pCurInstances;
    pInstances = ( *itr_m_sRegions ).pInstances;
    //   cout << "New instance found:" << itr_m_sRegions->cTopName << endl;
    for ( itr_pInstances = pInstances.begin();itr_pInstances != pInstances.end();itr_pInstances++ )
    {
      curInstance = ( *itr_pInstances );
      if ( !( strcmp( curInstance->cInstanceName, "edff_column" ) == 0 || strcmp( curInstance->cInstanceName, "ecell_column" ) == 0 || strcmp( curInstance->cInstanceName, "reg_file_column" ) == 0 || strcmp( curInstance->cInstanceName, "bram_column" ) == 0 ) )
      {
        vector<SRegions>::iterator itr_vSRegions;
        vector<pSAreaInstance>::iterator itr_pLastInstances;
        vector<pSAreaInstance> pLastInstances;
        for ( itr_vSRegions = vSRegions.begin();itr_vSRegions != vSRegions.end();itr_vSRegions++ )
        {
          pLastInstances = ( *itr_vSRegions ).pInstances;
          if ( strcmp( itr_vSRegions->cTopName, curInstance->cInstanceName ) == 0 )
          {
            //       cout << "Name matched!! Now do processing" << endl;
            for ( itr_pLastInstances = pLastInstances.begin();itr_pLastInstances != pLastInstances.end(); itr_pLastInstances++ )
            {
              pSAreaInstance tempInstance;
              tempInstance = ( pSAreaInstance ) malloc( sizeof( SAreaInstances ) );
              ( *tempInstance ).cInstanceName = strdup(( *itr_pLastInstances )->cInstanceName );
              ( *tempInstance ).iStartX = curInstance->iStartX + ( *itr_pLastInstances )->iStartX;
              ( *tempInstance ).iStartY = curInstance->iStartY + ( *itr_pLastInstances )->iStartY;
              pCurInstances.push_back( tempInstance );
            }
          }
        }
      }
      else
      {
        //     cout << "test test\n";
        pCurInstances.push_back( curInstance );
      }
    }
    curRegion.pInstances = pCurInstances;
    vSRegions.push_back( curRegion );
  }


  for ( itr_m_sRegions = vSRegions.begin();itr_m_sRegions != vSRegions.end();itr_m_sRegions++ )
  {
    pInstances = ( *itr_m_sRegions ).pInstances;
    for ( itr_pInstances = pInstances.begin();itr_pInstances != pInstances.end();itr_pInstances++ )
    {
      curInstance = ( *itr_pInstances );
  //    cout << itr_m_sRegions->cTopName << ":instance found:" << curInstance->cInstanceName << ":Name:" << curInstance->iStartX << ":Value:" << curInstance->iStartY << endl;

    }
  }

  SRegions topRegion = vSRegions[0];
  pInstances = topRegion.pInstances;
  for ( itr_pInstances = pInstances.begin();itr_pInstances != pInstances.end();itr_pInstances++ )
  {
    curInstance = ( *itr_pInstances );
  //     cout << itr_m_sRegions->cTopName << ":anurag instance found:" << curInstance->cInstanceName << ":Name:" << curInstance->iStartX << ":Value:" << curInstance->iStartY << endl;
  }
  */
  return 0;

}

double CDesign::GetHPWL(string msg)
{
  double hpwl = 0.0;
  for (unsigned int i = 0; i < mAllNets.size(); ++i)
  {
    double wl = mAllNets[i]->GetHPWL();
//     cout << "Net " << i << " wirelength = " << wl << endl;
    hpwl += wl;
  }

  assert(hpwl >= 0.0);

  if (msg != "")
    cout << "  - HPWL (" << msg << "):  Total wirelength = " << hpwl << endl;

  return hpwl;
}

int CDesign::ReadPLFile(string aPlFileName, bool readFixed)
{
  cout << "STAT:  Reading PL file " << aPlFileName << endl;

  std::ifstream in;
  in.open(aPlFileName.c_str());
  int numNodesLocated = 0;

  if (!in)
  {
    cout << "  EROR: ** Could not open file " << aPlFileName << endl;
    assert(false);
  }

  string line;
  vector<string> tok;
  while (!in.eof())
  {
    getline(in, line);
    Tokenize(line, tok, " \t\n");

    // The valid lines in the .pl file has the property that the
    // third token is ":".  Check for it and process the line.

    if (tok.size() >= 5 && ":" == tok[3] && "#" != tok[0])
    {
//       cout << "Looking for node whose name is " << tok[0] << endl;
      CNode *tmpNode = mMapNameToNode[tok[0]];
      assert(NULL != tmpNode);
      if (true == tmpNode->Fixed() && false == readFixed)
      {
      }
      else
      {
        tmpNode->PutAt(atof(tok[1].c_str()), atof(tok[2].c_str()));
      }
      ++numNodesLocated;
    }

    tok.clear();
  }

  in.close();

  cout << "STAT:  Read " << numNodesLocated << " node's location" << endl;
  return 0;
}

void CDesign::WriteImage(string fileBaseName)
{
  cout << "STAT:  Entering CDesign::WriteImage() (to file " <<
  fileBaseName << ")" << endl;

  string dataFileName = fileBaseName;
  dataFileName += ".dat";

  std::ofstream out;
  out.open(dataFileName.c_str());

  ostringstream os;
  os << "set style rectangle back fc rgb \"white\" fs solid 1.0 noborder\n";

  int numObjectWritten = 0;
  for (unsigned int i = 0; i < mAllNodes.size(); ++i)
  {
    switch (mAllNodes[i]->Type())
    {
    case 0: /* ECell */
      out << mAllNodes[i]->X() << " " << mAllNodes[i]->Y() << endl;
      break;
    case 1: /* EDff */
      out << mAllNodes[i]->X() << " " << mAllNodes[i]->Y() << endl;
      break;
    case 2: /* Reg File - size 8 x 32 */
      os << "set object " << numObjectWritten + 1 << " rectangle from " <<
      mAllNodes[i]->X() << "," << mAllNodes[i]->Y() << " to " <<
      mAllNodes[i]->X() + 8 << "," << mAllNodes[i]->Y() + 32 <<
      " fc rgb \"green\"\n";
      numObjectWritten++;
      break;
    case 3: /* BRAM - size 16 x 64 */
      os << "set object " << numObjectWritten + 1 << " rectangle from " <<
      mAllNodes[i]->X() << "," << mAllNodes[i]->Y() << " to " <<
      mAllNodes[i]->X() + 16 << "," << mAllNodes[i]->Y() + 64 <<
      " fc rgb \"blue\"\n";
      numObjectWritten++;
      break;
    }
  }

  out.close();

  string gplotName = fileBaseName;
  gplotName += ".plt";

  out.open(gplotName.c_str());

  out << "set terminal gif large enhanced size 1280,1024" << endl;
//   out << "set terminal postscript color" << endl;
  out << "set xrange [" << 0 << ":" << mWidth << "]" << endl;
  out << "set yrange [" << 0 << ":" << mHeight << "]" << endl;
  out << "set output '" << fileBaseName << ".gif'" << endl;
//   out << "set output '" << fileBaseName << ".ps'" << endl;
  out << os.str();
  out << "plot '" << dataFileName << "' with dots" << endl;
  out << "quit" << endl;

#if 0
  // Make the gif
  string cmd = "./rp.csh eda09 gnuplot ";
  cmd += gplotName;
  system(cmd.c_str());
#else
  cout << "INFO:  Run gnuplot " << gplotName << " to generate placement image\n";
#endif
}

// Fastplace style global swapping.  However, we do not worry
// about the exactly physical legality of the solution at this
// stage since those things are taken care of by the Legalize()
// function.
struct compareNodePtr
{
  bool operator()(const CNode* a, const CNode* b) const
  {
    return (a->mDistFromBestBox < b->mDistFromBestBox);
  }
};

void CDesign::GlobalSwap()
{
  cout << "STAT:  Entering CDesign::GlobalSwap()\n";

  // Compute the best bounding box of each node
  cout << "  - Computing best bounding box\n";
  for (int i = mECells.size() - 1; i >= 0; --i)
    mAllNodes[mECells[i]]->ComputeOptimalBBox();
  for (int i = mDffs.size() - 1; i >= 0; --i)
    mAllNodes[mDffs[i]]->ComputeOptimalBBox();

  // Make a duplicate of the vector of nodes.  We will need to run
  // sort() in this function and if we do that without copying, then
  // the markers of mECells, mDffs etc become invalid.
  vector<CNode*> vecNodes(mECells.size() + mDffs.size());
  for (int i=0; i<mECells.size(); i++)
    vecNodes[i] = mAllNodes[mECells[i]];
  for (int i=0; i<mDffs.size(); i++)
    vecNodes[mECells.size()+i] = mAllNodes[mDffs[i]];

  // Sort the nodes w.r.t. the distance from their best BB
  cout << "  - Sorting nodes w.r.t. distance from best BB";
  sort(vecNodes.begin(), vecNodes.end(), cmpNodeBBDistance);
  cout << "  DONE\n";
  assert(vecNodes[0]->DistToBB() >= vecNodes[vecNodes.size()-1]->DistToBB());

  // Get how many nodes are out from their BB
  int numNodesBad = 0;
  for (numNodesBad = vecNodes.size() - 1; numNodesBad >= 0; --numNodesBad)
    if (vecNodes[numNodesBad]->DistToBB() > DBL_ZERO_ABS)
      break;
  cout << "  - " << numNodesBad << " cells out of best BB\n";

  // For all the nodes that are away from their best BB, try
  // to put them in their best BB.
  cout << "  - Fixing nodes to be inside their BB\n";
  vector<CTile*> v;

  int numIntraTileImprove = 0;
  int numInterTileImprove = 0;
  int numSeqSearchImprove = 0;
  for (int i = 0; i < vecNodes.size(); ++i)
  {
    assert(false == vecNodes[i]->Fixed());
    assert(1 >= vecNodes[i]->Type());
    if (NULL == vecNodes[i]->BestBB())
      continue;

// #define _VERBOSE
#ifdef _VERBOSE
    cout << "    " << vecNodes[i]->Name() << " dist " <<
    vecNodes[i]->DistToBB() << endl;
#endif

    assert(2 > vecNodes[i]->Type());
    assert(0 <= vecNodes[i]->Type());

    CNode *currNode = vecNodes[i];
    currNode->ComputeOptimalBBox();

    // Dont do anything for nodes whose distance from best BB is
    // either very small or for which the BB is not correctly
    // defined now (because the defining BB's one cell has moved).
    if (currNode->DistToBB() < 1)
      continue;

    CTile* currTile = GetParentTile(currNode->X(), currNode->Y());
    assert(true == currTile->Contains(currNode));

    // If the node is in a tile which has a part common with the bounding
    // box, find the intersection and place (for now) the node at the
    // middle of the intersection.  Note that we dont need to check
    // "CanTake(...)" for the tile because since the node is already in the
    // tile, CanTake(...) would return true.
    CBBox *b = currNode->BestBB()->GetIntersectionBB(currTile);

    double xOrig = currNode->X();
    double yOrig = currNode->Y();
    double wlIncdntOrig = currNode->GetIncidentWL();
    
    if (NULL != b)
    {
      // Current tile has good region for this node.  Put it there.
      currTile->RemoveFrom(int(currNode->X() - currTile->X()), 
        int(currNode->Y() - currTile->Y()), currNode);

      bool foundAnotherPlace = false;
      if (0 == currNode->Type())
      {
        foundAnotherPlace = currTile->rotateAndPlace(currNode, 
            int(((b->x() + b->X()) / 2.0) - currTile->X())   ,
            int(((b->y() + b->Y()) / 2.0) - currTile->Y())   );
      }
      else 
      {
        foundAnotherPlace = currTile->rotateAndPlaceWithClock(currNode, 
            int(((b->x() + b->X()) / 2.0) - currTile->X())   ,
            int(((b->y() + b->Y()) / 2.0) - currTile->Y()))  ;
      }

      if (foundAnotherPlace == true)
      {
        numIntraTileImprove++;
//         cout << " - Moved " << currNode->Name() << " from " << currNode->X() <<
//           ","  << currNode->Y() << endl;
        if (wlIncdntOrig < currNode->GetIncidentWL())
        {
          numIntraTileImprove--;
//           cout << "    BUT moved it back (wl increased by " <<
//             currNode->GetIncidentWL() - wlIncdntOrig << endl;
          currTile->RemoveFrom(int(currNode->X() - currTile->X()), 
            int(currNode->Y() - currTile->Y()), currNode);
          currTile->PutAt( int(xOrig - currTile->X()), 
            int(yOrig - currTile->Y()), currNode);
        }
//         else
//         {
//           cout << "    WL reduced by " << wlIncdntOrig - currNode->GetIncidentWL() << endl;
//         }
      }
      else
      {
        currTile->PutAt(int(currNode->X() - currTile->X()), 
          int(currNode->Y() - currTile->Y()), currNode);
      }

      delete b;
      b = NULL;
      continue;
    }

    // If the current tile does not intersect with the best bounding
    // box, go to the one which does.
    assert(NULL == b);

    // Iterate over the list of possible tiles and insert
    // it in the first available tile which can accept it.
    v.clear();
    GetTilesInBBox(currNode->BestBB(), v);
//     if (v.size() > 0)
//       cout << "For node " << currNode->Name() << " " << v.size() << " tile options\n";

    bool foundAnotherPlace = false;
    for (int j = 0; j < v.size(); ++j)
    {
      if (true == v[j]->CanTake(currNode))
      {
//         cout << " - Trying " << j << " tile\n";
        b = currNode->BestBB()->GetIntersectionBB(v[j]);
        if (b == NULL)
        {
//           cout << " WTF!  Should not come here\n";
          continue;
        }
        assert(NULL != b);

        if (0 == currNode->Type())
        {
          foundAnotherPlace = v[j]->rotateAndPlace(currNode, 
              int(((b->x() + b->X()) / 2.0) - v[j]->X())   ,
              int(((b->y() + b->Y()) / 2.0) - v[j]->Y())   );
        }
        else
        {
          assert(1 == currNode->Type());
          foundAnotherPlace = v[j]->rotateAndPlaceWithClock(currNode, 
              int(((b->x() + b->X()) / 2.0) - v[j]->X())   ,
              int(((b->y() + b->Y()) / 2.0) - v[j]->Y())   );
        }

        if (true == foundAnotherPlace)
        {
          numInterTileImprove++;
          if (wlIncdntOrig < currNode->GetIncidentWL())
          {
            numInterTileImprove--;
            v[j]->RemoveFrom(int(currNode->X() - v[j]->X()),
              int(currNode->Y() - v[j]->Y()), currNode);
            currNode->PutAt(xOrig, yOrig);
            foundAnotherPlace = false;
          }
          else
          {
            currTile->RemoveFrom(int(xOrig - currTile->X()), 
              int(yOrig - currTile->Y()), currNode);
          }
        }

        if (true == foundAnotherPlace)
        {
//           cout << "   - WL reduced by " << wlIncdntOrig - currNode->GetIncidentWL() << endl;
          v[j]->AddNode(currNode);
          currTile->RemoveNode(currNode);
          break;
        }
      }
    }

    if (true == foundAnotherPlace)
      continue;

    if (currNode->Name() == "o70197" || currNode->Name() == "o70191")
      cout << "Putting Node3: " << currNode->Name() << " with orig:" <<  xOrig << "," << yOrig << endl;
    // If control comes here, none of the tiles in vector v could
    // provide space for currNode.  It can happen if they are crowded
    // or if by chance the optimal area happens to be in-between the
    // tiles (in the no-man's land).  In this case, try to insert as
    // close to the destination as possible.
    double targetXDist = currNode->BestBB()->x() - currNode->X();
    double targetYDist = currNode->BestBB()->y() - currNode->Y();
    int numSteps = abs(static_cast<int>(
                         std::max(targetXDist, targetYDist) / 3.0));
    double xstep = targetXDist / static_cast<double>(numSteps);
    double ystep = targetYDist / static_cast<double>(numSteps);

    for (int i = numSteps - 1; i > 0; --i)
    {
      double possPosnX = currNode->X() + static_cast<double>(i) * xstep;
      double possPosnY = currNode->Y() + static_cast<double>(i) * ystep;
      CTile *tmpTile = GetParentTile(possPosnX, possPosnY);
      
      if (tmpTile == GetParentTile(currNode->X(), currNode->Y()))
        continue;

      if (tmpTile->CanTake(currNode))
      {
//         cout << "Trying to put in tile " << tmpTile << endl;
        /*
         foundAnotherPlace = true;
         currNode->PutAt(possPosnX, possPosnY);
         tmpTile->AddNode(currNode);
         currTile->RemoveNode(currNode);
         break;
        */
        #if 0
         b = currNode->BestBB()->GetIntersectionBB(tmpTile);
         if (b == NULL)
         {
           cout << " WTF!  Should not come here\n";
           continue;
         }
         assert(NULL != b);
         #endif

         if (0 == currNode->Type())
         {
           foundAnotherPlace = tmpTile->rotateAndPlace(currNode, 
               std::min(gDesign->mTileWidth - 1, std::max(0, int(possPosnX - tmpTile->X())))   ,
               std::min(gDesign->mTileHeight - 1, std::max(0, int(possPosnY - tmpTile->Y())))   );
         }
         else
         {
           assert(1 == currNode->Type());
           foundAnotherPlace = tmpTile->rotateAndPlaceWithClock(currNode, 
               std::min(gDesign->mTileWidth - 1, std::max(0, int(possPosnX - tmpTile->X())))   ,
               std::min(gDesign->mTileHeight - 1, std::max(0, int(possPosnY - tmpTile->Y())))   );
         }

         if (true == foundAnotherPlace)
         {
           numInterTileImprove++;
           if (wlIncdntOrig < currNode->GetIncidentWL())
           {
             numInterTileImprove--;
             tmpTile->RemoveFrom(int(currNode->X() - tmpTile->X()),
               int(currNode->Y() - tmpTile->Y()), currNode);
             currNode->PutAt(xOrig, yOrig);
             foundAnotherPlace = false;
           }
           else
           {
             currTile->RemoveFrom(int(xOrig - currTile->X()), 
               int(yOrig - currTile->Y()), currNode);
           }
         }
      }

      if (true == foundAnotherPlace)
      {
//         cout << "   - WL (3rd case) reduced by " << wlIncdntOrig - currNode->GetIncidentWL() << endl;
        tmpTile->AddNode(currNode);
        currTile->RemoveNode(currNode);
        numSeqSearchImprove++;
        break;
      }
    }

    // If we still can't find place for the node, its very unfortunate.
    // We will need to check what to do.  For now, ignore it.

#ifdef _VERBOSE
    cout << "  - Moved to " << currNode->X() << "," << currNode->Y() << endl;
#endif
  }
  cout << "  - Done.  " << numInterTileImprove << " inter-tile, " << 
    numIntraTileImprove << " intra-tile and " << numSeqSearchImprove << 
    " sequential movement\n";
}

// Read the regions file of Platform
int CDesign::readPlatformScl(string aFileName)
{
  cout << "STAT:  Enter readPlatformScl" << endl;

  string line;
  vector <string> tok;

  std::ifstream in;
  in.open(aFileName.c_str());

  if (!in)
  {
    cout << "EROR: ** Could not open nets file " << aFileName << endl;
    assert(false);
  }

  // Get platform information
  int numRowRead = 0;
  int numRows = 0;
  int allRowHeight = -100;
  int allRowSiteWidth = -100;
  int allRowSiteSpace = -100;
  int allRowNumSites  = -100;

  bool isFirstCordinate = true;
  int  minXOfCoreArea = -1;
  int  minYOfCoreArea = -1;
  int  maxYOfCoreArea = -1;

  while (!in.eof())
  {
    getline(in, line);
    Tokenize(line, tok, "\n\t ");

    if (0 == tok.size())
      continue;

    if ("NumRows" == tok[0])
    {
      assert(":" == tok[1]);
      numRows = atoi(tok[2].c_str());
      mAllRows.reserve(numRows);
      mHeight = numRows;
    }
    else if ("CoreRow" == tok[0] && "Horizontal" == tok[1])
    {
      ++numRowRead;
      CRow *tmpRow = new CRow();

      while ("End" != tok[0])
      {
        tok.clear();
        getline(in, line);
        Tokenize(line, tok, "\n\t ");

        if (0 == tok.size())
          continue;

        if ("Coordinate" == tok[0])
        {
          if (isFirstCordinate)
          {
            minYOfCoreArea = atoi(tok[2].c_str());
            isFirstCordinate = false;
          }
          maxYOfCoreArea = atoi(tok[2].c_str());
          tmpRow->Coordinate(atoi(tok[2].c_str()));
        }
        else if ("Height" == tok[0])
        {
          if (-100 != allRowHeight)
          {
            assert(atoi(tok[2].c_str()) == CRow::mHeight);
          }
          else
          {
            allRowHeight = atoi(tok[2].c_str());
            CRow::mHeight = allRowHeight;
          }
        }
        else if ("Sitewidth" == tok[0])
        {
          if (-100 != allRowSiteWidth)
            assert(atoi(tok[2].c_str()) == CRow::mSiteWidth);
          else
          {
            allRowSiteWidth = atoi(tok[2].c_str());
            CRow::mSiteWidth = allRowSiteWidth;
          }
        }
        else if ("Sitespace" == tok[0])
        {
          if (-100 != allRowSiteSpace)
            assert(atoi(tok[2].c_str()) == CRow::mSiteSpace);
          else
          {
            allRowSiteSpace = atoi(tok[2].c_str());
            CRow::mSiteSpace = allRowSiteSpace;
          }
        }
        else if ("SubrowOrigin" == tok[0])
        {
          if (-100 != allRowNumSites)
          {
            assert(atoi(tok[5].c_str()) == CRow::mNumSites);
            minXOfCoreArea = atoi(tok[2].c_str());
          }
          else
          {
            allRowNumSites = atoi(tok[5].c_str());
            CRow::mNumSites = allRowNumSites;
            mWidth = allRowNumSites;
          }
        }
      }

      mAllRows.push_back(tmpRow);
    }

    tok.clear();
  }

  mIOSpaceX          = minXOfCoreArea;
  mIOSpaceY          = minYOfCoreArea;

  mCoreWidth         = allRowNumSites;
  mCoreHeight        = numRows;

  mWidth             = (mIOSpaceX << 1) + mCoreWidth;
  mHeight            = (mIOSpaceY << 1) + mCoreHeight;

  assert(numRowRead == numRows);
  assert((int) mAllRows.size() == numRows);

  cout << "For first element, numSites = " << mAllRows[0]->mNumSites << endl;

  cout << "STAT:  Exiting readPlatformScl" << endl;
  return 0;
}

void CDesign::TetrisLegalize()
{
  cout << "STAT:  Enter TetrisLegalize()" << endl;

  int rowWidth = CRow::mNumSites;

  vector<int> widthInRow;
  widthInRow.reserve(mAllRows.size());

  for (unsigned int i = 0; i < mAllRows.size(); ++i)
    widthInRow[i] = 0;

  for (unsigned int i = 0; i < mAllNodes.size(); ++i)
  {
//     if (mAllNodes[i]->IsFixed())
//       continue;

    double x = mAllNodes[i]->X();
    double y = mAllNodes[i]->Y();

    double minDist = 1000000000.0;
    int bestRow = -1;

    for (unsigned int j = 0; j < mAllRows.size(); ++j)
    {
      // Prevent overflow out of the row
      if (widthInRow[j] == rowWidth)
        continue;

      double dist = (x - widthInRow[j]) * (x - widthInRow[j]) + \
                    (y - j * CRow::mHeight) * (y - j * CRow::mHeight);

      if (dist < minDist)
      {
        minDist = dist;
        bestRow = j;
      }
    }

    assert(-1 != bestRow);

#if 0
    cout << "Cell " << mAllNodes[i]->Name() << " moved from " <<
    mAllNodes[i]->X() << "," << mAllNodes[i]->Y() << " to ";
#endif

    mAllNodes[i]->X(widthInRow[bestRow]);
    mAllNodes[i]->Y(CRow::mHeight*bestRow);
    mAllRows[bestRow]->AddCell(mAllNodes[i]);
    widthInRow[bestRow] += mAllNodes[i]->Width();

    if (1 != mAllNodes[i]->Width())
    {
      assert(1 == mAllNodes[i]->Width());
    }

#if 0
    cout << mAllNodes[i]->X() << "," << mAllNodes[i]->Y() << endl;

#endif
  }

  cout << "STAT:  Exiting TetrisLegalize()" << endl;
}

bool CDesign::checkSanity()
{
#if 0
  cout << "STAT:  Enter checkSanity()" << endl;

  cout << "  - Checking connectivity" << endl;
  int numUnconnected = 0;

  for (unsigned int i = 0; i < mAllNodes.size(); i++)
  {
    if (0 == mAllNodes[i]->GetNumIncidentNets())
    {
      numUnconnected++;
//       cout << "    ** WARN ** Node " << mAllNodes[i]->Name() <<
//       " totally unconnected.  Fixed = " << mAllNodes[i]->Fixed() << endl;
    }
  }

  if (0 != numUnconnected)
    cout << "    Total " << numUnconnected << " unconnected cells found (out of "
    << mAllNodes.size() << ")" << endl;


  cout << "  - Checking net's incident notes" << endl;
  numUnconnected = 0;
  for (unsigned int i = 0; i < mAllNets.size(); i++)
  {
    if (0 == mAllNets[i]->GetDegree())
    {
      numUnconnected++;
//       cout << "    ** WARN ** Net " << mAllNodes[i]->Name() <<
//       " has 0 nodes connected" << endl;
    }
  }

  if (0 != numUnconnected)
    cout << "    Total " << numUnconnected <<
    " unconnected nets found" << endl;


  cout << "  - Checking Feasibility" << endl;

  vector<int> areaCellOfType(mMapCellTypeToInt.size());
  vector<int> areaPlatformOfType(mMapCellTypeToInt.size());
  for (unsigned int i = 0; i < mMapCellTypeToInt.size(); i++)
  {
    areaCellOfType[i] = 0;
    areaPlatformOfType[i] = 0;
  }

  for (unsigned int i = 0; i < mAllNodes.size(); i++)
    areaCellOfType[mAllNodes[i]->Type()] += mAllNodes[i]->Area();

  for (unsigned int i = 0; i < gRegionsVec.size(); i++)
    areaPlatformOfType[gRegionsVec[i]->Type()] += gRegionsVec[i]->Area();

  for (unsigned int i = 0; i < mMapCellTypeToInt.size(); i++)
  {
    cout << "    Type " << i + 1 << ": Space=" << areaPlatformOfType[i] <<
    " Required=" << areaCellOfType[i] << "  (utilization = ";

    if (0 != areaPlatformOfType[i])
      cout << (double)areaCellOfType[i] / (double)areaPlatformOfType[i]
      << ")" <<  endl;
    else
      cout << "infinity)" << endl;
  }

  cout << "STAT:  Exiting checkSanity()" << endl;

#endif
  return true;
}

void CDesign::drawPlatformPic()
{
  return;
  double xmin = DBL_MAX;
  double xmax = DBL_MIN;
  double ymin = DBL_MAX;
  double ymax = DBL_MIN;

  assert(gRegionsVec.size() == gRegionXOff.size());
  assert(gRegionYOff.size() == gRegionXOff.size());
  assert(0 < gRegionXOff.size());
  cout << " Total regions = " << gRegionXOff.size() << endl;

  std::ofstream out;
  out.open("platform.plt");

  out << "set style rectangle back fc rgb \"white\" fs solid 1.0 noborder\n";
  for (unsigned int i = 0; i < gRegionsVec.size(); i++)
  {
    out << "set object " << i + 1 << " rectangle from " <<
    gRegionXOff[i] << "," << gRegionYOff[i] << " to " <<
    gRegionXOff[i] + gRegionsVec[i]->Width() << "," <<
    gRegionYOff[i] + gRegionsVec[i]->Height();

//     cout << gRegionsVec[i]->Name() << endl;
    switch (gRegionsVec[i]->Type())
    {
    case 0:
      out << " fc rgb \"red\"\n";
      break;
    case 1:
      out << " fc rgb \"green\"\n";
      break;
    case 2:
      out << " fc rgb \"blue\"\n";
      break;
    case 3:
      out << " fc rgb \"yellow\"\n";
      break;
    default:
      out << " fc rgb \"pink\"\n";
      break;
    }

    if (gRegionXOff[i] + gRegionsVec[i]->Width() > xmax)
      xmax = gRegionXOff[i] + gRegionsVec[i]->Width();

    if (gRegionXOff[i] < xmin)
      xmin = gRegionXOff[i];

    if (gRegionYOff[i] + gRegionsVec[i]->Height() > ymax)
      ymax = gRegionYOff[i] + gRegionsVec[i]->Height();

    if (gRegionYOff[i] < ymin)
      ymin = gRegionYOff[i];
  }

  out << "set xrange [" << xmin - mIOSpaceX << ":" << xmax + mIOSpaceY << "]" << endl;

  out << "set yrange [" << ymin - mIOSpaceY << ":" << ymax + mIOSpaceY << "]" << endl;
  out << "set terminal postscript color enhanced" << endl;
  out << "set output 'platform.ps'" << endl;
  out << "plot x with dots" << endl;
  out << "quit" << endl;

  out.close();
}

bool CDesign::SolveBRAMAssignment(bool bAddToTileAfterSolving)
{
  cout << "STAT: Enter SolveBRAMAssignment" << endl;

  double oldWl = 0.0;
  for (int i = 0; i < mBrams.size(); ++i)
    oldWl += mAllNodes[mBrams[i]]->GetIncidentWL();

  if (0 == mBrams.size())
  {
    cout << "BRAM:  0 BRAM in design.  Munkres not required\n";
    return false;
  }

  cout << "  #BRAM = " << mBrams.size() << " #Tiles = " << mNumTilesX*mNumTilesY << endl;

  // Create a matrix for storing weights (HPWL)
  Matrix<int> matrix(mBrams.size(), mNumTilesX * mNumTilesY);

  // Populate the weight matrix with HPWL cost
  for (int i = 0; i < mBrams.size(); ++i)
  {
    assert(3 == mAllNodes[mBrams[i]]->Type());
    // Store original position
    double xOld = mAllNodes[mBrams[i]]->X();
    double yOld = mAllNodes[mBrams[i]]->Y();

    for (int x = 0; x < mNumTilesX; ++x)
    {
      for (int y = 0; y < mNumTilesY; ++y)
      {
        mAllNodes[mBrams[i]]->PutAt(mTiles[x][y]->X() + 30, mTiles[x][y]->Y() + 0);
        int cost = (int)mAllNodes[mBrams[i]]->GetIncidentWL();
        matrix(i, x*mNumTilesY + y) = cost;
      }
    }

    mAllNodes[mBrams[i]]->X(xOld);
    mAllNodes[mBrams[i]]->Y(yOld);
  }

  // Make an instance of Munkres and solve it
  Munkres m;
  vector<int> results(matrix.rows());
  m.solve(matrix, results);
  assert(results.size() == mBrams.size());

#if 0
  for (int i = 0; i < mNumTilesX; i++)
    for (int j = 0; j < mNumTilesY; j++)
      mTiles[i][j]->RemoveBRAM();
#endif

  for (int i = 0; i < results.size(); ++i)
  {
    int xInd = results[i] / mNumTilesY;
    int yInd = results[i] - xInd * mNumTilesY;

    assert(xInd < mNumTilesX);
    assert(yInd < mNumTilesY);

    if (bAddToTileAfterSolving)
    {
      CTile *tmpTile =
        GetParentTile(mAllNodes[mBrams[i]]->X(), mAllNodes[mBrams[i]]->Y());
      assert(tmpTile->Contains(mAllNodes[mBrams[i]]));
      tmpTile->RemoveNode(mAllNodes[mBrams[i]]);
      mTiles[xInd][yInd]->AddNode(mAllNodes[mBrams[i]]);
    }

    mAllNodes[mBrams[i]]->PutAt(mTiles[xInd][yInd]->X() + 30, mTiles[xInd][yInd]->Y() + 0);

// #ifdef _VERBOSE
    cout << "BRAM " << i << " in mTile[" << xInd << "][" << yInd << "]\n";
// #endif
  }

  double newWl = 0.0;
  for (int i = 0; i < mBrams.size(); ++i)
    newWl += mAllNodes[mBrams[i]]->GetIncidentWL();

  cout << "     RegFile incident WL = " << newWl << " from " << oldWl << endl;
  cout << "STAT: Exit SolveBRAMAssignment" << endl;

  return (newWl < oldWl*0.999);
}


// Solve the assignment problem for register files.  The only catch is that
// each tile can have TWO REG FILES in it (unless something special in the
// benchmark changes drastically.  In this case, we will DOUBLE THE NUMBER
// OF COLUMNS (as if there are double workers).
bool CDesign::SolveRegFileAssignment(bool bAddToTileAfterSolving)
{
  cout << "STAT: Enter SolveRegFileAssignment" << endl;

  double oldWl = 0.0;
  for (int i = 0; i < mRegs.size(); ++i)
    oldWl += mAllNodes[mRegs[i]]->GetIncidentWL();

  if (0 == mRegs.size())
  {
    cout << "REGS:  0 REGFILES in design.  Munkres not required\n";
    return false;
  }

  // Each tile can have four Reg files.  Therefore, we make as if
  // there are 4*actual tiles number of positions available.
  static const int numOfTiles = mNumTilesX * mNumTilesY * 4;

  cout << "  #REG FILES = " << mRegs.size() << " #Posn = " <<  numOfTiles << endl;

#if 0
  for (int i = 0; i < mNumTilesX; i++)
  {
    for (int j = 0; j < mNumTilesY; j++)
    {
      cout << mTiles[i][j]->mRegs.size() << " ";
    }
    cout << endl;
  }
#endif

  // Create a matrix for storing weights (HPWL)
  Matrix<int> matrix(mRegs.size(), numOfTiles);

  // Populate the weight matrix with HPWL cost
  for (int i = 0; i < mRegs.size(); ++i)
  {
    assert(2 == mAllNodes[mRegs[i]]->Type());

    // Store original position
    double xOld = mAllNodes[mRegs[i]]->X();
    double yOld = mAllNodes[mRegs[i]]->Y();

    int cost = 0;
    for (int tileIdx = 0; tileIdx < numOfTiles; ++tileIdx)
    {
      int tilex = (tileIdx >> 2) / mNumTilesY;
      int tiley = (tileIdx >> 2) - tilex * mNumTilesY;
      switch (tileIdx % 4)
      {
      case 0:
        {
          mAllNodes[mRegs[i]]->PutAt(mTiles[tilex][tiley]->X() + 76,
                                     mTiles[tilex][tiley]->Y() + 0);
          break;
        }
      case 1:
        {
          mAllNodes[mRegs[i]]->PutAt(mTiles[tilex][tiley]->X() + 84,
                                     mTiles[tilex][tiley]->Y() + 0);
          break;
        }
      case 2:
        {
          mAllNodes[mRegs[i]]->PutAt(mTiles[tilex][tiley]->X() + 84,
                                     mTiles[tilex][tiley]->Y() + 32);
          break;
        }
      case 3:
        {
          mAllNodes[mRegs[i]]->PutAt(mTiles[tilex][tiley]->X() + 76,
                                     mTiles[tilex][tiley]->Y() + 32);
          break;
        }
      default:
        abort();
      }
      assert(-1 != tilex);
      assert(-1 != tiley);
      cost = (int)mAllNodes[mRegs[i]]->GetIncidentWL();

      matrix(i, tileIdx) = cost;
    }

    mAllNodes[mRegs[i]]->PutAt(xOld, yOld);
  }

#ifdef _VERBOSE
  dumpMatrix(matrix);
#endif

  // Make an instance of Munkres and solve it
  Munkres m;
  vector<int> results(matrix.rows());
  cout << " starting to solve\n";
  m.solve(matrix, results);
  cout << " done solving\n";
  assert(results.size() == mRegs.size());

#if 0
  for (int i = 0; i < mNumTilesX; i++)
    for (int j = 0; j < mNumTilesY; j++)
      mTiles[i][j]->RemoveRegFiles();
#endif

  for (int i = 0; i < results.size(); ++i)
  {
    int xInd = static_cast<int>((results[i] >> 2) / mNumTilesY);
    int yInd = (results[i] >> 2) - xInd * mNumTilesY;

    CTile *origTile = GetParentTile(mAllNodes[mRegs[i]]->X(),
                                    mAllNodes[mRegs[i]]->Y());
//     assert(origTile->Contains(mAllNodes[mRegs[i]]));

    assert(xInd < mNumTilesX);
    assert(yInd < mNumTilesY);

    switch (results[i] % 4)
    {
    case 0:
      {
        mAllNodes[mRegs[i]]->PutAt(mTiles[xInd][yInd]->X() + 76,
                                   mTiles[xInd][yInd]->Y() + 0);
        break;
      }
    case 1:
      {
        mAllNodes[mRegs[i]]->PutAt(mTiles[xInd][yInd]->X() + 84,
                                   mTiles[xInd][yInd]->Y() + 0);
        break;
      }
    case 2:
      {
        mAllNodes[mRegs[i]]->PutAt(mTiles[xInd][yInd]->X() + 84,
                                   mTiles[xInd][yInd]->Y() + 32);
        break;
      }
    case 3:
      {
        mAllNodes[mRegs[i]]->PutAt(mTiles[xInd][yInd]->X() + 76,
                                   mTiles[xInd][yInd]->Y() + 32);
        break;
      }
    default:
      abort();
    }

    if (true == bAddToTileAfterSolving)
    {
      CTile *tmpTile =
        GetParentTile(mAllNodes[mRegs[i]]->X(), mAllNodes[mRegs[i]]->Y());
      assert(origTile->Contains(mAllNodes[mRegs[i]]));
      origTile->RemoveNode(mAllNodes[mRegs[i]]);
      tmpTile->AddNode(mAllNodes[mRegs[i]]);
    }

// #ifdef _VERBOSE
    cout << "REG FILE " << i << " in mTile[" << xInd << "][" << yInd << "]\n";
// #endif
  }

  double newWl = 0.0;
  for (int i = 0; i < mRegs.size(); ++i)
    newWl += mAllNodes[mRegs[i]]->GetIncidentWL();

  cout << "        RegFile incident WL = " << newWl << " from " << oldWl << endl;

  cout << "STAT: Exit SolveRegFileAssignment" << endl;

#if 0
  int numRegs = 0;
  for (int i = 0; i < mNumTilesX; i++)
  {
    for (int j = 0; j < mNumTilesY; j++)
    {
      cout << mTiles[i][j]->mRegs.size() << " ";
      numRegs += mTiles[i][j]->mRegs.size();
    }
    cout << endl;
  }

  assert(numRegs == mRegs.size());
#endif

  return (newWl < oldWl);
}

void CDesign::ComputeTileStuff()
{
  cout << "STAT:  Entering ComputeTileStuff()" << endl;
  // We know this already (and its fixed)
  mTileWidth     = gWidthTile;
  mTileHeight    = gHeightTile;

  // Get the number of tiles in x direction
  // Sort the x cordinate and see how many times its repeated
  vector<double> nothing;
  nothing.clear();
  for (int i = 0; i < gTilesVec.size(); ++i)
    nothing.push_back(gTilesVec[i]->X());
  sort(nothing.begin(), nothing.end());
  double firstXOff = nothing[0];
  int numRows = 1;
  while (nothing[numRows] == firstXOff) ++numRows;
  cout << "Number of rows = " << numRows << endl;

  // Inter-tile space in x = maxX - minX of tile / (# tileX - 1)
  // The number of tiles in X direction is not known yet.  Will
  // be known in the next block.
  mInterTileSpaceX = (nothing[nothing.size()-1] - nothing[0]);

  nothing.clear();
  for (int i = 0; i < gTilesVec.size(); ++i)
    nothing.push_back(gTilesVec[i]->Y());
  sort(nothing.begin(), nothing.end());
  double firstYOff = nothing[0];
  int numCols = 1;
  while (nothing[numCols] == firstYOff) ++numCols;
  cout << "Number of cols = " << numCols << endl;

  // Inter-tile space in y = maxY - minY of tile / (# tileY - 1)
  // Ditto comments for mInterTileSpaceX hold for this.
  mInterTileSpaceY = (nothing[nothing.size()-1] - nothing[0]);

  mNumTilesX     = numCols;
  mNumTilesY     = numRows;

//   cout << "Num tiles x/y = " << mNumTilesX << " " << mNumTilesY << endl;

  // Now, we know # of tiles in X and Y direction, compute spacing
  mInterTileSpaceX -= (mNumTilesX - 1) * mTileWidth;
  mInterTileSpaceY -= (mNumTilesY - 1) * mTileHeight;

  mInterTileSpaceX = mInterTileSpaceX / (numCols - 1);
  mInterTileSpaceY = mInterTileSpaceY / (numRows - 1);

  cout << "intertile X=" << mInterTileSpaceX << "  intertile Y=" <<
  mInterTileSpaceY << endl;

  mInterTileSpaceY = (double)(static_cast<int>(mInterTileSpaceY * 100) / 100.0);

  nothing.clear();

  assert(gTilesVec.size() == numCols * numRows);

  // Allocate right amount of space for tiles
  vector< vector<CTile*> > tmpTiles(mNumTilesX, vector<CTile*>(mNumTilesY));

  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
      tmpTiles[i][j] = NULL;

  // Sort the tiles according to operator< in CTile.cpp
  sort(gTilesVec.begin(), gTilesVec.end(), cmpTiles);

  for (int i = 0; i < gTilesVec.size(); ++i)
  {
//     cout << gTilesVec[i]->X() << "," << gTilesVec[i]->Y();

    int xInd = static_cast<int>(i / mNumTilesY);
    int yInd = i - (xInd * mNumTilesY);

//     cout << " " << i << " : x/y Index = " << xInd << " " << yInd << endl;
    tmpTiles[xInd][yInd] = gTilesVec[i];
  }

  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
      assert(NULL != tmpTiles[i][j]);

  mTiles = tmpTiles;

  cout << "         Exiting ComputeTileStuff()" << endl;
}

void CDesign::Prepare()
{
  cout << "STAT:  Preparing the tile structure\n";
  // Get the dimension of each region
  mPlatformTop->ComputeDimension();

  // Get information about the tiles in the design
  gTilesVec.clear();
  mPlatformTop->GetTiles(mIOSpaceX, mIOSpaceY);
  assert(0 < gTilesVec.size());
  ComputeTileStuff();

  drawPlatformPic();

#ifdef _DEBUG
  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
      cout << i << "," << j << "  " << mTiles[i][j] << endl;
#endif
}

#define bramWidth 16
#define regfileWidth 8
#define ecelldffWidth 30
double CDesign::scaleInX(double in) const
{
  double relativeCord = in - mIOSpaceX;
  static const double tilePitch  = mTileWidth + mInterTileSpaceX;
  int tileNumber = static_cast<int>(relativeCord / tilePitch);
  double offsetInTile = relativeCord - tileNumber * tilePitch;

  in -= tileNumber * (mInterTileSpaceX + bramWidth + (regfileWidth << 1));
  double multiplier = offsetInTile / 92.0;
  double oneminusmult = 1.0 - multiplier;
  in -= offsetInTile * oneminusmult;

  assert(in > 0.0);
  assert(in < mWidth);
  return in;
}

double CDesign::scaleInY(double in) const
{
  double relativeCord = in - mIOSpaceY;
  static const double tilePitch  = mTileHeight + mInterTileSpaceY;
  int tileNumber = static_cast<int>(relativeCord / tilePitch);
  double offsetInTile = relativeCord - tileNumber * tilePitch;

  in -= tileNumber * mInterTileSpaceY;
  double multiplier = offsetInTile / mTileHeight;
  double oneminusmult = 1.0 - multiplier;
  in -= offsetInTile * oneminusmult;

  assert(in > 0.0);
  assert(in < mHeight);

  return in;
}

int CDesign::unscaleInX(double in)
{
#if 0
  static double space = double(mWidth - mReducedWidth) / (double)mNumTilesX;
  static double oneUnit = (double)mTileWidth + space;
  static double onebyunit = 1.00 / oneUnit;

  int numUnit = static_cast<int>(in * onebyunit);
  return static_cast<int>(in + numUnit*space);
#endif
}

int CDesign::unscaleInY(double in)
{
#if 0
  static double space = double(mHeight - mReducedHeight) / (double)mNumTilesY;
  static int temp = static_cast<int>(space * 100);
  space = (double)temp / 100.0;
  static double oneUnit = (double)mTileHeight + space;
  static double onebyunit = 1.00 / oneUnit;

  abort();
#endif
}

void CDesign::ShrinkSize()
{
  abort();
  cout << "STAT:  Entering CDesign::ShrinkSize()";

  // Change the tile's cordinates
  for (int i = mNumTilesX - 1; i >= 0; --i)
  {
    for (int j = mNumTilesY - 1; j >= 0; --j)
    {
#ifdef _VERBOSE
      cout << "Tile[" << i << "][" << j << "] from " <<
      mTiles[i][j]->X() << "," << mTiles[i][j]->Y() << " to ";
#endif
      mTiles[i][j]->X(i*mTileWidth + mIOSpaceX);
      mTiles[i][j]->Y(j*mTileHeight + mIOSpaceY);
#ifdef _VERBOSE
      cout << mTiles[i][j]->X() << "," << mTiles[i][j]->Y() << endl;
#endif
    }
  }

  // Change node's cordinates
  for (int i = mAllNodes.size() - 1; i >= 0; --i)
  {
//     if (true == mAllNodes[i]->Fixed())
//       continue;

#ifdef _VERBOSE
    cout << "Node " << mAllNodes[i]->Name() << " from " <<
    mAllNodes[i]->X() << "," << mAllNodes[i]->Y() << " to ";
#endif

    double x = mAllNodes[i]->X() - mIOSpaceX + mInterTileSpaceX / 2.0;
    double y = mAllNodes[i]->Y() - mIOSpaceY + mInterTileSpaceY / 2.0;

    int xInd = static_cast<int>(x / (mTileWidth + mInterTileSpaceX));
    int yInd = static_cast<int>(y / (mTileHeight + mInterTileSpaceY));

    double xMid = (xInd + 0.5) * (mTileWidth + mInterTileSpaceX);
    double yMid = (yInd + 0.5) * (mTileHeight + mInterTileSpaceY);

    double xDisFromMid = x - xMid;
    double yDisFromMid = y - yMid;

    xDisFromMid *= (mTileWidth / (mTileWidth + mInterTileSpaceX));
    yDisFromMid *= (mTileHeight / (mTileHeight + mInterTileSpaceY));

    mAllNodes[i]->X(xMid + xDisFromMid - (xInd * mInterTileSpaceX));
    mAllNodes[i]->Y(yMid + yDisFromMid - (yInd * mInterTileSpaceY));

#ifdef _VERBOSE
    cout << mAllNodes[i]->X() << "," << mAllNodes[i]->Y() << endl;
#endif

  }

  // Modify the width and height of design
  mWidth  = 2 * mIOSpaceX + mNumTilesX * mTileWidth;
  mHeight = 2 * mIOSpaceY + mNumTilesY * mTileHeight;

  mIsShrunk = true;

  cout << "         Exiting CDesign::ShrinkSize()" << endl;
}

void CDesign::UnshrinkSize()
{
  cout << "STAT:  Entering CDesign::UnshrinkSize()" << endl;

  cout << 100 << "=" << unscaleInX(scaleInX(100)) << endl;

  for (int i = 0; i < mAllNodes.size(); ++i)
  {
    mAllNodes[i]->X(unscaleInX(mAllNodes[i]->X()));
    mAllNodes[i]->Y(unscaleInY(mAllNodes[i]->Y()));
  }

  cout << "         Exiting  CDesign::UnshrinkSize()" << endl;
}

void CDesign::RunGlobalPlace()
{
  cout << "STAT:  Entering CDesign::RunGlobalPlace()" << endl;

  cout << "  NOT YET IMPLEMENTED (Anurag will put here)" << endl;
  cout << "         Exiting  CDesign::RunGlobalPlace()" << endl;
}

void CDesign::Legalize()
{
  cout << "STAT:  Entering CDesign::Legalize()" << endl;

  // Find how many tiles are broken right now
  int numTilesBroken = 0;
  int numInTiles = 0;

  for (int i = 0; i < mNumTilesX; ++i)
  {
    for (int j = 0; j < mNumTilesY; ++j)
    {
      if (false == mTiles[i][j]->IsDensityOK())
        ++numTilesBroken;
      numInTiles += mTiles[i][j]->NumCellsInside();
    }
  }

  cout << " Total " << numTilesBroken << " out of " <<
  mNumTilesX * mNumTilesY << " tiles not ok" << endl;

  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
    {
//       mTiles[i][j]->WriteImage(string("before_")+makeString(i)+"_"+makeString(j));  
//       cout << "legalizing tile: " << i << "," << j << endl;
      mTiles[i][j]->Legalize();
//       mTiles[i][j]->WriteImage(string("after_")+makeString(i)+"_"+makeString(j));  
    }

  cout << "         Exiting  CDesign::Legalize()" << endl;
}


// Check if the design is legal.  Each type of cell is put in a vector,
// sorted according to y cordinate and in this sorted list, it is made
// sure that no x,y pair is repeated.  For cells with unit width
// and unit height, this is sufficient.
bool CDesign::checkIfEverythingOK()
{
  cout << "STAT:  Entering CDesign::checkIfEverythingOK()" << endl;

  vector<CNode*> tmp;

  for (int i = 0; i < mECells.size(); ++i)
    tmp.push_back(mAllNodes[mECells[i]]);
  sort(tmp.begin(), tmp.end(), cmpNodes);
  for (int i = 0; i < tmp.size() - 1; ++i)
  {
    if (tmp[i]->Y() == tmp[i+1]->Y())
      if (tmp[i]->X() == tmp[i+1]->X())
        cout << "CHECK OVERLAP ERROR (ECELL): " << tmp[i]->Name() << "@" << tmp[i]->X() << "," << 
          tmp[i]->Y() << "  overlaps with " << tmp[i+1]->Name() << "@" << tmp[i+1]->X() 
          << "," << tmp[i+1]->Y() << endl; 
  }

  tmp.clear();

  if (mDffs.size() == 0)
    return true;
  for (int i = 0; i < mDffs.size(); ++i)
    tmp.push_back(mAllNodes[mDffs[i]]);
  sort(tmp.begin(), tmp.end(), cmpNodes);
  for (int i = 0; i < tmp.size() - 1; ++i)
  {
    if (tmp[i]->Y() == tmp[i+1]->Y())
    {
      if (tmp[i]->X() == tmp[i+1]->X())
        cout << "CHECK OVERLAP ERROR (DFF): " << tmp[i]->Name() << "@" << tmp[i]->X() << "," << 
          tmp[i]->Y() << "  overlaps with " << tmp[i+1]->Name() << "@" << tmp[i+1]->X() 
          << "," << tmp[i+1]->Y() << endl; 
    }
  }

  cout << "         Entering CDesign::checkIfEverythingOK()" << endl;
  return true;
}

void CDesign::WritePLFile(string aName)
{
  cout << "STAT:  Entering CDesign::WritePLFile(" << aName << ")" << endl;
  std::ofstream out;
  out.open(aName.c_str());

  if (!out)
  {
    cout << "EROR:  Could not open file " << aName << " to dump placement\n";
    abort();
  }

  out << "UCLA pl 1.0\n";
  out << "# Created \n";
  out << "# Platform \n";
  out << "\n";

  assert(mECells.size() + mBrams.size() + mDffs.size() + mRegs.size() + mFixeds.size() == mAllNodes.size());

  for (int i = mECells.size() - 1; i >= 0; --i)
    out << mAllNodes[mECells[i]]->Name() << " " << mAllNodes[mECells[i]]->X() << " " << mAllNodes[mECells[i]]->Y() << " : N" << endl;
  for (int i = mDffs.size() - 1; i >= 0; --i)
    out << mAllNodes[mDffs[i]]->Name() << " " << mAllNodes[mDffs[i]]->X() << " " << mAllNodes[mDffs[i]]->Y() << " : N" << endl;
  for (int i = mBrams.size() - 1; i >= 0; --i)
    out << mAllNodes[mBrams[i]]->Name() << " " << mAllNodes[mBrams[i]]->X() << " " << mAllNodes[mBrams[i]]->Y() << " : N" << endl;
  for (int i = mRegs.size() - 1; i >= 0; --i)
    out << mAllNodes[mRegs[i]]->Name() << " " << mAllNodes[mRegs[i]]->X() << " " << mAllNodes[mRegs[i]]->Y() << " : N" << endl;
  for (int i = mFixeds.size() - 1; i >= 0; --i)
    out << mAllNodes[mFixeds[i]]->Name() << " " << mAllNodes[mFixeds[i]]->X() << " " << mAllNodes[mFixeds[i]]->Y() << " : N /FIXED" << endl;

  out.close();
}

void CDesign::WriteNetsFileWithReducedDegree(string aName, int maxDegree)
{
  cout << "STAT:  Entering CDesign::WriteNetsFile(" << aName << ")" << endl;
  std::ofstream out;
  out.open(aName.c_str());

  if (!out)
  {
    cout << "EROR:  Could not open file " << aName << " to dump Netsfile\n";
    abort();
  }

  int numNets = 0;
  int numPins   = 0;
  for (unsigned int i=0; i<mAllNets.size(); i++)
  {
    vector <CNode*> nodes = mAllNets[i]->GetCellsInNet();
    if (nodes.size() > maxDegree)
      continue;

      numNets++;
      numPins += nodes.size();

  }

  out << "UCLA nets 1.0\n";
  out << "# Created : Anurag Kumar\n";
  out << "# Platform :\n";
  out << "\n\n";
  out << "NumNets : " << numNets << endl;
  out << "NumPins : " << numPins << endl;
  out << endl;
  int netNumber = 0;
  for (unsigned int i = 0; i < mAllNets.size(); i++)
  {
    vector <CNode*> nodes = mAllNets[i]->GetCellsInNet();
    if (nodes.size() > maxDegree)
      continue;

    out << "NetDegree : " << nodes.size() << " o" << netNumber << endl;
    netNumber++;

    for (unsigned int j = 0; j < nodes.size(); j++)
    {
      out << "    " << nodes[j]->Name() << " O : " << nodes[j]->X() << "  " << nodes[j]->Y() << endl;
    }
  }

  out.close();
}

void CDesign::WriteNetsFile(string aName)
{
  cout << "STAT:  Entering CDesign::WriteNetsFile(" << aName << ")" << endl;
  std::ofstream out;
  out.open(aName.c_str());

  if (!out)
  {
    cout << "EROR:  Could not open file " << aName << " to dump Netsfile\n";
    abort();
  }

  out << "UCLA nets 1.0\n";
  out << "# Created : Anurag Kumar\n";
  out << "# Platform :\n";
  out << "\n\n";
  out << "NumNets : " << mAllNodes.size() << endl;
  out << "NumPins : " << mNumPins << endl;
  out << endl;
  for (unsigned int i = 0; i < mAllNets.size(); i++)
  {
    vector <CNode*> nodes = mAllNets[i]->GetCellsInNet();
    out << "NetDegree : " << nodes.size() << " " << mAllNets[i]->Name() << endl;

    for (unsigned int j = 0; j < nodes.size(); j++)
    {
      out << "    " << nodes[j]->Name() << " O : " << nodes[j]->X() << "  " << nodes[j]->Y() << endl;
    }
  }

  out.close();
}

void dumpMatrix(Matrix<int>& m)
{
  cout << "STAT:  Enter dumpMatrix()" << endl;
  cout << "Matrix row = " << m.rows() << " columns = " << m.columns() << endl;

  for (int i = 0; i < m.rows(); ++i)
  {
    cout << "row " << i << " : ";
    for (int j = 0; j < m.columns(); ++j)
      cout << m(i, j) << " ";
    cout << endl;
  }

  cout << "         Exiting dumpMatrix()" << endl;
}

static const int NEGLOW = -10000000;
void CDesign::writeTransportation()
{
  int tileId = -1;

  bool bNoTPRequired = true;
  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
      if (false == mTiles[i][j]->IsDensityOK())
        bNoTPRequired = false;

  if (true == bNoTPRequired)
  {
    cout << "     Transportation problem not needed\n";
    return;
  }

  // currCellType shows the type of the cell for which we are solving
  // the transportation problem for.  Type 0 = ecell.  Type 1 = edff.
  // Type 2 = reg_file.  Type 3 = bram.
  int currCellType = 0;

#if 0
  mNumTilesX = 3;
  mNumTilesY = 3;
#else
  for (int i = 0; i < mNumTilesX; ++i)
  {
    for (int j = 0; j < mNumTilesY; ++j)
    {
      ++tileId;
#if 0
      cout << "tileId: " << tileId << " @ [" << mTiles[i][j]->X() <<
      "][" << mTiles[i][j]->Y() << "] : n " << tileId << " " <<
      mTiles[i][j]->ExtraOfType(currCellType) << endl;
#else
  cout << "n " << tileId << " " <<
  mTiles[i][j]->ExtraOfType(currCellType) << endl;
#endif
    }
  }
#endif

  tileId = -1;
  for (int i = 0; i < mNumTilesX; ++i)
  {
    for (int j = 0; j < mNumTilesY; ++j)
    {
      ++tileId;
//       if (false == mTiles[i][j]->IsDensityOK(currCellType))
//       {
      writeArcForTileAt(i, j, tileId);
      cout << endl;
//       }
    }
  }
}

#define BIGINT 1000000
void CDesign::writeArcForTileAt(int x, int y, int tileId)
{
  // 2 steps on left and 2 steps on right
  if (0 <= x - 1)
    cout << "a " << tileId << " " << tileId - mNumTilesY << " " <<
    0 << BIGINT << " " << 1 << endl;

  if (mNumTilesX > x + 1)
    cout << "a " << tileId << " " << tileId + mNumTilesY << " " <<
    0 << BIGINT << " " << 1 << endl;

  if (0 <= x - 2)
    cout << "a " << tileId << " " << tileId - 2*mNumTilesY << " " <<
    0 << BIGINT << " " << 2 << endl;

  if (mNumTilesX > x + 2)
    cout << "a " << tileId << " " << tileId + 2*mNumTilesY << " " <<
    0 << BIGINT << " " << 2 << endl;


  // 2 steps on down and 2 steps on up
  if (0 <= y - 1)
    cout << "a " << tileId << " " << tileId - 1 << " " <<
    0 << BIGINT << " " << 1 << endl;

  if (mNumTilesY > y + 1)
    cout << "a " << tileId << " " << tileId + 1 << " " <<
    0 << BIGINT << " " << 1 << endl;

  if (0 <= y - 2)
    cout << "a " << tileId << " " << tileId - 2 << " " <<
    0 << BIGINT << " " << 2 << endl;

  if (mNumTilesY > y + 2)
    cout << "a " << tileId << " " << tileId + 2 << " " <<
    0 << BIGINT << " " << 2 << endl;


  // One step right one up.  Etc.  Four choices
  if (0 <= y - 1 && 0 <= x - 1)
    cout << "a " << tileId << " " << tileId - mNumTilesY - 1 << " " <<
    0 << BIGINT << " " << 2 << endl;

  if (mNumTilesY > y + 1 && mNumTilesX > x + 1)
    cout << "a " << tileId << " " << tileId + mNumTilesY + 1 << " " <<
    0 << BIGINT << " " << 2 << endl;

  if (0 <= y - 1 && mNumTilesX > x + 1)
    cout << "a " << tileId << " " << tileId + mNumTilesX - 1 << " " <<
    0 << BIGINT << " " << 2 << endl;

  if (mNumTilesY > y + 1 && 0 <= x - 1)
    cout << "a " << tileId << " " << tileId - mNumTilesY + 1 << " " <<
    0 << BIGINT << " " << 2 << endl;
}

void CDesign::OptimizeLegal()
{
#if 0
  // Put in legal places inside the column vectors in the each tile
  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
      mTiles[i][j]->AssignCellsToColumns();
  GetHPWL("After Assign To Columns");
  WritePLFile("AfterColumnAssign.pl");
  CheckAllCellsInCore();
#endif

  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
      mTiles[i][j]->LocalReordering();

  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
      mTiles[i][j]->LocalReordering();

  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
      mTiles[i][j]->LocalReordering();

  GetHPWL("After LocalReorder");
  WritePLFile("AfterLocalReorder1.pl");
  CheckAllCellsInCore();
  checkIfEverythingOK();


#if 0
  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
      mTiles[i][j]->LinearPlacement();

  GetHPWL("After linear place");
  WritePLFile("AfterLinearPlace1.pl");
  CheckAllCellsInCore();

  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
      mTiles[i][j]->LocalReordering();

  GetHPWL("After LocalReorder");
  WritePLFile("AfterLocalReorder1.pl");
  CheckAllCellsInCore();

  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
      mTiles[i][j]->LocalReordering();

  GetHPWL("After LocalReorder");
  WritePLFile("AfterLocalReorder2.pl");
  CheckAllCellsInCore();
#endif

  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
      mTiles[i][j]->Anneal();

  GetHPWL("After Annealing");
  WritePLFile("AfterAnneal1.pl");
  CheckAllCellsInCore();

  checkIfEverythingOK();

  return;
#if 0
  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
      mTiles[i][j]->LocalReordering();

  GetHPWL("After LocalReorder");
  WritePLFile("AfterLocalReorder2.pl");
  CheckAllCellsInCore();

  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
      mTiles[i][j]->Anneal(0.02);

  GetHPWL("After Annealing");
  WritePLFile("AfterAnneal2.pl");
  CheckAllCellsInCore();

  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
      mTiles[i][j]->LinearPlacement();

  GetHPWL("After linear place");
  WritePLFile("AfterLinearPlace2.pl");
  CheckAllCellsInCore();
#endif

  assert(true == checkIfEverythingOK());
}

void CDesign::GetTilesInBBox(CBBox* bb, vector<CTile*>& v)
{
  // Get the extremeties of the bounding box
  double x = bb->x() - mIOSpaceX;
  double X = bb->X() - mIOSpaceX;
  double y = bb->y() - mIOSpaceY;
  double Y = bb->Y() - mIOSpaceY;

#if 0
  assert(0.1 > mCoreWidth  - mNumTilesX * (mTileWidth + mInterTileSpaceX));
  assert(0.1 > mCoreHeight - mNumTilesY * (mTileHeight + mInterTileSpaceY));
#endif

  // If part of the bounding box is outside the core placeable
  // area, we discard those parts by clipping them.
  x = std::max(0.0, std::min(x, mNumTilesX * (mTileWidth + mInterTileSpaceX) - 1));
  X = std::max(0.0, std::min(X, mNumTilesX * (mTileWidth + mInterTileSpaceX) - 1));
  y = std::max(0.0, std::min(y, mNumTilesY * (mTileWidth + mInterTileSpaceY) - 1));
  Y = std::max(0.0, std::min(Y, mNumTilesY * (mTileWidth + mInterTileSpaceX) - 1));

#if 0
  x = (x >= 0) ? x : 0;
  x = (x > mNumTilesX * mTileWidth - 1)  ? mNumTilesX * mTileWidth - 1 : x;
  X = (X >= 0) ? X : 0;
  X = (X > mNumTilesX * mTileWidth - 1)  ? mNumTilesX * mTileWidth  - 1 : X;
  y = (y >= 0) ? y : 0;
  y = (y > mNumTilesY * mTileHeight - 1)  ? mNumTilesY * mTileHeight - 1 : y;
  Y = (Y >= 0) ? Y : 0;
  Y = (Y > mNumTilesY * mTileHeight - 1)  ? mNumTilesY * mTileHeight - 1 : Y;
#endif

  assert(x >= 0.0);
  assert(x <= X);
  assert(y >= 0.0);
  assert(y <= Y);

  int tileIdxLeft;
  int tileIdxRight;
  int tileIdxBottom;
  int tileIdxTop;

  mIsShrunk = false;
  if (mIsShrunk)
  {
    tileIdxLeft  = static_cast<int>(x / mTileWidth);
    tileIdxRight = static_cast<int>(X / mTileWidth);
    tileIdxBottom  = static_cast<int>(y / mTileHeight);
    tileIdxTop     = static_cast<int>(Y / mTileHeight);
  }
  else
  {
    tileIdxLeft  = static_cast<int>(x / (mTileWidth + mInterTileSpaceX));
    tileIdxRight = static_cast<int>(X / (mTileWidth + mInterTileSpaceX));
    tileIdxBottom  = static_cast<int>(y / (mTileHeight + mInterTileSpaceY));
    tileIdxTop     = static_cast<int>(Y / (mTileHeight + mInterTileSpaceY));
  }

  assert(tileIdxLeft >= 0);
  assert(tileIdxRight < mNumTilesX);
  assert(tileIdxBottom >= 0);
  assert(tileIdxTop < mNumTilesY);

#ifdef _VERBOSE
  cout << *bb << " lies in tiles " << tileIdxLeft << "," << tileIdxBottom <<
  " and " << tileIdxRight << "," << tileIdxTop << endl;
#endif

  v.clear();

  for (int i = tileIdxLeft; i <= tileIdxRight; ++i)
    for (int j = tileIdxBottom; j <= tileIdxTop; ++j)
      v.push_back(mTiles[i][j]);

  // In buffer region above top idx
  if (y - tileIdxTop*(mTileHeight + mInterTileSpaceY) > mTileHeight)
    for (int i = tileIdxLeft; i <= tileIdxRight; i++)
      v.push_back(mTiles[i][tileIdxTop+1]);

  // In buffer region on right of right idx
  if (x - tileIdxRight*(mTileWidth + mInterTileSpaceX) > mTileWidth)
    for (int i = tileIdxBottom; i <= tileIdxTop; i++)
      v.push_back(mTiles[tileIdxRight+1][i]);

  assert(0 < v.size());
}

// Get pointer to the tile which has the point x,y
CTile* CDesign::GetParentTile(const double x, const double y) const
{
  double currx = x - mIOSpaceX;
  double curry = y - mIOSpaceY;

  int tileXIdx = -1;
  int tileYIdx = -1;
  if (mIsShrunk)
  {
    tileXIdx = static_cast<int>(currx / mTileWidth);
    tileYIdx = static_cast<int>(curry / mTileHeight);
  }
  else
  {
    tileXIdx = static_cast<int>(currx / (mTileWidth + mInterTileSpaceX));
    tileYIdx = static_cast<int>(curry / (mTileHeight + mInterTileSpaceY));
  }
  assert(-1 != tileXIdx);
  assert(-1 != tileYIdx);
  assert(tileXIdx < mNumTilesX);
  assert(tileYIdx < mNumTilesY);
  
  return mTiles[tileXIdx][tileYIdx];
}

void CDesign::FixBRAMs()
{
  abort();
  for (int i = mBrams.size() - 1; i >= 0; --i)
    mAllNodes[mBrams[i]]->Fixed(true);
}

void CDesign::FixRegFiles()
{
  abort();
  for (int i = mRegs.size() - 1; i >= 0; --i)
    mAllNodes[mRegs[i]]->Fixed(true);
}

void CDesign::DumpShrunkPlAndSclAndNodes(string aPlName, string aSclName, string aNodesName)
{
  map <CNode*, double> mapXOfFixed;
  map <CNode*, double> mapYOfFixed;

  for (int i = mFixeds.size() - 1; i >= 0; --i)
  {
    CNode *tmp = mAllNodes[mFixeds[i]];
    double xpos = tmp->X();
    double ypos = tmp->Y();
    mapXOfFixed[tmp] = xpos;
    mapYOfFixed[tmp] = ypos;
  }

  for (int i = mFixeds.size() - 1; i >= 0; --i)
  {
    CNode *tmp = mAllNodes[mFixeds[i]];
    double xpos = tmp->X();
    double ypos = tmp->Y();

    xpos -= mIOSpaceX;
    ypos -= mIOSpaceY;

    int numXtile = static_cast<int>(xpos / (mTileWidth + mInterTileSpaceX));
    int numYtile = static_cast<int>(ypos / (mTileHeight + mInterTileSpaceY));

    double deltax = xpos - numXtile * (mTileWidth + mInterTileSpaceX);
    double deltay = ypos - numYtile * (mTileHeight + mInterTileSpaceY);

    // Scale linearly the cordinates within tile
    double newxpos = numXtile * 36.0;
    newxpos += deltax * 36.0 / 100.0;
    newxpos += mIOSpaceX;

    double newypos = numYtile * mTileHeight;
    newypos += deltay * (mTileHeight) / (mTileHeight + mInterTileSpaceY);
    newypos += mIOSpaceY;

    newxpos = std::max<double>(0, newxpos);
    newypos = std::max<double>(0, newypos);

    if (xpos < 20)
      newxpos = xpos + mIOSpaceX;
    if (ypos < 20)
      newypos = ypos + mIOSpaceY;

    assert(newxpos <= tmp->X());
    assert(newypos <= tmp->Y());

    tmp->X(newxpos);
    tmp->Y(newypos);
  }

  WritePLFile(aPlName);

  for (int i = mFixeds.size() - 1; i >= 0; --i)
  {
    CNode *tmp = mAllNodes[mFixeds[i]];

    assert(mapXOfFixed.find(tmp) != mapXOfFixed.end());
    assert(mapYOfFixed.find(tmp) != mapYOfFixed.end());

    tmp->X(mapXOfFixed[tmp]);
    tmp->Y(mapYOfFixed[tmp]);
  }

  mapXOfFixed.clear();
  mapYOfFixed.clear();

  // Write SCL file
  std::ofstream out;
  out.open(aSclName.c_str());
  if (!out)
    cout << "ERROR: Could not open new scl file for writing\n";

  assert(0 < mAllRows.size());
  int newNumRow = static_cast<int>(mAllRows.size() * mTileHeight / (mTileHeight + mInterTileSpaceY));
  int newNumSites = CRow::mNumSites * 36 / 100;
  int firstSite = mIOSpaceY;

  out << "UCLA scl 1.0\n";
  out << "# Created       :\n";
  out << " # User          :       eASIC\n";
  out << "\n";
  out << "NumRows : " << newNumRow << endl;
  out << "\n";
  for (int i = 0; i < newNumRow; i++)
  {
    out << "CoreRow Horizontal\n";
    out << " Coordinate   :   " << firstSite + i << endl;
    out << " Height       :   1\n";
    out << " Sitewidth    :   1\n";
    out << " Sitespacing  :   1\n";
    out << " Siteorient   :   1\n";
    out << " Sitesymmetry :   1\n";
    out << " SubrowOrigin :   20    NumSites  :  " << newNumSites << endl;
    out << "End\n";
  }


  // Write nodes here
  std::ofstream outNodes;
  outNodes.open(aNodesName.c_str());
  if (!outNodes)
    cout << "ERROR: Could not open new scl file for writing\n";

  outNodes << "UCLA nodes 1.0\n";
  outNodes << "# Created       :\n";
  outNodes << "# User          :       eASIC netlist2nodes\n";
  outNodes << "\n";

  outNodes << "NumNodes : "     << mAllNodes.size() << endl;
  outNodes << "NumTerminals : " << mFixeds.size() << endl;
  outNodes << "\n";

  for (int i = 0; i < mECells.size(); i++)
    outNodes << mAllNodes[mECells[i]]->Name() << " 1 1\n";
  for (int i = 0; i < mDffs.size(); i++)
    outNodes << mAllNodes[mDffs[i]]->Name() << " 1 1\n";
  for (int i = 0; i < mRegs.size(); i++)
    outNodes << mAllNodes[mRegs[i]]->Name() << " 1 1\n";
  for (int i = 0; i < mBrams.size(); i++)
    outNodes << mAllNodes[mBrams[i]]->Name() << " 1 1\n";

  for (int i = 0; i < mFixeds.size(); i++)
    outNodes << mAllNodes[mFixeds[i]]->Name() << " 1 1 terminal\n";

  outNodes.close();
}

map< pair<double, double>, string> gMap;
void CDesign::ReadShrunkPlAndExpand(string aPlFile)
{
  cout << "STAT:  Entering ReadShrunkPlAndExpand()\n";

  bool readFixed = false;
  ReadPLFile(aPlFile, readFixed);

  GetHPWL("BeforeUnshrink");
  for (int i = mAllNodes.size() - 1; i >= 0; --i)
  {
    if (true == mAllNodes[i]->Fixed())
      continue;

    double xpos = mAllNodes[i]->X();
    double ypos = mAllNodes[i]->Y();

    xpos -= mIOSpaceX;
    ypos -= mIOSpaceY;

    int numTileX = static_cast<int>(xpos / 36.0);
    int numTileY = static_cast<int>(ypos / mTileHeight);

    double extraX = xpos - numTileX * 36.0;
    double extraY = ypos - numTileY * mTileHeight;
    if (18 > extraX)
    {
      int num3Block = static_cast<int>(extraX / 3.0);
      int leftOver = static_cast<int>(extraX) - num3Block * 3;
      extraX = (3 + 2) * num3Block;
      extraX += leftOver;
      assert(extraX < 30);
      extraX += 1;
    }
    else
    {
      extraX -= 18;
      assert(extraX < 18);
      int num3Block = static_cast<int>(extraX / 3.0);
      int leftOver = static_cast<int>(extraX) - num3Block * 3;
      extraX = (3 + 2) * num3Block;
      extraX += leftOver;
      assert(extraX < 30);
      extraX += 1;
      extraX += 46;
    }

    xpos = numTileX * (mInterTileSpaceX + mTileWidth) + extraX;
    ypos = numTileY * (mInterTileSpaceY + mTileHeight) + extraY;

    xpos += mIOSpaceX;
    ypos += mIOSpaceY;

    mAllNodes[i]->PutAt(xpos, ypos);
#if 1
    if (gMap.find(make_pair<double>(xpos, ypos)) != gMap.end())
    {
      cout << mAllNodes[i]->Name() << " clashes with " <<
      gMap[make_pair<double>(xpos, ypos)] << " at " <<
      xpos << "," << ypos << endl;
    }
    gMap[make_pair<double>(xpos, ypos)] = mAllNodes[i]->Name();
#endif
  }
  cout << "         Exiting ReadShrunkPlAndExpand()\n";
}

void CDesign::CheckAllCellsInCore()
{
//   static int firstTime = true;
//   if (firstTime == true)
//   {
//     cout << "WARN:  Prevented checking of CheckAllCellsInCore()\n";
//     firstTime = false;
//   }
//   return;

  cout << "STAT:  Checking all cells in Core area\n";
  for (int i = 0; i < mAllNodes.size(); i++)
  {
    if (true == mAllNodes[i]->Fixed())
      continue;

    double xpos = mAllNodes[i]->X();
    double ypos = mAllNodes[i]->Y();

    if (false == (xpos >= mIOSpaceX && xpos <= mWidth-mIOSpaceX-1) ||
        false == (ypos >= mIOSpaceY && ypos <= mHeight-mIOSpaceY-1))
    {
      cout << "***EROR***:  Cell " << mAllNodes[i]->Name() << " at " <<
      xpos << "," << ypos << " is outside core\n";
    }
  }
}

void CDesign::CheckDensity()
{
  cout << "STAT:  Checking if density of tiles are OK\n";
  int numTilesBroken = 0;
  for (int i = 0; i < mNumTilesX; ++i)
    for (int j = 0; j < mNumTilesY; ++j)
      if (false == mTiles[i][j]->IsDensityOK())
        ++numTilesBroken;
  if (0 < numTilesBroken)
    cout << "  - **EROR**: " << numTilesBroken << " out of " << mNumTilesX*mNumTilesY << " broken\n";
}


bool CDesign::RunMinCutBasedPlacer(string benchName)
{
  string newDirName = string("temp") + benchName;
  system((string("mkdir -p ") + newDirName).c_str());

  cout << "STAT:  Generating virtual benchmark in directory " << newDirName << endl;

  string newSclFileName  = newDirName + "/design.scl";
  string newPlFileName   = newDirName + "/design.pl";
  string newNodeFileName = newDirName + "/design.nodes";

  cout << "       - Dumping shrunk PL and SCL and NODES file\n";
  DumpShrunkPlAndSclAndNodes(newPlFileName, newSclFileName, newNodeFileName);

  cout << "       - Removing all nets more than 100 pins big\n";
  WriteNetsFileWithReducedDegree(newDirName + "/design.nets", 100);

  cout << "       - Generating AUX file for virtual benchmark\n";
  std::ofstream out;
  string newAuxName = newDirName + "/design.aux";
  out.open(newAuxName.c_str());
  out << "RowBasedPlacement : " <<
    "design.nodes"  << " " <<
    "design.nets"   << " " <<
    "design.scl" << " " <<
    "design.pl"  << endl;
  out.close();

  // Run partition based placer on the newDirName directory
  // Command line = "-f design.aux -faster -save"
  int numArgs = 5;
  const char *args[5] = {"place.exe", "-f", strdup((newDirName + "/design.aux").c_str()), "-faster", "-save"};
  cout << "STAT:  Starting Global placer now\n";
  for (int i=0; i<5; i++)
  {
    cout << "   " << args[i] << endl;
  }

// #define _DONT_RUNGP
#ifdef _DONT_RUNGP
#else
  cout << "STAT:  Calling MetaPlacer() (Capo's function)\n";
  MetaPlacer(numArgs, args);

  cout << "STAT:  Placement on virtual platform finished\n";
  system((string("mv out.pl ") + benchName + "_afterGP_MinCut.pl").c_str());
#endif

  ReadPLFile(string(benchName) + "_afterGP_MinCut.pl", false);
  WriteImage(string(benchName) + "_shrunk");
  GetHPWL(string(benchName) + "_shrunk");

  ReadShrunkPlAndExpand(string(benchName) + "_afterGP_MinCut.pl");
  checkIfEverythingOK();

  WriteImage(string(benchName) + "_expanded");
  GetHPWL(string(benchName) + "_expanded");
  return true;
}

bool CDesign::RunAnalyticalPlacer(string benchName)
{
  string newDirName = string("temp") + benchName;
  system((string("mkdir -p ") + newDirName).c_str());

  string newSclFileName  = newDirName + "/design.scl";
  string newPlFileName   = newDirName + "/design.pl";
  string newNodeFileName = newDirName + "/design.nodes";

  DumpShrunkPlAndSclAndNodes(newPlFileName, newSclFileName, newNodeFileName);
  WriteNetsFileWithReducedDegree(newDirName + "/design.nets", 100);

  std::ofstream out;
  string newAuxName = newDirName + "/design.aux";
  out.open(newAuxName.c_str());
  out << "RowBasedPlacement : " <<
    newDirName + "/design.nodes"  << " " <<
    newDirName + "/design.nets"   << " " <<
    newDirName + "/design.scl" << " " <<
    newDirName + "/design.pl"  << endl;
  out.close();

  // Run analytical placer on the newDirName directory
  // Command line = "executable -d aux -mvcycle 2 -cluster ratio 0.25 -writepl out file
  cout << "STAT:  Starting Global placer now\n";
  string cmd = "./bin/analytical.exe -d ";
  cmd += newDirName + "/design.aux -mvcycle 2 -cluster ratio 0.25 -writePl ";
  cmd += string(benchName) + "_afterGP_Analytical.pl";
  cout << "  - Running " << cmd << endl;
  system(cmd.c_str());

  system((string("mv out.pl ") + string(benchName) + "_afterGP_Analytical.pl").c_str());

  ReadPLFile(string(benchName) + "_afterGP_Analytical.pl", false);
//   WriteImage(string(benchName) + "_shrunk");
  GetHPWL(string(benchName) + "_shrunk");

  ReadShrunkPlAndExpand(string(benchName) + "_afterGP_Analytical.pl");
  checkIfEverythingOK();

//   WriteImage(string(benchName) + "_expanded");
  GetHPWL(string(benchName) + "_expanded");
  return true;
}

bool CDesign::RunGordianPlacer(string benchName)
{
  bool status = DoGlobalPlaceFixPoint();
  this->WritePLFile(benchName + "_afterGP_Gordian.pl");
  return status;
}
