/*
***********************************************************************
LICENSE AGREEMENT
***********************************************************************
 READ THIS LICENSE AGREEMENT CAREFULLY BEFORE USING THIS PRODUCT. BY 
 USING THIS PRODUCT YOU INDICATE YOUR ACCEPTANCE OF THE TERMS OF THE 
 FOLLOWING AGREEMENT. THESE TERMS APPLY TO YOU AND ANY SUBSEQUENT 
 LICENSEE OF THIS PRODUCT.

               License Agreement for RegularPlace 1.0

Copyright (c)  2008 by 
Ashutosh Chakraborty, Anurag Kumar, David Z. Pan
All rights reserved

ATTRIBUTION ASSURANCE LICENSE
Redistribution of the any code, with or without modification, are permitted 
provided that the conditions below are met. These conditions require a 
modest attribution to Ashutosh Chakraborty, Anurag Kumar and Dr. 
David Z. Pan (the "Author").

1. Redistributions of any code, with or without modification (the "Code"), 
must be accompanied by any documentation and, each time the resulting 
executable program or a program dependent thereon is launched, a 
prominent display (e.g., splash screen or banner text) of the Author's 
attribution information, which includes:
  (a) Ashutosh Chakraborty, Anurag Kumar, and Dr. David Z. Pan ("AUTHOR"),
  (b) The University of Texas at Austin ("PROFESSIONAL IDENTIFICATION"), and
  (c) http://www.cerc.utexas.edu/utda/ ("URL").

2. Use is limited to academic research groups only. Users who are interested 
in industry or commercial purposes must notify Author and request 
separate license agreement.  

3. Neither the name nor any trademark of the Author may be used to 
endorse or promote products derived from this software without specific 
prior written permission.

4. Users are entirely responsible, to the exclusion of the Author and 
any other persons, for compliance with (a) regulations set by owners or 
administrators of employed equipment, (b) licensing terms of any other 
software, and (c) local, national, and international regulations regarding 
use, including those regarding import, export, and use of encryption software. 

THIS FREE SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR 
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE AUTHOR OR ANY CONTRIBUTOR BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, EFFECTS OF UNAUTHORIZED OR MALICIOUS NETWORK 
ACCESS; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/
#ifndef __DESIGN_H__
#define __DESIGN_H__

#include "Util.h"
#include "Data.h"

class CNet;
class CNode;
class CGrid;
class CPropType;
class CRow;
class CRegion;
class CTile;
class CBBox;

#ifdef COMPILING_GLOBAL_PLACER
extern "C"
{
#include <laspack/errhandl.h>
#include <laspack/vector.h>
#include <laspack/matrix.h>
#include <laspack/qmatrix.h>
#include <laspack/operats.h>
#include <laspack/factor.h>
#include <laspack/precond.h>
#include <laspack/eigenval.h>
#include <laspack/rtc.h>
#include <laspack/itersolv.h>
#include <laspack/mlsolv.h>
#include <laspack/version.h>
#include <laspack/copyrght.h>
#include <laspack/vector.h>
#include <laspack/qmatrix.h>
#include <laspack/itersolv.h>
#include <laspack/copyrght.h>
}
#else
class QMatrix;
class Vector;
#endif

extern vector<CRegion*> gRegionsVec;
extern vector<double>   gRegionXOff;
extern vector<double>   gRegionYOff;

extern void Tokenize(const string &str,
                       vector<string>& tokens,
                       const string& delim = " \t\n");

class CDesign
{
  public:
    CDesign(void);
    ~CDesign(void);

  public:
    int  ReadPLFile(string aFileName, bool readFixed = true);

    void Read(string nodesFileName, string netFileName,
              string propFileName, string platRegionName,
              string platSclName, string initPlaceName);

    void WriteImage(string aFileName);
    double GetHPWL(string aMessage = "");
    void GlobalSwap();
    void TetrisLegalize();
    void Refine();
    int  GetNumClkDomains()
    {
      return mClockDomains.size();
    }
    void RunGlobalPlace();
    void DoIroning();
    void Legalize();
    void OptimizeLegal();
    void WritePLFile(string name);
    void WriteNetsFile(string name);
    void WriteNetsFileWithReducedDegree(string aName, int maxDegree);
    void GetTilesInBBox(CBBox* aBox, vector<CTile*> & v);
#if 0
//     CTile* GetParentTile(CNode* aNode) const
//     { return GetParentTile(aNode->X(), aNode->Y()); }
#endif
    CTile* GetParentTile(const double x, const double y) const;

    // Takes a bad node and returns (x,y) of the nearest tile it can be placed.
    pair<int, int> findLegalTileForPlacing(CNode *) const;

  private:
    void writeTransportation();
    void writeArcForTileAt(int x, int y, int id);
    int readNetsFile(string aNetsFileName);
    int readNodesFile(string aNodeFileName);
    int readPropsFile(string aPropsFileName);
    int readPlatformInfo(string aFileName);
    int readPlatformRegion(string aFileName);
    int readPlatformScl(string aFileName);
    void dumpInputForHmetis(string aHmetisInputFileName);
    void dumpPartitionedFiles(string aFileName, int iPartitionId);
    bool checkSanity();
    void drawPlatformPic();

  public:
    bool checkIfEverythingOK();
    bool SolveBRAMAssignment(bool bAddToTileAfterSolving = true);
    bool SolveRegFileAssignment(bool bAddToTileAfterSolving = true);
    void ShrinkSize();
    void UnshrinkSize();

  public:
    void FixBRAMs();
    void FixRegFiles();

  public:  // Tiles related stuff
    int mTileWidth;
    int mTileHeight;
    int mNumTilesX;
    int mNumTilesY;
    double mInterTileSpaceX;
    double mInterTileSpaceY;

    vector< vector< CTile* > > mTiles;
    void ComputeTileStuff();
    void Prepare();
    map <string, CNode*> mMapNameToNode;

  public:
    double scaleInX(double in) const;
    double scaleInY(double in) const;
    int    unscaleInX(double in);
    int    unscaleInY(double in);
    void   AssignNodesToTiles();
    void   AssignNodesToTiles(vector <CNode *>);

  public:
    vector <CNode*> mAllNodes;

  private:
    map <string, int> mMapRegionNameToWidth;
    map <string, int> mMapRegionNameToHeight;
    CRegion *mPlatformTop;   // Pointer to the top level platform instance
    string mTopInstName;

    // Info regarding clock domains
    vector <string> mClockDomains;
    map <string, int> mMapClockDomainNameToInt;

    vector <string> mCellTypes;
    int mNumCellTypes;
    map <string, int> mMapCellTypeToInt;

    // Nodes, nets, rows present in the circuit

    vector <int> mECells;
    vector <int> mDffs;
    vector <int> mRegs;
    vector <int> mBrams;
    vector <int> mFixeds;

    vector <CNet*>  mAllNets;
    vector <CRow *> mAllRows;


    int mNumNodes;
    int mNumNets;
    int mNumPins;
    int mNumTerms;

    bool mIsShrunk;

  public:
    // Platform specific information
    int mWidth;
    int mHeight;
    int mIOSpaceX;
    int mIOSpaceY;
    int mCoreWidth;
    int mCoreHeight;
    int mNumGroups;

  private:
    void doIroningOfEcells();
    void doIroningOfEcellsForTile(CTile *tile, bool direction);//direction = 0 for vertical, 1 for horizontal.
    void doIroningOfDff();
    void doIroningOfClocks();
    void doForceIroningOfEcells(); // Forces density equalization
    void doForceIroningOfDffs(); // Forces density equalization
  public:
    map<string, int> m_mNameClockDomain;
    //  bool FindTarget(int iCurrentGridX, int iCurrentGridY, int iEndGridX, int iEndGridY, vector<CGrid *>* pPath);
    int MapInfoToGrid();
    void makeDensityOK();
    void makeDensityOKForTile(int, int);

  public:
    void DumpShrunkPlAndSclAndNodes(string a, string b, string c);
    void ReadShrunkPlAndExpand(string a);
    void CheckAllCellsInCore();
    void CheckDensity();


    // For new attempt of new global placer
    void ComputeHessAndRHS(QMatrix& q, Vector& x, Vector& y);
    void SolveHessnStoreResults(QMatrix& q, Vector& sol, Vector& rhs);
    void UpdatePosn(Vector& x, Vector& y, bool updateX = true, bool updateY = true);
    void PopulateHessianAndRHS();
#if 0
    void AssignPartnID(const int aCurrIter, vector<int> numCell,
                       vector<double> xcom, vector<double> ycom);
#endif
    int  AssignNeighbors();

#if 0
    void RecomputeHess(int aNumPartn,
                       QMatrix& newHess, Vector& newrhsX, Vector& newrhsY,
                       QMatrix& q, Vector& rx, Vector& ry);
#else
    void RecomputeHess(vector<int>& numInPartns, vector<double>& CoMPartn,
                       QMatrix& newHess, Vector& newRhs, bool isForX);

    void RecomputeHessFixPoint(vector<int>& numInPartns, vector<double>& CoMPartn,
                               QMatrix& newHess, Vector& newRhs, bool isForX);
#endif

    void AssignPartnID(const int aCurrIter, vector<int>& xNumCell,
                       vector<int>& yNumCell, vector<double>& xcom, vector<double>& ycom);

    void FindPartitionVectors( vector <CNode*> &vec1, int partX1, int partY1,
                                    vector <CNode*> &vec2, int partX2, int partY2);

    void createHessian(bool isForX);
    bool DoGlobalPlace();
    bool DoGlobalPlaceFixPoint();
    int  GetCellsToPutInHessian();

    vector <CNode*> mNodesInHessian;
    void PrintCoM(vector<double>& xcom, vector<double>& ycom);

    void assignXPartnID(vector<double>& vecXPositions);
    void assignYPartnID(vector<double>& vecYPositions);

    void AssignPartnIDNew(const int aCurrIter, vector<double>& xcom, vector<double>& ycom);

    void assignXPartnID(CNode *n, vector<double>& xPartnPositions);
    void assignYPartnID(CNode *n, vector<double>& yPartnPositions);

    bool RunAnalyticalPlacer(string benchName);
    bool RunMinCutBasedPlacer(string benchName);
    bool RunGordianPlacer(string benchName);
};

#endif
