/*
***********************************************************************
LICENSE AGREEMENT
***********************************************************************
 READ THIS LICENSE AGREEMENT CAREFULLY BEFORE USING THIS PRODUCT. BY 
 USING THIS PRODUCT YOU INDICATE YOUR ACCEPTANCE OF THE TERMS OF THE 
 FOLLOWING AGREEMENT. THESE TERMS APPLY TO YOU AND ANY SUBSEQUENT 
 LICENSEE OF THIS PRODUCT.

               License Agreement for RegularPlace 1.0

Copyright (c)  2008 by 
Ashutosh Chakraborty, Anurag Kumar, David Z. Pan
All rights reserved

ATTRIBUTION ASSURANCE LICENSE
Redistribution of the any code, with or without modification, are permitted 
provided that the conditions below are met. These conditions require a 
modest attribution to Ashutosh Chakraborty, Anurag Kumar and Dr. 
David Z. Pan (the "Author").

1. Redistributions of any code, with or without modification (the "Code"), 
must be accompanied by any documentation and, each time the resulting 
executable program or a program dependent thereon is launched, a 
prominent display (e.g., splash screen or banner text) of the Author's 
attribution information, which includes:
  (a) Ashutosh Chakraborty, Anurag Kumar, and Dr. David Z. Pan ("AUTHOR"),
  (b) The University of Texas at Austin ("PROFESSIONAL IDENTIFICATION"), and
  (c) http://www.cerc.utexas.edu/utda/ ("URL").

2. Use is limited to academic research groups only. Users who are interested 
in industry or commercial purposes must notify Author and request 
separate license agreement.  

3. Neither the name nor any trademark of the Author may be used to 
endorse or promote products derived from this software without specific 
prior written permission.

4. Users are entirely responsible, to the exclusion of the Author and 
any other persons, for compliance with (a) regulations set by owners or 
administrators of employed equipment, (b) licensing terms of any other 
software, and (c) local, national, and international regulations regarding 
use, including those regarding import, export, and use of encryption software. 

THIS FREE SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR 
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE AUTHOR OR ANY CONTRIBUTOR BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, EFFECTS OF UNAUTHORIZED OR MALICIOUS NETWORK 
ACCESS; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include <iostream>
#include <vector>
#include <string>
#include <map>
#include <fstream>
#include <math.h>

#include "Design.h"
#include "BBox.h"
#include "Row.h"
#include "Region.h"
#include "munkres.h"
#include "Tile.h"
using namespace std;

pair<int, int> getTileLocation(int, int, int);
typedef struct edge
{
  int edgeNumber;
  int startNode;
  int endNode;
  int minCapacity;
  int maxCapacity;
  int cost;
} Edge;

typedef struct clockNodes
{
  vector <CNode *> nodes;
} ClockNodes;

void CDesign::Refine()
{
  cout << "STATUS: Entering CDesign::Refine()" << endl;

#if 0
  cout << "BAD stuff for testing move to tile\n";
  for (int i = 0; i < mNumTilesX; i++)
    for (int j = 0; j < mNumTilesY; j++)
    {
      mTiles[i][j]->mECells.clear();
      mTiles[i][j]->mDffs.clear();
      mTiles[i][j]->mRegs.clear();
      mTiles[i][j]->mBrams.clear();
      mTiles[i][j]->mNumCellsOfClk.clear();
    }
  for (int i = 0; i < mAllNodes.size(); i++)
  {
    if (false == mAllNodes[i]->Fixed())
    {
      mAllNodes[i]->X(mTiles[10][10]->X() + 5);
      mAllNodes[i]->Y(mTiles[10][10]->Y() + 5);
      mTiles[10][10]->AddNode(mAllNodes[i]);
    }
  }
#endif

  //   writeTransportation();
  //   makeDensityOK();
  DoIroning();
  GetHPWL("After ironing");
  // Fix things between tiles here.  Once it is done, all we need
  // to do is to legalize the stuff within each tile.

  cout << "TODO:  Density calculation should come here" << endl;

  cout << "        Exiting CDesign::Refine()" << endl;
  return;
}

void CDesign::AssignNodesToTiles()
{
  for (int i = 0; i < mAllNodes.size(); i++)
  {
    if (mAllNodes[i]->Fixed())
      continue;

    double x = mAllNodes[i]->X();
    double y = mAllNodes[i]->Y();

    int xInd = -1;
    int yInd = -1;
    if (true == mIsShrunk)
    {
      xInd = static_cast<int>((x - mIOSpaceX) / mTileWidth);
      yInd = static_cast<int>((y - mIOSpaceY) / mTileHeight);
    }
    else
    {
      xInd = static_cast<int>((x - mIOSpaceX) / (mTileWidth + mInterTileSpaceX));
      yInd = static_cast<int>((y - mIOSpaceY) / (mTileHeight + mInterTileSpaceY));
    }
    assert(-1 != xInd);
    assert(-1 != yInd);
    assert(xInd < mTiles.size());
    assert(yInd < mTiles[0].size());
    mTiles[xInd][yInd]->AddNode(mAllNodes[i]);
  }
}

void CDesign::AssignNodesToTiles(vector <CNode *> vecNodes)
{
  for (int i = 0; i < vecNodes.size(); i++)
  {
    if (vecNodes[i]->Fixed())
      continue;

    int xInd = static_cast<int>((vecNodes[i]->X() - mIOSpaceX) /
                                (mTileWidth + mInterTileSpaceX));
    int yInd = static_cast<int>((vecNodes[i]->Y() - mIOSpaceY) /
                                (mTileHeight + mInterTileSpaceY));

    assert(xInd < mTiles.size());
    assert(yInd < mTiles[0].size());
    mTiles[xInd][yInd]->AddNode(vecNodes[i]);
  }
}

// Takes a bad node and returns (x,y) of nearest tile it can be placed in.
pair<int, int> CDesign::findLegalTileForPlacing(CNode *badNode) const
{
  int countOfTraversal, count;
//   i = (int)badNode->X()/mTileWidth;
//   j = (int)badNode->Y()/mTileHeight;
  // Tiles are being calculated locally, which may go out of sink later. Must be taken care of.
  int i = static_cast<int>((badNode->X() - mIOSpaceX) /
                           (mTileWidth + mInterTileSpaceX));
  int j = static_cast<int>((badNode->Y() - mIOSpaceY) /
                           (mTileHeight + mInterTileSpaceY));
  countOfTraversal = 1;
//   cout << i << " " << j << " Number of tiles \n";
  while (countOfTraversal < mNumTilesX || countOfTraversal < mNumTilesY)
  {
    for (count = 0; count < countOfTraversal; count++)
    {
      j--;
      //Do operation here
      if (i < mNumTilesX && j < mNumTilesY && i >= 0 && j >= 0)
      {
        if (mTiles[i][j]->CanTake(badNode))
        {
          pair<int, int> coordinate(i, j);
          return coordinate;
        }
        //     cout << i << " " << j << endl;
      }
    }
    for (count = 0; count < countOfTraversal; count++)
    {
      i++;
      //Do operation here
      //    cout << i << " " << j << endl;
      if (i < mNumTilesX && j < mNumTilesY && i >= 0 && j >= 0)
      {
        if (mTiles[i][j]->CanTake(badNode))
        {
          pair<int, int> coordinate(i, j);
          return coordinate;
        }
        //     cout << i << " " << j << endl;
      }
    }
    countOfTraversal += 1;
    for (count = 0; count < countOfTraversal; count++)
    {
      j++;
      //Do operation here
      //    cout << i << " " << j << endl;
      if (i < mNumTilesX && j < mNumTilesY && i >= 0 && j >= 0)
      {
        if (mTiles[i][j]->CanTake(badNode))
        {
          pair<int, int> coordinate(i, j);
          return coordinate;
        }
        //     cout << i << " " << j << endl;
      }
    }
    for (count = 0; count < countOfTraversal; count++)
    {
      i--;
      //Do operation here
      //    cout << i << " " << j << endl;
      if (i < mNumTilesX && j < mNumTilesY && i >= 0 && j >= 0)
      {
        if (mTiles[i][j]->CanTake(badNode))
        {
          pair<int, int> coordinate(i, j);
          return coordinate;
        }
        //     cout << i << " " << j << endl;
      }
    }
    countOfTraversal += 1;
  }
}

void CDesign::makeDensityOK()
{
  int tileId = 0;
  std::ofstream out;
  out.open("test.inp");
  int i, j;

  for (i = 0; i < mNumTilesX; i++)
    for (j = 0; j < mNumTilesY; j++)
    {
      //       if (false == mTiles[i][j]->IsDensityOK())
      //        return; // Density already ok.

      mTiles[i][j]->mTileId = tileId;
      tileId += 1;
    }

  for (i = 0; i < mNumTilesX; i++)
  {
    for (j = 0; j < mNumTilesY; j++)
    {
      if (false == mTiles[i][j]->IsDensityOK(0))
      {
        cout << "n " << i << " " << j << " " << mTiles[i][j]->mTileId <<
        " " << mTiles[i][j]->mECells.size() << endl;
        makeDensityOKForTile(i, j);
        //           makeDensityOKForTile(10, 9);

        for (int countj = 0; countj < mNumTilesY; countj++)
        {
          for (int counti = 0; counti < mNumTilesX; counti++)
          {
            //                         cout << mTiles[counti][countj]->mECells.size() << " ";
            //                         printf("%4d ",mTiles[counti][countj]->NumCellsInside());
            printf("%5d ", mTiles[counti][countj]->mECells.size());
          }
          cout << endl;
        }
        //     exit(0);
      }
    }
  }


#if 0
  vector <int> nodes;
  vector <Edge> edges;
  // Assign node to each Tile where movement can happen
  int nodeNumber = 1;
  i = 10;
  j = 9;

  vector <int> tiles;
  vector <int> cells;
  map <int, int> mapNodeNumberToTileId;
  map <int, int> mapNodeNumberToCellId;
  mapNodeNumberToTileId[nodeNumber] = mTiles[i][j]->mTileId;
  tiles.push_back(nodeNumber);
  nodeNumber += 1;
  if (mNumTilesX > i + 1)
  {
    tiles.push_back(nodeNumber);
    mapNodeNumberToTileId[nodeNumber] = mTiles[i+1][j]->mTileId;
    nodeNumber += 1;
  }
  if (0 <= i - 1)
  {
    tiles.push_back(nodeNumber);
    mapNodeNumberToTileId[nodeNumber] = mTiles[i-1][j]->mTileId;
    nodeNumber += 1;
  }
  if (mNumTilesY > j + 1)
  {
    tiles.push_back(nodeNumber);
    mapNodeNumberToTileId[nodeNumber] = mTiles[i][j+1]->mTileId;
    nodeNumber += 1;
  }
  if (0 <= j - 1)
  {
    tiles.push_back(nodeNumber);
    mapNodeNumberToTileId[nodeNumber] = mTiles[i][j-1]->mTileId;
    nodeNumber += 1;
  }
  if (0 <= j - 1 && 0 <= i - 1)
  {
    tiles.push_back(nodeNumber);
    mapNodeNumberToTileId[nodeNumber] = mTiles[i-1][j-1]->mTileId;
    nodeNumber += 1;
  }
  if (0 <= j - 1 && mNumTilesX > i + 1)
  {
    tiles.push_back(nodeNumber);
    mapNodeNumberToTileId[nodeNumber] = mTiles[i+1][j-1]->mTileId;
    nodeNumber += 1;
  }
  if (mNumTilesY > j + 1 && 0 <= i - 1)
  {
    tiles.push_back(nodeNumber);
    mapNodeNumberToTileId[nodeNumber] = mTiles[i-1][j+1]->mTileId;
    nodeNumber += 1;
  }
  if (mNumTilesY > j + 1 && mNumTilesX > i + 1)
  {
    tiles.push_back(nodeNumber);
    mapNodeNumberToTileId[nodeNumber] = mTiles[i+1][j+1]->mTileId;
    nodeNumber += 1;
  }

  int startingNode, endingNode, currentNode;
  startingNode = nodeNumber;
  nodeNumber += 1;
  endingNode   = nodeNumber;

  // Assign nodeNumber to each node present in the current tile
  int countOfEdges = 0;
  double currentHPWL, newHPWL;
  pair <int, int> tileLocation;
  for (counti = 0; counti < mTiles[i][j]->mECells.size(); counti++)
  {
    cells.push_back(nodeNumber);
    mapNodeNumberToCellId[nodeNumber] = counti;
    currentNode = nodeNumber;
    nodeNumber += 1;

    // Create edge from each tile to this node
    for (countj = 0; countj < tiles.size(); countj++)
    {
      Edge newEdge;
      newEdge.edgeNumber = countOfEdges;
      newEdge.startNode  = currentNode;
      newEdge.endNode    = tiles[countj];
      newEdge.minCapacity = 0;
      newEdge.maxCapacity = 1;
      //      newEdge.cost        = 1; // to be replaced by HPWL
      // Calculate cost
      currentHPWL = mTiles[i][j]->mECells[counti]->GetIncidentWL();
      double xOld = mTiles[i][j]->mECells[counti]->X();
      double yOld = mTiles[i][j]->mECells[counti]->Y();
      int curTileId = mapNodeNumberToTileId[countj];
      tileLocation = getTileLocation(curTileId, mNumTilesX, mNumTilesY);
      double xNew = (double)mTiles[tileLocation.first][tileLocation.second]->X() + mTileWidth / 2;
      double yNew = (double)mTiles[tileLocation.first][tileLocation.second]->Y() + mTileHeight / 2;
      mTiles[i][j]->mECells[counti]->PutAt(xNew, yNew);
      newHPWL = mTiles[i][j]->mECells[counti]->GetIncidentWL();
      mTiles[i][j]->mECells[counti]->PutAt(xOld, yOld);
      newEdge.cost = (int)(newHPWL - currentHPWL);

      edges.push_back(newEdge);
      countOfEdges += 1;  // Increment the count of edges
    }

    Edge first;
    first.edgeNumber = countOfEdges;
    first.startNode  = startingNode;
    first.endNode    = currentNode;
    first.minCapacity = 0;
    first.maxCapacity = 1;
    first.cost        = 0;
    edges.push_back(first);
    countOfEdges += 1;  // Increment the count of edges

  }

  for (counti = 0; counti < tiles.size(); counti++)
  {
    Edge edgeEndNode;
    edgeEndNode.edgeNumber = countOfEdges;
    edgeEndNode.startNode  = tiles[counti];
    edgeEndNode.endNode    = endingNode;
    edgeEndNode.minCapacity = 0;
    int curTileId = mapNodeNumberToTileId[counti];
    tileLocation = getTileLocation(curTileId, mNumTilesX, mNumTilesY);
    assert(mTiles[tileLocation.first][tileLocation.second]->mTileId == curTileId);

    int capacity = mTiles[tileLocation.first][tileLocation.second]->getCapacity(0);
    if (capacity > 0)
      edgeEndNode.maxCapacity = capacity;
    else
      edgeEndNode.maxCapacity = 0;
    edgeEndNode.cost        = 0;
    edges.push_back(edgeEndNode);
    countOfEdges += 1;
  }

  out << "p min " << nodeNumber << " " << countOfEdges << endl;
  out << "n " << startingNode << " " << mTiles[i][j]->mECells.size() << endl;
  out << "n " << endingNode   << " -" << mTiles[i][j]->mECells.size() << endl;
  for (counti = 0; counti < edges.size(); counti++)
  {
    out << "a " << edges[counti].startNode << " " << edges[counti].endNode << " " << edges[counti].minCapacity << " " << edges[counti].maxCapacity << " " << edges[counti].cost << endl;
  }

#endif
}

void CDesign::makeDensityOKForTile(int tileIndexX, int tileIndexY)
{
  int counti, countj;
  vector <int> nodes;
  vector <Edge> edges;
  std::ofstream out;
  out.open("cs2.inp");
  // Assign node to each Tile where movement can happen
  int nodeNumber = 1;
  int i = tileIndexX;
  int j = tileIndexY;

  vector <int> tiles;
  vector <int> cells;
  map <int, int> mapNodeNumberToTileId;
  map <int, int> mapNodeNumberToCellId;
  map <int, CNode *> mapNodeNumberToCNode;
  mapNodeNumberToTileId[nodeNumber] = mTiles[i][j]->mTileId;
  tiles.push_back(nodeNumber); // nodeNumber 1 assigned to main tile.
  int mainTile = nodeNumber;
  nodeNumber += 1;
  if (mNumTilesX > i + 1)
  {
    tiles.push_back(nodeNumber);
    mapNodeNumberToTileId[nodeNumber] = mTiles[i+1][j]->mTileId;
    nodeNumber += 1;
  }
  if (0 <= i - 1)
  {
    tiles.push_back(nodeNumber);
    mapNodeNumberToTileId[nodeNumber] = mTiles[i-1][j]->mTileId;
    nodeNumber += 1;
  }
  if (mNumTilesY > j + 1)
  {
    tiles.push_back(nodeNumber);
    mapNodeNumberToTileId[nodeNumber] = mTiles[i][j+1]->mTileId;
    nodeNumber += 1;
  }
  if (0 <= j - 1)
  {
    tiles.push_back(nodeNumber);
    mapNodeNumberToTileId[nodeNumber] = mTiles[i][j-1]->mTileId;
    nodeNumber += 1;
  }
  if (0 <= j - 1 && 0 <= i - 1)
  {
    tiles.push_back(nodeNumber);
    mapNodeNumberToTileId[nodeNumber] = mTiles[i-1][j-1]->mTileId;
    nodeNumber += 1;
  }
  if (0 <= j - 1 && mNumTilesX > i + 1)
  {
    tiles.push_back(nodeNumber);
    mapNodeNumberToTileId[nodeNumber] = mTiles[i+1][j-1]->mTileId;
    nodeNumber += 1;
  }
  if (mNumTilesY > j + 1 && 0 <= i - 1)
  {
    tiles.push_back(nodeNumber);
    mapNodeNumberToTileId[nodeNumber] = mTiles[i-1][j+1]->mTileId;
    nodeNumber += 1;
  }
  if (mNumTilesY > j + 1 && mNumTilesX > i + 1)
  {
    tiles.push_back(nodeNumber);
    mapNodeNumberToTileId[nodeNumber] = mTiles[i+1][j+1]->mTileId;
    nodeNumber += 1;
  }

  int startingNode, endingNode, currentNode;
  startingNode = nodeNumber;
  nodeNumber += 1;
  endingNode   = nodeNumber;

  // Assign nodeNumber to each node present in the current tile
  int countOfEdges = 0;
  double currentHPWL, newHPWL;
  pair <int, int> tileLocation;
  vector <CNode *> nodesInTile;
  for (counti = 0; counti < mTiles[i][j]->mECells.size(); counti++)
  {
    cells.push_back(nodeNumber);
    mapNodeNumberToCellId[nodeNumber] = counti + 1; // Incremented value stored to distinguish
    mapNodeNumberToCNode[nodeNumber] = mTiles[i][j]->mECells[counti];
    currentNode = nodeNumber;
    nodeNumber += 1;

    // Create edge from each tile to this node
    for (countj = 0; countj < tiles.size(); countj++)
    {
      Edge newEdge;
      newEdge.edgeNumber = countOfEdges;
      newEdge.startNode  = currentNode;
      newEdge.endNode    = tiles[countj];
      newEdge.minCapacity = 0;
      newEdge.maxCapacity = 1;
      //      newEdge.cost        = 1; // to be replaced by HPWL
      // Calculate cost
      if (tiles[countj] == mainTile)
      {
        newEdge.cost = 0;
      }
      else
      {
        currentHPWL = mTiles[i][j]->mECells[counti]->GetIncidentWL();
        double xOld = mTiles[i][j]->mECells[counti]->X();
        double yOld = mTiles[i][j]->mECells[counti]->Y();
        int curTileId = mapNodeNumberToTileId[tiles[countj]];
        tileLocation = getTileLocation(curTileId, mNumTilesX, mNumTilesY);
        double xNew = (double)mTiles[tileLocation.first][tileLocation.second]->X() + mTileWidth / 2;
        double yNew = (double)mTiles[tileLocation.first][tileLocation.second]->Y() + mTileHeight / 2;
        mTiles[i][j]->mECells[counti]->PutAt(xNew, yNew);
        newHPWL = mTiles[i][j]->mECells[counti]->GetIncidentWL();
        mTiles[i][j]->mECells[counti]->PutAt(xOld, yOld);
        newEdge.cost = (int)(newHPWL - currentHPWL);
      }
      edges.push_back(newEdge);
      countOfEdges += 1;  // Increment the count of edges
    }

    Edge first;
    first.edgeNumber = countOfEdges;
    first.startNode  = startingNode;
    first.endNode    = currentNode;
    first.minCapacity = 0;
    first.maxCapacity = 1;
    first.cost        = 0;
    edges.push_back(first);
    countOfEdges += 1;  // Increment the count of edges

  }

  for (counti = 0; counti < tiles.size(); counti++)
  {
    Edge edgeEndNode;
    edgeEndNode.edgeNumber = countOfEdges;
    edgeEndNode.startNode  = tiles[counti];
    edgeEndNode.endNode    = endingNode;
    edgeEndNode.minCapacity = 0;
    int curTileId = mapNodeNumberToTileId[tiles[counti]];
    tileLocation = getTileLocation(curTileId, mNumTilesX, mNumTilesY);
    assert(mTiles[tileLocation.first][tileLocation.second]->mTileId == curTileId);

    int capacity = mTiles[tileLocation.first][tileLocation.second]->getCapacity(0);
    if (capacity > 0)
      edgeEndNode.maxCapacity = capacity;
    else
      edgeEndNode.maxCapacity = 0;
    edgeEndNode.cost        = 0;
    edges.push_back(edgeEndNode);
    countOfEdges += 1;

    cout << curTileId << tiles[counti] << " " << tileLocation.first << " " << tileLocation.second << endl;
  }

  out << "p min " << nodeNumber << " " << countOfEdges << endl;
  out << "n " << startingNode << " " << mTiles[i][j]->mECells.size() << endl;
  out << "n " << endingNode   << " -" << mTiles[i][j]->mECells.size() << endl;
  for (counti = 0; counti < edges.size(); counti++)
  {
    out << "a " << edges[counti].startNode << " " << edges[counti].endNode << " " << edges[counti].minCapacity << " " << edges[counti].maxCapacity << " " << edges[counti].cost << endl;
  }

  out.close();

  system("./binaries/cs2-4.3/cs2.exe < cs2.inp");

  //start reading the input file

  std::ifstream in;
  string line;
  in.open("cs2.out");
  if (!in)
  {
    cout << "ERROR: Data not dumped by network flow solver" << endl;
    return;
  }

  cout << "Size of current tile " << mTiles[i][j]->mECells.size() << endl;
  int startNodeIn, endNodeIn, flowIn, cellID, tileID;
  while (!in.eof())
  {
    getline(in, line);
    if (line.find_first_not_of("\n\t ", 0) == string::npos)
      continue;
    if (line.c_str() == NULL)
    {
      continue;
    }
    sscanf(line.c_str(), "%d %d %d", &startNodeIn, &endNodeIn, &flowIn);
    //        cout << startNodeIn << " " << endNodeIn << " " << flowIn << " " << mapNodeNumberToCellId[startNodeIn] << endl;

    if (mapNodeNumberToCellId[startNodeIn])
    {
      cellID = mapNodeNumberToCellId[startNodeIn] - 1; // One subtracted since it was added while adding
      tileID = mapNodeNumberToTileId[endNodeIn];
      if (mTiles[i][j]->mTileId != tileID)      // If current tile id is different move the node to this tile.
      {
        //          cout << startNodeIn << " " << cellID << endl;
        tileLocation = getTileLocation(tileID, mNumTilesX, mNumTilesY);
        //          CNode *currentNode = mTiles[i][j]->mECells[cellID];
        mTiles[i][j]->RemoveNode(mapNodeNumberToCNode[startNodeIn]);
        mTiles[tileLocation.first][tileLocation.second]->AddNode(mapNodeNumberToCNode[startNodeIn]);
        //          cout << endNodeIn << " " << tileID << " " << tileLocation.first << " " << tileLocation.second << endl;
        double xNew = (double)mTiles[tileLocation.first][tileLocation.second]->X() + mTileWidth / 2;
        double yNew = (double)mTiles[tileLocation.first][tileLocation.second]->Y() + mTileHeight / 2;
        // this needs to be improved. Not really good :(. Increases the wire-length to almost double.
        mapNodeNumberToCNode[startNodeIn]->X(xNew);
        mapNodeNumberToCNode[startNodeIn]->Y(yNew);
      }
    }
  }

  cout << "Size of current tile " << mTiles[i][j]->mECells.size() << endl;
  in.close();
  system("rm cs2.out");

}

void CDesign::DoIroning()
{
  cout << "STATUS:  Entering DoIroning()\n";

#if 1
//   doIroningOfEcells();
  cout << "Number of ecells before ironing : Max allowed: " << gMaxECellAllowed << "\n";
  for (int i=0; i<mNumTilesX; i++)
  {
    for (int j=0; j<mNumTilesY; j++)
    {
      cout << mTiles[i][j]->mECells.size() << " ";
    }
    cout << endl;
  }
  
  cout << "\nNumber of dffs before ironing : Max allowed: " << gMaxDffAllowed << "\n";
  for (int i=0; i<mNumTilesX; i++)
  {
    for (int j=0; j<mNumTilesY; j++)
    {
      cout << mTiles[i][j]->mDffs.size() << " ";
    }
    cout << endl;
  }

  cout << "\n\n";

  doForceIroningOfEcells();

//   doIroningOfDff();

  cout << "Ironing of dffs being done here\n";
  doForceIroningOfDffs();

  doIroningOfClocks();
#endif

  cout << "check the effect \n\n";
  cout << "STATUS:  Exiting DoIroning()\n";
}

void CDesign::doIroningOfDff()
{
  cout << "STATUS:  Entering doIroningOfDff()\n";
  int i, j;
  int countCells;
  int counti, countj;
  double xTotal = 0;
  int total = 0;
  int iTotalCells, iMaxCells;
  CNode *currentNode;
  int countOfIroning = 0;
  vector <CTile *> vecTiles;
  vector <CNode *> vecNodes;
  vector <double> nodeTransformedCordinates;
  int start;
  double iMinimumX, iMaximumX;
  double expFactor;
  double multFactor = 0.3;
  double mean = 0;
  for (countj = 0; countj < mNumTilesY; countj++)
  {
    for (counti = 0; counti < mNumTilesX; counti++)
    {
      printf("%5d ", mTiles[counti][countj]->mDffs.size());
    }
    cout << endl;
  }

  int stopIroning = 0;
  while (countOfIroning < 50)
  {
#if 1
    for (j = 0; j < mNumTilesY; j++)
    {
      start = 0;
      //       cout << "size of vector tile: " << vecTiles.size() << endl;

      for (i = 0; i < mNumTilesX; i++)
      {
        //      cout << mTiles[i][j]->X() << " " << mTiles[i][j]->Y() << endl;
        if (mTiles[i][j]->mDffs.size() > gMaxDffAllowed)
        {
          start = 1;
          vecTiles.push_back(mTiles[i][j]);
        }
        else
        {
          if (start == 1)
          {
            break;
          }
        }
      }
      //       cout << "size of vector tile: " << vecTiles.size() << endl;
      if (vecTiles.size() == 0)
        continue;
      xTotal = 0;
      total  = 0;
      iTotalCells = 0;
      iMaxCells = 0;
      for (counti = 0; counti < vecTiles.size(); counti++)
      {
        cout << vecTiles[counti]->X() << " ";
        for (countCells = 0; countCells < vecTiles[counti]->mDffs.size(); countCells++)
        {
          xTotal += vecTiles[counti]->mDffs[countCells]->X();
          total  += 1;
          vecNodes.push_back(vecTiles[counti]->mDffs[countCells]);
          //           vecTiles[counti]->RemoveNode(vecTiles[counti]->mDffs[countCells]);
          // Node removed from current tile will be added later
          // Replace this function by another function which removes a whole vector instead of one node.
        }

        iTotalCells += vecTiles[counti]->mDffs.size();
        iMaxCells   += gMaxDffAllowed;
      }
      cout << endl;
      mean = xTotal / total;
      double oldX;
      iMinimumX = 1000000;
      iMaximumX = -1000000;
      expFactor = multFactor * iTotalCells / iMaxCells;
      for (counti = 0; counti < vecNodes.size(); counti++)
      {
        oldX = vecNodes[counti]->X();
        nodeTransformedCordinates.push_back(oldX);
        nodeTransformedCordinates[counti] = expFactor * (oldX - mean) + mean;
//         vecNodes[counti]->X( 0.1*(iTotalCells/gMaxDffAllowed)*(oldX - mean) + mean);
        if (iMinimumX > nodeTransformedCordinates[counti])
          iMinimumX = nodeTransformedCordinates[counti];
        if (iMaximumX < nodeTransformedCordinates[counti])
          iMaximumX = nodeTransformedCordinates[counti];
      }

      cout << iMinimumX << ":min " << iMaximumX << ":max " << mCoreWidth << " " << vecNodes.size() << endl;


      if ((iMaximumX > mCoreWidth && iMinimumX < 0) || (iMaximumX - iMinimumX > mCoreWidth))
      {
        vecNodes.clear();
        vecTiles.clear();
        nodeTransformedCordinates.clear();
        continue;
      }
      else if (iMinimumX < 0)
      {
        iMinimumX = fabs(iMinimumX);
        for (counti = 0; counti < vecNodes.size(); counti++)
        {
          oldX = nodeTransformedCordinates[counti];
          vecNodes[counti]->X(oldX + iMinimumX);
        }
      }
      else if (iMaximumX > mCoreWidth)
      {
        for (counti = 0; counti < vecNodes.size(); counti++)
        {
          oldX = nodeTransformedCordinates[counti];
          vecNodes[counti]->X(oldX - iMaximumX + mCoreWidth);
        }
      }
      else
      {
        for (counti = 0; counti < vecNodes.size(); counti++)
        {
          oldX = nodeTransformedCordinates[counti];
          vecNodes[counti]->X(oldX);
        }
      }

      for (counti = 0; counti < vecTiles.size(); counti++)
      {
//         for (countCells=0; countCells<vecTiles[counti]->mDffs.size(); countCells++)
        for (countCells = vecTiles[counti]->mDffs.size() - 1; countCells >= 0; countCells--)
        {
          vecTiles[counti]->RemoveNode(vecTiles[counti]->mDffs[countCells]);
        }
      }

      AssignNodesToTiles(vecNodes);


      vecNodes.clear();
      vecTiles.clear();
      nodeTransformedCordinates.clear();
      nodeTransformedCordinates.clear();
      for (countj = 0; countj < mNumTilesY; countj++)
      {
        for (counti = 0; counti < mNumTilesX; counti++)
        {
          printf("%5d ", mTiles[counti][countj]->mDffs.size());
        }
        cout << endl;
      }
    }
#endif

#if 1
    for (i = 0; i < mNumTilesX; i++)
    {
      start = 0;
      //       cout << "size of vector tile: " << vecTiles.size() << endl;
      for (j = 0; j < mNumTilesY; j++)
      {

        //      cout << mTiles[i][j]->X() << " " << mTiles[i][j]->Y() << endl;
        if (mTiles[i][j]->mDffs.size() > gMaxDffAllowed)
        {
          start = 1;
          vecTiles.push_back(mTiles[i][j]);
        }
        else
        {
          if (start == 1)
          {
            break;
          }
        }

        //   for(countCells=0; countCells<mTiles[i][j]->mECells.size(); countCells++) {
        //       currentNode = mTiles[i][j]->mECells[countCells];
        //             xTotal += currentNode->X();
        //    total  += 1;
        //   }
      }
      //       cout << "size of vector tile: " << vecTiles.size() << endl;
      if (vecTiles.size() == 0)
        continue;
      xTotal = 0;
      total  = 0;
      iTotalCells = 0;
      iMaxCells = 0;
      for (counti = 0; counti < vecTiles.size(); counti++)
      {
        cout << vecTiles[counti]->Y() << " " ;
        for (countCells = 0; countCells < vecTiles[counti]->mDffs.size();
             countCells++)
        {
          xTotal += vecTiles[counti]->mDffs[countCells]->Y();
          total  += 1;
          vecNodes.push_back(vecTiles[counti]->mDffs[countCells]);
          vecTiles[counti]->RemoveNode(vecTiles[counti]->mDffs[countCells]);
          // Node removed from current tile will be added later
          // Replace this function by another function which removes a
          // whole vector instead of one node.
        }

        iTotalCells += vecTiles[counti]->mDffs.size();
        iMaxCells   += gMaxDffAllowed;
      }

      cout << endl;

      mean = xTotal / total;
      double oldX;
      iMinimumX = 1000000;
      iMaximumX = -1000000;
      expFactor = multFactor * iTotalCells / iMaxCells;
      for (counti = 0; counti < vecNodes.size(); counti++)
      {
        oldX = vecNodes[counti]->Y();
        nodeTransformedCordinates.push_back(oldX);
        nodeTransformedCordinates[counti] = expFactor * (oldX - mean) + mean;
//         vecNodes[counti]->Y(0.1*(iTotalCells/gMaxDffAllowed)*(oldX - mean) + mean);
        if (iMinimumX > nodeTransformedCordinates[counti])
          iMinimumX = nodeTransformedCordinates[counti];
        if (iMaximumX < nodeTransformedCordinates[counti])
          iMaximumX = nodeTransformedCordinates[counti];
      }

      if ((iMaximumX > mCoreHeight && iMinimumX < 0) || (iMaximumX - iMinimumX > mCoreHeight))
      {
        vecNodes.clear();
        vecTiles.clear();
        nodeTransformedCordinates.clear();
        continue;
      }
      else if (iMinimumX < 0)
      {
        iMinimumX = fabs(iMinimumX);
        for (counti = 0; counti < vecNodes.size(); counti++)
        {
          oldX = nodeTransformedCordinates[counti];
          vecNodes[counti]->Y(oldX + iMinimumX);
        }
      }
      else if (iMaximumX > mCoreHeight)
      {
        for (counti = 0; counti < vecNodes.size(); counti++)
        {
          oldX = nodeTransformedCordinates[counti];
          vecNodes[counti]->Y(oldX - iMaximumX + mCoreHeight);
        }
      }
      else
      {
        for (counti = 0; counti < vecNodes.size(); counti++)
        {
          oldX = nodeTransformedCordinates[counti];
          vecNodes[counti]->Y(oldX);
        }
      }

      for (counti = 0; counti < vecTiles.size(); counti++)
      {
//         for (countCells=0; countCells<vecTiles[counti]->mDffs.size(); countCells++)
        for (countCells = vecTiles[counti]->mDffs.size() - 1; countCells >= 0; countCells--)
        {
          vecTiles[counti]->RemoveNode(vecTiles[counti]->mDffs[countCells]);
        }
      }

      AssignNodesToTiles(vecNodes);

      vecNodes.clear();
      vecTiles.clear();
      nodeTransformedCordinates.clear();
      stopIroning = 1;
      for (countj = 0; countj < mNumTilesY; countj++)
      {
        for (counti = 0; counti < mNumTilesX; counti++)
        {
          printf("%5d ", mTiles[counti][countj]->mDffs.size());
          if (mTiles[counti][countj]->mDffs.size() > gMaxDffAllowed)
            stopIroning = 0;
        }
        cout << endl;
      }
      if (stopIroning)
        break;
    }
#endif
    countOfIroning += 1;
  }
  for (countj = 0; countj < mNumTilesY; countj++)
  {
    for (counti = 0; counti < mNumTilesX; counti++)
    {
      printf("%5d ", (mTiles[counti][countj]->mNumCellsOfClk.size()));
    }
    cout << endl;
  }

  cout << "STATUS:  Exiting doIroningOfDff()\n";
  //  exit(0);
}

void CDesign::doIroningOfEcellsForTile(CTile *tile, bool direction)
{
  vector <CNode *> vecNodes;
  vector <double> nodeTransformedCordinates;
  int counti;
  double sum = 0;
  double oldX, iMinimumX, iMaximumX;
  double maxHeight;

  if (direction)
    maxHeight = mCoreWidth;
  else
    maxHeight = mCoreHeight;
  for (counti = 0; counti < tile->mECells.size(); counti++)
  {
    vecNodes.push_back(tile->mECells[counti]);
    if (direction)
      sum += tile->mECells[counti]->X();
    else
      sum += tile->mECells[counti]->Y();
  }
  double mean = sum / tile->mECells.size();
  iMinimumX = 1000000;
  iMaximumX = -1000000;
  for (counti = 0; counti < vecNodes.size(); counti++)
  {
    if (direction)
      oldX = vecNodes[counti]->X();
    else
      oldX = vecNodes[counti]->Y();
    nodeTransformedCordinates.push_back(oldX);
    nodeTransformedCordinates[counti] = 5 * (tile->mECells.size() / gMaxECellAllowed) * (oldX - mean) + mean;
    if (iMinimumX > nodeTransformedCordinates[counti])
      iMinimumX = nodeTransformedCordinates[counti];//vecNodes[counti]->X();
    if (iMaximumX < nodeTransformedCordinates[counti])
      iMaximumX = nodeTransformedCordinates[counti];//vecNodes[counti]->X();
  }
  cout << iMinimumX << ":min " << iMaximumX << ":max " << mCoreWidth << " " << vecNodes.size() << endl;


  if (iMinimumX < 0)
  {
    iMinimumX = fabs(iMinimumX);
    for (counti = 0; counti < vecNodes.size(); counti++)
    {
      oldX = nodeTransformedCordinates[counti];//vecNodes[counti]->X();
      if (direction)
        vecNodes[counti]->X(oldX + iMinimumX);
      else
        vecNodes[counti]->Y(oldX + iMinimumX);
    }
  }
  else if (iMaximumX > maxHeight)
  {
    for (counti = 0; counti < vecNodes.size(); counti++)
    {
      oldX = nodeTransformedCordinates[counti];//vecNodes[counti]->X();
      if (direction)
        vecNodes[counti]->X(oldX - iMaximumX + mCoreWidth);
      else
        vecNodes[counti]->Y(oldX - iMaximumX + mCoreWidth);
    }
  }
  else
  {
    cout << " Reached last else\n";
    for (counti = 0; counti < vecNodes.size(); counti++)
    {
      oldX = nodeTransformedCordinates[counti];//vecNodes[counti]->X();
      if (direction)
        vecNodes[counti]->X(oldX);
      else
        vecNodes[counti]->Y(oldX);
    }
  }

  for (counti = tile->mECells.size() - 1; counti >= 0; counti--)
  {
    tile->RemoveNode(tile->mECells[counti]);
  }
  AssignNodesToTiles(vecNodes);

  vecNodes.clear();
  nodeTransformedCordinates.clear();
}

void CDesign::doIroningOfEcells()
{
  cout << "STATUS:  Entering doIroningOfEcells()\n";
  int i, j;
  int countCells;
  int counti, countj;
  double xTotal = 0;
  int total = 0;
  double iTotalCells, iMaxCells;
  double expFactor;
  double maxAllowedCells;
  double multFactor = 1.0;
  CNode *currentNode;
  int countOfIroning = 0;
  vector <CTile *> vecTiles;
  vector <CNode *> vecNodes;
  vector <double> nodeTransformedCordinates;
  int start;
  double iMinimumX, iMaximumX;
  double mean = 0;
  for (countj = 0; countj < mNumTilesY; countj++)
  {
    for (counti = 0; counti < mNumTilesX; counti++)
    {
      printf("%5d ", mTiles[counti][countj]->mECells.size());
    }
    cout << endl;
  }

  int stopIroning = 0;
  while (countOfIroning < 10)
  {
#if 1
    for (j = 0; j < mNumTilesY; j++)
    {
      start = 0;
      //       cout << "size of vector tile: " << vecTiles.size() << endl;

      for (i = 0; i < mNumTilesX; i++)
      {
        //      cout << mTiles[i][j]->X() << " " << mTiles[i][j]->Y() << endl;
        if (mTiles[i][j]->mECells.size() > gMaxECellAllowed)
        {
          start = 1;
          vecTiles.push_back(mTiles[i][j]);
        }
        else
        {
          if (start == 1)
          {
            break;
          }
        }
      }

      maxAllowedCells = 0;
      for (i = 0; i < mNumTilesX; i++)
      {
        maxAllowedCells += gMaxECellAllowed;
      }

//       cout << "size of vector tile: " << vecTiles.size() << endl;
      if (vecTiles.size() == 0)
        continue;
      xTotal = 0;
      total  = 0;
      iTotalCells = 0;
      iMaxCells = 0;
      for (counti = 0; counti < vecTiles.size(); counti++)
      {
        cout << vecTiles[counti]->X() << " ";
        for (countCells = 0; countCells < vecTiles[counti]->mECells.size(); countCells++)
        {
          xTotal += vecTiles[counti]->mECells[countCells]->X();
          total  += 1;
          vecNodes.push_back(vecTiles[counti]->mECells[countCells]);
//           vecTiles[counti]->RemoveNode(vecTiles[counti]->mECells[countCells]);
          // Node removed from current tile will be added later
          // Replace this function by another function which removes a whole vector instead of one node.
        }

        iTotalCells += vecTiles[counti]->mECells.size();
        iMaxCells   += gMaxECellAllowed;
      }
      cout << "iTotalCells: " << iTotalCells << ": iMaxCells:" << iMaxCells << endl;
      mean = xTotal / total;
      double oldX;
      iMinimumX = 1000000;
      iMaximumX = -1000000;
      expFactor = multFactor * iTotalCells / iMaxCells;
      if (iTotalCells > maxAllowedCells)
      {
        cout << "changed expFactor :" << expFactor << " " << maxAllowedCells / iMaxCells << endl;
        expFactor = multFactor * maxAllowedCells / iMaxCells;
      }
      for (counti = 0; counti < vecNodes.size(); counti++)
      {
        oldX = vecNodes[counti]->X();
        nodeTransformedCordinates.push_back(oldX);
        nodeTransformedCordinates[counti] = expFactor * (oldX - mean) + mean;
//         vecNodes[counti]->X( 0.1*(iTotalCells/gMaxECellAllowed)*(oldX - mean) + mean);
        if (iMinimumX > nodeTransformedCordinates[counti])
          iMinimumX = nodeTransformedCordinates[counti];//vecNodes[counti]->X();
        if (iMaximumX < nodeTransformedCordinates[counti])
          iMaximumX = nodeTransformedCordinates[counti];//vecNodes[counti]->X();
      }

      cout << iMinimumX << ":min " << iMaximumX << ":max " << mCoreWidth << " " << vecNodes.size() << endl;


      if ((iMaximumX > mCoreWidth && iMinimumX < 0) || (iMaximumX - iMinimumX > mCoreWidth))
      {
//         vecNodes.clear();
//         vecTiles.clear();
//   nodeTransformedCordinates.clear();
//   cout << "Did not do anything" << endl;
//   continue;

        for (counti = 0; counti < vecTiles.size(); counti++)
        {
          iMinimumX = fabs(iMinimumX);
          oldX = nodeTransformedCordinates[counti];
          if (oldX < 0)
            vecNodes[counti]->X(oldX + iMinimumX);
          else if (oldX > mCoreHeight)
            vecNodes[counti]->X(oldX - iMaximumX + mCoreHeight);
          else
            vecNodes[counti]->X(oldX);
        }
      }
      else if (iMinimumX < 0)
      {
        iMinimumX = fabs(iMinimumX);
        for (counti = 0; counti < vecNodes.size(); counti++)
        {
          oldX = nodeTransformedCordinates[counti];//vecNodes[counti]->X();
          vecNodes[counti]->X(oldX + iMinimumX);
        }
      }
      else if (iMaximumX > mCoreWidth)
      {
        for (counti = 0; counti < vecNodes.size(); counti++)
        {
          oldX = nodeTransformedCordinates[counti];//vecNodes[counti]->X();
          vecNodes[counti]->X(oldX - iMaximumX + mCoreWidth);
        }
      }
      else
      {
        cout << " Reached last else\n";
        for (counti = 0; counti < vecNodes.size(); counti++)
        {
          oldX = nodeTransformedCordinates[counti];//vecNodes[counti]->X();
          vecNodes[counti]->X(oldX);
        }
      }


      for (counti = 0; counti < vecTiles.size(); counti++)
      {
        cout << "Number of ecells: " << vecTiles[counti]->mECells.size() << endl;
//         for (countCells=0; countCells<vecTiles[counti]->mECells.size(); countCells++)
        for (countCells = vecTiles[counti]->mECells.size() - 1; countCells >= 0; countCells--)
        {
          vecTiles[counti]->RemoveNode(vecTiles[counti]->mECells[countCells]);
        }
      }

      AssignNodesToTiles(vecNodes);
      for (countj = 0; countj < mNumTilesY; countj++)
      {
        for (counti = 0; counti < mNumTilesX; counti++)
        {
          printf("%5d ", mTiles[counti][countj]->mECells.size());
        }
        cout << endl;
      }

      if (vecTiles.size() > 6)
      {
        cout << "single tile ironing \n\n\n";
        for (counti = 0; counti < vecTiles.size(); counti++)
          doIroningOfEcellsForTile(vecTiles[counti], 0);
      }

      vecNodes.clear();
      vecTiles.clear();
      nodeTransformedCordinates.clear();
      for (countj = 0; countj < mNumTilesY; countj++)
      {
        for (counti = 0; counti < mNumTilesX; counti++)
        {
          printf("%5d ", mTiles[counti][countj]->mECells.size());
        }
        cout << endl;
      }
    }
#endif

#if 1
    for (i = 0; i < mNumTilesX; i++)
    {
      start = 0;
      //       cout << "size of vector tile: " << vecTiles.size() << endl;
      for (j = 0; j < mNumTilesY; j++)
      {

        //      cout << mTiles[i][j]->X() << " " << mTiles[i][j]->Y() << endl;
        if (mTiles[i][j]->mECells.size() > (unsigned)gMaxECellAllowed)
        {
          start = 1;
          vecTiles.push_back(mTiles[i][j]);
        }
        else
        {
          if (start == 1)
          {
            break;
          }
        }
      }
      if (vecTiles.size() == 0)
        continue;

      maxAllowedCells = 0;
      for (i = 0; i < mNumTilesY; i++)
      {
        maxAllowedCells += gMaxECellAllowed;
      }

      xTotal = 0;
      total  = 0;
      iTotalCells = 0;
      iMaxCells = 0;
      for (counti = 0; counti < (signed)vecTiles.size(); counti++)
      {
        cout << vecTiles[counti]->Y() << " " ;
        for (countCells = 0; countCells < (signed)vecTiles[counti]->mECells.size(); countCells++)
        {
          xTotal += vecTiles[counti]->mECells[countCells]->Y();
          total  += 1;
          vecNodes.push_back(vecTiles[counti]->mECells[countCells]);
        }

        iTotalCells += vecTiles[counti]->mECells.size();
        iMaxCells   += gMaxECellAllowed;
      }

      cout << endl;

      mean = xTotal / total;
      double oldX;
      iMinimumX = 1000000;
      iMaximumX = -1000000;
      expFactor = multFactor * iTotalCells / iMaxCells;
      if (iTotalCells > maxAllowedCells)
        expFactor = multFactor * maxAllowedCells / iMaxCells;
      for (counti = 0; counti < (signed)vecNodes.size(); counti++)
      {
        oldX = vecNodes[counti]->Y();
        nodeTransformedCordinates.push_back(oldX);
        nodeTransformedCordinates[counti] = expFactor * (oldX - mean) + mean;
//   vecNodes[counti]->Y( 0.1*(iTotalCells/gMaxECellAllowed)*(oldX - mean) + mean);
        if (iMinimumX > nodeTransformedCordinates[counti])
          iMinimumX = nodeTransformedCordinates[counti];
        if (iMaximumX < nodeTransformedCordinates[counti])
          iMaximumX = nodeTransformedCordinates[counti];
      }

      cout << "Processing columns ..." << endl;
      if ((iMaximumX > mCoreHeight && iMinimumX < 0) || (iMaximumX - iMinimumX > mCoreHeight))
      {
//         vecNodes.clear();
//         vecTiles.clear();
//   nodeTransformedCordinates.clear();
//   cout << "Did not do anything" << endl;
//         continue;
        for (counti = 0; counti < vecTiles.size(); counti++)
        {
          iMinimumX = fabs(iMinimumX);
          oldX = nodeTransformedCordinates[counti];
          if (oldX < 0)
            vecNodes[counti]->Y(oldX + iMinimumX);
          else if (oldX > mCoreHeight)
            vecNodes[counti]->Y(oldX - iMaximumX + mCoreHeight);
          else
            vecNodes[counti]->Y(oldX);
        }
      }
      else if (iMinimumX < 0)
      {
        iMinimumX = fabs(iMinimumX);
        for (counti = 0; counti < vecNodes.size(); counti++)
        {
          oldX = nodeTransformedCordinates[counti];
          vecNodes[counti]->Y(oldX + iMinimumX);
        }
      }
      else if (iMaximumX > mCoreHeight)
      {
        for (counti = 0; counti < vecNodes.size(); counti++)
        {
          oldX = nodeTransformedCordinates[counti];
          vecNodes[counti]->Y(oldX - iMaximumX + mCoreHeight);
        }
      }
      else
      {
        for (counti = 0; counti < vecNodes.size(); counti++)
        {
          oldX = nodeTransformedCordinates[counti];
          vecNodes[counti]->Y(oldX);
        }

      }

      for (counti = 0; counti < vecTiles.size(); counti++)
      {
//         for (countCells=0; countCells<vecTiles[counti]->mECells.size(); countCells++)
        cout << "Number of ecells: " << vecTiles[counti]->mECells.size() << endl;
        for (countCells = vecTiles[counti]->mECells.size() - 1; countCells >= 0; countCells--)
        {
          vecTiles[counti]->RemoveNode(vecTiles[counti]->mECells[countCells]);
        }
      }
      AssignNodesToTiles(vecNodes);

      if (vecTiles.size() > 6)
      {
        cout << "single tile ironing \n\n\n";
        for (counti = 0; counti < vecTiles.size(); counti++)
          doIroningOfEcellsForTile(vecTiles[counti], 1);
      }


      vecNodes.clear();
      vecTiles.clear();
      nodeTransformedCordinates.clear();
      stopIroning = 1;
      for (countj = 0; countj < mNumTilesY; countj++)
      {
        for (counti = 0; counti < mNumTilesX; counti++)
        {
          printf("%5d ", mTiles[counti][countj]->mECells.size());
          if (mTiles[counti][countj]->mECells.size() > gMaxECellAllowed)
            stopIroning = 0;
        }
        cout << endl;
      }
      if (stopIroning)
        break;
    }
#endif
    countOfIroning += 1;
  }
  //  exit(0);
  cout << "STATUS:  Exiting doIroningOfEcells()\n";
}

void CDesign::doForceIroningOfEcells()
{
  CTile *replacementTile;
  bool foundPlace;

  for (int countj = 0; countj < mNumTilesY; countj++)
  {
    for (int counti = 0; counti < mNumTilesX; counti++)
    {
      CTile *tmpTile = mTiles[counti][countj];
      int numExtraCells = tmpTile->mECells.size() - gMaxECellAllowed;
#if 0
      cout << "In tile [" << counti << "][" << countj << "] has " <<
      tmpTile->mECells.size()- gMaxECellAllowed << " extra ECells (" <<
      tmpTile->mECells.size() << " - " << gMaxECellAllowed << ")\n";
#endif
      if (numExtraCells > 0)
      {


        for (int currIdx = 0; currIdx < numExtraCells; currIdx++)
        {
          CNode *tmpNode = tmpTile->mECells.back();
//           cout << " Removed " << tmpNode->Name() << endl;
          assert(tmpTile->Contains(tmpNode));

#if 0
          static int numCalled = 0;
          if (0 == numCalled++)
            cout << "WARN:  Suboptimal jhaadu done\n";
#endif

//           cout << " doing here\n";
          pair<int, int> coordinateOfLegalTile;
          coordinateOfLegalTile = findLegalTileForPlacing(tmpNode);
          mTiles[coordinateOfLegalTile.first][coordinateOfLegalTile.second]->AddNode(tmpNode);
          tmpNode->X(mTiles[coordinateOfLegalTile.first][coordinateOfLegalTile.second]->X() + mTileWidth / 2);
          tmpNode->Y(mTiles[coordinateOfLegalTile.first][coordinateOfLegalTile.second]->Y() + mTileHeight / 2);

#if 0
          foundPlace = false;
          for (int i = 0; i < mNumTilesX; i++)
          {
            for (int j = 0; j < mNumTilesY; j++)
            {
              if (mTiles[i][j]->mECells.size() < gMaxECellAllowed)
              {
                mTiles[i][j]->AddNode(tmpNode);
                foundPlace = true;
                break;
              }
            }
            if (true == foundPlace)
              break;
          }
          assert(true == foundPlace);
#endif
          tmpTile->mECells.pop_back();
        }
      }
    }
  }

  for (int countj = 0; countj < mNumTilesY; countj++)
  {
    for (int counti = 0; counti < mNumTilesX; counti++)
    {
      printf("%5d ", mTiles[counti][countj]->mECells.size());
    }
    cout << endl;
  }
}

void CDesign::doForceIroningOfDffs()
{
  CTile *replacementTile;
  bool foundPlace;

  for (int countj = 0; countj < mNumTilesY; countj++)
  {
    for (int counti = 0; counti < mNumTilesX; counti++)
    {
      CTile *tmpTile = mTiles[counti][countj];
      int numExtraCells = tmpTile->mDffs.size() - gMaxDffAllowed;

      if (numExtraCells > 0)
      {
        for (int currIdx = 0; currIdx < numExtraCells; currIdx++)
        {
          CNode *tmpNode = tmpTile->mDffs.back();
          assert(tmpTile->Contains(tmpNode));

#if 0
          static int numCalled = 0;
          if (0 == numCalled++)
            cout << "WARN:  Suboptimal jhaadu done\n";
#endif

          pair<int, int> coordinateOfLegalTile;
          coordinateOfLegalTile = findLegalTileForPlacing(tmpNode);
          mTiles[coordinateOfLegalTile.first][coordinateOfLegalTile.second]->AddNode(tmpNode);
          tmpNode->X(mTiles[coordinateOfLegalTile.first][coordinateOfLegalTile.second]->X() + mTileWidth / 2);
          tmpNode->Y(mTiles[coordinateOfLegalTile.first][coordinateOfLegalTile.second]->Y() + mTileHeight / 2);
          tmpTile->RemoveNode(tmpNode);
        }
      }

      while (tmpTile->mColumnsEDff.size() < tmpTile->NumColumnReqd())
      {
        cout << "Trying to fix number of column violation\n";
        CNode *tmpNode = tmpTile->mDffs.back();
        assert(tmpTile->Contains(tmpNode));

#if 0
        static int numCalled = 0;
        if (0 == numCalled++)
          cout << "WARN:  Suboptimal jhaadu done\n";
#endif

        pair<int, int> coordinateOfLegalTile;
        coordinateOfLegalTile = findLegalTileForPlacing(tmpNode);
        mTiles[coordinateOfLegalTile.first][coordinateOfLegalTile.second]->AddNode(tmpNode);
        tmpNode->X(mTiles[coordinateOfLegalTile.first][coordinateOfLegalTile.second]->X() + mTileWidth / 2);
        tmpNode->Y(mTiles[coordinateOfLegalTile.first][coordinateOfLegalTile.second]->Y() + mTileHeight / 2);
        tmpTile->RemoveNode(tmpNode);
      }
    }
  }

  for (int countj = 0; countj < mNumTilesY; countj++)
  {
    for (int counti = 0; counti < mNumTilesX; counti++)
    {
      printf("%5d ", mTiles[counti][countj]->mDffs.size());
    }
    cout << endl;
  }
}

int WhichLocationInTileForRegFile(CTile *t)
{
  bool firFree = true;
  bool secFree = true;
  bool trdFree = true;
  bool forFree = true;
  for (int i = 0; i < t->mRegs.size(); i++)
  {
    if (t->mRegs[i]->X() == 76 && t->mRegs[i]->Y() == 0)
      firFree = false;
    if (t->mRegs[i]->X() == 84 && t->mRegs[i]->Y() == 0)
      secFree = false;
    if (t->mRegs[i]->X() == 84 && t->mRegs[i]->Y() == 32)
      trdFree = false;
    if (t->mRegs[i]->X() == 76 && t->mRegs[i]->Y() == 32)
      forFree = false;
  }

  if (firFree)
    return 0;
  if (secFree)
    return 1;
  if (trdFree)
    return 2;
  if (forFree)
    return 3;

  return 4;
}

void CDesign::doIroningOfClocks()
{
  int counti, countj, countNodes, i, j;
//  int countOfClocks[50];
  CTile *currentTile;
  CNode *currentNode;

  int countOfWhileLoop = 0;
  cout << "before clock ironing\n";

#if 0
  for (int i = 0; i < mNumTilesX; i++)
  {
    for (int j = 0; j < mNumTilesY; j++)
    {
      CTile *tmpTile = mTiles[i][j];

      vector<int> vecClkInTile = tmpTile->GetClockIdsInside();
      if (vecClkInTile.size() <= gMaxClockInTile)
        continue;

      vector<int> numOfCellsOfClkType(vecClkInTile.size());
      for (int currCell = 0; currCell <= tmpTile->mDffs.size(); i++)
      {
        int matchPos = -1;
      }

    }
  }
#endif

//   for ( countj=0; countj<mNumTilesY; countj++)
//   {
//     for ( counti=0; counti<mNumTilesX; counti++)
//     {
//       printf("%5d ",(mTiles[counti][countj]->mNumCellsOfClk.size()));
//     }
//     cout << endl;
//   }

  while (countOfWhileLoop < 1000)
  {
    for (countj = 0; countj < mNumTilesY; countj++)
    {
      for (counti = 0; counti < mNumTilesX; counti++)
      {
        ClockNodes clkNodes[50];
        currentTile = mTiles[counti][countj];
        vector <int> clockIds = currentTile->GetClockIdsInside();
        if (clockIds.size() > gMaxClockInTile)
        {

          for (countNodes = 0; countNodes < (signed)currentTile->mDffs.size(); countNodes++)
          {
            currentNode = currentTile->mDffs[countNodes];
            for (int numClk = 0; numClk < currentNode->mClkDom.size(); numClk++)
              clkNodes[currentNode->mClkDom[numClk]].nodes.push_back(currentNode);
          }
          for (countNodes = 0; countNodes < (signed)currentTile->mBrams.size(); countNodes++)
          {
            currentNode = currentTile->mBrams[countNodes];
            for (int numClk = 0; numClk < currentNode->mClkDom.size(); numClk++)
              clkNodes[currentNode->mClkDom[numClk]].nodes.push_back(currentNode);
          }
          for (countNodes = 0; countNodes < (signed)currentTile->mRegs.size(); countNodes++)
          {
            currentNode = currentTile->mRegs[countNodes];
            for (int numClk = 0; numClk < currentNode->mClkDom.size(); numClk++)
              clkNodes[currentNode->mClkDom[numClk]].nodes.push_back(currentNode);
          }

          vector <int> sortedClockIds;
          vector <int> clockNodeSize;
          vector <bool> clockIdVisit;
          map <int, int> mapClockNodeSizeToClockId;

          for (i = 0; i < clockIds.size(); i++)
          {
            clockNodeSize.push_back(clkNodes[clockIds[i]].nodes.size());
            clockIdVisit.push_back(0);
            mapClockNodeSizeToClockId[clkNodes[clockIds[i]].nodes.size()] = clockIds[i];
//             cout << clkNodes[clockIds[i]].nodes.size() << ":" << clockIds[i] << " " ;
          }
          sort(clockNodeSize.begin(), clockNodeSize.end());

          for (i = 0; i < clockNodeSize.size(); i++)
          {
//             cout << clockNodeSize[i] << ":" << mapClockNodeSizeToClockId[clockNodeSize[i]] << " ";
            for (j = 0; j < clockIds.size(); j++)
            {
              if (clockNodeSize[i] == clkNodes[clockIds[j]].nodes.size() && !clockIdVisit[j])
              {
                sortedClockIds.push_back(clockIds[j]);
                clockIdVisit[j] = 1;
              }
            }
          }

          assert(clockIds.size() == sortedClockIds.size());
          // Remove lowest number of clocks from the current tile

          map <CNode*, bool> gNothing;
          while (currentTile->mNumCellsOfClk.size() > gMaxClockInTile)
          {
            for (i = 0; i < sortedClockIds.size(); i++)
            {
              int currentClockId = sortedClockIds[i];
              for (j = 0; j < clkNodes[currentClockId].nodes.size(); j++)
              {
                currentNode = clkNodes[currentClockId].nodes[j]; //remove all these nodes

                if (gNothing.find(currentNode) != gNothing.end())
                  continue;

                pair<int, int> coordinateOfLegalTile;
                coordinateOfLegalTile = findLegalTileForPlacing(currentNode);

                CTile *newTile;
                newTile = mTiles[coordinateOfLegalTile.first][coordinateOfLegalTile.second];

                currentTile->RemoveNode(currentNode);
                newTile->AddNode(currentNode);
                gNothing[currentNode] = true;

                if (currentNode->Type() < 2)
                {
                  currentNode->X(newTile->X() + mTileWidth / 2);
                  currentNode->Y(newTile->Y() + mTileHeight / 2);
                }
                else if (currentNode->Type() == 2)
                {
                  // reg files
                  switch (WhichLocationInTileForRegFile(newTile))
                  {
                  case 0:
                    currentNode->PutAt(newTile->X() + 76, newTile->Y() + 0);
                    break;
                  case 1:
                    currentNode->PutAt(newTile->X() + 84, newTile->Y() + 0);
                    break;
                  case 2:
                    currentNode->PutAt(newTile->X() + 84, newTile->Y() + 32);
                    break;
                  case 3:
                    currentNode->PutAt(newTile->X() + 76, newTile->Y() + 32);
                    break;
                  default:
                    abort();
                  }
                }
                else if (currentNode->Type() == 3)
                {
                  currentNode->PutAt(newTile->X() + 30, newTile->Y() + 0);
                }
                else
                {
                  cout << "Rotating what?\n";
                  currentNode->X(newTile->X() + mTileWidth / 2);
                  currentNode->Y(newTile->Y() + mTileHeight / 2);
                }
              }
            }
          }
          gNothing.clear();
        }
      }
    }

    bool breakCondition = true;

    for (countj = 0; countj < mNumTilesY; countj++)
    {
      for (counti = 0; counti < mNumTilesX; counti++)
      {
        if (mTiles[counti][countj]->mNumCellsOfClk.size() > gMaxClockInTile)
          breakCondition = false;
      }
    }
    if (breakCondition)
      break;

    countOfWhileLoop += 1;
  }
  if (countOfWhileLoop == 100)
    cout << "WARN:  Iron of Clock stopped due to while Loop termination\n";

  for (countj = 0; countj < mNumTilesY; countj++)
  {
    for (counti = 0; counti < mNumTilesX; counti++)
    {
      printf("%5d ", (mTiles[counti][countj]->mNumCellsOfClk.size()));
    }
    cout << endl;
  }
//
//   for ( countj=0; countj<mNumTilesY; countj++)
//   {
//     for ( counti=0; counti<mNumTilesX; counti++)
//     {
//       printf("%5d ",(mTiles[counti][countj]->mDffs.size()));
//     }
//     cout << endl;
//   }

}

pair<int, int> getTileLocation(int tileId, int lengthTileX, int lengthTileY)
{
  int x = (int)tileId / lengthTileY;
  int y = tileId - lengthTileY * x;
  pair <int, int> tileLocation(x, y);
  return tileLocation;
}
