/************************************************************************
LICENSE AGREEMENT
***********************************************************************
 READ THIS LICENSE AGREEMENT CAREFULLY BEFORE USING THIS PRODUCT. BY 
 USING THIS PRODUCT YOU INDICATE YOUR ACCEPTANCE OF THE TERMS OF THE 
 FOLLOWING AGREEMENT. THESE TERMS APPLY TO YOU AND ANY SUBSEQUENT 
 LICENSEE OF THIS PRODUCT.

               License Agreement for RegularPlace 1.0

Copyright (c)  2008 by 
Ashutosh Chakraborty, Anurag Kumar, David Z. Pan
All rights reserved

ATTRIBUTION ASSURANCE LICENSE
Redistribution of the any code, with or without modification, are permitted 
provided that the conditions below are met. These conditions require a 
modest attribution to Ashutosh Chakraborty, Anurag Kumar and Dr. 
David Z. Pan (the "Author").

1. Redistributions of any code, with or without modification (the "Code"), 
must be accompanied by any documentation and, each time the resulting 
executable program or a program dependent thereon is launched, a 
prominent display (e.g., splash screen or banner text) of the Author's 
attribution information, which includes:
  (a) Ashutosh Chakraborty, Anurag Kumar, and Dr. David Z. Pan ("AUTHOR"),
  (b) The University of Texas at Austin ("PROFESSIONAL IDENTIFICATION"), and
  (c) http://www.cerc.utexas.edu/utda/ ("URL").

2. Use is limited to academic research groups only. Users who are interested 
in industry or commercial purposes must notify Author and request 
separate license agreement.  

3. Neither the name nor any trademark of the Author may be used to 
endorse or promote products derived from this software without specific 
prior written permission.

4. Users are entirely responsible, to the exclusion of the Author and 
any other persons, for compliance with (a) regulations set by owners or 
administrators of employed equipment, (b) licensing terms of any other 
software, and (c) local, national, and international regulations regarding 
use, including those regarding import, export, and use of encryption software. 

THIS FREE SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR 
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE AUTHOR OR ANY CONTRIBUTOR BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, EFFECTS OF UNAUTHORIZED OR MALICIOUS NETWORK 
ACCESS; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


/*
 * =======================================================================
 *
 *       Filename:  Global.cpp
 *
 *    Description:  Global Placement
 *
 *        Version:  1.0
 *        Created:  11/21/2008 10:59:23 AM CST
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ashutosh Chakraborty (), ashutosh@mail.utexas.edu
 *        Company:  University of Texas, Austin
 *
 * =======================================================================
 */

#include <iostream>
#include <vector>
#include <string>
#include <numeric>
#include <queue>
#include <map>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <cfloat>

using namespace std;

#include <math.h>
#include "Net.h"

#define COMPILING_GLOBAL_PLACER
#include "Design.h"
#undef COMPILING_GLOBAL_PLACER

vector<CNode*> GetConnectedCell(int &numFixed, int &numMovable);
extern int NumMovableCellsHigherThan(vector<CNode *>& v, CNode *n);
static const int gMaxNetFanoutLimitForHessian = 100;

void CheckSoln(QMatrix& hess, Vector& soln, Vector& rhs);

void MinCutForType(int aType, vector<CNode*>& left, vector<CNode*>& right, 
    bool isX, double partitionLocation, double widthToConsider, int disBal);

void MinCutForTypeNew(int aType, vector<CNode*>& left, vector<CNode*>& right, 
    bool isX, double partitionLocation, double widthToConsider, int disBal, 
    int leftId, int rightId);

// Global original hessian matrix.  This is populated in PopulateHessianAndRHS
// and stays constant for the rest of the program (when it is copied into
// successive hessians.  Similarly, the original RHS-X and RHS-Y are also globals.
QMatrix gOrigHessian;
Vector  gRhsX;
Vector  gRhsY;

vector <vector<int> > gPartitions20;
vector <vector<int> > gPartitions22;
vector <vector<int> > gPartitions24;

string makeString(int x)
{
  ostringstream o;
  bool success = o << x;

  if (false == success)
    cout << "EROR:  Could not convert " << x << " to string\n";

  return o.str();
}

void InitPartitionSizes()
{
  cout << "STATUS:  Initializing partition sizes\n";
  vector<int> dummy;

  // For size 20
  dummy.clear();
  dummy.push_back(10);
  dummy.push_back(10);
  gPartitions20.push_back(dummy);

  dummy.clear();
  dummy.push_back(5);
  dummy.push_back(5);
  dummy.push_back(5);
  dummy.push_back(5);
  gPartitions20.push_back(dummy);

  dummy.clear();
  dummy.push_back(2);
  dummy.push_back(3);
  dummy.push_back(2);
  dummy.push_back(3);
  dummy.push_back(2);
  dummy.push_back(3);
  dummy.push_back(2);
  dummy.push_back(3);
  gPartitions20.push_back(dummy);

  dummy.clear();
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(2);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(2);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(2);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(2);
  gPartitions20.push_back(dummy);

  dummy.clear();
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  gPartitions20.push_back(dummy);


  // For size 22
  dummy.clear();
  dummy.push_back(11);
  dummy.push_back(11);
  gPartitions22.push_back(dummy);

  dummy.clear();
  dummy.push_back(5);
  dummy.push_back(6);
  dummy.push_back(5);
  dummy.push_back(6);
  gPartitions22.push_back(dummy);

  dummy.clear();
  dummy.push_back(2);
  dummy.push_back(3);
  dummy.push_back(3);
  dummy.push_back(3);
  dummy.push_back(2);
  dummy.push_back(3);
  dummy.push_back(3);
  dummy.push_back(3);
  gPartitions22.push_back(dummy);

  dummy.clear();
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(2);
  dummy.push_back(1);
  dummy.push_back(2);
  dummy.push_back(1);
  dummy.push_back(2);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(2);
  dummy.push_back(1);
  dummy.push_back(2);
  dummy.push_back(1);
  dummy.push_back(2);
  gPartitions22.push_back(dummy);

  dummy.clear();
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  gPartitions22.push_back(dummy);


  // For size 24
  dummy.clear();
  dummy.push_back(12);
  dummy.push_back(12);
  gPartitions24.push_back(dummy);

  dummy.clear();
  dummy.push_back(6);
  dummy.push_back(6);
  dummy.push_back(6);
  dummy.push_back(6);
  gPartitions24.push_back(dummy);

  dummy.clear();
  dummy.push_back(3);
  dummy.push_back(3);
  dummy.push_back(3);
  dummy.push_back(3);
  dummy.push_back(3);
  dummy.push_back(3);
  dummy.push_back(3);
  dummy.push_back(3);
  gPartitions24.push_back(dummy);

  dummy.clear();
  dummy.push_back(1);
  dummy.push_back(2);
  dummy.push_back(1);
  dummy.push_back(2);
  dummy.push_back(1);
  dummy.push_back(2);
  dummy.push_back(1);
  dummy.push_back(2);
  dummy.push_back(1);
  dummy.push_back(2);
  dummy.push_back(1);
  dummy.push_back(2);
  dummy.push_back(1);
  dummy.push_back(2);
  dummy.push_back(1);
  dummy.push_back(2);
  gPartitions24.push_back(dummy);

  dummy.clear();
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(1);
  dummy.push_back(0);
  dummy.push_back(1);
  dummy.push_back(1);
  gPartitions24.push_back(dummy);

  dummy.clear();

  for (int i = 0; i < gPartitions24.size(); i++)
  {
    dummy = gPartitions20[i];
    assert(20 == std::accumulate(gPartitions20[i].begin(), gPartitions20[i].end(), 0));
    assert(22 == std::accumulate(gPartitions22[i].begin(), gPartitions22[i].end(), 0));
    assert(24 == std::accumulate(gPartitions24[i].begin(), gPartitions24[i].end(), 0));
  }

  cout << "STATUS:  Done\n";
}

void MinCutOverVectors(vector<CNode*>& left, vector<CNode*>& right,
                       bool isX, double partitionLocation, double widthToConsider, int disBal)
{

  cout << "STATUS:  Performing min cut on vectors of size " <<
  left.size() << " " << right.size() << endl;

  if (left.size() == 0 || right.size() == 0)
    return;

  int leftId;

  int rightId;

  if (true == isX)
  {
    leftId = left[0]->PartitionIDX();
    rightId = right[0]->PartitionIDX();
  }
  else
  {
    leftId = left[0]->PartitionIDY();
    rightId = right[0]->PartitionIDY();
  }

  const int multiplier = gMaxNetFanoutLimitForHessian;

  vector<pair< pair<CNode*, CNode*>, int> > pairsToPrint;


  // Map of all nodes present in the two vectors we have.  This will
  // be used to determine true neighbors for partitioning purpose
  map <CNode*, bool> mapNodesPresent;

  map <CNode*, bool> mapNodesNotPresent;

  for (int i = 0; i < left.size(); i++)
  {
    if (isX)
    {
      if (fabs(partitionLocation - left[i]->X()) <= widthToConsider)
      {
        assert(mapNodesPresent.find(left[i]) == mapNodesPresent.end());
        mapNodesPresent[left[i]] = true;
        //         cout << " added : " << left[i]->X() << endl;
      }
    }
    else
    {
      if (fabs(partitionLocation - left[i]->Y()) <= widthToConsider)
      {
        assert(mapNodesPresent.find(left[i]) == mapNodesPresent.end());
        mapNodesPresent[left[i]] = true;
        //         cout << " added : " << left[i]->Y() << endl;
      }
    }
  }

  for (int i = 0; i < right.size(); i++)
  {
    if (isX)
    {
      if (fabs(right[i]->X() - partitionLocation) <= widthToConsider)
      {
        assert(mapNodesPresent.find(right[i]) == mapNodesPresent.end());
        mapNodesPresent[right[i]] = true;
        //         cout << " added : " << right[i]->X() << endl;
      }
    }
    else
    {
      if (fabs(right[i]->Y() - partitionLocation) <= widthToConsider)
      {
        assert(mapNodesPresent.find(right[i]) == mapNodesPresent.end());
        mapNodesPresent[right[i]] = true;
        //         cout << " added : " << right[i]->Y() << endl;
      }
    }
  }

  int pairsInPartition = 0;

  int pairsNotInPartition = 0;
  cout << "  - Doing left\n";

  for (int i = 0; i < left.size(); i++)
  {
    CNode *n = left[i];
    //     cout << n->Name() << endl;

    if (mapNodesPresent.find(n) == mapNodesPresent.end())
      continue;

    vector<CNode*> neigh = n->GetNeighborCells();

    vector<double> wts   = n->GetNeighborWts();

    for (int j = 0; j < neigh.size(); j++)
    {
      if (mapNodesPresent.find(neigh[j]) != mapNodesPresent.end())
      {
        //         cout << " added : " << n->Name() << " " << neigh[j]->Name() << endl;
        pairsToPrint.push_back(make_pair(make_pair(n, neigh[j]), int(wts[j]*multiplier)));
        pairsInPartition += 1;
      }
      else
      {
        //         cout << " added external node: " << n->Name() << " " << neigh[j]->Name() << endl;
        pairsToPrint.push_back(make_pair(make_pair(n, neigh[j]), int(wts[j]*multiplier)));
        mapNodesNotPresent[neigh[j]] = true;
        pairsNotInPartition += 1;
      }
    }
  }

  cout << "  - Doing right\n";

  for (int i = 0; i < right.size(); i++)
  {
    CNode *n = right[i];
    //     cout << n->Name() << endl;

    if (mapNodesPresent.find(n) == mapNodesPresent.end())
      continue;

    vector<CNode*> neigh = n->GetNeighborCells();

    vector<double> wts   = n->GetNeighborWts();

    for (int j = 0; j < neigh.size(); j++)
    {
      if (mapNodesPresent.find(neigh[j]) != mapNodesPresent.end())
      {
        pairsToPrint.push_back(make_pair(make_pair(n, neigh[j]), int(wts[j]*multiplier)));
        pairsInPartition += 1;
      }
      else
      {
        //         cout << " added external node: " << n->Name() << " " << neigh[j]->Name() << endl;
        pairsToPrint.push_back(make_pair(make_pair(n, neigh[j]), int(wts[j]*multiplier)));
        mapNodesNotPresent[neigh[j]] = true;
        pairsNotInPartition += 1;
      }
    }
  }

  std::ofstream fixFile;

  fixFile.open("partition.fix");

  if (false == fixFile)
    cout << "EROR:  Could not write to file partition.fix\n";

  map <CNode*, int> mapNodeToSeqInt;

  map <int, CNode*> mapSeqIntToNode;

  map <CNode*, bool>::iterator it = mapNodesPresent.begin();

  int currAdded = 0;

  while (it != mapNodesPresent.end())
  {
    mapNodeToSeqInt[it->first] = currAdded;
    mapSeqIntToNode[currAdded] = it->first;
    currAdded++;
    it++;
    fixFile << "-1" << endl;
  }

  it = mapNodesNotPresent.begin();

  while (it != mapNodesNotPresent.end())
  {
    mapNodeToSeqInt[it->first] = currAdded;
    //     mapSeqIntToNode[currAdded] = it->first;

    if (isX)
    {
      if (it->first->X() < partitionLocation)
        fixFile << "0" << endl;
      else
        fixFile << "1" << endl;
    }
    else
    {
      if (it->first->Y() < partitionLocation)
        fixFile << "0" << endl;
      else
        fixFile << "1" << endl;
    }

    currAdded++;

    it++;
  }

  fixFile.close();

  cout << "  - Printing to hgr file now\n";
  std::ofstream out;
  out.open("partition.hgr");

  if (false == out)
    cout << "EROR:  Could not write to file partition.hgr\n";

  //   cout << "pairsToPrint.size() is " << pairsToPrint.size() << endl;
  assert(pairsInPartition % 2 == 0);

  cout << "Movable Nodes in partition file: " << mapNodesPresent.size() << endl;
  cout << "Unmovable Nodes in partition file: " << mapNodesNotPresent.size() << endl;
  cout << "Nets in partition file: " << (pairsInPartition / 2) + pairsNotInPartition << endl;

  out << (pairsInPartition / 2) + pairsNotInPartition << " " << 
    mapNodesPresent.size() + mapNodesNotPresent.size() << " 1\n";

  for (int i = 0; i < pairsToPrint.size(); i++)
  {
    CNode *left = pairsToPrint[i].first.first;
    CNode *right = pairsToPrint[i].first.second;
    int wt = pairsToPrint[i].second;
    assert(wt > 0);

    if (mapNodeToSeqInt[left] < mapNodeToSeqInt[right])
      out << wt << " " << mapNodeToSeqInt[left] + 1 << " " <<
      mapNodeToSeqInt[right] + 1 << endl;
  }

  out.close();

  string cmd = "./bin/shmetis partition.hgr partition.fix 2 ";
  cmd += makeString(disBal);
  cmd += " >& /dev/null";
  cout << "  - Running " << cmd.c_str() << endl;

  cout << "hmetis starts ... \n";
  system(cmd.c_str());
  cout << "hmetis ends ... \n";

//   std::ifstream in;
//   in.open("partition.hgr.part.2");

//   if (false == in)
//   {
//     cout << "EROR:  Cannot read output file of hmetis\n";
//     abort();
//   }

  std::ifstream in, test1, test2;

  test2.open("partition.hgr.part.2");

  if (test2)
  {
    //     cout << "    Two partitions were made\n";
    in.open("partition.hgr.part.2");
  }
  else
  {
    test1.open("partition.hgr.part.1");

    if (test1)
    {
      //       cout << "    One partitions were made.  Will use it\n";
      in.open("partition.hgr.part.1");
    }
    else
    {
      cout << "    Could not open 1 or 2 partition's out file\n";
      return;
    }
  }

  assert(in);

  string line;

  int currIdx = 0;

  while (false == in.eof())
  {
    getline(in, line);

    if (line == "")
      continue;

    int partnNum = atoi(line.c_str());

    CNode *n = mapSeqIntToNode[currIdx];

    //     cout << currIdx << " should be " << mapNodeToSeqInt[mapSeqIntToNode[currIdx]] << endl;
    if (true == isX)
      n->PartitionIDX(partnNum == 0 ? leftId : rightId);
    else
      n->PartitionIDY(partnNum == 0 ? leftId : rightId);

    //     cout << "Node " << n->Name() << " got Parn Id = " << n->PartitionIDX() << endl;
    currIdx++;

    if (currIdx == mapNodesPresent.size())
      break;
  }

  assert(currIdx == mapNodesPresent.size());

  system("mv partition.hgr partition.hgr.old");
  system("mv partition.hgr.part.2 partition.hgr.part.2.old");
  system("mv partition.fix partition.fix.old");
}

void PrintMatrix(QMatrix& q)
{
  cout << "Printing Hessian\n";
  int dim = Q_GetDim(&q);

  for (int i = 0; i < dim; i++)
  {
    for (int j = 0; j < dim; j++)
    {
      //       cout << Q_GetEl(&q, i+1, j+1) << " ";
      printf("%4.2f ", Q_GetEl(&q, i + 1, j + 1));
      assert(Q_GetEl(&q, i + 1, j + 1) == Q_GetEl(&q, j + 1, i + 1));
    }

    cout << endl;
  }
}

void CopyMatrix(QMatrix& q, QMatrix &v)
{
  int dim = Q_GetDim(&q);

  Q_Constr(&v, "abcd", dim, True,   Rowws, Normal, True);

  for (int i = 0; i < dim; i++)
  {
    int numItems = Q_GetLen(&q, i + 1);
    Q_SetLen(&v, i + 1, numItems);

    for (int j = 0; j < numItems; j++)
    {
      int currPos = Q_GetPos(&q, i + 1, j);
      double currVal = Q_GetVal(&q, i + 1, j);
      Q_SetEntry(&v, i + 1, j, currPos, currVal);
    }
  }
}

void copyVector(Vector &q, Vector &v)
{
  int size = V_GetDim(&q);
  V_Constr(&v,  "newrhsY", size, Normal, True);

  for (int i = 0; i < size; i++)
  {
    V_SetCmp(&v, i + 1, V_GetCmp(&q, i + 1));
  }
}

bool compareVector(Vector &q, Vector &v)
{
  assert(V_GetDim(&q) == V_GetDim(&v));

  int size = V_GetDim(&q);
  for (int i = 0; i < size; i++)
  {
    assert(V_GetCmp(&q, i + 1) == V_GetCmp(&v, i + 1));
  }
}

bool compareMatrix(QMatrix &q, QMatrix &v)
{
  assert(Q_GetDim(&q) == Q_GetDim(&v));

  int size = Q_GetDim(&q);
  for (int i = 0; i < size; i++)
  {
    int length = Q_GetLen(&q, i + 1);
    assert(Q_GetLen(&q, i + 1) == Q_GetLen(&v, i + 1));

    for (int j = 0; j < length; j++)
    {
      int currPos = Q_GetPos(&q, i + 1, j);
      assert(Q_GetEl(&q, i + 1, currPos) == Q_GetEl(&v, i + 1, currPos));
    }
  }
}

void PrintMatrixNonZero(QMatrix& q, string outFile, bool invert)
{
  cout << "Printing Hessian\n";
  std::ofstream out;
  out.open(outFile.c_str());
  int dim = Q_GetDim(&q);

  int currPos;

  for (int i = 0; i < dim; i++)
  {
    int  length = Q_GetLen(&q, i + 1);
    if (true == invert)
    {
      for (int j = length - 1; j >= 0; j--)
      {
        currPos = Q_GetPos(&q, i + 1, j);
        out << Q_GetVal(&q, i + 1, j) << "," << currPos << " " ;
      }
    }
    else
    {
      for (int j = 0; j < length; j++)
      {
        currPos = Q_GetPos(&q, i + 1, j);
        out << Q_GetVal(&q, i + 1, j) << "," << currPos << " " ;
      }
    }
    out << endl;
  }
  out.close();
}

void PrintVector(Vector& v)
{
  cout << "Printing Vector\n";
  int dim = V_GetDim(&v);

  for (int i = 0; i < dim; i++)
  {
    cout << V_GetCmp(&v, i + 1) << endl;
  }
}

void CDesign::PopulateHessianAndRHS()
{
  cout << "STATUS:  Populating Hessian and RHS vectors\n";

  int numNonZeroInHessian = 0;

  for (int i = 0; i < mNodesInHessian.size(); i++)
  {
    CNode *tmpNode = mNodesInHessian[i];

    vector<CNode *> neigh = tmpNode->GetNeighborCells();
    vector<double>  wts   = tmpNode->GetNeighborWts();
    assert(neigh.size() == wts.size());

    int numHigherMovableNeighbor = NumMovableCellsHigherThan(neigh, tmpNode);

    Q_SetLen(&gOrigHessian, i + 1, numHigherMovableNeighbor + 1);
    // Neighbour (higher) entries

    double cumulFixedXPosn = 0.0;
    double cumulFixedYPosn = 0.0;
    int fixedNeighborSeen = 0;

    double myTotalWtConn = std::accumulate(wts.begin(), wts.end(), 0.0);

    double totalWtConn = 0.0;
    int countOfEntries = 0;
    for (int j = 0; j < neigh.size(); j++)
    {
      assert(neigh[j]->GetCurLocInHess() != tmpNode->GetCurLocInHess());
      totalWtConn += wts[j];

      if (true == neigh[j]->Fixed())
      {
        fixedNeighborSeen++;
        cumulFixedXPosn += wts[j] * neigh[j]->X();
        cumulFixedYPosn += wts[j] * neigh[j]->Y();
        continue;
      }

      assert(false == neigh[j]->Fixed());

      if (neigh[j]->GetCurLocInHess() < tmpNode->GetCurLocInHess())
        continue;

      assert(wts[j] > 0.001);

      assert(0 == Q_GetEl(&gOrigHessian, i + 1, neigh[j]->GetCurLocInHess() + 1));
      Q_SetEntry(&gOrigHessian, i + 1, countOfEntries,
                 neigh[j]->GetCurLocInHess() + 1, -1*wts[j]);
      assert(Q_GetEl(&gOrigHessian, i + 1, neigh[j]->GetCurLocInHess() + 1) == -1*wts[j]);

      assert(i < neigh[j]->GetCurLocInHess());

      numNonZeroInHessian++;
      countOfEntries++;
    }

    assert(totalWtConn > 0);
    assert(countOfEntries == numHigherMovableNeighbor);
    assert(myTotalWtConn == totalWtConn);

    Q_SetEntry(&gOrigHessian, i + 1, countOfEntries, i + 1, totalWtConn);
    numNonZeroInHessian++;

    assert(fabs(totalWtConn) > 0.1);

    V_SetCmp(&gRhsX, i + 1, cumulFixedXPosn);
    V_SetCmp(&gRhsY, i + 1, cumulFixedYPosn);
  }

// #define _CHECK
#ifdef _CHECK
  for (int i = 0; i < mNodesInHessian.size(); i++)
  {
    bool negFound = false;
    bool posFound = false;

    for (int j = 0; j < mNodesInHessian.size(); j++)
    {
      if (Q_GetEl(&gOrigHessian, i + 1, j + 1) > 0)
      {
        posFound = true;
        assert(i + 1 == j + 1);

        if (negFound == true)
          break;
      }
      else if (Q_GetEl(&gOrigHessian, i + 1, j + 1) < 0)
      {
        negFound = true;

        if (posFound == true)
          break;
      }
    }

    assert(negFound == true);

    assert(posFound == true);
  }

#endif

  cout << "  " << numNonZeroInHessian << " non-zero entries in Hessian\n";
}

#if 0
void CDesign::PopulateHessianAndRHS(QMatrix& q, Vector& rhsX, Vector& rhsY)
{
  cout << "Entering " << __PRETTY_FUNCTION__ << endl;

  for (int i = 0; i < mAllNodes.size(); i++)
  {
    CNode *tmpNode = mAllNodes[i];

    if (true == tmpNode->Fixed())
      continue;

    int numMovable = 0;

    int numFixed = 0;

    vector<CNode*> connectedCells =
      tmpNode->GetConnectedCellSorted(numMovable, numFixed);

    vector<int> indexOfConnCell(numMovable);

    vector<int> wtOfConnection(numMovable);

    int lengthOfRow = 0;

    for (int j = 0; j < numMovable; j++)
    {
      if (connectedCells[j]->ID() > tmpNode->ID())
        lengthOfRow += 1;
    }

    //     cout << "length of Row " << lengthOfRow << endl;

    Q_SetLen(&q, i + 1, lengthOfRow + 1);

    // Self entry
    Q_SetEntry(&q, i + 1, 0, i + 1,
               numFixed + numMovable);

    //     cout << "(" << i + 1 << "," << i + 1 << ")" <<
    //       Q_GetVal(&A, i + 1, 0) << "\n ";

    // Connected movable cell's entries
    int countOfEntries = 0;

    for (int other = 0; other < numMovable; other++)
    {
      if (connectedCells[other]->ID() <= tmpNode->ID())
        continue;

      Q_SetEntry(&q, i + 1, countOfEntries + 1,
                 connectedCells[other]->ID() + 1, -1);

      countOfEntries += 1;

      assert(false == connectedCells[other]->Fixed());
    }

    // Connected fixed cell's entries
    double accumFixedPosX = 0.0;

    double accumFixedPosY = 0.0;

    for (int other = 0; other < numFixed; other++)
    {
      accumFixedPosX += connectedCells[other + numMovable]->X();
      accumFixedPosY += connectedCells[other + numMovable]->Y();
      assert(true == connectedCells[other+numMovable]->Fixed());
    }

    V_SetCmp(&rhsX, i + 1, accumFixedPosX);

    V_SetCmp(&rhsY, i + 1, accumFixedPosY);
  }
}

#endif

void CDesign::SolveHessnStoreResults(QMatrix& hess, Vector& soln, Vector& rhs)
{
  cout << "STATUS:  Solving Hessian and Storing Results\n";

  int matDim = Q_GetDim(&hess);

  for (int i = 0; i < matDim; i++)
    assert(0 < Q_GetEl(&hess, i + 1, i + 1));

  QMatrix q;
  Vector  v;

  CopyMatrix(hess, q);
  copyVector(rhs, v);
  // #define _USE_ILUP
#define _USE_SSORP
#ifdef _USE_ILUP
  cout << "  Using ILUPrecond\n";

  CGIter(&q, &soln, &v, 100, ILUPrecond, 1.2);

#else
#ifdef _USE_SSORP
  cout << "  Using SSORPrecond\n";

  CGIter(&q, &soln, &v, 100, SSORPrecond, 1.2);

#endif
#endif

  if (LASResult() != LASOK)
  {
    cout << "  Solver Failed\n";
    WriteLASErrDescr(stderr);

    CGIter(&hess, &soln, &rhs, 100, ILUPrecond, 1.2);
    if (LASResult() != LASOK)
    {
      //     cout << " JacobiPrecond failed! " << endl;
      WriteLASErrDescr(stderr);
      //         CGIter(&hess, &soln, &rhs, 100, JacobiPrecond, 1.2);
      //     if (LASResult() != LASOK)
      //     {
      //        cout << " SSORPrecond failed! " << endl;
      //           WriteLASErrDescr(stderr);
      //     }
      abort();
    }
  }

  assert(0 < Q_GetDim(&hess));

  assert(0 < V_GetDim(&rhs));
  assert(0 < V_GetDim(&soln));

  CheckSoln(hess, soln, rhs);
}

void CDesign::UpdatePosn(Vector& solnX, Vector& solnY, bool updateX, bool updateY)
{
  cout << "Entering " << __PRETTY_FUNCTION__ << endl;

  int numVal = Q_GetDim(&gOrigHessian);
  assert(numVal == mNodesInHessian.size());
  assert(numVal <= V_GetDim(&solnX));
  assert(numVal <= V_GetDim(&solnY));
  double val = 0.0;

  for (int i = 0; i < numVal; i++)
  {
    assert(false == mNodesInHessian[i]->Fixed());

    if (true == updateX)
    {
      val = V_GetCmp(&solnX, i + 1);
#if 0

      if (val < mIOSpaceX)
        val = mIOSpaceX;
      else if (val > mCoreWidth + mIOSpaceX)
        val = mCoreWidth + mIOSpaceX;

#endif
      mNodesInHessian[i]->X(val);
    }

    if (true == updateY)
    {
      val = V_GetCmp(&solnY, i + 1);
#if 0

      if (val < mIOSpaceY)
        val = mIOSpaceY;
      else if (val > mCoreHeight + mIOSpaceY)
        val = mCoreHeight + mIOSpaceY;

#endif
      mNodesInHessian[i]->Y(val);
    }
  }

  cout << endl;

  if (mNodesInHessian.size() != V_GetDim(&solnX))
  {
    cout << "Printing Lambda X and Y\n";

    for (int i = mNodesInHessian.size(); i < V_GetDim(&solnX); i++)
      cout << V_GetCmp(&solnX, i + 1) << " " << V_GetCmp(&solnY, i + 1) << endl;

    cout << "Lambda printed\n";

    cout << endl;
  }
}

// Takes current X and Y partition sizes and updates the hessian
// corresponding by center of gravity constraints.
void UpdateHessian(vector<int> aXPartnSizes, vector<int>aYPartnSizes)
{
  // Use previous hessian to copy into new one
  // Add new rows and columns in it based on current x/y position of cells
}


// Function related to neighbor computation.  For computing
// the hessian, we store the neighbor nodes to each node
// in the vector mNeighbors.  Following function performs this.
//
// NOTE:  If a net has more than gMaxNetFanoutLimitForHessian
// nodes in it, neighbor are not computed for such a net.  This
// is to make the solver faster and under the assumption that
// such nets would be very few so screwing their wirelength
// is immaterial.
//
// Function AssignNeighbors() uses the functions ClearNeighbors()
// AddNeighbor() and GetNumNeighbor() of the CNode class to assign
// neighbors.
int CDesign::AssignNeighbors()
{
  cout << "STATUS:  Assigning neighbors\n";

  static bool alreadyCalled = false;

  if (false == alreadyCalled)
    alreadyCalled = true;
  else
    cout << "WARN: AssignNeighbors() running second time.  All OK?\n";

  // Clear the existing neighbors of each cell
  for (int i = 0; i < mAllNodes.size(); i++)
    mAllNodes[i]->ClearNeighbors();

  // Count number of neighbours added and those which were ignored
  int numNeighborAdded = 0;

  int numNeighborIgnor = 0;

  // Go one net by one net.  Append to each node in the net, the address
  // of all other nodes.  This builds up the neighbor list.
  for (int currNetCnt = 0; currNetCnt < mAllNets.size(); currNetCnt++)
  {
    CNet *tmpNet = mAllNets[currNetCnt];

    vector<CNode*> nodes = tmpNet->GetCellsInNet();

    // If a net has more than gMaxNetFanoutLimitForHessian nets, ignore
    // its neighbor update step (as it is too big and would make solver slow)

    if (gMaxNetFanoutLimitForHessian < nodes.size())
    {
      numNeighborIgnor += nodes.size() * (nodes.size() - 1);
      continue;
    }
    else
    {
      numNeighborAdded += nodes.size() * (nodes.size() - 1);
    }

    // Use star model for now.  For a net with 'p' nodes, the weight
    // of each of the edges is equal to 2/p so that the total weight
    // is (p-1) over all the star nets.
    assert(nodes.size() != 1);

    assert(nodes.size() > 0);

    double currNetWt = 1.0 / (nodes.size() - 1);

    for (int currNodCnt = 0; currNodCnt < nodes.size(); currNodCnt++)
    {
      // We are building up the neighbor list of currNode
      CNode *currNode = nodes[currNodCnt];

      // Go over all the nodes in vector nodes and insert those which
      // are not currNode itself.

      for (int iterNodeCnt = 0; iterNodeCnt < nodes.size(); iterNodeCnt++)
        //         if (iterNodeCnt != currNodCnt)
        if (currNode->ID() != nodes[iterNodeCnt]->ID())
          currNode->AddNeighbor(nodes[iterNodeCnt], currNetWt);
    }
  }

  for (int i = 0; i < mAllNodes.size(); i++)
  {
    //     if (true == mAllNodes[i]->Fixed())
    //       continue;
    mAllNodes[i]->UniquifyNeighbor();
  }

#ifdef _DEBUG
  int testNumNeighbor = 0;

  for (int i = 0; i < mAllNodes.size(); i++)
    testNumNeighbor += mAllNodes[i]->GetNumNeighbor();

  assert(numNeighborAdded == testNumNeighbor);

#endif

  cout << "        - Added " << numNeighborAdded << " neighbors " <<
  "and IGNORED " << numNeighborIgnor << " possible neighbor pairs\n";

  return numNeighborAdded;
}

static const double gRTCAccuracyForSolver = 1e-15;
void AllocateHessAndVectors(const int aSize, Vector& solX, Vector& solY)
{
  cout << "STATUS:  Allocating Hess and Vector space\n";

  if (aSize == 0)
    cout << "WARN:  Zero sized hessian requested\n";

  static int hessNum = 0;
  hessNum++;

  // Each time, generate different names by appending the stringified
  // version of hessNum
  string iterName = makeString(hessNum);
  string hessName = string("hess") + iterName;
  string solXName = string("solX") + iterName;
  string solYName = string("solY") + iterName;
  string rhsXName = string("rhsX") + iterName;
  string rhsYName = string("rhsY") + iterName;

  SetRTCAccuracy(gRTCAccuracyForSolver);

  // Q is of size aSize, symmetric matrix
  Q_Constr(&gOrigHessian, (char*)hessName.c_str(), aSize, True,   Rowws, Normal, True);
  V_Constr(&solX, (char*)solXName.c_str(), aSize, Normal, True);
  V_Constr(&solY, (char*)solYName.c_str(), aSize, Normal, True);
  V_Constr(&gRhsX, (char*)rhsXName.c_str(), aSize, Normal, True);
  V_Constr(&gRhsY, (char*)rhsYName.c_str(), aSize, Normal, True);
  V_SetAllCmp(&solX, 0.0);
  V_SetAllCmp(&solY, 0.0);
  V_SetAllCmp(&gRhsX, 0.0);
  V_SetAllCmp(&gRhsY, 0.0);
}

bool CDesign::DoGlobalPlaceFixPoint()
{
  for (int i = 0; i < 5; i++)
    cout << "\n";

  cout << "******* STARTING GLOBAL PLACE *******\n";

  cout << "\n";

  InitPartitionSizes();
  AssignNeighbors();

  int numInHessian = GetCellsToPutInHessian();

  assert(mNodesInHessian.size() > 0);

  // Make hessian of size equal to movable cells in the circuit
  int hessSize = mNodesInHessian.size();

  Vector  solX, solY;

  // Allocate the space for hessian and the vectors
  AllocateHessAndVectors(hessSize, solX, solY);

  // Populate Hessian and the rhs vectors
  PopulateHessianAndRHS();

  //   PrintMatrixNonZero (gOrigHessian, "orig.txt", false);
  // Call the solver for first time with the global matrix and rhs
  SolveHessnStoreResults(gOrigHessian, solX, gRhsX);
  SolveHessnStoreResults(gOrigHessian, solY, gRhsY);

  if (Q_GetDim(&gOrigHessian) == 0)
    cout << "ERROR happened\n";

  //   PrintMatrix(gOrigHessian);
  //   PrintVector(solX);
  //   PrintVector(solY);
  //   PrintVector(gRhsX);
  //   PrintVector(gRhsY);

  UpdatePosn(solX, solY, true, true);

  GetHPWL("Initial");
  WriteImage("Initial");
  bool done = false;

  int currPartnIterNum = 0;
  while (false == done)
  {
    // Make these variables local so that they destroy and recreate each time
    QMatrix newHessX, newHessY;
    Vector  newrhsX, newrhsY, asolnX, asolnY;
    vector<int> numCellInPartnX, numCellInPartnY;
    vector<double> xCoMForPartns, yCoMForPartns;

#if 1
    AssignPartnID(currPartnIterNum, numCellInPartnX, numCellInPartnY,
                  xCoMForPartns, yCoMForPartns);
#else
    AssignPartnIDNew(currPartnIterNum, xCoMForPartns, yCoMForPartns);
#endif

    V_Constr(&asolnX,     "asolnX", hessSize , Normal, True);
    V_Constr(&asolnY,     "asolnY", hessSize , Normal, True);

    RecomputeHessFixPoint(numCellInPartnX, xCoMForPartns, newHessX, newrhsX, true);
    RecomputeHessFixPoint(numCellInPartnY, yCoMForPartns, newHessY, newrhsY, false);

    //   cout << " Comparing with X matrix\n";
    //   compareMatrix(gOrigHessian, newHessX);
    //   cout << " Comparing with Y matrix\n";
    //   compareMatrix(gOrigHessian, newHessY);

    //   cout << " Comparing with X rhs\n";
    //   compareVector(gRhsX, newrhsX);
    //   cout << " Comparing with Y rhs\n";
    //   compareVector(gRhsY, newrhsY);

    assert(0 < Q_GetDim(&newHessX) * Q_GetDim(&newHessY) *
           V_GetDim(&newrhsX) * V_GetDim(&newrhsY));

    // Solve actual matrices
    SolveHessnStoreResults(newHessX, asolnX, newrhsX);
    assert(0 < Q_GetDim(&newHessX)*V_GetDim(&asolnX)*V_GetDim(&newrhsX));
    WriteLASErrDescr(stdout);

    SolveHessnStoreResults(newHessY, asolnY, newrhsY);
    assert(0 < Q_GetDim(&newHessY)*V_GetDim(&asolnY)*V_GetDim(&newrhsY));
    WriteLASErrDescr(stdout);

    //     PrintMatrix(newHessX);
    //     PrintVector(asolnX);
    //     PrintVector(newrhsX);
    //     PrintVector(asolnY);
    //     PrintVector(newrhsY);
    //     exit(0);

    UpdatePosn(asolnX, asolnY, true, true);
    WriteImage(string("afterIter") + makeString(currPartnIterNum + 1));
//     WriteNetsFile(string("afterIter") + makeString(currPartnIterNum + 1) + ".nets");
    GetHPWL(string("afterIter") + makeString(currPartnIterNum + 1));

    PrintCoM(xCoMForPartns, yCoMForPartns);

    if (Q_GetDim(&newHessX) == 0 || 0 == Q_GetDim(&newHessY))
      cout << "ERROR happened\n";

    currPartnIterNum++;

    if (5 == currPartnIterNum)
      done = true;
  }

  return true;
}

#if 0
bool CDesign::DoGlobalPlace()
{
  InitPartitionSizes();
  int numNeighborAdded = AssignNeighbors();

  int numInHessian = GetCellsToPutInHessian();
  assert(mNodesInHessian.size() > 0);

  // Make hessian of size equal to movable cells in the circuit
  int hessSize = mNodesInHessian.size();
  Vector  solX;
  Vector  solY;

  // Allocate the space for hessian and the vectors
  AllocateHessAndVectors(hessSize, solX, solY);

  // Populate Hessian and the rhs vectors
  PopulateHessianAndRHS();

  // Call the solver for first time with the global matrix and rhs
  SolveHessnStoreResults(gOrigHessian, solX, gRhsX);
  SolveHessnStoreResults(gOrigHessian, solY, gRhsY);

  if (Q_GetDim(&gOrigHessian) == 0)
    cout << "ERROR happened\n";

  //   PrintMatrix(gOrigHessian);
  //   PrintVector(solX);
  //   PrintVector(solY);
  //   PrintVector(gRhsX);
  //   PrintVector(gRhsY);

  UpdatePosn(solX, solY, true, true);

  //   GetHPWL("AfterFirstPartition");

  WriteImage("afterFirstIter");

  bool done = false;

  int currPartnIterNum = 0;

  while (false == done)
  {
    // Make these variables local so that they destroy and recreate each time
    QMatrix newHessX, newHessY;
    Vector  newrhsX, newrhsY, asolnX, asolnY;
    vector<int> numCellInPartnX, numCellInPartnY;
    vector<double> xCoMForPartns, yCoMForPartns;

    AssignPartnID(currPartnIterNum, numCellInPartnX, numCellInPartnY,
                  xCoMForPartns, yCoMForPartns);

    V_Constr(&asolnX,     "asolnX", hessSize + xCoMForPartns.size(), Normal, True);
    V_Constr(&asolnY,     "asolnY", hessSize + yCoMForPartns.size(), Normal, True);

    RecomputeHess(numCellInPartnX, xCoMForPartns, newHessX, newrhsX, true);
    RecomputeHess(numCellInPartnY, yCoMForPartns, newHessY, newrhsY, false);

    assert(0 < Q_GetDim(&newHessX) * Q_GetDim(&newHessY) *
           V_GetDim(&newrhsX) * V_GetDim(&newrhsY));

    // Solve actual matrices
    SolveHessnStoreResults(newHessX, asolnX, newrhsX);
    assert(0 < Q_GetDim(&newHessX)*V_GetDim(&asolnX)*V_GetDim(&newrhsX));

    SolveHessnStoreResults(newHessY, asolnY, newrhsY);
    assert(0 < Q_GetDim(&newHessY)*V_GetDim(&asolnY)*V_GetDim(&newrhsY));

    UpdatePosn(asolnX, asolnY, true, true);
    WriteImage(string("afterIter") + makeString(currPartnIterNum + 1));
    WriteNetsFile(string("afterIter") + makeString(currPartnIterNum + 1) + ".nets");
    PrintCoM(xCoMForPartns, yCoMForPartns);

    if (Q_GetDim(&newHessX) == 0 || 0 == Q_GetDim(&newHessY))
      cout << "ERROR happened\n";

    currPartnIterNum++;

    if (5 == currPartnIterNum)
      done = true;
  }

  return true;
}
#endif


// With partitioning done
void CDesign::RecomputeHess(vector<int>& aNodesInPartn, vector<double>& aCoMPartn,
                            QMatrix& newHess, Vector& newrhs, bool isForX)
{
  static int numTimesEntered = 0;
  numTimesEntered++;

  assert(aNodesInPartn.size() == aCoMPartn.size());
  int numPartn = aCoMPartn.size();

  cout << "STATUS:  Performing RecomputeHessian().  " << numTimesEntered <<
  " time.  # Partns = " << numPartn << endl;

  int origDim = Q_GetDim(&gOrigHessian);
  int newDim = origDim + numPartn;

  if (true == isForX)
  {
    Q_Constr(&newHess, "newHessX", newDim, True,   Rowws, Normal, True);
    V_Constr(&newrhs,  "newrhsY", newDim, Normal, True);
  }
  else
  {
    Q_Constr(&newHess, "newHessY", newDim, True,   Rowws, Normal, True);
    V_Constr(&newrhs,  "newrhsY", newDim, Normal, True);
  }

  V_SetAllCmp(&newrhs, 0.0);

  // Copy old values from last hessian

  for (int i = 0; i < origDim; i++)
  {
    int numItems = Q_GetLen(&gOrigHessian, i + 1);
    Q_SetLen(&newHess, i + 1, numItems + 1);

    for (int j = 0; j < numItems; j++)
    {
      int currPos = Q_GetPos(&gOrigHessian, i + 1, j);
      double currVal = Q_GetVal(&gOrigHessian, i + 1, j);
      Q_SetEntry(&newHess, i + 1, j, currPos, currVal);

      if (Q_GetDim(&newHess) == 0)
        cout << "****** Hess Size(X) i = " << i << " j = " << j << endl;
    }

    if (true == isForX)
    {
      Q_SetEntry(&newHess, i + 1, numItems,
                 origDim + mNodesInHessian[i]->PartitionIDX() + 1, -0.5);

      if (Q_GetDim(&newHess) == 0)
        cout << "****** Hess Size(X) i = " << i << " location = " <<
        origDim + mNodesInHessian[i]->PartitionIDX() + 1 << endl;

      assert(0 <= mNodesInHessian[i]->PartitionIDX());

      assert(mNodesInHessian[i]->PartitionIDX() < numPartn);
    }
    else
    {
      Q_SetEntry(&newHess, i + 1, numItems,
                 origDim + mNodesInHessian[i]->PartitionIDY() + 1, -0.5);

      if (Q_GetDim(&newHess) == 0)
        cout << "****** Hess Size(X) i = " << i << " location = " <<
        origDim + mNodesInHessian[i]->PartitionIDY() + 1 << endl;

      assert(0 <= mNodesInHessian[i]->PartitionIDY());

      assert(mNodesInHessian[i]->PartitionIDY() < numPartn);
    }

    // Copy RHS parts
    double val = 0.0;

    if (true == isForX)
      val = V_GetCmp(&gRhsX, i + 1);
    else
      val = V_GetCmp(&gRhsY, i + 1);

    V_SetCmp(&newrhs, i + 1, val);
  }

  for (int i = 0; i < numPartn; i++)
    V_SetCmp(&newrhs, origDim + i + 1, -1.0*aNodesInPartn[i]*aCoMPartn[i] / 2.0);

#if 1
  // Assign last entries (of size #partitions) at end of hessian
  for (unsigned i = 0; i < numPartn; i++)
  {
    Q_SetLen(&newHess, origDim + i + 1, 1);
    Q_SetEntry(&newHess, origDim + i + 1, 0, origDim + i + 1, (1e-5));

    if (Q_GetDim(&newHess) == 0)
      cout << "****** Hess Size(X) i = " << i << " location = self entry\n";
  }

#endif
}

//Recompute after fixing a new node to center of mass and connecting all nodes to it
static double gWtToCoM = 1.0;
void CDesign::RecomputeHessFixPoint(vector<int>& aNodesInPartn, vector<double>& aCoMPartn, 
    QMatrix& newHess, Vector& newrhs, bool isForX)
{
  static int numTimesEntered = 0;
  numTimesEntered++;

  //   gWtToCoM = 1.0;
//   assert(aNodesInPartn.size() == aCoMPartn.size());
  int numPartn = aCoMPartn.size();

  cout << "STATUS:  Performing RecomputeHessianFoxPoint().  " << numTimesEntered <<
  " time.  # Partns = " << numPartn << endl;

  int origDim = Q_GetDim(&gOrigHessian);
  int newDim = origDim;

  if (true == isForX)
  {
    Q_Constr(&newHess, "newHessX", newDim, True,   Rowws, Normal, True);
    V_Constr(&newrhs,  "newrhsY", newDim, Normal, True);
  }
  else
  {
    Q_Constr(&newHess, "newHessY", newDim, True,   Rowws, Normal, True);
    V_Constr(&newrhs,  "newrhsY", newDim, Normal, True);
  }

  V_SetAllCmp(&newrhs, 0.0);

  // Copy old values from last hessian

  for (int i = 0; i < origDim; i++)
  {
    int numItems = Q_GetLen(&gOrigHessian, i + 1);
    Q_SetLen(&newHess, i + 1, numItems);

    for (int j = 0; j < numItems; j++)
    {
      int currPos = Q_GetPos(&gOrigHessian, i + 1, j);
      double currVal = Q_GetVal(&gOrigHessian, i + 1, j);

      if (currPos == i + 1)
        Q_SetEntry(&newHess, i + 1, j, currPos, currVal + gWtToCoM);
      else
        Q_SetEntry(&newHess, i + 1, j, currPos, currVal);

      if (Q_GetDim(&newHess) == 0)
        cout << "****** Hess Size(X) i = " << i << " j = " << j << endl;
    }

    // Copy RHS parts
    double val = 0.0;

    if (true == isForX)
      val = V_GetCmp(&gRhsX, i + 1);
    else
      val = V_GetCmp(&gRhsY, i + 1);

    if (val != 0.0)
      V_SetCmp(&newrhs, i + 1, val);
  }

  FILE *fp;
  if (true == isForX)
  {
    fp = fopen("rhsFile", "w");
  }
  for (int i = 0; i < origDim; i++)
  {
    double val = V_GetCmp(&newrhs, i + 1);

    if (true == isForX)
    {
      double newVal = val +  gWtToCoM * aCoMPartn[mNodesInHessian[i]->PartitionIDX()];
      if (newVal != 0.0)
        V_SetCmp(&newrhs, i + 1, newVal);
      double newvalue = V_GetCmp(&newrhs, i + 1);
      fprintf(fp, "rhs : %f\n", newvalue);
    }
    else
    {
      double newVal = val +  gWtToCoM * aCoMPartn[mNodesInHessian[i]->PartitionIDY()];
      if (newVal != 0.0)
        V_SetCmp(&newrhs, i + 1, newVal);
    }
  }
  if (true == isForX)
    fclose(fp);
}

#if 0
void CDesign::RecomputeHess(int aNumPartn,
                            QMatrix& newHess, Vector& newrhsX, Vector& newrhsY,
                            QMatrix& q, Vector& rx, Vector& ry)
{
  cout << "Entering " << __PRETTY_FUNCTION__ << endl;
  int oldDim = Q_GetDim(&gOrigHessian);

  int newDim = oldDim + aNumPartn;

#if 0
  QMatrix newHess;
  Vector  newrhsX;
  Vector  newrhsY;
#endif

  Q_Constr(&newHess,  "newHess", newDim, True,   Rowws, Normal, True);
  V_Constr(&newrhsX,  "newrhsX", newDim, Normal, True);
  V_Constr(&newrhsY,  "newrhsY", newDim, Normal, True);

  V_SetAllCmp(&newrhsX, 0.0);
  V_SetAllCmp(&newrhsY, 0.0);

  // Copy old values from last hessian

  for (int i = 0; i < oldDim; i++)
  {
    if (0 == Q_GetDim(&newHess))
      cout << "matrix vanished at i=" << i << endl;

    int numItems = Q_GetLen(&gOrigHessian, i + 1);

    Q_SetLen(&newHess, i + 1, numItems + 1);

    for (int j = 0; j < numItems; j++)
    {
      int currPos = Q_GetPos(&gOrigHessian, i + 1, j);
      double currVal = Q_GetVal(&gOrigHessian, i + 1, j);
      Q_SetEntry(&newHess, i + 1, j, currPos, currVal);
    }

    Q_SetEntry(&newHess, i + 1, numItems,

               oldDim + mNodesInHessian[i]->PartitionId() + 1, 0.5);

    assert(0 == mNodesInHessian[i]->PartitionId() ||
           1 == mNodesInHessian[i]->PartitionId());

    // Copy RHS-X parts
    double val = V_GetCmp(&rx, i + 1);
    V_SetCmp(&newrhsX, i + 1, val);

    // Copy RHS-Y parts
    val = V_GetCmp(&ry, i + 1);
    V_SetCmp(&newrhsY, i + 1, val);
  }

  V_SetCmp(&newrhsX, oldDim + 1, (mNodesInHessian.size() / 2.0)*500.0 / 2.0);

  V_SetCmp(&newrhsX, oldDim + 2, (mNodesInHessian.size() / 2.0)*1500.0 / 2.0);

  V_SetCmp(&newrhsY, oldDim + 1, (mNodesInHessian.size() / 2.0)*500.0 / 2.0);
  V_SetCmp(&newrhsY, oldDim + 2, (mNodesInHessian.size() / 2.0)*1500.0 / 2.0);

#if 1
  // Assign last entries (of size #partitions) at end of hessian

  for (unsigned i = 0; i < aNumPartn; i++)
  {
    Q_SetLen(&newHess, oldDim + i + 1, 1);

    if (0 == Q_GetDim(&newHess))
      cout << "matrix vanished at (dummy entry set len) with i=" << i << endl;

    Q_SetEntry(&newHess, oldDim + i + 1, 0, oldDim + i + 1, 1e-5);

    if (0 == Q_GetDim(&newHess))
      cout << "matrix vanished at (dummy entry set entry) with i=" << i << endl;
  }

#endif
}

#endif

int GetPartnNum(double pos, vector<int>& accumTiles, double tileSize)
{
  double tileNumLowerBound = pos / tileSize;

  if (pos <= 0)
    return 0;

  for (int i = 1; i < accumTiles.size(); i++)
  {
    if (accumTiles[i] >= tileNumLowerBound)
      return i -1;
  }

  return accumTiles.size() - 2;
}

// For returning which position is least larger number in vector than
// passed value
int GetPartnNum(double value, vector<double>& decisionPts)
{
  for (int i = 0; i < decisionPts.size(); i++)
  {
    if (decisionPts[i] > value)
      return i;
  }

  assert(value == decisionPts[decisionPts.size()-1]);

  return decisionPts.size() - 1;
}

class PartitionRegion
{

  public:
    void push_back(CNode *n)
    {
      v.push_back(n);
    }

  public:
    vector<CNode *> v;
    int xSize;
    int ySize;
    int xIndex;
    int yIndex;
};

static const int gOffsetForPartitionId = 100;
#if 1
// Partitions a given vector into two X and two Y regions (4 regions).
// The imbalance factor for hMetis is computed as follows:
// If m = number of cells in the parent region
//    p = number of cells which can go (physically) in left reigon
//    q = number of cells which can go (physically) in right region
// then, disbalance factor b is given by
//    b1 = 50 - 100 * (1 - min(p, m)/m) (for the left region) and
//    b2 = 50 - 100 * (1 - min(q, m)/m) (for the right region)
// Finally, b is computed as min(b1, b2) capped to 49 and 0 as per
// hMetis requirement
vector<PartitionRegion *> PartitionTheseInXY(
  vector<CNode*>& vec, int lastXPartitions, int lastYPartitions,
  int xLastIndex, int yLastIndex)
{
  if (vec.size() == 0)
    return vector<PartitionRegion *>();

  vector <double> xPos(vec.size());

  int firstXPartSize = lastXPartitions - lastXPartitions / 2;
  int secndXPartSize = lastXPartitions - firstXPartSize;

  int firstYPartSize = lastYPartitions - lastYPartitions / 2;
  int secndYPartSize = lastYPartitions - firstYPartSize;

  for (int i = 0; i < vec.size(); i++)
    xPos[i] = vec[i]->X();

  // sort this vector
  std::sort(xPos.begin(), xPos.end());

#define _POSN_WRT_PLACEMENT_OF_CELLS
// #define _POSN_WRT_PLATOFRM_INFO

#ifdef _POSN_WRT_PLACEMENT_OF_CELLS
  double xPartLoc = xPos[int((xPos.size()-1)*(double)firstXPartSize/
      double(firstXPartSize + secndXPartSize))];
#else
#ifdef _POSN_WRT_PLATOFRM_INFO
  double xPartLoc = 380;
#endif
#endif

  cout << "Deciding partitions in X based on x-pos at " << xPartLoc << endl;
  vector< vector <CNode *> >nodesInXPartitions(2);
  for (int i = 0; i < vec.size(); i++)
  {
    if (vec[i]->X() < xPartLoc)
    {
      vec[i]->PartitionIDX(2*xLastIndex);
      nodesInXPartitions[0].push_back(vec[i]);
    }
    else
    {
      vec[i]->PartitionIDX(2*xLastIndex + 1);
      nodesInXPartitions[1].push_back(vec[i]);
    }
  }

  double xMin = xPos[0];
  double xMax = xPos[xPos.size() - 1];

  double disBalDuringPartnOne = 50 - 100 * (1 - std::min<double>(firstXPartSize * lastYPartitions * (gMaxECellAllowed + gMaxDffAllowed + gMaxRegsAllowed + gMaxBramsAllowed), vec.size()) / vec.size());
  double disBalDuringPartnTwo = 50 - 100 * (1 - std::min<double>(secndXPartSize * lastYPartitions * (gMaxECellAllowed + gMaxDffAllowed + gMaxRegsAllowed + gMaxBramsAllowed), vec.size()) / vec.size());

  int disBalDuringPartn = static_cast<int>(std::min(disBalDuringPartnOne, disBalDuringPartnTwo));
  disBalDuringPartn = std::min(49, std::max(1, disBalDuringPartn));

// #define _AllTypesTogether
#ifdef _AllTypesTogether
  disBalDuringPartn = 1; 
  MinCutOverVectors(nodesInXPartitions[0], nodesInXPartitions[1], true, xPartLoc, (xMax - xMin)/20.0, disBalDuringPartn);
#else
  disBalDuringPartn = 1;
  MinCutForType(3, nodesInXPartitions[0], nodesInXPartitions[1], true, xPartLoc, (xMax - xMin) / 1.0, disBalDuringPartn);
  MinCutForType(2, nodesInXPartitions[0], nodesInXPartitions[1], true, xPartLoc, (xMax - xMin) / 1.0, disBalDuringPartn);
  MinCutForType(1, nodesInXPartitions[0], nodesInXPartitions[1], true, xPartLoc, (xMax - xMin) / 200.0, disBalDuringPartn);
  MinCutForType(0, nodesInXPartitions[0], nodesInXPartitions[1], true, xPartLoc, (xMax - xMin) / 200.0, disBalDuringPartn);
#endif

  vector <PartitionRegion *> vecRgns(4);
  for (int i = 0; i < 4; i++)
    vecRgns[i] = new PartitionRegion();
  assert(4 == vecRgns.size());
  assert(2 == nodesInXPartitions.size());

  cout << "  - Will min-cut in Y for each X's min-cut results now\n";
  for (int i = 0; i < nodesInXPartitions.size(); i++)
  {
    assert(4 == vecRgns.size());
    // Sort in Y
    vector <double> yPos(nodesInXPartitions[i].size());

    for (int j = 0; j < nodesInXPartitions[i].size(); j++)
      yPos[j] = nodesInXPartitions[i][j]->Y();

    std::sort(yPos.begin(), yPos.end());

    // Find inflexion point
#ifdef _POSN_WRT_PLACEMENT_OF_CELLS
    double yPartLoc = yPos[int((yPos.size()-1)* 
        double(firstYPartSize)/double((firstYPartSize + secndYPartSize)))];
#else
#ifdef _POSN_WRT_PLATOFRM_INFO
    double yPartLoc = (i == 0) ?  372 : 1076;
#endif
#endif

    cout << "  yPartLoc = " << yPartLoc << " got by finding ratio " <<
    double(firstYPartSize) / double((firstYPartSize + secndYPartSize)) << endl;

    int match = 0;
    for (int k = 0; k < vec.size(); k++)
      if (vec[k]->Y() == yPartLoc)
        match++;
    cout << "    - " << match << " nodes are at critical point\n";

    // Assign Y partition Id in the current X partition
    for (int j = 0; j < nodesInXPartitions[i].size(); j++)
    {
      if (nodesInXPartitions[i][j]->Y() < yPartLoc)
      {
        nodesInXPartitions[i][j]->PartitionIDY(2*yLastIndex);
        vecRgns[2*i + 0]->push_back(nodesInXPartitions[i][j]);
      }
      else
      {
        nodesInXPartitions[i][j]->PartitionIDY(2*yLastIndex + 1);
        vecRgns[2*i + 1]->push_back(nodesInXPartitions[i][j]);
      }
    }

    double yMin = yPos[0];
    double yMax = yPos[yPos.size() - 1];

    double disBalDuringPartnOne = 50 - 100 * (1 - std::min<double>(firstYPartSize * 
          (0 == i ? firstXPartSize : secndXPartSize) * 
          (gMaxECellAllowed + gMaxDffAllowed + gMaxRegsAllowed + gMaxBramsAllowed), 
          nodesInXPartitions[i].size()) / nodesInXPartitions[i].size());
    double disBalDuringPartnTwo = 50 - 100 * (1 - std::min<double>(secndYPartSize * 
          (0 == i ? firstXPartSize : secndXPartSize) * 
          (gMaxECellAllowed + gMaxDffAllowed + gMaxRegsAllowed + gMaxBramsAllowed), 
          nodesInXPartitions[i].size()) / nodesInXPartitions[i].size());
    int disBalDuringPartn = static_cast<int>(std::min(disBalDuringPartnOne, disBalDuringPartnTwo));

    disBalDuringPartn = std::min(49, std::max(1, disBalDuringPartn));

#ifdef _AllTypesTogether
    disBalDuringPartn = 1;
    MinCutOverVectors(vecRgns[2*i]->v, vecRgns[2*i + 1]->v, false, yPartLoc, (yMax - yMin)/20.0, disBalDuringPartn);
#else
    disBalDuringPartn = 1;
    MinCutForType(3, vecRgns[2*i]->v, vecRgns[2*i + 1]->v, false, yPartLoc, (yMax - yMin) / 1.0, disBalDuringPartn);
    MinCutForType(2, vecRgns[2*i]->v, vecRgns[2*i + 1]->v, false, yPartLoc, (yMax - yMin) / 1.0, disBalDuringPartn);
    MinCutForType(1, vecRgns[2*i]->v, vecRgns[2*i + 1]->v, false, yPartLoc, (yMax - yMin) / 200.0, disBalDuringPartn);
    MinCutForType(0, vecRgns[2*i]->v, vecRgns[2*i + 1]->v, false, yPartLoc, (yMax - yMin) / 200.0, disBalDuringPartn);
#endif
  }

  cout << "   VecRegn[0][0] got " <<  vecRgns[0]->v.size() << " cells\n";

  cout << "   VecRegn[0][1] got " <<  vecRgns[1]->v.size() << " cells\n";
  cout << "   VecRegn[1][0] got " <<  vecRgns[2]->v.size() << " cells\n";
  cout << "   VecRegn[1][1] got " <<  vecRgns[3]->v.size() << " cells\n";

  vecRgns[0]->xSize = firstXPartSize;
  vecRgns[0]->ySize = firstYPartSize;

  vecRgns[1]->xSize = firstXPartSize;
  vecRgns[1]->ySize = secndYPartSize;

  vecRgns[2]->xSize = secndXPartSize;
  vecRgns[2]->ySize = firstYPartSize;

  vecRgns[3]->xSize = secndXPartSize;
  vecRgns[3]->ySize = secndYPartSize;

  vecRgns[0]->xIndex = 2 * (xLastIndex) + 0;
  vecRgns[0]->yIndex = 2 * (yLastIndex) + 0;

  vecRgns[1]->xIndex = 2 * (xLastIndex) + 0;
  vecRgns[1]->yIndex = 2 * (yLastIndex) + 1;

  vecRgns[2]->xIndex = 2 * (xLastIndex) + 1;
  vecRgns[2]->yIndex = 2 * (yLastIndex) + 0;

  vecRgns[3]->xIndex = 2 * (xLastIndex) + 1;
  vecRgns[3]->yIndex = 2 * (yLastIndex) + 1;


  return vecRgns;
}
#endif

void CDesign::assignYPartnID(CNode *n, vector<double>& yPartnPositions)
{
  assert(n->Type() > 1); // Call only for bram and reg file

  double smaller = 0.0;
  double yval = n->Y();
  for (int j=0; j<yPartnPositions.size(); j++)
  {
    if (yval > smaller && yval < yPartnPositions[j])
    {
      n->PartitionIDY(j);
      break;
    }
    else
    {
      smaller = yPartnPositions[j];
    }
  }

  if (-1 == n->PartitionIDY())
    n->PartitionIDY(yPartnPositions.size() - 1);
}

void CDesign::assignXPartnID(CNode *n, vector<double>& xPartnPositions)
{
  assert(n->Type() > 1); // Call only for bram and reg file

  double smaller = 0.0;
  double xval = n->X();
  for (int j=0; j<xPartnPositions.size(); j++)
  {
    if (xval > smaller && xval < xPartnPositions[j])
    {
      n->PartitionIDX(j);
      break;
    }
    else
    {
      smaller = xPartnPositions[j];
    }
  }

  if (-1 == n->PartitionIDX())
    n->PartitionIDX(xPartnPositions.size() - 1);
}

void CDesign::assignXPartnID(vector<double>& xPartnPositions)
{
  for (int i=0; i<mNodesInHessian.size(); i++)
  {
    double smaller = 0.0;
    double xval = mNodesInHessian[i]->X();
    for (int j=0; j<xPartnPositions.size(); j++)
    {
      if (xval > smaller && xval < xPartnPositions[j])
      {
        mNodesInHessian[i]->PartitionIDX(j);
        break;
      }
      else
      {
        smaller = xPartnPositions[j];
      }
    }

    if (-1 == mNodesInHessian[i]->PartitionIDX())
      mNodesInHessian[i]->PartitionIDX(xPartnPositions.size() - 1);

//     cout << "Node : " << mNodesInHessian[i]->Name() << " X : " <<
//       mNodesInHessian[i]->X() << " X Partn ID: " << 
//       mNodesInHessian[i]->PartitionIDX() << endl;
  }
}

void CDesign::assignYPartnID(vector<double>& yPartnPositions)
{
  for (int i=0; i<mNodesInHessian.size(); i++)
  {
    double smaller = 0.0;
    double yval = mNodesInHessian[i]->Y();
    for (int j=0; j<yPartnPositions.size(); j++)
    {
      if (yval > smaller && yval < yPartnPositions[j])
      {
        mNodesInHessian[i]->PartitionIDY(j);
        break;
      }
      else
      {
        smaller = yPartnPositions[j];
      }
    }

    if (-1 == mNodesInHessian[i]->PartitionIDY())
      mNodesInHessian[i]->PartitionIDY(yPartnPositions.size() - 1);

//     cout << "Node : " << mNodesInHessian[i]->Name() << " Y : " <<
//       mNodesInHessian[i]->Y() << " Y Partn ID: " << 
//       mNodesInHessian[i]->PartitionIDY() << endl;
  }
}

class NodesInPartn
{
  public:
    NodesInPartn() { vec.clear(); }
    vector<CNode *> vec;
};

void CDesign::AssignPartnIDNew(int aCurrIter, 
    vector<double>& xcom, vector<double>&ycom)
{
  cout << " mWidth is " << mWidth << " perhaps should be 2000\n";

#define _PLATFORMA
#ifdef _PLATFORMA
  vector<int> xPartitions = gPartitions20[aCurrIter];
  vector<int> yPartitions = gPartitions22[aCurrIter];

  int totalXPartns = 20;
  int totalYPartns = 22;
#else
#ifdef _PLATFORMB
  vector<int> xPartitions = gPartitions22[aCurrIter];
  vector<int> yPartitions = gPartitions24[aCurrIter];
  int totalXPartns = 22;
  int totalYPartns = 24;
#else
  abort();
  ABCD
#endif
#endif

#undef _SHRUNK
#ifdef _SHRUNK
  double effTileWidth  = 36.0;
  double effTileHeight = 64.0;
#else
  double effTileWidth  = mTileWidth + mInterTileSpaceX;
  double effTileHeight = mTileHeight + mInterTileSpaceY;
#endif

  // Compute COM for each partition
  xcom.clear();
  ycom.clear();
  xcom.resize(xPartitions.size());
  ycom.resize(yPartitions.size());

  /* Computing accumulate vector and center of mass vector for
     X and Y directions */
  vector<int> accumCntX;
  accumCntX.push_back(0);

  cout << "  - Computing center of masses for partitions at this level\n";
  for (int i = 0; i < xPartitions.size(); i++)
  {
    xcom[i] = ((double)accumCntX[i] / (double)totalXPartns) * (totalXPartns * effTileWidth) +
              ((double)xPartitions[i] / (double)totalXPartns) * (0.5 * totalXPartns * effTileWidth);
    xcom[i] += mIOSpaceX;
    accumCntX.push_back(xPartitions[i] + accumCntX[i]);
    assert(xcom[i] > 0.0);
  }
  assert(accumCntX.size() == xPartitions.size() + 1);

  vector<int> accumCntY;
  accumCntY.push_back(0);
  for (int i = 0; i < yPartitions.size(); i++)
  {
    ycom[i] = ((double)accumCntY[i] / (double)totalYPartns) * totalYPartns * effTileHeight +
              ((double)yPartitions[i] / (double)totalYPartns) * (0.5 * totalYPartns * effTileHeight);
    ycom[i] += mIOSpaceY;
    accumCntY.push_back(yPartitions[i] + accumCntY[i]);
    assert(ycom[i] > 0.0);
  }
  assert(accumCntY.size() == yPartitions.size() + 1);

  cout << "  - Expected X COM" << endl << "    " ;
  for (int i = 0; i < xcom.size(); i++)
    cout << xcom[i] << " ";
  cout << endl;

  cout << "  - Expected Y COM" << endl << "    " ;
  for (int i = 0; i < ycom.size(); i++)
    cout << ycom[i] << " ";
  cout << endl;

  for (int i=0; i<mNodesInHessian.size(); i++)
  {
    mNodesInHessian[i]->PartitionIDX(-1);
    mNodesInHessian[i]->PartitionIDY(-1);
  }

  vector<double> xPartnLocation(xPartitions.size());
  double sumXPartnSizes = std::accumulate(xPartitions.begin(), xPartitions.end(), 0.0);
  double sumXUntilPrev = 0.0;
  for (int i=0; i<xPartnLocation.size(); i++)
  {
    xPartnLocation[i] = sumXUntilPrev + (double)xPartitions[i]*(double)mCoreWidth/(double)sumXPartnSizes;
    if (i == 0)
      xPartnLocation[i] += mIOSpaceX;
    sumXUntilPrev += (double)xPartitions[i]*(double)mCoreWidth/(double)sumXPartnSizes;
    if (i == 0)
      sumXUntilPrev += mIOSpaceX;
  }

  vector<double> yPartnLocation(yPartitions.size());
  double sumYPartnSizes = std::accumulate(yPartitions.begin(), yPartitions.end(), 0.0);
  double sumYUntilPrev = 0.0;
  for (int i=0; i<yPartnLocation.size(); i++)
  {
    yPartnLocation[i] = sumYUntilPrev + (double)yPartitions[i]*(double)mCoreHeight/(double)sumYPartnSizes;
    if (i == 0)
      yPartnLocation[i] += mIOSpaceY;
    sumYUntilPrev += (double)yPartitions[i]*(double)mCoreHeight/(double)sumYPartnSizes;
    if (i == 0)
      sumYUntilPrev += mIOSpaceY;
  }

  cout << "X Partitioning locations are (total = " << xPartnLocation.size() << ")" << endl;
  for (int i=0; i<xPartnLocation.size(); i++)
    cout << "  - " << xPartnLocation[i] << endl;

  cout << "Y Partitioning locations are (total = " << yPartnLocation.size() << ")"  << endl;
  for (int i=0; i<yPartnLocation.size(); i++)
    cout << "  - " << yPartnLocation[i] << endl;

  assignXPartnID(xPartnLocation);
  assignYPartnID(yPartnLocation);

  for (int i=0; i<mNodesInHessian.size(); i++)
  {
    assert(-1 != mNodesInHessian[i]->PartitionIDX());
    assert(-1 != mNodesInHessian[i]->PartitionIDY());
  }

#if 0
  vector< vector<NodesInPartn *> >nodesInPos(xPartitions.size(), 
      vector<NodesInPartn*>(yPartitions.size(), new NodesInPartn()));
#endif

  vector< vector<NodesInPartn *> >nodesInPos;
  for (int i=0; i<xPartitions.size(); i++)
  {
    vector<NodesInPartn *> v;
    for (int j=0; j<yPartitions.size(); j++)
    {
      v.push_back(new NodesInPartn());
    }
    nodesInPos.push_back(v);
  }

  for (int i=0; i<mNodesInHessian.size(); i++)
  {
    int myXParnNum = mNodesInHessian[i]->PartitionIDX();
    int myYParnNum = mNodesInHessian[i]->PartitionIDY();
    nodesInPos[myXParnNum][myYParnNum]->vec.push_back(mNodesInHessian[i]);
  }

  cout << "Number of cells in each partitions before hmetis:" << endl;
  for (int i=0; i<xPartitions.size(); i++)
  {
    for (int j=0; j<yPartitions.size(); j++)
    {
      cout << nodesInPos[i][j]->vec.size() << " "; 
    }
    cout << endl;
  }

  static map<int, bool> mapAlreadyPartitionedInY; 
  map<int, bool> mapCurrentAlreadyPartitionedInY; 
  for (int i=0; i<nodesInPos.size(); i++)
  {
    for (int j=0; j<nodesInPos[0].size() - 1; j++)
    {
      if (mapAlreadyPartitionedInY.find(int(yPartnLocation[j])) != mapAlreadyPartitionedInY.end())
      {
        cout << "Not doing partition for Location "<< yPartnLocation[j] << endl;
        continue;
      }
      mapCurrentAlreadyPartitionedInY[int(yPartnLocation[j])] = true;
      double firstPartSize = yPartnLocation[j+1] - yPartnLocation[j];
      vector<CNode*> toBePartitionedVec1;
      vector<CNode*> toBePartitionedVec2;
      FindPartitionVectors( toBePartitionedVec1, i, j, toBePartitionedVec2, i, j+1); 
      MinCutForTypeNew(0, toBePartitionedVec1, toBePartitionedVec2, false, 
          yPartnLocation[j], 10000, 1, j, j+1);
      
      toBePartitionedVec1.clear();
      toBePartitionedVec2.clear();
      FindPartitionVectors( toBePartitionedVec1, i, j, toBePartitionedVec2, i, j+1); 
      MinCutForTypeNew(1, toBePartitionedVec1, toBePartitionedVec2, false, 
          yPartnLocation[j], 10000, 1, j, j+1);

#if 0
      MinCutForTypeNew(0, nodesInPos[i][j]->vec, nodesInPos[i][j+1]->vec, false, 
          yPartnLocation[j], 10000, 1, j, j+1);
      MinCutForTypeNew(1, nodesInPos[i][j]->vec, nodesInPos[i][j+1]->vec, false, 
          yPartnLocation[j], 10000, 1, j, j+1);
#endif
//       MinCutForTypeNew(2, nodesInPos[i][j]->vec, nodesInPos[i][j+1]->vec, false, 
//           yPartnLocation[j], 10000, 1, j, j+1);
//       MinCutForTypeNew(3, nodesInPos[i][j]->vec, nodesInPos[i][j+1]->vec, false, 
//           yPartnLocation[j], 10000, 1, j, j+1);
    }
  }

  static map<int, bool> mapAlreadyPartitionedInX; 
  map<int, bool> mapCurrentAlreadyPartitionedInX; 
  for (int j=0; j<nodesInPos[0].size(); j++)
  {
    for (int i=0; i<nodesInPos.size() - 1; i++)
    {
      if (mapAlreadyPartitionedInX.find(int(xPartnLocation[j])) != mapAlreadyPartitionedInX.end())
      {
        cout << "Not doing partition for Location "<< yPartnLocation[j] << endl;
        continue; 
      }
      mapCurrentAlreadyPartitionedInX[int(xPartnLocation[j])] = true;
      double firstPartSize = xPartnLocation[i+1] - xPartnLocation[i];
      vector<CNode*> toBePartitionedVec1;
      vector<CNode*> toBePartitionedVec2;
      FindPartitionVectors( toBePartitionedVec1, i, j, toBePartitionedVec2, i+1, j); 
      MinCutForTypeNew(0, toBePartitionedVec1, toBePartitionedVec2, true, 
          xPartnLocation[j], 10000, 1, i, i+1);
      
      toBePartitionedVec1.clear();
      toBePartitionedVec2.clear();
      FindPartitionVectors( toBePartitionedVec1, i, j, toBePartitionedVec2, i+1, j); 
      MinCutForTypeNew(1, toBePartitionedVec1, toBePartitionedVec2, true, 
          xPartnLocation[j], 10000, 1, i, i+1);
#if 0      
      MinCutForTypeNew(0, nodesInPos[i][j]->vec, nodesInPos[i+1][j]->vec, true, 
          xPartnLocation[j], 10000, 1, i, i+1);
      MinCutForTypeNew(1, nodesInPos[i][j]->vec, nodesInPos[i+1][j]->vec, true, 
          xPartnLocation[j], 10000, 1, i, i+1);
#endif
//       MinCutForTypeNew(2, nodesInPos[i][j]->vec, nodesInPos[i+1][j]->vec, true, 
//           xPartnLocation[j], 10000, 1, i, i+1);
//       MinCutForTypeNew(3, nodesInPos[i][j]->vec, nodesInPos[i+1][j]->vec, true, 
//           xPartnLocation[j], 10000, 1, i, i+1);
    }
  }

  map<int, bool>::iterator itX = mapCurrentAlreadyPartitionedInX.begin();
  while (itX != mapCurrentAlreadyPartitionedInX.end())
  {
    mapAlreadyPartitionedInX[itX->first] = true;
    itX++;
  }
  
  map<int, bool>::iterator itY = mapCurrentAlreadyPartitionedInY.begin();
  while (itY != mapCurrentAlreadyPartitionedInY.end())
  {
    mapAlreadyPartitionedInY[itY->first] = true;
    itY++;
  }
  
  for (int i=0; i<mNodesInHessian.size(); i++)
  {
    assert(mNodesInHessian[i]->PartitionIDY() >= 0 && 
        mNodesInHessian[i]->PartitionIDY() < yPartitions.size());
    assert(mNodesInHessian[i]->PartitionIDX() >= 0 && 
        mNodesInHessian[i]->PartitionIDX() < xPartitions.size());
  }

  SolveBRAMAssignment(false);
  SolveRegFileAssignment(false);

  for (int i=0; i<mBrams.size(); i++)
  {
    assignXPartnID(mAllNodes[mBrams[i]], xPartnLocation);
    assignYPartnID(mAllNodes[mBrams[i]], yPartnLocation);
  }
  for (int i=0; i<mRegs.size(); i++)
  {
    assignXPartnID(mAllNodes[mRegs[i]], xPartnLocation);
    assignYPartnID(mAllNodes[mRegs[i]], yPartnLocation);
  }

  vector< vector<int> > nodesInPartitions (xPartitions.size(), vector<int>(yPartitions.size(), int()));
  for (int i=0; i<mNodesInHessian.size(); i++)
  {
    int myXParnNum = mNodesInHessian[i]->PartitionIDX();
    int myYParnNum = mNodesInHessian[i]->PartitionIDY();
    nodesInPartitions[myXParnNum][myYParnNum]++;
  }
  cout << "Number of cells in each partitions after hmetis:" << endl;
  for (int i=0; i<xPartitions.size(); i++)
  {
    for (int j=0; j<yPartitions.size(); j++)
    {
      cout << nodesInPartitions[i][j] << " "; 
    }
    cout << endl;
  }

}

void CDesign::FindPartitionVectors( vector <CNode*> &vec1, int partX1, int partY1,
                                    vector <CNode*> &vec2, int partX2, int partY2)
{
  
  for (int i=0; i<mNodesInHessian.size(); i++)
  {
    int partX = mNodesInHessian[i]->PartitionIDX();
    int partY = mNodesInHessian[i]->PartitionIDY();

    if (partX == partX1 && partY == partY1)
    {
      vec1.push_back(mNodesInHessian[i]);
    }
    else if (partX == partX2 && partY == partY2)
    {
      vec2.push_back(mNodesInHessian[i]);
    }
  }

}

#if 1
void CDesign::AssignPartnID(const int aCurrIter, vector<int>& xNumCell,
                            vector<int>& yNumCell, vector<double>& xcom, vector<double>& ycom)
{
  cout << "STATUS:  Entering AssignPartnID()\n";

#define _PLATFORMA
#ifdef _PLATFORMA
  vector<int> xPartitions = gPartitions20[aCurrIter];
  vector<int> yPartitions = gPartitions22[aCurrIter];

  int totalXPartns = 20;
  int totalYPartns = 22;
#else
#ifdef _PLATFORMB
  vector<int> xPartitions = gPartitions22[aCurrIter];
  vector<int> yPartitions = gPartitions24[aCurrIter];
  int totalXPartns = 22;
  int totalYPartns = 24;
#else
  abort();
  ABCD
#endif
#endif

// #define _SHRUNK
#undef _SHRUNK
#ifdef _SHRUNK
  double effTileWidth  = 36.0;
  double effTileHeight = 64.0;
#else
  double effTileWidth  = mTileWidth + mInterTileSpaceX;
  double effTileHeight = mTileHeight + mInterTileSpaceY;
#endif

  // Clear the previous vectors and fill NULL in them
  xcom.clear();
  ycom.clear();
  xNumCell.clear();
  yNumCell.clear();

  xcom.resize(xPartitions.size());
  ycom.resize(yPartitions.size());
  xNumCell.resize(xPartitions.size());
  yNumCell.resize(yPartitions.size());

  std::fill(xNumCell.begin(), xNumCell.end(), 0);
  std::fill(yNumCell.begin(), yNumCell.end(), 0);

  vector<int> accumCntX;
  vector<int> accumCntY;
  accumCntX.push_back(0);
  accumCntY.push_back(0);


  /* Computing accumulate vector and center of mass vector for
     X and Y directions */
  cout << "  - Computing center of masses for partitions at this level\n";
  for (int i = 0; i < xPartitions.size(); i++)
  {
    xcom[i] = ((double)accumCntX[i] / (double)totalXPartns) * (totalXPartns * effTileWidth) +
              ((double)xPartitions[i] / (double)totalXPartns) * (0.5 * totalXPartns * effTileWidth);
    xcom[i] += mIOSpaceX;
    accumCntX.push_back(xPartitions[i] + accumCntX[i]);
    assert(xcom[i] > 0.0);
  }
  assert(accumCntX.size() == xPartitions.size() + 1);

  for (int i = 0; i < yPartitions.size(); i++)
  {
    ycom[i] = ((double)accumCntY[i] / (double)totalYPartns) * totalYPartns * effTileHeight +
              ((double)yPartitions[i] / (double)totalYPartns) * (0.5 * totalYPartns * effTileHeight);
    ycom[i] += mIOSpaceY;
    accumCntY.push_back(yPartitions[i] + accumCntY[i]);
    assert(ycom[i] > 0.0);
  }
  assert(accumCntY.size() == yPartitions.size() + 1);

  cout << "  - Expected X COM" << endl << "    " ;
  for (int i = 0; i < xcom.size(); i++)
    cout << xcom[i] << " ";
  cout << endl;
  cout << "  - Expected Y COM" << endl << "    " ;
  for (int i = 0; i < ycom.size(); i++)
    cout << ycom[i] << " ";
  cout << endl;

  // Somehow populated
  static vector <PartitionRegion *> doNextTime;
  static int count = 0;
  if (0 == count)
  {
    cout << "  - First Iteration.  Initializing \"doNextTime\" vector\n";
    PartitionRegion *pr = new PartitionRegion();
    pr->v = mNodesInHessian;
    pr->xSize = 20;
    pr->ySize = 22;
    pr->xIndex = 0;
    pr->yIndex = 0;
    doNextTime.push_back(pr);
    count++;
  }

  vector <PartitionRegion *> tmpPartnRegions;
  for (int i = 0; i < doNextTime.size(); i++)
  {
    cout << "  - Processing " << i + 1 << " of " << doNextTime.size() << " partitions\n";
    vector<PartitionRegion *> vp = PartitionTheseInXY(doNextTime[i]->v, 
        doNextTime[i]->xSize, doNextTime[i]->ySize, 
        doNextTime[i]->xIndex, doNextTime[i]->yIndex);

    if (vp.size() == 0)
    {
    }
    else
    {
      assert(4 == vp.size());
      assert(doNextTime[i]->xSize == vp[0]->xSize + vp[2]->xSize);
      assert(doNextTime[i]->ySize == vp[0]->ySize + vp[1]->ySize);
    }

    for (int k = 0; k < vp.size(); k++)
    {
      assert(vp[k]->xSize > 0 || doNextTime[i]->xSize == 1);
      assert(vp[k]->ySize > 0 || doNextTime[i]->ySize == 1);
      tmpPartnRegions.push_back(vp[k]);
    }
  }

  cout << "  - Copying temporary vector to doNextTime vector\n";

  doNextTime.clear();

  for (int i = 0; i < tmpPartnRegions.size(); i++)
    doNextTime.push_back(tmpPartnRegions[i]);
  assert(doNextTime.size() == tmpPartnRegions.size());
  tmpPartnRegions.clear();

  vector< vector< int> > numInside(xPartitions.size(), vector<int>(yPartitions.size(), int()));

  for (int i = 0; i < mNodesInHessian.size(); i++)
  {
    numInside[mNodesInHessian[i]->PartitionIDX()][mNodesInHessian[i]->PartitionIDY()]++;
  }

  cout << "  - Printing # cells in MATRIX partitions\n";
  for (int j = numInside[0].size() - 1; j >= 0; j--)
  {
    for (int i = 0; i < numInside.size(); i++)
      cout << "        " << numInside[i][j] << " ";

    cout << endl;
  }

#if 0
  for (int i = 0; i < xNumCell.size(); i++)
    cout << "  X Partn #" << i + 1 << " : " << xNumCell[i] << endl;

  cout << "\nPrinting # cells in Y partitions\n";

  for (int i = 0; i < yNumCell.size(); i++)
    cout << "  Y Partn #" << i + 1 << " : " << yNumCell[i] << endl;

  assert(std::accumulate(xNumCell.begin(), xNumCell.end(), 0) ==
         std::accumulate(yNumCell.begin(), yNumCell.end(), 0));

#endif
}
#endif

int CDesign::GetCellsToPutInHessian()
{
  cout << "STATUS:  Getting valid cells to put in Hessian\n";
  mNodesInHessian.clear();

  // All nodes which have non-zero neighbours arre added into
  // mNodesInHessian vector.

  for (int i = 0; i < mAllNodes.size(); i++)
  {
    if (true == mAllNodes[i]->Fixed())
      continue;

    if (mAllNodes[i]->GetNumNeighbors() == 0)
      continue;

    if (true == mAllNodes[i]->hasNoMovableNeighbor)
      continue;

    mAllNodes[i]->SetCurLocInHess(mNodesInHessian.size());

    mNodesInHessian.push_back(mAllNodes[i]);
  }

  return mNodesInHessian.size();
}

void CDesign::PrintCoM(vector<double>& xcom, vector<double>& ycom)
{
  int partnX = xcom.size();
  int partnY = ycom.size();

  vector <double> acumPartnX(partnX);
  vector <double> acumPartnY(partnY);
  vector <int> countEntrX(partnX);
  vector <int> countEntrY(partnY);

  std::ofstream out;
  out.open("OnePartitionX.txt");

  for (int i = 0; i < mNodesInHessian.size(); i++)
  {
    if (mNodesInHessian[i]->PartitionIDX() == 0)
    {
      out << mNodesInHessian[i]->X() << endl;
    }

    acumPartnX[mNodesInHessian[i]->PartitionIDX()] += mNodesInHessian[i]->X();

    countEntrX[mNodesInHessian[i]->PartitionIDX()] += 1;
    acumPartnY[mNodesInHessian[i]->PartitionIDY()] += mNodesInHessian[i]->Y();
    countEntrY[mNodesInHessian[i]->PartitionIDY()] += 1;
  }

  cout << endl;

  cout << "          EXPECTED PARTITIONS   " << partnX <<  " & " << partnY << endl;

  for (int i = 0; i < partnX; i++)
  {
    cout << "** xcom (computed) = " << acumPartnX[i] / countEntrX[i] <<
    " expected = " << xcom[i] << " (over " << countEntrX[i] <<  " cells)\n";
  }

  cout << endl;

  for (int i = 0; i < partnY; i++)
  {
    cout << "** ycom (computed) = " << acumPartnY[i] / countEntrY[i] << " expected = " <<
    ycom[i] << " (over " << countEntrY[i] << " cells)\n";
  }

  cout << endl;
}

void MinCutForTypeNew(int aType, vector<CNode*>& left, vector<CNode*>& right, 
    bool isX, double partitionLocation, double widthToConsider, int disBal, int leftId, int rightId)
{
  return;
  cout << "STATUS:  Performing min cut on vectors of size " <<
  left.size() << " " << right.size() << endl;

  if (left.size() == 0 && right.size() == 0)
    return;

#if 0
  int leftId;
  int rightId;
  if (true == isX)
  {
    if (left.size() != 0)
      leftId = left[0]->PartitionIDX();
    if (right.size() != 0)
      rightId = right[0]->PartitionIDX();
  }
  else
  {
    if (left.size() != 0)
      leftId = left[0]->PartitionIDY();
    if (right.size() != 0)
      rightId = right[0]->PartitionIDY();
  }
#endif

  const int multiplier = gMaxNetFanoutLimitForHessian;
  vector<pair< pair<CNode*, CNode*>, int> > pairsToPrint;

  // Map of all nodes present in the two vectors we have.  This will
  // be used to determine true neighbors for partitioning purpose
  map <CNode*, bool> mapNodesPresent;
  map <CNode*, bool> mapNodesNotPresent;

  for (int i = 0; i < left.size(); i++)
  {
    if (isX)
    {
      if (fabs(partitionLocation - left[i]->X()) <= widthToConsider && left[i]->Type() == aType)
      {
        assert(mapNodesPresent.find(left[i]) == mapNodesPresent.end());
        mapNodesPresent[left[i]] = true;
        //         cout << " added : " << left[i]->X() << endl;
      }
    }
    else
    {
      if (fabs(partitionLocation - left[i]->Y()) <= widthToConsider && left[i]->Type() == aType)
      {
        assert(mapNodesPresent.find(left[i]) == mapNodesPresent.end());
        mapNodesPresent[left[i]] = true;
        //         cout << " added : " << left[i]->Y() << endl;
      }
    }
  }

  for (int i = 0; i < right.size(); i++)
  {
    if (isX)
    {
      if (fabs(right[i]->X() - partitionLocation) <= widthToConsider && right[i]->Type() == aType)
      {
        assert(mapNodesPresent.find(right[i]) == mapNodesPresent.end());
        mapNodesPresent[right[i]] = true;
        //         cout << " added : " << right[i]->X() << endl;
      }
    }
    else
    {
      if (fabs(right[i]->Y() - partitionLocation) <= widthToConsider && right[i]->Type() == aType)
      {
        assert(mapNodesPresent.find(right[i]) == mapNodesPresent.end());
        mapNodesPresent[right[i]] = true;
        //         cout << " added : " << right[i]->Y() << endl;
      }
    }
  }

  if (0 == mapNodesPresent.size())
  {
    cout << "Returning from partition function.  Nothing to do\n";
    return;
  }

  int pairsInPartition = 0;

  int pairsNotInPartition = 0;
  //   cout << "  - Doing left\n";

  for (int i = 0; i < left.size(); i++)
  {
    CNode *n = left[i];
    //     cout << n->Name() << endl;

    if (mapNodesPresent.find(n) == mapNodesPresent.end())
      continue;

    vector<CNode*> neigh = n->GetNeighborCells();

    vector<double> wts   = n->GetNeighborWts();

    for (int j = 0; j < neigh.size(); j++)
    {
      if (mapNodesPresent.find(neigh[j]) != mapNodesPresent.end())
      {
        //         cout << " added : " << n->Name() << " " << neigh[j]->Name() << endl;
        pairsToPrint.push_back(make_pair(make_pair(n, neigh[j]), int(wts[j]*multiplier)));
        pairsInPartition += 1;
      }
      else
      {
        //         cout << " added external node: " << n->Name() << " " << neigh[j]->Name() << endl;
        pairsToPrint.push_back(make_pair(make_pair(n, neigh[j]), int(wts[j]*multiplier)));
        mapNodesNotPresent[neigh[j]] = true;
        pairsNotInPartition += 1;
      }
    }
  }

  //   cout << "  - Doing right\n";
  for (int i = 0; i < right.size(); i++)
  {
    CNode *n = right[i];
    //     cout << n->Name() << endl;

    if (mapNodesPresent.find(n) == mapNodesPresent.end())
      continue;

    vector<CNode*> neigh = n->GetNeighborCells();

    vector<double> wts   = n->GetNeighborWts();

    for (int j = 0; j < neigh.size(); j++)
    {
      if (mapNodesPresent.find(neigh[j]) != mapNodesPresent.end())
      {
        //         cout << " added : " << n->Name() << " " << neigh[j]->Name() << endl;
        pairsToPrint.push_back(make_pair(make_pair(n, neigh[j]), int(wts[j]*multiplier)));
        pairsInPartition += 1;
      }
      else
      {
        //         cout << " added external node: " << n->Name() << " " << neigh[j]->Name() << endl;
        pairsToPrint.push_back(make_pair(make_pair(n, neigh[j]), int(wts[j]*multiplier)));
        mapNodesNotPresent[neigh[j]] = true;
        pairsNotInPartition += 1;
      }
    }
  }

  std::ofstream fixFile;

  fixFile.open("partition.fix");

  if (false == fixFile)
    cout << "EROR:  Could not write to file partition.fix\n";

  map <CNode*, int> mapNodeToSeqInt;

  map <int, CNode*> mapSeqIntToNode;

  map <CNode*, bool>::iterator it = mapNodesPresent.begin();

  int currAdded = 0;

  while (it != mapNodesPresent.end())
  {
    mapNodeToSeqInt[it->first] = currAdded;
    mapSeqIntToNode[currAdded] = it->first;
    currAdded++;
    it++;
    fixFile << "-1" << endl;
  }

  it = mapNodesNotPresent.begin();

  int countFixInFirst = 0;
  int countFixInSecond = 0;
  while (it != mapNodesNotPresent.end())
  {
    mapNodeToSeqInt[it->first] = currAdded;
    //     mapSeqIntToNode[currAdded] = it->first;

    if (isX)
    {
      if (it->first->X() < partitionLocation)
      {
        fixFile << "0" << endl;
        countFixInFirst += 1;
      }
      else
      {
        fixFile << "1" << endl;
        countFixInSecond += 1;
      }
    }
    else
    {
      if (it->first->Y() < partitionLocation)
      {
        fixFile << "0" << endl;
        countFixInFirst += 1;
      }
      else
      {
        fixFile << "1" << endl;
        countFixInSecond += 1;
      }
    }

    currAdded++;

    it++;
  }

  int countOfPrint = abs(countFixInSecond - countFixInFirst);
  if (countFixInSecond > countFixInFirst)
  {
    for (int i = 0; i < countOfPrint; i++)
    {
      fixFile << "0" << endl;
    }
  }
  else
  {
    for (int i = 0; i < countOfPrint; i++)
    {
      fixFile << "1" << endl;
    }
  }
  fixFile.close();

  std::ofstream out;
  out.open("partition.hgr");

  if (false == out)
    cout << "EROR:  Could not write to file partition.hgr\n";

#if 0
  for (int i = 0; i < pairsToPrint.size(); i++)
  {
    cout << "HEHE " << pairsToPrint[i].first.first->Name() << " " <<
    pairsToPrint[i].first.second->Name() << endl;
  }

#endif

  //   cout << "pairsToPrint.size() is " << pairsToPrint.size() << endl;
  assert(pairsInPartition % 2 == 0);

  //   cout << "Movable Nodes in partition file: " << mapNodesPresent.size() << endl;
  if (mapNodesPresent.size() == 4 && mapNodesNotPresent.size() == 84)
    cout << "trest\n";

  //   cout << "Unmovable Nodes in partition file: " << mapNodesNotPresent.size() << endl;
  //   cout << "Nets in partition file: " << (pairsInPartition/2) + pairsNotInPartition << endl;

  out << (pairsInPartition / 2) + pairsNotInPartition << " " <<
  mapNodesPresent.size() + mapNodesNotPresent.size() + countOfPrint << " 1\n";

  for (int i = 0; i < pairsToPrint.size(); i++)
  {
    CNode *left = pairsToPrint[i].first.first;
    CNode *right = pairsToPrint[i].first.second;
    int wt = pairsToPrint[i].second;
    assert(wt > 0);

    if (mapNodeToSeqInt[left] < mapNodeToSeqInt[right])
      out << wt << " " << mapNodeToSeqInt[left] + 1 << " " <<
      mapNodeToSeqInt[right] + 1 << endl;
  }

  out.close();

//   string cmd = "./shmetis partition.hgr partition.fix 2 ";
  string cmd = "./hmetis partition.hgr partition.fix 2 ";
  cmd += makeString(disBal);
  cmd += " 1 1 3 2 0 0 ";
  // hmetic command
  // hmetis HGraphFile FixFile Nparts UBfactor Nruns CType RType Vcycle Reconst dgblvl
  cmd += " >& /dev/null";
  cout << "  - Running " << cmd.c_str() << endl;

  //   cout << "hmetis starts ... \n";
  system(cmd.c_str());
  //   cout << "hmetis ends ... \n";

  std::ifstream in, test1, test2;

  test2.open("partition.hgr.part.2");

  if (test2)
  {
    //     cout << "    Two partitions were made\n";
    in.open("partition.hgr.part.2");
  }
  else
  {
    test1.open("partition.hgr.part.1");

    if (test1)
    {
      //       cout << "    One partitions were made.  Will use it\n";
      in.open("partition.hgr.part.1");
    }
    else
    {
      cout << "    Could not open 1 or 2 partition's out file\n";
      return;
    }
  }

  assert(in);

#if 0

  if (false == in)
  {
    cout << "EROR:  Cannot read output file of hmetis\n";
    in.close();
    in.open("partition.hgr.part.1");

    if (false == in)
    {
      in.open("partition.hgr");
      cout << "EROR: Cannot read output file of hmetis: partition.hgr.part.1\n";
    }
  }

#endif

  string line;
  int currIdx = 0;
  while (false == in.eof())
  {
    getline(in, line);
    if (line == "")
      continue;

    int partnNum = atoi(line.c_str());
    CNode *n = mapSeqIntToNode[currIdx];

    //     cout << currIdx << " should be " << mapNodeToSeqInt[mapSeqIntToNode[currIdx]] << endl;
    if (true == isX)
      n->PartitionIDX(partnNum == 0 ? leftId : rightId);
    else
      n->PartitionIDY(partnNum == 0 ? leftId : rightId);

    //     cout << "Node " << n->Name() << " got Parn Id = " << n->PartitionIDX() << endl;
    currIdx++;
    if (currIdx == mapNodesPresent.size())
      break;
  }

  assert(currIdx == mapNodesPresent.size());

  system("ls partition.hgr && mv partition.hgr partition.hgr.old");
  system("ls partition.hgr.part.2 >& /dev/null && mv partition.hgr.part.2 partition.hgr.part.2.old");
  system("ls partition.hgr.part.1 >& /dev/null && mv partition.hgr.part.1 partition.hgr.part.1.old");
  system("ls partition.fix && mv partition.fix partition.fix.old");
}

#if 1
void MinCutForType(int aType, vector<CNode*>& left, vector<CNode*>& right, 
    bool isX, double partitionLocation, double widthToConsider, int disBal)
{
  return;
  cout << "STATUS:  Performing min cut on vectors of size " <<
  left.size() << " " << right.size() << endl;

  if (left.size() == 0 || right.size() == 0)
    return;

  int leftId;
  int rightId;

  if (true == isX)
  {
    leftId = left[0]->PartitionIDX();
    rightId = right[0]->PartitionIDX();
  }
  else
  {
    leftId = left[0]->PartitionIDY();
    rightId = right[0]->PartitionIDY();
  }

  const int multiplier = gMaxNetFanoutLimitForHessian;
  vector<pair< pair<CNode*, CNode*>, int> > pairsToPrint;

  // Map of all nodes present in the two vectors we have.  This will
  // be used to determine true neighbors for partitioning purpose
  map <CNode*, bool> mapNodesPresent;
  map <CNode*, bool> mapNodesNotPresent;

  for (int i = 0; i < left.size(); i++)
  {
    if (isX)
    {
      if (fabs(partitionLocation - left[i]->X()) <= widthToConsider && left[i]->Type() == aType)
      {
        assert(mapNodesPresent.find(left[i]) == mapNodesPresent.end());
        mapNodesPresent[left[i]] = true;
        //         cout << " added : " << left[i]->X() << endl;
      }
    }
    else
    {
      if (fabs(partitionLocation - left[i]->Y()) <= widthToConsider && left[i]->Type() == aType)
      {
        assert(mapNodesPresent.find(left[i]) == mapNodesPresent.end());
        mapNodesPresent[left[i]] = true;
        //         cout << " added : " << left[i]->Y() << endl;
      }
    }
  }

  for (int i = 0; i < right.size(); i++)
  {
    if (isX)
    {
      if (fabs(right[i]->X() - partitionLocation) <= widthToConsider && right[i]->Type() == aType)
      {
        assert(mapNodesPresent.find(right[i]) == mapNodesPresent.end());
        mapNodesPresent[right[i]] = true;
        //         cout << " added : " << right[i]->X() << endl;
      }
    }
    else
    {
      if (fabs(right[i]->Y() - partitionLocation) <= widthToConsider && right[i]->Type() == aType)
      {
        assert(mapNodesPresent.find(right[i]) == mapNodesPresent.end());
        mapNodesPresent[right[i]] = true;
        //         cout << " added : " << right[i]->Y() << endl;
      }
    }
  }

  if (0 == mapNodesPresent.size())
  {
    cout << "Returning from partition function.  Nothing to do\n";
    return;
  }

  int pairsInPartition = 0;

  int pairsNotInPartition = 0;
  //   cout << "  - Doing left\n";

  for (int i = 0; i < left.size(); i++)
  {
    CNode *n = left[i];
    //     cout << n->Name() << endl;

    if (mapNodesPresent.find(n) == mapNodesPresent.end())
      continue;

    vector<CNode*> neigh = n->GetNeighborCells();

    vector<double> wts   = n->GetNeighborWts();

    for (int j = 0; j < neigh.size(); j++)
    {
      if (mapNodesPresent.find(neigh[j]) != mapNodesPresent.end())
      {
        //         cout << " added : " << n->Name() << " " << neigh[j]->Name() << endl;
        pairsToPrint.push_back(make_pair(make_pair(n, neigh[j]), int(wts[j]*multiplier)));
        pairsInPartition += 1;
      }
      else
      {
        //         cout << " added external node: " << n->Name() << " " << neigh[j]->Name() << endl;
        pairsToPrint.push_back(make_pair(make_pair(n, neigh[j]), int(wts[j]*multiplier)));
        mapNodesNotPresent[neigh[j]] = true;
        pairsNotInPartition += 1;
      }
    }
  }

  //   cout << "  - Doing right\n";
  for (int i = 0; i < right.size(); i++)
  {
    CNode *n = right[i];
    //     cout << n->Name() << endl;

    if (mapNodesPresent.find(n) == mapNodesPresent.end())
      continue;

    vector<CNode*> neigh = n->GetNeighborCells();

    vector<double> wts   = n->GetNeighborWts();

    for (int j = 0; j < neigh.size(); j++)
    {
      if (mapNodesPresent.find(neigh[j]) != mapNodesPresent.end())
      {
        //         cout << " added : " << n->Name() << " " << neigh[j]->Name() << endl;
        pairsToPrint.push_back(make_pair(make_pair(n, neigh[j]), int(wts[j]*multiplier)));
        pairsInPartition += 1;
      }
      else
      {
        //         cout << " added external node: " << n->Name() << " " << neigh[j]->Name() << endl;
        pairsToPrint.push_back(make_pair(make_pair(n, neigh[j]), int(wts[j]*multiplier)));
        mapNodesNotPresent[neigh[j]] = true;
        pairsNotInPartition += 1;
      }
    }
  }

  std::ofstream fixFile;

  fixFile.open("partition.fix");

  if (false == fixFile)
    cout << "EROR:  Could not write to file partition.fix\n";

  map <CNode*, int> mapNodeToSeqInt;

  map <int, CNode*> mapSeqIntToNode;

  map <CNode*, bool>::iterator it = mapNodesPresent.begin();

  int currAdded = 0;

  while (it != mapNodesPresent.end())
  {
    mapNodeToSeqInt[it->first] = currAdded;
    mapSeqIntToNode[currAdded] = it->first;
    currAdded++;
    it++;
    fixFile << "-1" << endl;
  }

  it = mapNodesNotPresent.begin();

  int countFixInFirst = 0;
  int countFixInSecond = 0;
  while (it != mapNodesNotPresent.end())
  {
    mapNodeToSeqInt[it->first] = currAdded;
    //     mapSeqIntToNode[currAdded] = it->first;

    if (isX)
    {
      if (it->first->X() < partitionLocation)
      {
        fixFile << "0" << endl;
        countFixInFirst += 1;
      }
      else
      {
        fixFile << "1" << endl;
        countFixInSecond += 1;
      }
    }
    else
    {
      if (it->first->Y() < partitionLocation)
      {
        fixFile << "0" << endl;
        countFixInFirst += 1;
      }
      else
      {
        fixFile << "1" << endl;
        countFixInSecond += 1;
      }
    }

    currAdded++;

    it++;
  }

  int countOfPrint = abs(countFixInSecond - countFixInFirst);
  if (countFixInSecond > countFixInFirst)
  {
    for (int i = 0; i < countOfPrint; i++)
    {
      fixFile << "0" << endl;
    }
  }
  else
  {
    for (int i = 0; i < countOfPrint; i++)
    {
      fixFile << "1" << endl;
    }
  }
  fixFile.close();

  std::ofstream out;
  out.open("partition.hgr");

  if (false == out)
    cout << "EROR:  Could not write to file partition.hgr\n";

#if 0
  for (int i = 0; i < pairsToPrint.size(); i++)
  {
    cout << "HEHE " << pairsToPrint[i].first.first->Name() << " " <<
    pairsToPrint[i].first.second->Name() << endl;
  }

#endif

  //   cout << "pairsToPrint.size() is " << pairsToPrint.size() << endl;
  assert(pairsInPartition % 2 == 0);

  //   cout << "Movable Nodes in partition file: " << mapNodesPresent.size() << endl;
  if (mapNodesPresent.size() == 4 && mapNodesNotPresent.size() == 84)
    cout << "trest\n";

  //   cout << "Unmovable Nodes in partition file: " << mapNodesNotPresent.size() << endl;
  //   cout << "Nets in partition file: " << (pairsInPartition/2) + pairsNotInPartition << endl;

  out << (pairsInPartition / 2) + pairsNotInPartition << " " <<
  mapNodesPresent.size() + mapNodesNotPresent.size() + countOfPrint << " 1\n";

  for (int i = 0; i < pairsToPrint.size(); i++)
  {
    CNode *left = pairsToPrint[i].first.first;
    CNode *right = pairsToPrint[i].first.second;
    int wt = pairsToPrint[i].second;
    assert(wt > 0);

    if (mapNodeToSeqInt[left] < mapNodeToSeqInt[right])
      out << wt << " " << mapNodeToSeqInt[left] + 1 << " " <<
      mapNodeToSeqInt[right] + 1 << endl;
  }

  out.close();

  string cmd = "./shmetis partition.hgr partition.fix 2 ";
  cmd += makeString(disBal);
//   cmd += " >& /dev/null";
  cout << "  - Running " << cmd.c_str() << endl;

  //   cout << "hmetis starts ... \n";
  system(cmd.c_str());
  //   cout << "hmetis ends ... \n";

  std::ifstream in, test1, test2;

  test2.open("partition.hgr.part.2");

  if (test2)
  {
    //     cout << "    Two partitions were made\n";
    in.open("partition.hgr.part.2");
  }
  else
  {
    test1.open("partition.hgr.part.1");

    if (test1)
    {
      //       cout << "    One partitions were made.  Will use it\n";
      in.open("partition.hgr.part.1");
    }
    else
    {
      cout << "    Could not open 1 or 2 partition's out file\n";
      return;
    }
  }

  assert(in);

#if 0

  if (false == in)
  {
    cout << "EROR:  Cannot read output file of hmetis\n";
    in.close();
    in.open("partition.hgr.part.1");

    if (false == in)
    {
      in.open("partition.hgr");
      cout << "EROR: Cannot read output file of hmetis: partition.hgr.part.1\n";
    }
  }

#endif

  string line;
  int currIdx = 0;
  while (false == in.eof())
  {
    getline(in, line);
    if (line == "")
      continue;

    int partnNum = atoi(line.c_str());
    CNode *n = mapSeqIntToNode[currIdx];

    //     cout << currIdx << " should be " << mapNodeToSeqInt[mapSeqIntToNode[currIdx]] << endl;
    if (true == isX)
      n->PartitionIDX(partnNum == 0 ? leftId : rightId);
    else
      n->PartitionIDY(partnNum == 0 ? leftId : rightId);

    //     cout << "Node " << n->Name() << " got Parn Id = " << n->PartitionIDX() << endl;
    currIdx++;
    if (currIdx == mapNodesPresent.size())
      break;
  }

  assert(currIdx == mapNodesPresent.size());

  system("ls partition.hgr && mv partition.hgr partition.hgr.old");
  system("ls partition.hgr.part.2 >& /dev/null && mv partition.hgr.part.2 partition.hgr.part.2.old");
  system("ls partition.hgr.part.1 >& /dev/null && mv partition.hgr.part.1 partition.hgr.part.1.old");
  system("ls partition.fix && mv partition.fix partition.fix.old");
}
#endif

void CheckSoln(QMatrix& hess, Vector& soln, Vector& rhs)
{
  int dim = Q_GetDim(&hess);
  int num = rand() % dim;
  num = 1;

  while (0.001 >= V_GetCmp(&rhs, num++));
  num--;

  cout << "V_GetCmp of rhs at num = " << V_GetCmp(&rhs, num) << endl;

  double sum = 0;
  for (int i = 0; i < dim; i++)
  {
    double val1 = Q_GetEl(&hess, num, i + 1);
    double val2 = V_GetCmp(&soln, i + 1);
    double val = val1 * val2;
//     if (val != 0)
//       cout << "  At location " << i+1 << " value = " << val1 << " * " << val2 << endl;
    sum += val;
  }

  cout << "Q x = " << sum << endl;

  double expectedVal = V_GetCmp(&rhs, num);

  cout << ((sum - expectedVal) / sum)*100 << "\% error in solver\n";
}
