/*
***********************************************************************
LICENSE AGREEMENT
***********************************************************************
 READ THIS LICENSE AGREEMENT CAREFULLY BEFORE USING THIS PRODUCT. BY 
 USING THIS PRODUCT YOU INDICATE YOUR ACCEPTANCE OF THE TERMS OF THE 
 FOLLOWING AGREEMENT. THESE TERMS APPLY TO YOU AND ANY SUBSEQUENT 
 LICENSEE OF THIS PRODUCT.

               License Agreement for RegularPlace 1.0

Copyright (c)  2008 by 
Ashutosh Chakraborty, Anurag Kumar, David Z. Pan
All rights reserved

ATTRIBUTION ASSURANCE LICENSE
Redistribution of the any code, with or without modification, are permitted 
provided that the conditions below are met. These conditions require a 
modest attribution to Ashutosh Chakraborty, Anurag Kumar and Dr. 
David Z. Pan (the "Author").

1. Redistributions of any code, with or without modification (the "Code"), 
must be accompanied by any documentation and, each time the resulting 
executable program or a program dependent thereon is launched, a 
prominent display (e.g., splash screen or banner text) of the Author's 
attribution information, which includes:
  (a) Ashutosh Chakraborty, Anurag Kumar, and Dr. David Z. Pan ("AUTHOR"),
  (b) The University of Texas at Austin ("PROFESSIONAL IDENTIFICATION"), and
  (c) http://www.cerc.utexas.edu/utda/ ("URL").

2. Use is limited to academic research groups only. Users who are interested 
in industry or commercial purposes must notify Author and request 
separate license agreement.  

3. Neither the name nor any trademark of the Author may be used to 
endorse or promote products derived from this software without specific 
prior written permission.

4. Users are entirely responsible, to the exclusion of the Author and 
any other persons, for compliance with (a) regulations set by owners or 
administrators of employed equipment, (b) licensing terms of any other 
software, and (c) local, national, and international regulations regarding 
use, including those regarding import, export, and use of encryption software. 

THIS FREE SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR 
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE AUTHOR OR ANY CONTRIBUTOR BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, EFFECTS OF UNAUTHORIZED OR MALICIOUS NETWORK 
ACCESS; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <iostream>
#include <vector>
#include <string>
#include <map>
#include <fstream>

using namespace std;

#include "Design.h"
#include "Row.h"

int CRow::mHeight    = -1;
int CRow::mSiteWidth = -1;
int CRow::mSiteSpace = -1;
int CRow::Siteorient = -1;
int CRow::mNumSites  = -1;

extern string makeString(int x);

void makeBanner()
{
  cout << "\n";
  cout << "                     RegPlace [release 1.0]                \n";
  cout << "========================================================== \n";
  cout << "Authors:  Ashutosh Chakraborty, Anurag Kumar, David Z. Pan \n";
  cout << "          University of Texas at Austin, Austin, TX 78703  \n";
  cout << "See LICENSE.txt for licence information.\n";
  cout << "Home: http://www.cerc.utexas.edu/utda/download/RegPlace.htm\n";
  cout << "\n";
}

#include <sys/stat.h>


// Copied from http://www.techbytes.ca/techbyte103.html
bool FileExists(string strFilename) 
{
  struct stat stFileInfo;
  bool blnReturn;
  int intStat;
  intStat = stat(strFilename.c_str(),&stFileInfo);

  if(intStat == 0) 
  {     
    blnReturn = true;
  } 
  else 
  {     
    blnReturn = false;
  }

  return(blnReturn);
}

// Arguments
// 0 -> Program name
// 1 -> Node file name
// 2 -> Nets file name
// 3 -> Pl file name
// 4 -> Props file name
// 5 -> SCL file name
// 6 -> Regions file name
int main(int argc, char* argv[])
{
  CDesign Design;

  makeBanner();

  if (false == FileExists("./bin/glpsol"))
  {
    cout << "\n";
    cout << "WARN:  This program assumes the binary glpsol (http://www.gnu.org/software/glpk/)\n";
    cout << "       If you dont have bin directory, please create it and copy glpsol there\n";
    cout << "       Will stop the placer now.  Please fix this and rerun\n";
    exit(1);
  }

  if (argc != 7)
  {
    cout << "argc is " << argc << endl;
    cout << "EROR: Invalid invocation.  Please call the binary as follows\n";
    cout << "      place.exe a.nodes a.nets a.pl a.prop a.scl a.regions\n";
    cout << "\n";
    exit(1);
  }

  string inodeFileName = argv[1];
  string inetsFileName = argv[2];
  string iPlFileName   = argv[3];
  string iPropFileName = argv[4];
  string iSCLFileName  = argv[5];
  string iRegnFileName = argv[6];

  string::size_type slashPos = inodeFileName.find_last_of("/");
  string::size_type dotPos   = inodeFileName.find_last_of(".");

  if (dotPos == string::npos)
    dotPos = inodeFileName.length();

  string benchName;
  if (slashPos == string::npos)
    benchName = inodeFileName.substr(0, dotPos - slashPos - 1);
  else
    benchName = inodeFileName.substr(slashPos + 1, dotPos - slashPos - 1);
  cout << inodeFileName << " " << benchName << endl;

  if (benchName.length() == 0)
    benchName = "CurrDesign";

  string newSclFileName  = string(benchName) + "_shrunk.scl";
  string newPlFileName   = string(benchName) + "_shrunk.pl";
  string newNodeFileName = string(benchName) + "_shrunk.nodes";

  Design.Read(
      inodeFileName,
      inetsFileName,
      iPropFileName,
      iRegnFileName,
      iSCLFileName,
      iPlFileName);

  Design.Prepare();

  /***********************************************************/
  // Global placer runs here.
  /***********************************************************/
  // argv[7] decides which GP to run.  
  // 0 means Capo, 1 means mPL, 2 means Gordian
  int gpChoice;
  if (argc < 8)
    gpChoice = 0;
  else
    gpChoice = atoi(argv[7]);
  bool gpSuccessful = false;
  switch (gpChoice)
  {
    case 0:
      gpSuccessful = Design.RunMinCutBasedPlacer(benchName);
      break;
    case 1:
      gpSuccessful = Design.RunAnalyticalPlacer(benchName);
      break;
    case 2:
      gpSuccessful = Design.RunGordianPlacer(benchName);
      break;
    default:
      gpSuccessful = Design.RunMinCutBasedPlacer(benchName);
      break;
  }

  if (false == gpSuccessful)
    cout << "**EROR** in global placement.  Check\n";


  /***********************************************************/
  // Detailed placer runs here.
  /***********************************************************/

  Design.CheckDensity();
  Design.CheckAllCellsInCore();
  Design.AssignNodesToTiles();

  Design.Refine();
  Design.CheckDensity();
  Design.CheckAllCellsInCore();

  Design.SolveBRAMAssignment(true);
  Design.CheckDensity();
  Design.CheckAllCellsInCore();
  Design.GetHPWL("PostBRAM");

  Design.SolveRegFileAssignment(true);
  Design.CheckDensity();
  Design.CheckAllCellsInCore();
  Design.GetHPWL("PostREGF");

  //   Design.WriteImage(string(benchName)+"PostRegBramSolve");
  Design.WritePLFile(string(benchName) + "PostRegBramSolve.pl");
  Design.GetHPWL("PostRegBramSolve");

  Design.Refine();
  Design.WriteImage("AfterCompleteRefinement");
  Design.Legalize();
  Design.CheckDensity();
  Design.GetHPWL("AfterLegalize1");
  Design.CheckAllCellsInCore();
  Design.WriteImage(string(benchName)+"AfterLegalize1");
  Design.WritePLFile(string(benchName) + "AfterLegalize1.pl");
  Design.checkIfEverythingOK();

  int globalSwapCount = 0;
  while (true)
  {
    cout << "Running GS for " << globalSwapCount << " times\n";
    double oldWl = Design.GetHPWL();
    Design.GlobalSwap();
    Design.checkIfEverythingOK();

    if (Design.GetHPWL(string("GS")+makeString(globalSwapCount)) > 0.999*oldWl)
      break;
    oldWl = Design.GetHPWL();
    globalSwapCount++;
  }
  Design.WritePLFile(string(benchName) + "AfterGlobalSwap.pl");

#if 0
  Design.GlobalSwap();
  Design.CheckAllCellsInCore();
  Design.CheckDensity();
  Design.GetHPWL("PostFirstGSwap");
  Design.WriteImage(string(benchName)+"PostFirstGSwap");
  Design.WritePLFile(string(benchName) + "PostFirstGSwap.pl");
  Design.checkIfEverythingOK();
  exit(1);

  Design.GlobalSwap();
  Design.CheckDensity();
  Design.CheckAllCellsInCore();
  Design.GetHPWL("PostSecondGSwap");
  Design.WriteImage(string(benchName)+"PostSecondGSwap");
  Design.WritePLFile(string(benchName) + "PostSecondGSwap.pl");
  Design.checkIfEverythingOK();
  exit(1);
#endif

  //   Design.Legalize();
  //   Design.GetHPWL("AfterLegalize2");
  Design.CheckDensity();
  Design.CheckAllCellsInCore();
  //   Design.WriteImage(string(benchName)+"AfterLegalize2");
  Design.WritePLFile(string(benchName) + "AfterLegalize2.pl");

  Design.OptimizeLegal();
  Design.GetHPWL("IntraTileOptim");
  Design.CheckDensity();
  Design.CheckAllCellsInCore();
  Design.WriteImage(string(benchName)+"IntraTileOptim");
  Design.WritePLFile(string(benchName) + "IntraTileOptim.pl");

  Design.WritePLFile(benchName + "_final.pl");

  return 0;
}
