/************************************************************************
LICENSE AGREEMENT
***********************************************************************
 READ THIS LICENSE AGREEMENT CAREFULLY BEFORE USING THIS PRODUCT. BY 
 USING THIS PRODUCT YOU INDICATE YOUR ACCEPTANCE OF THE TERMS OF THE 
 FOLLOWING AGREEMENT. THESE TERMS APPLY TO YOU AND ANY SUBSEQUENT 
 LICENSEE OF THIS PRODUCT.

               License Agreement for RegularPlace 1.0

Copyright (c)  2008 by 
Ashutosh Chakraborty, Anurag Kumar, David Z. Pan
All rights reserved

ATTRIBUTION ASSURANCE LICENSE
Redistribution of the any code, with or without modification, are permitted 
provided that the conditions below are met. These conditions require a 
modest attribution to Ashutosh Chakraborty, Anurag Kumar and Dr. 
David Z. Pan (the "Author").

1. Redistributions of any code, with or without modification (the "Code"), 
must be accompanied by any documentation and, each time the resulting 
executable program or a program dependent thereon is launched, a 
prominent display (e.g., splash screen or banner text) of the Author's 
attribution information, which includes:
  (a) Ashutosh Chakraborty, Anurag Kumar, and Dr. David Z. Pan ("AUTHOR"),
  (b) The University of Texas at Austin ("PROFESSIONAL IDENTIFICATION"), and
  (c) http://www.cerc.utexas.edu/utda/ ("URL").

2. Use is limited to academic research groups only. Users who are interested 
in industry or commercial purposes must notify Author and request 
separate license agreement.  

3. Neither the name nor any trademark of the Author may be used to 
endorse or promote products derived from this software without specific 
prior written permission.

4. Users are entirely responsible, to the exclusion of the Author and 
any other persons, for compliance with (a) regulations set by owners or 
administrators of employed equipment, (b) licensing terms of any other 
software, and (c) local, national, and international regulations regarding 
use, including those regarding import, export, and use of encryption software. 

THIS FREE SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR 
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE AUTHOR OR ANY CONTRIBUTOR BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, EFFECTS OF UNAUTHORIZED OR MALICIOUS NETWORK 
ACCESS; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <iostream>
#include <vector>
#include <string>
#include <map>
#include <cfloat>
#include <fstream>

using namespace std;

#include "BBox.h"
#include "Net.h"
#include "Design.h"

CNet::CNet(void) { }

CNet::~CNet(void) { }

CNet::CNet(string aName)
    : mName(aName), mIsDirty(false), mWirelength(0.0), mBBox(NULL)
{
}

// Add a node to this net.  Set the flag of dirty to be true.
void CNet::AddNode(CNode* aNode)
{
  mNodes.push_back(aNode);
  mIsDirty = true;
}

// Get HPWL of this net.
double CNet::GetHPWL()
{
  if (true == mIsDirty)
  {
    computeBB();
    assert(false == mIsDirty);
    mWirelength = mBBox->HPWL();
  }

  return mWirelength;
}

// Get the BBox of this net.  This updates the mBBox variable and
// returns its address.  Dont let anyone change the mBBox variable
// directly.  Thus return a const reference.
const CBBox* CNet::GetBBox()
{
  computeBB();
  return mBBox;
}

#if 0
// Compute the BBox of this net and store in the variable mBBox
// NOTE:  Only this function can unset the mIsDirty flag
void CNet::computeBB()
{
  // Early exit if nothing good has been done in this net
  // Good means adding nodes, or moving nodes.
  if (false == mIsDirty)
    return;

  mBBox = computeBBwithout(sInvalidNodeID);
  mIsDirty = false;
}
#endif

// Returns the bounding box of this net IF the node with id aID
// was not part of it.  This function is (at least) used in
// computing the Optimal location of a node during global cell swap.
#define max(a, b) a>b ? a : b
#define min(a, b) a<b ? a : b
CBBox* CNet::computeBBwithout(const int aSkipThisID)
{
  assert(mNodes.size() > 0);

  double gxMin = DBL_MAX;
  double gxMax = DBL_MIN;
  double gyMin = DBL_MAX;
  double gyMax = DBL_MIN;

  for (int i = mNodes.size() - 1; i >= 0; --i)
  {
    if (aSkipThisID == mNodes[i]->ID())
      continue;
    gxMin = (gxMin > mNodes[i]->X()) ? mNodes[i]->X() : gxMin;
    gxMax = (gxMax < mNodes[i]->X()) ? mNodes[i]->X() : gxMax;
    gyMin = (gyMin > mNodes[i]->Y()) ? mNodes[i]->Y() : gyMin;
    gyMax = (gyMax < mNodes[i]->Y()) ? mNodes[i]->Y() : gyMax;
  }

  assert(gxMin <= gxMax);
  assert(gyMin <= gyMax);

  CBBox* tmpBB = new CBBox(gxMin, gxMax, gyMin, gyMax);

  return tmpBB;
}

int CNet::GetNumNodesInPartition(int iPartitionId)
{
  int count = 0;
  for (int i = 0;i < mNodes.size();i++)
  {
    if (mNodes[i]->PartitionId() == iPartitionId)
      count++;
  }
  return count;
}
