/*
***********************************************************************
LICENSE AGREEMENT
***********************************************************************
 READ THIS LICENSE AGREEMENT CAREFULLY BEFORE USING THIS PRODUCT. BY 
 USING THIS PRODUCT YOU INDICATE YOUR ACCEPTANCE OF THE TERMS OF THE 
 FOLLOWING AGREEMENT. THESE TERMS APPLY TO YOU AND ANY SUBSEQUENT 
 LICENSEE OF THIS PRODUCT.

               License Agreement for RegularPlace 1.0

Copyright (c)  2008 by 
Ashutosh Chakraborty, Anurag Kumar, David Z. Pan
All rights reserved

ATTRIBUTION ASSURANCE LICENSE
Redistribution of the any code, with or without modification, are permitted 
provided that the conditions below are met. These conditions require a 
modest attribution to Ashutosh Chakraborty, Anurag Kumar and Dr. 
David Z. Pan (the "Author").

1. Redistributions of any code, with or without modification (the "Code"), 
must be accompanied by any documentation and, each time the resulting 
executable program or a program dependent thereon is launched, a 
prominent display (e.g., splash screen or banner text) of the Author's 
attribution information, which includes:
  (a) Ashutosh Chakraborty, Anurag Kumar, and Dr. David Z. Pan ("AUTHOR"),
  (b) The University of Texas at Austin ("PROFESSIONAL IDENTIFICATION"), and
  (c) http://www.cerc.utexas.edu/utda/ ("URL").

2. Use is limited to academic research groups only. Users who are interested 
in industry or commercial purposes must notify Author and request 
separate license agreement.  

3. Neither the name nor any trademark of the Author may be used to 
endorse or promote products derived from this software without specific 
prior written permission.

4. Users are entirely responsible, to the exclusion of the Author and 
any other persons, for compliance with (a) regulations set by owners or 
administrators of employed equipment, (b) licensing terms of any other 
software, and (c) local, national, and international regulations regarding 
use, including those regarding import, export, and use of encryption software. 

THIS FREE SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR 
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE AUTHOR OR ANY CONTRIBUTOR BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, EFFECTS OF UNAUTHORIZED OR MALICIOUS NETWORK 
ACCESS; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

#ifndef __NET_H__
#define __NET_H__

// #include "MFRouter.h"
// #include "Grid.h"
#include "Util.h"
#include "Data.h"

// class CGrid;
// class CNode;
// class CBBox;

class CNet
{
  public:
    CNet(string aName);
    ~CNet(void);

  private: // No no-name nets
    CNet(void);

  public:
    void AddNode(CNode *aNode);
    double GetHPWL();

  public:  // Get set functions
    string Name()
    {
      return mName;
    }
    bool   Dirty()
    {
      return mIsDirty;
    }
    void   Dirty(bool aIsDirty)
    {
      assert(aIsDirty);
      mIsDirty = true;
    }
    int    GetDegree()
    {
      return mNodes.size();
    }
    int    GetNumNodesInPartition(int iPartitionId);

    // Return a reference to the vector containing
    // the nodes which the current net has
    const vector <CNode*>& GetCellsInNet() const
    {
      return mNodes;
    }

  public:  // Wirelength related functions

    // Get the BBox of this net.  This updates the mBBox variable and
    // returns its address.  Dont let anyone change the mBBox variable
    // directly.  Thus return a const reference.
    const CBBox* GetBBox();

    // Get the BBox of this net without the node whose ID is
    // aSkipThisID.  Since this variable is temporary, we dont
    // mind if someone changes it.  Therefore non-const is returned
    CBBox* GetBBoxWithoutID(const int aSkipThisID)
    {
      return computeBBwithout(aSkipThisID);
    }

    // Same as above.  But the exclusion is based on the pointer
    // to the Node which needs to be excluded (instead of its ID).
    // Since this variable is temporary, we dont mind if someone
    // changes it.  Therefore non-const is returned
    CBBox* GetBBoxWithoutNode(const CNode *aNode)
    {
      return computeBBwithout(aNode->ID());
    }

  private:
    static const int sInvalidNodeID = -1;
    // Compute the BBox of this net and store in the variable mBBox
    // NOTE:  Only this function can unset the mIsDirty flag
    void computeBB()
    {
      if (false == mIsDirty)
        return;
      if (mBBox != NULL)
      {
        delete mBBox;
        mBBox = NULL;
      }
      mBBox = computeBBwithout(sInvalidNodeID);
      mIsDirty = false;
    }

    // Returns the bounding box of this net IF the node with id aID
    // was not part of it.  This function is (at least) used in
    // computing the Optimal location of a node during global cell swap.
    CBBox* computeBBwithout(const int aSkipThisID);

  public:
    vector <CNode*> mNodes;

  private:
    double mWirelength;
    CBBox*  mBBox;
    string mName;
    bool mIsDirty;
};

#endif
