/*
***********************************************************************
LICENSE AGREEMENT
***********************************************************************
 READ THIS LICENSE AGREEMENT CAREFULLY BEFORE USING THIS PRODUCT. BY 
 USING THIS PRODUCT YOU INDICATE YOUR ACCEPTANCE OF THE TERMS OF THE 
 FOLLOWING AGREEMENT. THESE TERMS APPLY TO YOU AND ANY SUBSEQUENT 
 LICENSEE OF THIS PRODUCT.

               License Agreement for RegularPlace 1.0

Copyright (c)  2008 by 
Ashutosh Chakraborty, Anurag Kumar, David Z. Pan
All rights reserved

ATTRIBUTION ASSURANCE LICENSE
Redistribution of the any code, with or without modification, are permitted 
provided that the conditions below are met. These conditions require a 
modest attribution to Ashutosh Chakraborty, Anurag Kumar and Dr. 
David Z. Pan (the "Author").

1. Redistributions of any code, with or without modification (the "Code"), 
must be accompanied by any documentation and, each time the resulting 
executable program or a program dependent thereon is launched, a 
prominent display (e.g., splash screen or banner text) of the Author's 
attribution information, which includes:
  (a) Ashutosh Chakraborty, Anurag Kumar, and Dr. David Z. Pan ("AUTHOR"),
  (b) The University of Texas at Austin ("PROFESSIONAL IDENTIFICATION"), and
  (c) http://www.cerc.utexas.edu/utda/ ("URL").

2. Use is limited to academic research groups only. Users who are interested 
in industry or commercial purposes must notify Author and request 
separate license agreement.  

3. Neither the name nor any trademark of the Author may be used to 
endorse or promote products derived from this software without specific 
prior written permission.

4. Users are entirely responsible, to the exclusion of the Author and 
any other persons, for compliance with (a) regulations set by owners or 
administrators of employed equipment, (b) licensing terms of any other 
software, and (c) local, national, and international regulations regarding 
use, including those regarding import, export, and use of encryption software. 

THIS FREE SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR 
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE AUTHOR OR ANY CONTRIBUTOR BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, EFFECTS OF UNAUTHORIZED OR MALICIOUS NETWORK 
ACCESS; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

/*
 * =======================================================================
 *
 *       Filename:  Region.cpp
 *
 *    Description:  Implementation of Regions
 *
 *        Version:  1.0
 *        Created:  09/02/2008 04:10:59 PM CDT
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ashutosh Chakraborty (), ashutosh@mail.utexas.edu
 *        Company:  University of Texas, Austin
 *
 * =======================================================================
 */

#include <string>
#include <iostream>
#include <sstream>
#include <vector>
#include <cfloat>
#include <map>
#include "math.h"
#include "assert.h"

using namespace std;

#include "Region.h"
#include "Design.h"
#include "Tile.h"

extern map <string, CRegion*> gMapRegionNameToPtr;

CRegion::CRegion()
    : mName(""), mHeight(-1), mWidth(-1), mType(-1), mNumClkDom(-1)
{
  mRegions.clear();
}

CRegion::CRegion(const CRegion& r)
{
  cout << "Copy constructor called... Unexpectedly" << endl;
  abort();
  mName = r.mName;
  mHeight = r.mHeight;
  mWidth = r.mWidth;
  mType = r.mType;
  mNumClkDom = r.mNumClkDom;
  mRegions = r.mRegions;
}

vector<CRegion*> gRegionsVec;
vector<double>   gRegionXOff;
vector<double>   gRegionYOff;

int CRegion::Flatten(double xOff, double yOff, int recurDepth)
{
  // When dealing with the top-most module, clear the space for
  // holding the regions eventually
  if (0 == recurDepth)
  {
    gRegionYOff.clear();
    gRegionXOff.clear();
    gRegionsVec.clear();
  }

  if (mRegions.size() == 0)
//   if ( "group" == mName )
  {
    gRegionsVec.push_back(this);
    gRegionXOff.push_back(xOff);
    gRegionYOff.push_back(yOff);
//     cout << this << " " << xOff << " " << yOff << endl;
  }

  for (int i = 0; i < mRegions.size(); i++)
    mRegions[i]->Flatten(xOff + mXOffset[i], yOff + mYOffset[i], recurDepth + 1);

  if (0 == recurDepth)
  {
    assert(gRegionYOff.size() == gRegionXOff.size());
    assert(gRegionXOff.size() == gRegionsVec.size());
    assert(gRegionXOff.size() > 0);

    for (int i = 0; i < gRegionsVec.size(); i++)
    {
      CRegion *tmp = gRegionsVec[i];
#if _VERBOSE
      cout << gRegionXOff[i] << "," << gRegionYOff[i] << " " <<
      gRegionXOff[i] + tmp->Width() << "," <<
      gRegionYOff[i] + tmp->Height() << " " << tmp->Name() << endl;
#endif
    }
  }

  return gRegionsVec.size();
}

string CRegion::Info(int count)
{
  ostringstream os;

  os << "Region " << mName << " Ht=" << mHeight << " Wd=" <<
  mWidth << " type=" << gMapRegionNameToPtr[mName] << " #Dom=" <<
  mNumClkDom << " #reg=" << mRegions.size();

  for (unsigned int i = 0; i < mRegions.size(); i++)
  {
    CRegion *child = mRegions[i];
    os << "\n";
    for (int j = 0; j < count + 1; j++)
      os << " ";
    os << "at " << mXOffset[i] << "," << mYOffset[i] << " - ";
    os << child->Info(count + 1);
  }

  return os.str();
}

extern vector<CTile*> gTilesVec;
int CRegion::GetTiles(double xOff, double yOff)
{
  assert(0 < xOff && 0 < yOff);
  if ("group" == mName)
    gTilesVec.push_back(new CTile(xOff, yOff));
  else
    for (int i = 0; i < mRegions.size(); i++)
      mRegions[i]->GetTiles(xOff + mXOffset[i], yOff + mYOffset[i]);
}

void CRegion::AddRegion(CRegion *aRegion, double xOffset, double yOffset)
{
  mRegions.push_back(aRegion);
  mXOffset.push_back(xOffset);
  mYOffset.push_back(yOffset);
}

void CRegion::ComputeDimension()
{
  // Termination condition for the recursive function
  if (0 == mRegions.size())
    return;

  // Early exit in case this block's dimension is already determined
  if (-1 != mWidth && -1 != mHeight)
    return;

  // Compute the children's dimensions
  for (int i = 0; i < mRegions.size(); i++)
    mRegions[i]->ComputeDimension();

  double x_m = DBL_MAX;
  double x_M = DBL_MIN;
  double y_m = DBL_MAX;
  double y_M = DBL_MIN;

  for (int i = 0; i < mRegions.size(); i++)
  {
    CRegion* tmpR = mRegions[i];

    double xleft = mXOffset[i];
    double xright = mXOffset[i] + mRegions[i]->Width();
    if (xleft < x_m) x_m = xleft;
    if (xright > x_M) x_M = xright;

    double ybottom = mYOffset[i];
    double ytop = mYOffset[i] + mRegions[i]->Height();
    if (ybottom < y_m) y_m = ybottom;
    if (ytop > y_M) y_M = ytop;
  }

  assert(x_m <= x_M);
  assert(y_m <= y_M);

  mWidth = x_M - x_m;
  mHeight = y_M - y_m;
}
