/************************************************************************
LICENSE AGREEMENT
***********************************************************************
 READ THIS LICENSE AGREEMENT CAREFULLY BEFORE USING THIS PRODUCT. BY 
 USING THIS PRODUCT YOU INDICATE YOUR ACCEPTANCE OF THE TERMS OF THE 
 FOLLOWING AGREEMENT. THESE TERMS APPLY TO YOU AND ANY SUBSEQUENT 
 LICENSEE OF THIS PRODUCT.

               License Agreement for RegularPlace 1.0

Copyright (c)  2008 by 
Ashutosh Chakraborty, Anurag Kumar, David Z. Pan
All rights reserved

ATTRIBUTION ASSURANCE LICENSE
Redistribution of the any code, with or without modification, are permitted 
provided that the conditions below are met. These conditions require a 
modest attribution to Ashutosh Chakraborty, Anurag Kumar and Dr. 
David Z. Pan (the "Author").

1. Redistributions of any code, with or without modification (the "Code"), 
must be accompanied by any documentation and, each time the resulting 
executable program or a program dependent thereon is launched, a 
prominent display (e.g., splash screen or banner text) of the Author's 
attribution information, which includes:
  (a) Ashutosh Chakraborty, Anurag Kumar, and Dr. David Z. Pan ("AUTHOR"),
  (b) The University of Texas at Austin ("PROFESSIONAL IDENTIFICATION"), and
  (c) http://www.cerc.utexas.edu/utda/ ("URL").

2. Use is limited to academic research groups only. Users who are interested 
in industry or commercial purposes must notify Author and request 
separate license agreement.  

3. Neither the name nor any trademark of the Author may be used to 
endorse or promote products derived from this software without specific 
prior written permission.

4. Users are entirely responsible, to the exclusion of the Author and 
any other persons, for compliance with (a) regulations set by owners or 
administrators of employed equipment, (b) licensing terms of any other 
software, and (c) local, national, and international regulations regarding 
use, including those regarding import, export, and use of encryption software. 

THIS FREE SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR 
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE AUTHOR OR ANY CONTRIBUTOR BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, EFFECTS OF UNAUTHORIZED OR MALICIOUS NETWORK 
ACCESS; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*
 * =======================================================================
 *
 *       Filename:  Tile.cpp
 *
 *    Description:  Implementation of Tile class
 *
 *        Version:  1.0
 *        Created:  10/01/2008 05:14:19 PM CDT
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ashutosh Chakraborty (), ashutosh@mail.utexas.edu
 *        Company:  University of Texas, Austin
 *
 * =======================================================================
 */

#include <iostream>
#include <map>
#include <sstream>
#include <algorithm>
#include <fstream>
#include <numeric>

#include "math.h"

#include "Util.h"
#include "Design.h"
#include "Data.h"
#include "Tile.h"
#include "munkres.h"
using namespace std;

map <int, int> CTile::mMapInitXOfDffColm;
map <int, int> CTile::mMapInitXOfECellColm;
map <int, int> CTile::sECMapXToColmNum;
map <int, int> CTile::sEDMapXToColmNum;

extern string makeString(int x);

void CTile::Clear()
{

  // Clear flags
  mIsLegal = false;
  mIsDirty = true;

  // Clear the vectors
  mECells.clear();
  mDffs.clear();
  mRegs.clear();
  mBrams.clear();

  // Clear individual vectors
  mColumnsECell.clear();
  mColumnsEDff.clear();
}

void CTile::AddNode(CNode* tmpNode)
{
  if (tmpNode->Type() < 0 || tmpNode->Type() > 3)
  {
    cout << "**EROR:  Node " << tmpNode->Name() << "'s type " <<
    "not identifyable.  Is it fixed node?  If so, dont add " <<
    "such a node to any tile" << endl;
    exit(1);
  }

  mIsDirty = true;

  switch (tmpNode->Type())
  {
  case 0:
    mECells.push_back(tmpNode);
    break;
  case 1:
    mDffs.push_back(tmpNode);
    break;
  case 2:
    mRegs.push_back(tmpNode);
    break;
  case 3:
    mBrams.push_back(tmpNode);
    break;
  default:
    abort();
  }

#if 0
  // Need to take care of clocks if the WHOLE design
  // has less clocks than each tile can hold.
  if (gDesign->GetNumClkDomains() <= gMaxClockInTile)
  {
    abort();
    return;
  }
#endif

  for (int i = 0; i < tmpNode->mClkDom.size(); i++)
  {
    int clkCell = tmpNode->mClkDom[i];
    if (mNumCellsOfClk.find(clkCell) == mNumCellsOfClk.end())
      mNumCellsOfClk[clkCell] = 1;
    else
      mNumCellsOfClk[clkCell]++;
  }
}

void CTile::removeFromVec(vector<CNode*>& v, const CNode* aNode)
{
  vector<CNode *>::iterator it = v.begin();

  bool isRemoved = false;
  while (it != v.end())
  {
    if ((*it)->ID() == aNode->ID())
//     if ((*it)->Name() == aNode->Name())
    {
      v.erase(it);
      isRemoved = true;
      break;
    }
    it++;
  }

  if (false == isRemoved)
  {
    cout << "  Could not remove node " << aNode->Name() << " " << aNode->X() << " " << aNode->Y() << ".  Whats up???\n";
    abort();
  }
}

void CTile::RemoveNode(const CNode *aNode)
{
//   cout << "STATUS:  Entering function RemoveNode()\n";

  // Remove from the vector of nodes
  switch (aNode->Type())
  {
  case 0:
    removeFromVec(mECells, aNode);
    break;
  case 1:
    removeFromVec(mDffs, aNode);
    break;
  case 2:
    removeFromVec(mRegs, aNode);
    break;
  case 3:
    removeFromVec(mBrams, aNode);
    break;
  default:
    abort();
  }

  // Remove its contribution to clocks
  for (int i = 0; i < aNode->mClkDom.size(); i++)
  {
    assert(mNumCellsOfClk.find(aNode->mClkDom[i]) !=
           mNumCellsOfClk.end());
    assert(0 < mNumCellsOfClk[aNode->mClkDom[i]]);

    mNumCellsOfClk[aNode->mClkDom[i]]--;

    if (mNumCellsOfClk[aNode->mClkDom[i]] == 0)
      mNumCellsOfClk.erase(mNumCellsOfClk.find(aNode->mClkDom[i]));
  }

//   cout << "         Exiting function RemoveNode()\n";
}

void CTile::RemoveNodes(const vector<CNode *>& v)
{
  abort();  // Need to write this function properly.  Dont use yet.
#if 0
  cout << "STATUS:  Entering function RemoveNodes()\n";

  // Add the ID of the nodes which we need to search and remove
  map <int, bool> nodeMap;
  for (int i = 0; i < v->size(); i++)
    map[(*v)[i]->ID()] = true;

  cout << "         Exiting function RemoveNodes()\n";
#endif
}

bool CTile::CanTake(int aType) const
{
  switch (aType)
  {
  case 0:
    return (mECells.size() < gMaxECellAllowed);
    break;
  case 1:
    return (mDffs.size() < gMaxDffAllowed);
    break;
  case 2:
    return (mRegs.size() < gMaxRegsAllowed);
    break;
  case 3:
    return (mBrams.size() < gMaxBramsAllowed);
    break;
  default:
    abort();
  }
}

bool CTile::CanTake(CNode* aNode) const
{
  if (false == (CanTake(aNode->Type()) && IsClockAcceptable(aNode)))
    return false;

  if (NumColumnReqd() < mColumnsEDff.size())
    return true;

  if (NumColumnReqd() > mColumnsEDff.size())
    return false;

  if (NumColumnReqd() == mColumnsEDff.size())
  {
    static int numTimes = 0;
    if (0 == numTimes)
    {
      numTimes++;
      cout << "**WARN**:  Returning 0 wrongly in CanTake()\n";
    }
    return false;
  }
}

bool CTile::IsClockAcceptable(CNode *aNode) const
{
  if (0 >= aNode->mClkDom.size())
    return true;

  if (mNumCellsOfClk.size() < gMaxClockInTile)
    return true;

  bool allClkInNodeAreAlreadyInThisTile = true;
  for (int i = 0; i < aNode->mClkDom.size(); i++)
    if (mNumCellsOfClk.find(aNode->mClkDom[i]) ==
        mNumCellsOfClk.end())
      allClkInNodeAreAlreadyInThisTile = false;

  return allClkInNodeAreAlreadyInThisTile;
}

void CTile::WriteImage(string fileBaseName)
{
  cout << "STATUS:  Entering CDesign::WriteImage() (to file " <<
  fileBaseName << ")" << endl;

  if (mECells.size() + mDffs.size() + mRegs.size() + mBrams.size() == 0)
    return;

  string dataFileName = fileBaseName;
  dataFileName += ".dat";

  std::ofstream out;
  out.open(dataFileName.c_str());

  ostringstream os;
  os << "set style rectangle back fc rgb \"white\" fs solid 1.0 noborder\n";

  int numObjectWritten = 0;
  for (unsigned int i = 0; i < mECells.size(); ++i)
  {
      os << "set object " << numObjectWritten + 1 << " rectangle from " <<
      mECells[i]->X() << "," << mECells[i]->Y() << " to " <<
      mECells[i]->X() + 1 << "," << mECells[i]->Y() + 1 <<
      " fc rgb \"green\"\n";
      numObjectWritten++;
  }
  
  for (unsigned int i = 0; i < mDffs.size(); ++i)
  {
      os << "set object " << numObjectWritten + 1 << " rectangle from " <<
      mDffs[i]->X() << "," << mDffs[i]->Y() << " to " <<
      mDffs[i]->X() + 1 << "," << mDffs[i]->Y() + 1 <<
      " fc rgb \"blue\"\n";
      numObjectWritten++;
  }
  
  for (unsigned int i = 0; i < mRegs.size(); ++i)
  {
      os << "set object " << numObjectWritten + 1 << " rectangle from " <<
      mRegs[i]->X() << "," << mRegs[i]->Y() << " to " <<
      mRegs[i]->X() + 8 << "," << mRegs[i]->Y() + 32 <<
      " fc rgb \"red\"\n";
      numObjectWritten++;
  }
  
  for (unsigned int i = 0; i < mBrams.size(); ++i)
  {
      os << "set object " << numObjectWritten + 1 << " rectangle from " <<
      mBrams[i]->X() << "," << mBrams[i]->Y() << " to " <<
      mBrams[i]->X() + 16 << "," << mBrams[i]->Y() + 64 <<
      " fc rgb \"yellow\"\n";
      numObjectWritten++;
  }
  out.close();

  string gplotName = fileBaseName;
  gplotName += ".plt";

  out.open(gplotName.c_str());

  out << "set terminal gif large enhanced size 1280,1024" << endl;
//   out << "set terminal postscript color" << endl;
  out << "set xrange [" << mX << ":" << 100+mX << "]" << endl;
  out << "set yrange [" << mY << ":" << 100+mY << "]" << endl;
  out << "set output '" << fileBaseName << ".gif'" << endl;
//   out << "set output '" << fileBaseName << ".ps'" << endl;
  out << os.str();
//   out << "plot '" << dataFileName << "'" << endl;
  out << "plot x" << endl;
  out << "quit" << endl;

#if 0
  // Make the gif
  string cmd = "./rp.csh eda09 gnuplot ";
  cmd += gplotName;
  cmd += " >& /dev/null";
#else
  cout << "INFO:  Run gnuplot " << gplotName << " to generate picture\n";
#endif

//   system(cmd.c_str());
}

vector<int>& CTile::GetClockIdsInside() const
{
  static vector<int> clockIdVec;

  if (false == mIsDirty)
    return clockIdVec;

  clockIdVec.clear();

  map<int, int>::const_iterator it;
  for (it = mNumCellsOfClk.begin(); it != mNumCellsOfClk.end(); it++)
    clockIdVec.push_back(it->first);

  assert(clockIdVec.size() == mNumCellsOfClk.size());
  return clockIdVec;
}

bool CTile::IsDensityOK(int aType) const
{
  if (-1 == aType)
  {
    return (
             (mECells.size() <= gMaxECellAllowed) &&
             (mDffs.size() <= gMaxDffAllowed) &&
             (mRegs.size() <= gMaxRegsAllowed) &&
             (mBrams.size() <= gMaxBramsAllowed) &&
             (mNumCellsOfClk.size() <= gMaxClockInTile)
           );
  }
  else
  {
    switch (aType)
    {
    case 0:
      return (mECells.size() <= gMaxECellAllowed);
      break;
    case 1:
      return (mDffs.size() <= gMaxDffAllowed);
      break;
    case 2:
      return (mRegs.size() <= gMaxRegsAllowed);
      break;
    case 3:
      return (mBrams.size() <= gMaxBramsAllowed);
      break;
    case 4:
      return (mNumCellsOfClk.size() <= gMaxClockInTile);
      break;
    default:
      abort();
    }
  }
}

int CTile::getCapacity(int aType) const
{
  if (-1 == aType)
  {
    return (
             (gMaxECellAllowed  - mECells.size()) +
             (gMaxDffAllowed    - mDffs.size())   +
             (gMaxRegsAllowed   - mRegs.size())   +
             (gMaxBramsAllowed  - mBrams.size())
           );
  }
  else
  {
    switch (aType)
    {
    case 0:
      return (gMaxECellAllowed  - mECells.size());
      break;
    case 1:
      return (gMaxDffAllowed    - mDffs.size());
      break;
    case 2:
      return (gMaxRegsAllowed   - mRegs.size());
      break;
    case 3:
      return (gMaxBramsAllowed  - mBrams.size());
      break;
    default:
      abort();
    }
  }
}

bool CTile::IsLegal() const
{
  if (false == mIsDirty)
    return mIsLegal;
  return false;
}


bool CTile::Legalize()
{
  cout << "Legalizing " << *this << endl;

  if (false == IsDensityOK())
  {
    cout << "  ** EROR:  Cannot legalize.  Density  not OK here" << endl;
    cout << "     CONTINUING for now\n";
    if (2304 < mECells.size())
        abort();

//     abort();
  }

#if 0
  for (int i = 0; i < mECells.size(); i++)
    mECells[i]->PutAt(mECells[i]->X() - mX, mECells[i]->Y() - mY);
  for (int i = 0; i < mDffs.size(); i++)
    mDffs[i]->PutAt(mDffs[i]->X() - mX, mDffs[i]->Y() - mY);
#endif

  LegalizeToClosestPosition(mECells);
  LegalizeToClosestPosition(mDffs);
  CheckECellAndDffPosnLegality();

  LegalizeDffsToClock();
//   assignECellsToBestColumn(mX, mY);
//   assignEDffsToBestColumn(mX, mY);

//   gDesign->GetHPWL("during legalizing");
  mIsDirty = false;
  mIsLegal = true;
  return mIsLegal;
}

void CTile::CheckECellAndDffPosnLegality()
{
  for (int i=0; i<mECells.size(); i++)
  {
    assert(sECMapXToColmNum.find(int(mECells[i]->X() - mX)) != sECMapXToColmNum.end());
    assert(mECells[i]->Y() - mY >= 0 && mECells[i]->Y() - mY < 64);
  }
  for (int i=0; i<mDffs.size(); i++)
  {
    assert(sEDMapXToColmNum.find(int(mDffs[i]->X() - mX)) != sEDMapXToColmNum.end());
    assert(mDffs[i]->Y() - mY >= 0 && mDffs[i]->Y() - mY < 64);
  }
}

#if 1
bool cmpTiles(const CTile* a, const CTile* b)
{
  if (a->mX < b->mX)
    return true;

  if (a->mX > b->mX)
    return false;

  if (a->mY < b->mY)
    return true;

  if (a->mY > b->mY)
    return false;

  return false;
}
#endif

void CTile::RemoveBRAM()
{
  abort();
//   mBrams.clear();
//   mBramsClks.clear();
}

void CTile::RemoveRegFiles()
{
  abort();
//   mRegs.clear();
//   mRegsClks.clear();
}

int CTile::ExtraOfType(int aType) const
{
  assert(0 == aType);
  switch (aType)
  {
  case 0:
    return (mECells.size() - gMaxECellAllowed);
    break;
  case 1:
    return (mDffs.size() - gMaxDffAllowed);
    break;
  case 2:
    return (mRegs.size() - gMaxRegsAllowed);
    break;
  case 3:
    return (mBrams.size() - gMaxBramsAllowed);
    break;
  default:
    abort();
  }
}

bool bIsCellsAssignedToColm = false;
void CTile::AssignCellsToColumns()
{
  bIsCellsAssignedToColm = true;
  assert(24 == mColumnsEDff.size());
  assert(36 == mColumnsECell.size());

  for (int i = 0; i < mColumnsEDff.size(); i++)
    mColumnsEDff[i].clear();

  for (int i = 0; i < mDffs.size(); i++)
  {
    int locationInt = static_cast<int>(mDffs[i]->X() - mX);
    assert(locationInt >= 0);
    int columnNum = sEDMapXToColmNum[locationInt];
    assert(sEDMapXToColmNum.find(locationInt) != sEDMapXToColmNum.end());
    mColumnsEDff[columnNum].push_back(mDffs[i]);
    assert(mColumnsEDff[columnNum].size() <= gMaxColmHeight);
  }

  for (int i = 0; i < mColumnsEDff.size(); i++)
    sort(mColumnsEDff[i].begin(), mColumnsEDff[i].end(), cmpNodes);

  for (int i = 0; i < mColumnsECell.size(); i++)
    mColumnsECell[i].clear();

  for (int i = 0; i < mECells.size(); i++)
  {
    int locationInt = static_cast<int>(mECells[i]->X() - mX);
    assert(locationInt >= 0);
    int columnNum = sECMapXToColmNum[locationInt];
    assert(sECMapXToColmNum.find(locationInt) != sECMapXToColmNum.end());
    mColumnsECell[columnNum].push_back(mECells[i]);
  }

  for (int i = 0; i < mColumnsECell.size(); i++)
    sort(mColumnsECell[i].begin(), mColumnsECell[i].end(), cmpNodes);

}

void CTile::LinearPlacement()
{
  cout << *this << endl;

  // Do the DFF.
  for (int i = 0; i < mColumnsEDff.size(); i++)
    linearPlaceColumn(mColumnsEDff[i]);

  // Do for the ecell
  for (int i = 0; i < mColumnsECell.size(); i++)
    linearPlaceColumn(mColumnsECell[i]);
}

#if 0
void CTile::LocalReordering()
{
  if (false == bIsCellsAssignedToColm)
  {
    cout << "** EROR:  Cannot do cell reordering before assinging to rows\n";
    return;
  }

  for (int i = 0; i < mColumnsEDff.size(); i++)
    localReorder(mColumnsEDff[i]);

  for (int i = 0; i < mColumnsECell.size(); i++)
    localReorder(mColumnsECell[i]);
}
#else
void CTile::LocalReordering()
{
  map<int, int>::iterator it = mMapInitXOfDffColm.begin();
  for ( ; it != mMapInitXOfDffColm.end(); it++)
    localReorderUp(it->second);

  it = mMapInitXOfECellColm.begin();
  for ( ; it != mMapInitXOfECellColm.end(); it++)
    localReorderUp(it->second);

  it = mMapInitXOfDffColm.begin();
  for ( ; it != mMapInitXOfDffColm.end(); it++)
    localReorderDown(it->second);
  it = mMapInitXOfECellColm.begin();
  for ( ; it != mMapInitXOfECellColm.end(); it++)
    localReorderDown(it->second);
}
#endif

void CTile::localReorderUp(int x)
{
  for (int i=0; i<64-1; i++)
  {
    if (true == vecBool[x][i].isOccupied)
    {
      double cost = vecBool[x][i].n->GetIncidentWL();
      if (false == vecBool[x][i+1].isOccupied)
      {
        vecBool[x][i].n->Y(mY+i+1);
        if (cost < vecBool[x][i].n->GetIncidentWL())
          vecBool[x][i].n->Y(mY+i);
        else
        {
          vecBool[x][i+1].n = vecBool[x][i].n;
          vecBool[x][i].n = NULL;
          vecBool[x][i].isOccupied = false;
          vecBool[x][i+1].isOccupied = true;
        }
      }
      else
      {
        // Swap these?
        double cost2 = vecBool[x][i+1].n->GetIncidentWL();
        vecBool[x][i].n->Y(mY+i+1);
        vecBool[x][i+1].n->Y(mY+i);
        if (cost + cost2 < vecBool[x][i].n->GetIncidentWL() + vecBool[x][i+1].n->GetIncidentWL())
        {
          vecBool[x][i].n->Y(mY+i);
          vecBool[x][i+1].n->Y(mY+i+1);
        }
        else
        {
          CNode *tmp = vecBool[x][i].n;
          vecBool[x][i].n = vecBool[x][i+1].n;
          vecBool[x][i+1].n = tmp;
        }
      }
    }
  }
}

void CTile::localReorderDown(int x)
{
  for (int i=63; i>0; i--)
  {
    if (true == vecBool[x][i].isOccupied)
    {
      double cost = vecBool[x][i].n->GetIncidentWL();
      if (false == vecBool[x][i-1].isOccupied)
      {
        vecBool[x][i].n->Y(mY+i-1);
        if (cost < vecBool[x][i].n->GetIncidentWL())
          vecBool[x][i].n->Y(mY+i);
        else
        {
          vecBool[x][i-1].n = vecBool[x][i].n;
          vecBool[x][i].n = NULL;
          vecBool[x][i].isOccupied = false;
          vecBool[x][i-1].isOccupied = true;
        }
      }
      else
      {
        // Swap these?
        double cost2 = vecBool[x][i-1].n->GetIncidentWL();
        vecBool[x][i].n->Y(mY+i-1);
        vecBool[x][i-1].n->Y(mY+i);
        if (cost + cost2 < vecBool[x][i].n->GetIncidentWL() + vecBool[x][i-1].n->GetIncidentWL())
        {
          vecBool[x][i].n->Y(mY+i);
          vecBool[x][i-1].n->Y(mY+i-1);
        }
        else
        {
          CNode *tmp = vecBool[x][i].n;
          vecBool[x][i].n = vecBool[x][i-1].n;
          vecBool[x][i-1].n = tmp;
        }
      }
    }
  }
}

void CTile::localReorder(vector<CNode *>& v)
{
  if (v.size() < 3)
    return;

  int bestconfig = -1;
  double bestcost = 10000000.0;
  double cost = 0;

  static int numTimesSwapHelped = 0;

  for (int i = 0; i < v.size() - 2; i++)
  {
    bestcost = 100000000.0;

    // A -> B -> C
    cost = v[i]->GetIncidentWL() + v[i+1]->GetIncidentWL() + v[i+2]->GetIncidentWL();
    if (cost < bestcost)
    {
      bestconfig = 1;
      bestcost   = cost;
    }

    // A -> C -> B
    v[i+1]->SwapLocation(v[i+2]);
    cost = v[i]->GetIncidentWL() + v[i+1]->GetIncidentWL() + v[i+2]->GetIncidentWL();
    if (cost < bestcost)
    {
      bestconfig = 2;
      bestcost   = cost;
    }

    // B -> C -> A
    v[i]->SwapLocation(v[i+2]);
    cost = v[i]->GetIncidentWL() + v[i+1]->GetIncidentWL() + v[i+2]->GetIncidentWL();
    if (cost < bestcost)
    {
      bestconfig = 3;
      bestcost   = cost;
    }

    // B -> A -> C
    v[i+1]->SwapLocation(v[i+2]);
    cost = v[i]->GetIncidentWL() + v[i+1]->GetIncidentWL() + v[i+2]->GetIncidentWL();
    if (cost < bestcost)
    {
      bestconfig = 4;
      bestcost   = cost;
    }

    // C -> A -> B
    v[i]->SwapLocation(v[i+2]);
    cost = v[i]->GetIncidentWL() + v[i+1]->GetIncidentWL() + v[i+2]->GetIncidentWL();
    if (cost < bestcost)
    {
      bestconfig = 5;
      bestcost   = cost;
    }

    // C -> B -> A
    v[i+1]->SwapLocation(v[i+2]);
    cost = v[i]->GetIncidentWL() + v[i+1]->GetIncidentWL() + v[i+2]->GetIncidentWL();
    if (cost < bestcost)
    {
      bestconfig = 6;
      bestcost   = cost;
    }

    if (bestconfig != 1)
      numTimesSwapHelped++;

    // Currently the configuration is C -> B -> A
    switch (bestconfig)
    {
    case 1:
      v[i]->SwapLocation(v[i+2]);
      break;
    case 2:
      v[i+1]->SwapLocation(v[i+2]);
      v[i+1]->SwapLocation(v[i]);
      break;
    case 3:
      v[i]->SwapLocation(v[i+1]);
      break;
    case 4:
      v[i]->SwapLocation(v[i+1]);
      v[i+1]->SwapLocation(v[i+2]);
      break;
    case 5:
      v[i+1]->SwapLocation(v[i+2]);
      break;
    case 6:
      break;
    default:
      abort();
    }
  }

}

void CTile::linearPlaceColumn(vector<CNode*>& v)
{
  if (v.size() == 0)
    return;

  assert(v.size() <= gMaxColmHeight);

  // Using assignment problem
  Matrix <int> matrix(v.size(), gMaxColmHeight);

  for (int i = 0; i < v.size(); i++)
  {
    CNode *tmp = v[i];

    int origY = static_cast<int>(tmp->Y());

    // TODO:  Just need to compute incident WL in Y direction
    for (int j = 0; j < gMaxColmHeight; j++)
    {
      tmp->Y(mY + j);
      int cost = static_cast<int>(tmp->GetIncidentWL());
      matrix(i, j) = cost;
    }

    tmp->Y(origY);
  }


  Munkres m;
  vector<int> results(matrix.rows());
  m.solve(matrix, results);
  assert(results.size() == v.size());

  for (int i = 0; i < results.size(); i++)
    v[i]->Y(mY + results[i]);
}

// Returns true if the node aNode is present in this tile.  The
// decision is based on existence of this node in one of the four
// vectors (for ecell, dff, reg, bram) and not on the geomtrical
// cordinates.
// Warning:  Terribly inefficient function.  :).  Needs to be
//           written with map if this becomes critical.
// Warning:  The "match" is performed by defined operator==
//           of class CNode.  Make sure its reasonably fast.
bool CTile::Contains(CNode *aNode) const
{
  vector<CNode*>::const_iterator it;

  it = find(mECells.begin(), mECells.end(), aNode);
  if (it != mECells.end())
    return true;

  it = find(mDffs.begin(), mDffs.end(), aNode);
  if (it != mDffs.end())
    return true;

  it = find(mRegs.begin(), mRegs.end(), aNode);
  if (it != mRegs.end())
    return true;

  it = find(mBrams.begin(), mBrams.end(), aNode);
  if (it != mBrams.end())
    return true;

  return false;
}

int gNumAnnealSuccess = 0;
int gNumAnnealTried   = 0;
void doAnnealOnVector(vector<CNode *>& v, bool isDffVec = false)
{
  if (v.size() < 3)
    return;

  for (int times = 0; times < 5; times++)
  {
    random_shuffle(v.begin(), v.end());
    for (int i = 0; i < v.size() - 1; i++)
    {
      CNode* fir = v[i];
      CNode *sec = v[i+1];

      // Check that two Dffs should be even tried to exchange
      // only if they are of the same clock domain type.
      if (true == isDffVec && (fir->mClkDom[0] != sec->mClkDom[0]))
          continue;

      double currCost = fir->GetIncidentWL() + sec->GetIncidentWL();
      fir->SwapLocation(sec);
      gNumAnnealTried++;
      gNumAnnealSuccess++;
      if (currCost <= fir->GetIncidentWL() + sec->GetIncidentWL())
      {
        fir->SwapLocation(sec);
        gNumAnnealSuccess--;
      }
    }
  }
}


// Make random movement.
void CTile::Anneal(double aMinSuccessRatio)
{
  if (false == this->IsDensityOK())
  {
    cout << "WARN:  Still some tiles have invalid densities\n";
    abort();
    return;
  }

  cout << "STATUS:  Annealing cells in " << *this << endl;
  double ratio = 0.0;

  // ECells anneal
  do
  {
    gNumAnnealTried = 0;
    gNumAnnealSuccess = 0;
    doAnnealOnVector(mECells);
    ratio = (double)gNumAnnealSuccess / ((double)gNumAnnealTried + 1.0);
  }
  while (ratio > aMinSuccessRatio);

  // dff anneal
  do
  {
    gNumAnnealTried = 0;
    gNumAnnealSuccess = 0;
    doAnnealOnVector(mDffs, true);
    ratio = (double)gNumAnnealSuccess / ((double)gNumAnnealTried + 1.0);
  }
  while (ratio > aMinSuccessRatio);
}

int CTile::NumColumnReqd() const
{
  int sumOfMins = 0;
  map<int, int> reqdColmOfType;
  map<int, int> colmOccBy;

  map<int, int>::const_iterator it = this->mNumCellsOfClk.begin();
  while (it != mNumCellsOfClk.end())
  {
    reqdColmOfType[it->first] = static_cast<int>((it->second) / gMaxColmHeight);
    if (it->second > gMaxColmHeight*reqdColmOfType[it->first])
      reqdColmOfType[it->first]++;
    sumOfMins += reqdColmOfType[it->first];
    it++;
  }
//   assert(sumOfMins <= mColumnsEDff.size());

  if (mDffs.size() > 0 || mBrams.size() > 0 || mRegs.size() > 0)
    assert(sumOfMins > 0);

  return sumOfMins;
}

bool CTile::rotateAndPlaceWithClock(
    CNode *n, int startXPoint, int startYPoint)
{
  if (n->Name() == "o46690")
    cout << " hehe\n";
//   cout << "Rotating and placing with clock\n";
  int countOfTraversal, count;
  countOfTraversal = 1;
  if (startXPoint >= gDesign->mTileWidth)
    startXPoint = gDesign->mTileWidth - 1;

  if (startYPoint >= gDesign->mTileHeight)
    startYPoint = gDesign->mTileHeight - 1;
  
  assert(24 == mClkIdInEachColm.size());

#if 0
  bool isAssgnOfClkToColmnDone = false;
  for (int i=0; i<mClkIdInEachColm.size(); i++)
    if (-1 != mClkIdInEachColm[i])
      isAssgnOfClkToColmnDone = true;
  if (false == isAssgnOfClkToColmnDone)
  {
    cout << "WARN:  Colmn were not assigned prefered clock.  Doing now\n";
    cout <<"        This is common if GlobalSwap runs without legalization\n";
    findClocksForColumn();
  }
#endif

  int origX = startXPoint;
  int origY = startYPoint;
//   cout << "origY: " << origY << endl;
  while (countOfTraversal < gDesign->mTileWidth * 4 || countOfTraversal < gDesign->mTileHeight * 4)
  {
//     cout << "count of Traversal = " << countOfTraversal << endl;
    for (count = 0; count < countOfTraversal; count++)
    {
      startYPoint--;
      if (startXPoint < 0 || startXPoint >= gDesign->mTileWidth ||
          startYPoint < 0 || startYPoint >= gDesign->mTileHeight)
        continue;
//       cout << "looking at" << startXPoint << "," << startYPoint << endl;
      //Do operation here
      if (false == isOccupied(startXPoint, startYPoint) &&
          n->Type() == typeAtLocation(startXPoint, startYPoint))
      {
        if (mClkIdInEachColm[sEDMapXToColmNum[startXPoint]] == n->mClkDom[0])
        {
          PutAt(startXPoint, startYPoint, n);
          RemoveFrom(origX, origY, n);
          return true;
        }
        else if (mClkIdInEachColm[sEDMapXToColmNum[startXPoint]] == -1)
        {
          PutAt(startXPoint, startYPoint, n);
          RemoveFrom(origX, origY, n);
          mClkIdInEachColm[sEDMapXToColmNum[startXPoint]] = n->mClkDom[0];
          return true;
        }
      }
    }
    for (count = 0; count < countOfTraversal; count++)
    {
      startXPoint++;
      if (startXPoint < 0 || startXPoint >= gDesign->mTileWidth ||
          startYPoint < 0 || startYPoint >= gDesign->mTileHeight)
        continue;

//       cout << "looking at" << startXPoint << "," << startYPoint << endl;
      //Do operation here
      if (false == isOccupied(startXPoint, startYPoint) &&
          n->Type() == typeAtLocation(startXPoint, startYPoint))
      {
        if (mClkIdInEachColm[sEDMapXToColmNum[startXPoint]] == n->mClkDom[0])
        {
          PutAt(startXPoint, startYPoint, n);
          RemoveFrom(origX, origY, n);
          return true;
        }
        else if (mClkIdInEachColm[sEDMapXToColmNum[startXPoint]] == -1)
        {
          PutAt(startXPoint, startYPoint, n);
          RemoveFrom(origX, origY, n);
          mClkIdInEachColm[sEDMapXToColmNum[startXPoint]] = n->mClkDom[0];
          return true;
        }
      }
//       cout << startXPoint << " " << startYPoint << endl;
    }

    countOfTraversal += 1;
    for (count = 0; count < countOfTraversal; count++)
    {
      startYPoint++;
      if (startXPoint < 0 || startXPoint >= gDesign->mTileWidth ||
          startYPoint < 0 || startYPoint >= gDesign->mTileHeight)
        continue;

//       cout << "looking at" << startXPoint << "," << startYPoint << endl;
      //Do operation here
      if (false == isOccupied(startXPoint, startYPoint) &&
          n->Type() == typeAtLocation(startXPoint, startYPoint))
      {
        if (mClkIdInEachColm[sEDMapXToColmNum[startXPoint]] == n->mClkDom[0])
        {
          PutAt(startXPoint, startYPoint, n);
          RemoveFrom(origX, origY, n);
          return true;
        }
        else if (mClkIdInEachColm[sEDMapXToColmNum[startXPoint]] == -1)
        {
          PutAt(startXPoint, startYPoint, n);
          RemoveFrom(origX, origY, n);
          mClkIdInEachColm[sEDMapXToColmNum[startXPoint]] = n->mClkDom[0];
          return true;
        }
      }
    }

    for (count = 0; count < countOfTraversal; count++)
    {
      startXPoint--;
      if (startXPoint < 0 || startXPoint >= gDesign->mTileWidth ||
          startYPoint < 0 || startYPoint >= gDesign->mTileHeight)
        continue;

//       cout << "looking at" << startXPoint << "," << startYPoint << endl;
      //Do operation here
      if (false == isOccupied(startXPoint, startYPoint) &&
          n->Type() == typeAtLocation(startXPoint, startYPoint))
      {
        if (mClkIdInEachColm[sEDMapXToColmNum[startXPoint]] == n->mClkDom[0])
        {
          PutAt(startXPoint, startYPoint, n);
          RemoveFrom(origX, origY, n);
          return true;
        }
        else if (mClkIdInEachColm[sEDMapXToColmNum[startXPoint]] == -1)
        {
          PutAt(startXPoint, startYPoint, n);
          RemoveFrom(origX, origY, n);
          mClkIdInEachColm[sEDMapXToColmNum[startXPoint]] = n->mClkDom[0];
          return true;
        }
      }
    }
    countOfTraversal += 1;
  }

  return false;
}

bool CTile::rotateAndPlace(CNode *n, int startXPoint, int startYPoint)
{
  int countOfTraversal, count;
  countOfTraversal = 1;

  while (countOfTraversal < gDesign->mTileWidth*2 || countOfTraversal < gDesign->mTileHeight*2)
  {
    for (count = 0; count < countOfTraversal; count++)
    {
      startYPoint--;
      if (startXPoint < 0 || startXPoint >= gDesign->mTileWidth ||
          startYPoint < 0 || startYPoint >= gDesign->mTileHeight)
        continue;
      //Do operation here
      if (false == isOccupied(startXPoint, startYPoint) &&
          n->Type() == typeAtLocation(startXPoint, startYPoint))
      {
        PutAt(startXPoint, startYPoint, n);
//         n->PutAt(mX + startXPoint, mY + startYPoint);
//         isOccupied(startXPoint, startYPoint) = true;
//         cout << "found place at " << startXPoint << " " << startYPoint << endl;
        return true;
      }
    }
    for (count = 0; count < countOfTraversal; count++)
    {
      startXPoint++;
      if (startXPoint < 0 || startXPoint >= gDesign->mTileWidth ||
          startYPoint < 0 || startYPoint >= gDesign->mTileHeight)
        continue;

      //Do operation here
      if (false == isOccupied(startXPoint, startYPoint) &&
          n->Type() == typeAtLocation(startXPoint, startYPoint))
      {
        PutAt(startXPoint, startYPoint, n);
//         n->PutAt(mX + startXPoint, mY + startYPoint);
//         isOccupied(startXPoint, startYPoint) = true;
//         cout << "found place at " << startXPoint << " " << startYPoint << endl;
        return true;
      }
//       cout << startXPoint << " " << startYPoint << endl;
    }

    countOfTraversal += 1;
    for (count = 0; count < countOfTraversal; count++)
    {
      startYPoint++;
      if (startXPoint < 0 || startXPoint >= gDesign->mTileWidth ||
          startYPoint < 0 || startYPoint >= gDesign->mTileHeight)
        continue;

      //Do operation here
      if (false == isOccupied(startXPoint, startYPoint) &&
          n->Type() == typeAtLocation(startXPoint, startYPoint))
      {
        PutAt(startXPoint, startYPoint, n);
//         n->PutAt(mX + startXPoint, mY + startYPoint);
//         isOccupied(startXPoint, startYPoint) = true;
//         cout << "found place at " << startXPoint << " " << startYPoint << endl;
        return true;
      }
    }

    for (count = 0; count < countOfTraversal; count++)
    {
      startXPoint--;
      if (startXPoint < 0 || startXPoint >= gDesign->mTileWidth ||
          startYPoint < 0 || startYPoint >= gDesign->mTileHeight)
        continue;

      //Do operation here
      if (false == isOccupied(startXPoint, startYPoint) &&
          n->Type() == typeAtLocation(startXPoint, startYPoint))
      {
        PutAt(startXPoint, startYPoint, n);
//         n->PutAt(mX + startXPoint, mY + startYPoint);
//         isOccupied(startXPoint, startYPoint) = true;
//         cout << "found place at " << startXPoint << " " << startYPoint << endl;
        return true;
      }
    }
    countOfTraversal += 1;
  }

  return false;
}

void CTile::LegalizeToClosestPosition(vector<CNode*>& v)
{
  if (v.size() == 0)
    return;

  for (int i=0; i<v.size(); i++)
  {
    assert(this->Contains(v[i]));

    double xposRel = v[i]->X() - mX;
    double yposRel = v[i]->Y() - mY;

    if (xposRel > gDesign->mTileWidth - 1)
      xposRel = gDesign->mTileWidth - 1;
    if (xposRel < 0)
      xposRel = 0;
    if (yposRel > gDesign->mTileHeight - 1)
      yposRel = gDesign->mTileHeight - 1;
    if (yposRel < 0)
      yposRel = 0;

    if (true == vecBool[(int)xposRel][(int)yposRel].isOccupied &&
        v[i] == vecBool[(int)xposRel][(int)yposRel].n &&
        v[i]->Type() == vecBool[(int)xposRel][(int)yposRel].cellType)
      continue;

    int intX = int(xposRel);
    int intY = int(yposRel);

    bool donePlacing = false;
    if (xposRel - intX < 0.5)
    {
      // lies more towards intX than next int
      if (yposRel - intY < 0.5)
      {
        // lies more towards intY than next int
        if (false == isOccupied(intX, intY) && v[i]->Type() == typeAtLocation(intX, intY))
        {
          PutAt(intX, intY, v[i]);
//           v[i]->PutAt(mX + intX, mY + intY);
//           isOccupied(intX, intY) = true;
          donePlacing = true;
        }
      }
      else
      {
        // lies more towards next int than intY
        if (false == isOccupied(intX, intY + 1) && v[i]->Type() == typeAtLocation(intX, intY + 1))
        {
          PutAt(intX, intY+1, v[i]);
//           v[i]->PutAt(mX + intX, mY + intY + 1);
//           isOccupied(intX, intY + 1) = true;
          donePlacing = true;
        }
      }
    }
    else
    {
      // lies more towards next int than intX
      if (yposRel - intY < 0.5)
      {
        // lies more towards intY than next int
        if (false == isOccupied(intX + 1, intY) && v[i]->Type() == typeAtLocation(intX + 1, intY))
        {
          PutAt(intX + 1, intY, v[i]);
//           v[i]->PutAt(mX + intX + 1, mY + intY);
//           isOccupied(intX + 1, intY) = true;
          donePlacing = true;
        }
      }
      else
      {
        // lies more towards next int than intY
        if (false == isOccupied(intX + 1, intY + 1) && v[i]->Type() == typeAtLocation(intX + 1, intY + 1))
        {
          PutAt(intX + 1, intY + 1, v[i]);
//           v[i]->PutAt(mX + intX + 1, mY + intY + 1);
//           isOccupied(intX + 1, intY + 1) = true;
          donePlacing = true;
        }
      }
    }

    if (false == donePlacing)
      donePlacing = rotateAndPlace(v[i], (int)xposRel, (int)yposRel);

//     if (true == donePlacing)
//       cout << "breaking\n";
    assert(true == donePlacing);

//     if (fabs(mX + xposRel - v[i]->X()) + fabs(mY + yposRel - v[i]->Y()) > 3)
//       cout << " Cell " << v[i]->Name() << " moved " << mX + xposRel << 
//         "," << mY + yposRel << " => " << v[i]->X() << "," << v[i]->Y() <<
//         " (distance = " << fabs(mX + xposRel - v[i]->X()) + 
//         fabs(mY + yposRel - v[i]->Y()) << ") \n";
  }
}

void CTile::LegalizeDffsToClock()
{
  if (mDffs.size() == 0 )
    return;

  if (mNumCellsOfClk.size() < 2)
  {
    initializeClocks();
    return;
  }
  map<int, int>::iterator it = mMapInitXOfDffColm.begin();

  bool checkHighClock = false;
  // Writing objective function here
  for ( ; it != mMapInitXOfDffColm.end(); it++)
  {
    int xIndexOfCol = it->second;
    int clockId = -1;
    for (int j=0; j<gDesign->mTileHeight; j++)
    {
      if (true == vecBool[xIndexOfCol][j].isOccupied)
      {
        assert(1 == vecBool[xIndexOfCol][j].cellType);
        assert(NULL != vecBool[xIndexOfCol][j].n);
        CNode *n = vecBool[xIndexOfCol][j].n;
        assert(1 == n->mClkDom.size());
        if (clockId == -1)
        {
          clockId = vecBool[xIndexOfCol][j].n->mClkDom[0]; 
          cout << "column " << xIndexOfCol << " has clock " << clockId << endl; 
        }
        else 
        {
          if (clockId != vecBool[xIndexOfCol][j].n->mClkDom[0])
          {
            checkHighClock = true;
            cout << "column " << xIndexOfCol << " has more than one clock " << vecBool[xIndexOfCol][j].n->mClkDom[0] << endl; 
            break;
          }
        }
      }
    }
    if (checkHighClock == true)
      break;
  }

  if (checkHighClock == false)
  {
    cout << "returning after initializing clocks\n";
    initializeClocks();
    return;
  }

  findClocksForColumn();

  legalizeDffsUsingClocksForColumns();
  checkIfLegalityInThisTileOK();
}

void CTile::checkIfLegalityInThisTileOK()
{
  bool checkHighClock = false;
  map<int,int>::iterator it = mMapInitXOfDffColm.begin();
  for ( ; it != mMapInitXOfDffColm.end(); it++)
  {
//     cout << "  - Checking column " << it->first << " (located at " << it->second << ")\n";

    int xIndexOfCol = it->second;
    int clockId = -1;

    for (int j=0; j<gDesign->mTileHeight; j++)
    {
//       cout << "    - Checking y position " << j << "   :";
      if (true == vecBool[xIndexOfCol][j].isOccupied)
      {
        assert(1 == vecBool[xIndexOfCol][j].cellType);
        assert(NULL != vecBool[xIndexOfCol][j].n);
        CNode *n = vecBool[xIndexOfCol][j].n;
        assert(1 == n->mClkDom.size());
        if (clockId == -1)
        {
          clockId = vecBool[xIndexOfCol][j].n->mClkDom[0]; 
//           cout << " has clock " << vecBool[xIndexOfCol][j].n->mClkDom[0] << endl;
        }
        else 
        {
          if (clockId != vecBool[xIndexOfCol][j].n->mClkDom[0])
          {
            checkHighClock = true;
//             cout << " has clock " << vecBool[xIndexOfCol][j].n->mClkDom[0] << endl;
//             cout << "EROR:  Which is duplicate\n";
            abort();
            break;
          }
        }
      }
      else
      {
//         cout << endl;
      }
    }
    if (checkHighClock == true)
      break;
  }
}

void CTile::legalizeDffsUsingClocksForColumns()
{
  for (int i=0; i<mClkIdInEachColm.size(); i++)
  {
    int colorForThisColm = mClkIdInEachColm[i];
    if (-1 == colorForThisColm)
    {
      cout << " Do what\n";
    }
    else
    {
      for (int j=0; j<gDesign->mTileHeight; j++)
      {
        if (true == vecBool[mMapInitXOfDffColm[i]][j].isOccupied &&
            colorForThisColm != vecBool[mMapInitXOfDffColm[i]][j].n->mClkDom[0])
        {
          RemoveFrom(mMapInitXOfDffColm[i], j, vecBool[mMapInitXOfDffColm[i]][j].n);
        }
      }
    }
  }

  for (int i=0; i<mDffs.size(); i++)
  {
    int columnNum = static_cast<int>(mDffs[i]->X() - mX);
    assert(columnNum < 92);
    assert(mDffs[i]->Y() - mY < 64);
    assert(1 == mDffs[i]->mClkDom.size());

    if (mClkIdInEachColm[sEDMapXToColmNum[columnNum]] == mDffs[i]->mClkDom[0])
      continue;

    bool done = rotateAndPlaceWithClock(
        mDffs[i], columnNum, static_cast<int>(mDffs[i]->Y() - mY));
    if (false == done)
      WriteImage("test");
    assert(true == done);
  }
}

void CTile::initializeClocks()
{
  map<int, int>::iterator it = mMapInitXOfDffColm.begin();

  for ( ; it != mMapInitXOfDffColm.end(); it++)
  {
    int xIndexOfCol = it->second;
    int clockId = -1;
    for (int j=0; j<gDesign->mTileHeight; j++)
    {
      if (true == vecBool[xIndexOfCol][j].isOccupied)
      {
        assert(1 == vecBool[xIndexOfCol][j].cellType);
        assert(NULL != vecBool[xIndexOfCol][j].n);
        CNode *n = vecBool[xIndexOfCol][j].n;
        assert(1 == n->mClkDom.size());
        clockId = vecBool[xIndexOfCol][j].n->mClkDom[0]; 
        mClkIdInEachColm[it->first] = clockId;
        break;
      }
    }
  }

}

void CTile::findClocksForColumn()
{
  vector <int> clocksInsideTile = GetClockIdsInside();
  assert (clocksInsideTile.size() <= 4);

  map<int, int> mapClockToInt;
  map<int, int> mapIntToClock;

//   cout << " clock inside tile size = " << clocksInsideTile.size() << endl;
  for (int i=0; i<clocksInsideTile.size(); i++)
  {
    mapClockToInt[clocksInsideTile[i]] = i;
    mapIntToClock[i] = clocksInsideTile[i];
//     cout << " mapped clock " << clocksInsideTile[i] << " to " << i << endl;
//     cout << " mapped number " << i << " to clock " << mapIntToClock[i] << endl;
  }

  map<int, int>::iterator it = mMapInitXOfDffColm.begin();

  ostringstream os;
  os << "Minimize\n";
  os << "Obj: ";
  bool isFirstTime = true;

  vector<vector<int> >numCellInColmForEachClk(24, vector<int>(clocksInsideTile.size(), int()));

  // Writing objective function here
  for ( ; it != mMapInitXOfDffColm.end(); it++)
  {
    int xIndexOfCol = it->second;
    for (int j=0; j<gDesign->mTileHeight; j++)
    {
      if (true == vecBool[xIndexOfCol][j].isOccupied)
      {
        assert(1 == vecBool[xIndexOfCol][j].cellType);
        assert(NULL != vecBool[xIndexOfCol][j].n);
        CNode *n = vecBool[xIndexOfCol][j].n;
        assert(1 == n->mClkDom.size());
        numCellInColmForEachClk[it->first][mapClockToInt[n->mClkDom[0]]]++;
//         cout << " Node " << n->Name() << " has clock " << n->mClkDom[0] << " mapped to " <<
//           mapClockToInt[n->mClkDom[0]] << " increased count to " << 
//           numCellInColmForEachClk[it->first][mapClockToInt[n->mClkDom[0]]] << endl;
      }
    }

    vector<int> sumOfOtherClocks(clocksInsideTile.size());
    int sum = std::accumulate(numCellInColmForEachClk[it->first].begin(), numCellInColmForEachClk[it->first].end(), 0);
    for (int i=0; i<sumOfOtherClocks.size(); i++)
      sumOfOtherClocks[i] = sum - numCellInColmForEachClk[it->first][i];

//     cout << " Printing ILP of cells of tile at " << mX << ", " << mY << endl;
    for (int i=0; i<clocksInsideTile.size(); i++)
    {
      if (isFirstTime)
        isFirstTime = false;
      else
        os << " + ";

      os << sumOfOtherClocks[i] << " X_" << xIndexOfCol << "_" << i;
    }
  }
  os << endl;

  // Equality constraints
  it = mMapInitXOfDffColm.begin();
  int numConstraint = 0;
  isFirstTime = true;
  os << "Subject To\n";
  for ( ; it != mMapInitXOfDffColm.end(); it++)
  {
    int xIndexOfCol = it->second;
    for (int j=0; j<clocksInsideTile.size(); j++)
    {

      if (j != 0)
        os << " + ";
      else
        os << "C" << numConstraint++ << ": ";

      os << "X_" << xIndexOfCol << "_" << j;
    }
    os << " = 1\n";
  }

  // Compute minimum columns required by each clock type
  vector<int> minReqdForClock(clocksInsideTile.size(), int());
  for (int i=0; i<clocksInsideTile.size(); i++)
  {
    int numCells = mNumCellsOfClk[clocksInsideTile[i]];
//     cout << "# cells of type " << clocksInsideTile[i] << " = " << numCells << endl;
    if (0 == numCells % 64)
    {
      minReqdForClock[mapClockToInt[clocksInsideTile[i]]] = numCells/64;
    }
    else
    {
      minReqdForClock[mapClockToInt[clocksInsideTile[i]]] = numCells/64 + 1;
    }

  }

  for (int i=0; i<clocksInsideTile.size(); i++)
    assert(0 < minReqdForClock[i] || mNumCellsOfClk[clocksInsideTile[i]] == 0);

  // Compute the inequality constraints
  for (int j=0; j<clocksInsideTile.size(); j++)
  {
    it = mMapInitXOfDffColm.begin();
    for ( ; it != mMapInitXOfDffColm.end(); it++)
    {
      int xIndexOfCol = it->second;
      if (it == mMapInitXOfDffColm.begin())
      {
        os << "C" << numConstraint++ << ": ";
      }
      else
      {
        os << " + ";
      }
      os << "X_" << xIndexOfCol << "_" << j;
    }
    os << " >= " << minReqdForClock[j] << endl;
  }


  // Dump the binary constraints
  os << "Binary\n";
  it = mMapInitXOfDffColm.begin();
  for ( ; it != mMapInitXOfDffColm.end(); it++)
  {
    int xIndexOfCol = it->second;
    for (int j=0; j<clocksInsideTile.size(); j++)
    {
      os << "X_" << xIndexOfCol << "_" << j << endl;
    }
  }

  os << "End\n";

//   cout << os.str();

  std::ofstream out;
  out.open(string("ilp.cplex").c_str());

  cout << " Writing ILP to file " << string("/tmp/ilp.cplex") << endl;
  out << os.str();
  
  system("./bin/glpsol --cpxlp ilp.cplex -o output.cplex");
  system("grep \"X_*\" output.cplex | awk '{print $2,$4}' | sed -e 's@X_@@g' | sed -e 's@_@ @g' > out_result.cplex");

  std::ifstream in;
  in.open("out_result.cplex");

  if (false == in)
  {
    cout << "EROR: Result could not be generated by ILP solver\n";
    abort();
  }
  
  it = mMapInitXOfDffColm.begin();
//   for ( ; it != mMapInitXOfDffColm.end(); it++)
//   {
//     cout << "mapped values " << it->first << " to " << it->second << endl;
//   }
  string line;
  vector<string> tok;

  it = mMapInitXOfDffColm.begin();
  for ( ; it != mMapInitXOfDffColm.end(); it++)
  {
    int xIndexOfCol = it->second;
    for (int j=0; j<clocksInsideTile.size(); j++)
    {
      getline(in, line);
      tok.clear();
      Tokenize(line, tok, "\n\t ");
      assert(3 == tok.size());
      int colmIdx = atoi(tok[0].c_str());

      int color;
      if (std::accumulate(
            numCellInColmForEachClk[it->first].begin(),
            numCellInColmForEachClk[it->first].end(), 0) > 0)
        color = atoi(tok[1].c_str());
      else
#if 0
        color = -1;
#else
        color = atoi(tok[1].c_str());
#endif

      int reserved = atoi(tok[2].c_str());
      assert(reserved == 0 || reserved == 1);
      assert(colmIdx == xIndexOfCol);
      assert(color == j || (color == -1 && 
        std::accumulate(numCellInColmForEachClk[it->first].begin(),
            numCellInColmForEachClk[it->first].end(), 0) == 0));

//       if (reserved == 1)
//       {
//         cout << "  From input line = " << line << endl;
//         cout << "  Column " << it->first << " reserved for clock " <<
//           mapIntToClock[color] << endl;
//       }

      if (reserved == 1)
      {
        assert(-1 == mClkIdInEachColm[it->first]);
        if (-1 == color)
          mClkIdInEachColm[it->first] = -1;
        else
          mClkIdInEachColm[it->first] = mapIntToClock[color];
      }
    }
  }

  it = mMapInitXOfDffColm.begin();

  for ( ; it != mMapInitXOfDffColm.end(); it++)
  {
    assert (mClkIdInEachColm[it->first] != -1);
  }

  system("mv ilp.cplex ilp.cplex.old");
  system("mv output.cplex output.cplex.old");
  system("mv out_result.cplex out_result.cplex.old");

#if 0
  it = mMapInitXOfDffColm.begin();
  int columNum = 0;
  for ( ; it != mMapInitXOfDffColm.end(); it++)
  {
    vector<int> numCellsOfClkType(clocksInsideTile.size(), int());

    int xIndexOfCol = it->second;
    for (int j=0; j<gDesign->mTileHeight; j++)
    {
      if (true == vecBool[xIndexOfCol][j].isOccupied)
      {
        CNode *n = vecBool[xIndexOfCol][j].n;
        numCellsOfClkType[mapClockToInt[n->mClkDom[0]]]++;
      }
    }
    
    cout << "In Colm " << columNum++ << endl;
    for (int i=0; i<clocksInsideTile.size(); i++)
    {
      cout << "  - Clock: " << mapIntToClock[i] << " #Cells: " << numCellsOfClkType[i] << endl;
    }
    cout << "  === Assigned ID = " << idOfClockInColm[it->first] << endl;
    cout << endl;
  }

  exit(1);
#endif
  return;
}

bool CTile::PutAt(int x, int y, CNode *node)
{
  
  assert(false == vecBool[x][y].isOccupied);
  assert(NULL  == vecBool[x][y].n);

  vecBool[x][y].isOccupied = true;
  vecBool[x][y].n = node;

  assert(vecBool[x][y].cellType == node->Type());
  node->PutAt(double(mX + x), double(mY + y));

  if (int(node->X()-mX) != x)
  {
    node->PutAt(node->X()+0.001 ,node->Y());
  }
  if (int(node->Y()-mY) != y)
  {
    node->PutAt(node->X(), node->Y()+0.001);
  }
  assert(int(node->X()-mX) == x);
  assert(int(node->Y()-mY) == y);
  return true;
}

CNode* CTile::RemoveFrom(int x, int y, CNode* node)
{
  if (node->Name() == "o63784" || node->Name() == "o96589")
    cout << "hehe" << endl;

  if (false == vecBool[x][y].isOccupied)
    return NULL;

  assert(true == vecBool[x][y].isOccupied);
  CNode *n = vecBool[x][y].n;
  
  if (node != n)
    return NULL;
  
  vecBool[x][y].isOccupied = false;
  vecBool[x][y].n = NULL;
//   vecBool[x][y].cellType = -1;
  return n;
}
