/*
***********************************************************************
LICENSE AGREEMENT
***********************************************************************
 READ THIS LICENSE AGREEMENT CAREFULLY BEFORE USING THIS PRODUCT. BY 
 USING THIS PRODUCT YOU INDICATE YOUR ACCEPTANCE OF THE TERMS OF THE 
 FOLLOWING AGREEMENT. THESE TERMS APPLY TO YOU AND ANY SUBSEQUENT 
 LICENSEE OF THIS PRODUCT.

               License Agreement for RegularPlace 1.0

Copyright (c)  2008 by 
Ashutosh Chakraborty, Anurag Kumar, David Z. Pan
All rights reserved

ATTRIBUTION ASSURANCE LICENSE
Redistribution of the any code, with or without modification, are permitted 
provided that the conditions below are met. These conditions require a 
modest attribution to Ashutosh Chakraborty, Anurag Kumar and Dr. 
David Z. Pan (the "Author").

1. Redistributions of any code, with or without modification (the "Code"), 
must be accompanied by any documentation and, each time the resulting 
executable program or a program dependent thereon is launched, a 
prominent display (e.g., splash screen or banner text) of the Author's 
attribution information, which includes:
  (a) Ashutosh Chakraborty, Anurag Kumar, and Dr. David Z. Pan ("AUTHOR"),
  (b) The University of Texas at Austin ("PROFESSIONAL IDENTIFICATION"), and
  (c) http://www.cerc.utexas.edu/utda/ ("URL").

2. Use is limited to academic research groups only. Users who are interested 
in industry or commercial purposes must notify Author and request 
separate license agreement.  

3. Neither the name nor any trademark of the Author may be used to 
endorse or promote products derived from this software without specific 
prior written permission.

4. Users are entirely responsible, to the exclusion of the Author and 
any other persons, for compliance with (a) regulations set by owners or 
administrators of employed equipment, (b) licensing terms of any other 
software, and (c) local, national, and international regulations regarding 
use, including those regarding import, export, and use of encryption software. 

THIS FREE SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR 
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE AUTHOR OR ANY CONTRIBUTOR BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, EFFECTS OF UNAUTHORIZED OR MALICIOUS NETWORK 
ACCESS; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*//*
 * =======================================================================
 *
 *       Filename:  TileLegalizeDff.cpp
 *
 *    Description:  Legalization and optimization of edffs in their columns
 *
 *        Version:  1.0
 *        Created:  10/03/2008 07:35:03 PM CDT
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ashutosh Chakraborty (), ashutosh@mail.utexas.edu
 *        Company:  University of Texas, Austin
 *
 * =======================================================================
 */


#include <iostream>
#include <cfloat>
#include <map>

#include "math.h"

#include "Util.h"
#include "Design.h"
#include "Data.h"
#include "Tile.h"

#undef _VERBOSE
// #define _VERBOSE

using namespace std;

void CTile::assignEDffsToBestColumn(double xOff, double yOff, bool bIsFinalAssign)
{
#ifdef _VERBOSE
  cout << "\nEntering CTile::assignEDffsToBestColumn with " << mDffs.size() << " dffs\n";
#endif

  // Assign a dirty value to clock in each column
  vector<int> clkInColm(mColumnsEDff.size(), int());
  for (int i = 0; i < mColumnsEDff.size(); i++)
  {
    mColumnsEDff[i].clear();
    clkInColm[i] = -1;
  }

  // The number of columns which are not committed for any particular
  // clock till now.  We initialize it as "all the columns".
  int numStillUncommitted = mColumnsEDff.size();

  // Compute how many columns of each clock type would be
  // required at the *minimum*.  Cell of clock type 'c' will require
  // reqdColmOfType['c'] columns after this.  The variable colmOccBy['c']
  // contains how manu columns has been occupied by type 'c'.
  int sumOfMins = 0;
  map<int, int> reqdColmOfType;
  map<int, int> colmOccBy;

  map<int, int>::iterator it = mNumCellsOfClk.begin();
  while (it != mNumCellsOfClk.end())
  {
    reqdColmOfType[it->first] = static_cast<int>((it->second) / gMaxColmHeight);
    if (it->second > gMaxColmHeight * reqdColmOfType[it->first])
      reqdColmOfType[it->first]++;
    colmOccBy[it->first]  = 0;
    sumOfMins += reqdColmOfType[it->first];
    it++;
  }
  assert(sumOfMins <= mColumnsEDff.size());


  // Compute how much is the sum of the minimum columns needed for
  // all columns except the current one.  This is sort of a constraint
  // for a particular clock type to respect to make sure it does not
  // eat too many columns.
  map<int, int> sumOfOtherMins;
  it = mNumCellsOfClk.begin();
  while (it != mNumCellsOfClk.end())
  {
    sumOfOtherMins[it->first] = sumOfMins - reqdColmOfType[it->first];
    it++;
  }

  // Sort all nodes according to their Y cordinate (and x within y)
  sort(mDffs.begin(), mDffs.end(), cmpNodes);

  vector <int> posnFilledInColm(mColumnsEDff.size());
  posnFilledInColm.clear();
  for (int i = 0; i < posnFilledInColm.size(); i++)
    posnFilledInColm[i] = 0;

  // Iterator over the sorted vector of dffs
  for (int i = 0; i < mDffs.size(); i++)
  {
    CNode* tmp = mDffs[i];
    int thisclk = tmp->mClkDom[0];
    assert(1 == tmp->mClkDom.size());

#ifdef _VERBOSE
    cout << "Doing node " << tmp->Name() << endl;

    // State of each type of clock cells
    it = mNumCellsOfClk.begin();
    cout << "  Clock State:\n";
    while (it != mNumCellsOfClk.end())
    {
      cout <<
      "    Clk: " << it->first <<
      " MinReqd:" << reqdColmOfType[it->first] <<
      " OtherMin:" << sumOfOtherMins[it->first] <<
      " NumColm:" << colmOccBy[it->first] << endl;
      it++;
    }

    // Print which columns are free right now
    cout << "  Free colm = ";
    for (int tmp = 0; tmp < mColumnsEDff.size(); tmp++)
      if (clkInColm[tmp] == -1)
        cout << tmp << " ";
    cout << endl;

#endif

    // Find where this cell should go.
    int bestIdx = -1;
    double bestCost = DBL_MAX;
    double cost = -1;

    // Iterate over all the possible columns
    for (int j = 0; j < mColumnsEDff.size(); j++)
    {
      cost = -1.0;
#ifdef _VERBOSE
      cout << "  Considering in column " << j;
#endif

      // Early exit (from this column) is it is full
      if (posnFilledInColm[j] == gMaxColmHeight)
        continue;

      // If the j-th column is yet unclaimed, check if occupying it would
      // make the design impossible to legalize.  The condition for the
      // latter is that there should be at least sumOfOtherMins[thisclk]
      // columns empty after this step.
      if (clkInColm[j] == -1)
      {
#ifdef _VERBOSE
        cout << " Check: avail=" << numStillUncommitted << " need=" <<
        sumOfOtherMins[thisclk];
#endif

        if (sumOfOtherMins[thisclk] >= numStillUncommitted)
        {
#ifdef _VERBOSE
          cout << "  Give up (would violate)" << endl;
#endif
          continue;
        }
      }
      else if (clkInColm[j] != thisclk)
      {
#ifdef _VERBOSE
        cout << "  Occupied by " << clkInColm[j] << endl;
#endif
        continue;
      }

#define _USE_WL_AS_COST
// #define _USE_DISTANCE_AS_COST
#ifdef _USE_WL_AS_COST

      // If the need for calling assignEDffsToBestColumn twice was not there,
      // we could remove the map posnFilledInColm[i] and modify it to
      // mColumnsEDff[i].size() instead.
      tmp->PutAt(xOff + mMapInitXOfDffColm[j], yOff + posnFilledInColm[j]);
      cost = tmp->GetIncidentWL();
#else
      cost = fabs(tmp->X() - mMapInitXOfDffColm[j]) +
             fabs(tmp->Y() + posnFilledInColm[j]);
#endif

      assert(-1.0 != cost);

      if (cost < bestCost)
      {
        bestCost = cost;
        bestIdx = j;
      }
#ifdef _VERBOSE
      cout << "***  cost=" << cost << " and best = " << bestCost << endl;
#endif
    }

    assert(-1 != bestIdx);
    assert(bestCost < DBL_MAX);

    // If this was a new column for thisclk, note this in the
    // in vector colmOccBy.
    if (-1 == clkInColm[bestIdx])
    {
      colmOccBy[thisclk]++;
      numStillUncommitted--;

      // If this is an "important" (needed) column for thisclk,
      // reduce the stringent requirements for other columns.
      if (colmOccBy[thisclk] <= reqdColmOfType[thisclk])
      {
        // Decrement the sum of other mins for all clocks except this
        // because thisclk has received one of its promised clocks
        it = mNumCellsOfClk.begin();
        while (it != mNumCellsOfClk.end())
        {
          if (it->first != thisclk)
            if (sumOfOtherMins[it->first] > 0)
              sumOfOtherMins[it->first]--;
          it++;
        }
      }
    }

    clkInColm[bestIdx] = thisclk;
    assert(thisclk == clkInColm[bestIdx]);

#ifdef _VERBOSE
    cout << "  putting at " << xOff + mMapInitXOfDffColm[bestIdx] << " and " <<
    yOff + posnFilledInColm[bestIdx] << endl;
#endif

    tmp->PutAt(xOff + mMapInitXOfDffColm[bestIdx],
               yOff + posnFilledInColm[bestIdx]);
    posnFilledInColm[bestIdx]++;

    if (bIsFinalAssign)
    {
      mColumnsEDff[bestIdx].push_back(tmp);
      assert(posnFilledInColm[bestIdx] = mColumnsEDff[bestIdx].size());
    }

#ifdef _VERBOSE
    cout << "  Best Idx=" << bestIdx << ".  Cell at " << tmp->X() <<
    "," << tmp->Y() << endl;
#endif
  }

#ifdef _VERBOSE
  cout << " starting dump\n";
  for (int i = 0; i < mDffs.size(); i++)
    cout << "   " << *(mDffs[i]) << endl;
  cout << " ending dump\n";
#endif

#ifdef _VERBOSE
  cout << "Exiting CTile::assignEDffsToBestColumn\n\n";
#endif
}
