#ifndef __UMPACK_CAPI
#define __UMPACK_CAPI

/**
 * C-linkage API for UMpack tools.
 * Provides a simple API for leveraging the UMpack tools in non-C++
 * environments. Logical languages include C, Java (via JNI), Python, etc.
 *
 * Eventual goal is to include lots of useful functions here.
 *
 * For now: provides an easy way to
 * adapt programs which already use the HMetis library
 *
 * Limitations: not very many functions (all focused on partitioning/clustering),
 *   only tested under Windows (built as DLL), likely not full tested, hardly documented.
 *
 * Original Author: Michael Wrighton (wrighton@caltech.edu)
 **/

#ifdef __cplusplus
#ifdef WIN32
#define EXTERN extern "C" __declspec(dllexport)
#else
#define EXTERN extern "C"
#endif
#else /* if not C++ */
#define EXTERN extern
#endif

EXTERN int UMpack_cluster(int nvtxs, int nhedges, double *vwgts,
                          int *eptr, int *eind,
                          double *edgeWeights, int nclusters,
                          int *part,
                          int debugLevel);

EXTERN int UMpack_mlpart(int nvtxs, int nhedges, double *vwgts,
                         int *eptr,
                         int *eind,
                         double *edgeWeights, int nparts,double *balanceArray ,
                         double tolerance, int *part,
                         int startsPerRun, int totalRuns,
                         int debug);

EXTERN int UMpack_shellpart(int nvtxs, int nhedges, double *vwgts,
                            int *eptr, int *eind,
                            double *edgeWeights, int nparts,
                            double *balanceArray,
                            double tolerance, int *part,
                            int bbPushLimit,
                            int startsPerRun,
                            int totalRuns,
                            int debug);

#endif

