#ifndef __CS2_H__
#define __CS2_H__

/* min-cost flow */
/* successive approximation algorithm */
/* Copyright C IG Systems, igsys@eclipse.com */
/* any use except for evaluation purposes requires a licence */

/************************************** constants  &  parameters ********/

/* definitions of types: node & arc */

#include "types_cs2.h"

int parse (
  /* all parameters are output */
  long    *n_ad,                 /* address of the number of nodes */
  long    *m_ad,                 /* address of the number of arcs */
  node    **nodes_ad,            /* address of the array of nodes */
  arc     **arcs_ad,             /* address of the array of arcs */
  long    *node_min_ad,          /* address of the minimal node */
  price_t *m_c_ad,               /* maximum arc cost */
  long    **cap_ad               /* array of capacities */
);

double timer ();

class CS2
{

private:

  node _dnd, *_dnode;

  arc     *_sa, *_sb;
  long    _d_cap;

public:

  long _n_push,
  _n_relabel,
  _n_discharge,
  _n_refine,
  _n_update,
  _n_scan,
  _n_prscan,
  _n_prscan1,
  _n_prscan2,
  _n_bad_pricein,
  _n_bad_relabel,
  _n_prefine;

private:

  long   _n,                    /* number of nodes */
  _m;                    /* number of arcs */

  long   *_cap;                 /* array containig capacities */

  node   *_nodes,               /* array of nodes */
  *_sentinel_node,       /* next after last */
  *_excq_first,          /* first node in push-queue */
  *_excq_last;           /* last node in push-queue */

  arc    *_arcs,                /* array of arcs */
  *_sentinel_arc;        /* next after last */

  bucket *_buckets,             /* array of buckets */
  *_l_bucket;            /* last bucket */
  long   _linf;                 /* number of l_bucket + 1 */

  int _time_for_price_in;

  price_t _epsilon,             /* quality measure */
  _dn;                  /* cost multiplier - number of nodes  + 1 */
  price_t _price_min,           /* lowest bound for prices */
  _mmc;                 /* multiplied maximal cost */
  double  _f_scale,             /* scale factor */
  _cut_off_factor,      /* multiplier to produce cut_on and cut_off
				 from n and epsilon */
  _cut_on,              /* the bound for returning suspended arcs */
  _cut_off;             /* the bound for suspending arcs */

  excess_t _total_excess;       /* total excess */

  long   _n_rel,                /* number of relabels from last price update */
  _n_ref,                /* current number of refines */
  _n_src;                /* current number of nodes with excess */

  int   _flag_price,            /* if = 1 - signal to start price-in ASAP -
				 maybe there is infeasibility because of
				 susoended arcs */
  _flag_updt;             /* if = 1 - update failed some sources are
				 unreachable: either the problem is
				 unfeasible or you have to return
                                 suspended arcs */

//long  _empty_push_bound;      /* maximal possible number of zero pushes
//                                 during one discharge */

  int   _snc_max;               /* maximal number of cycles cancelled
                                 during price refine */

  arc   _d_arc;                 /* dummy arc - for technical reasons */

  node  _d_node,                /* dummy node - for technical reasons */
  *_dummy_node;           /* the address of d_node */

#ifdef CHECK_SOLUTION
  long long int *_node_balance;
#endif

  /************************************************ abnormal finish **********/

  void err_end ( int cc );

  /************************************************* initialization **********/

  void cs_init (
    long    n_p,        /* number of nodes */
    long    m_p,        /* number of arcs */
    node    *nodes_p,   /* array of nodes */
    arc     *arcs_p,    /* array of arcs */
    long    f_sc,       /* scaling factor */
    price_t max_c,      /* maximal cost */
    long    *cap_p      /* array of capacities */
  );

  /********************************************** up_node_scan *************/

  void up_node_scan (
    node *i                        /* node for scanning */
  );

  /*************************************************** price_update *******/

  void  price_update ();

  /****************************************************** relabel *********/

  int relabel (
    register node *i          /* node for relabelling */
  );

  /***************************************************** discharge *********/

  void discharge (
    register node *i        /* node to be discharged */
  );

  /***************************************************** price_in *******/

  int price_in ();

  /************************************************** refine **************/

  void refine ();

  /*************************************************** price_refine **********/

  int price_refine ();

  /*************************************************** compute_prices ********/

  void compute_prices ();

  /***************************************************** price_out ************/

  void price_out ();

  /**************************************************** update_epsilon *******/
  /*----- decrease epsilon after epsilon-optimal flow is constructed */

  int update_epsilon();

  /* check complimentary slackness */
  int check_cs ();

  int check_eps_opt ();

  /*************************************************** finishup ***********/
  void finishup (
    double *obj_ad       /* objective */
  );

  /*********************************************** init_solution ***********/
  void init_solution ();

public:

  /******** constructor ***/

  CS2 ();

  /************************************************* cs2 - main program ***/

  void  cs2 (
    long    n_p,        /* number of nodes */
    long    m_p,        /* number of arcs */
    node    *nodes_p,   /* array of nodes */
    arc     *arcs_p,    /* array of arcs */
    long    f_sc,       /* scaling factor */
    price_t max_c,      /* maximal cost */
    long    *cap_p,     /* capacities */
    double  *obj_ad,    /* objective */
    bool    print = true
  );

private:

#ifdef COST_RESTART

  void cs_cost_reinit ();

  /******** restart after a cost update ***/

  void cs2_cost_restart (
    double *obj_ad    /* objective */
  );

#endif

public:

#ifdef PRINT_ANS

  void print_solution(node *ndp, arc *arp, long nmin);

#endif

};

#define UNFEASIBLE          2
#define ALLOCATION_FAULT    5
#define PRICE_OFL           6

/* parameters */

#define UPDT_FREQ      0.4
#define UPDT_FREQ_S    30

#define SCALE_DEFAULT  12.0

/* PRICE_OUT_START may not be less than 1 */
#define PRICE_OUT_START 1

#define CUT_OFF_POWER    0.44
#define CUT_OFF_COEF     1.5
#define CUT_OFF_POWER2   0.75
#define CUT_OFF_COEF2    1
#define CUT_OFF_GAP      0.8
#define CUT_OFF_MIN      12
#define CUT_OFF_INCREASE 4

/*
#define TIME_FOR_PRICE_IN    5
*/
#define TIME_FOR_PRICE_IN1    2
#define TIME_FOR_PRICE_IN2    4
#define TIME_FOR_PRICE_IN3    6

//#define EMPTY_PUSH_COEF      1.0
/*
#define MAX_CYCLES_CANCELLED 10
#define START_CYCLE_CANCEL   3
*/
#define MAX_CYCLES_CANCELLED 0
#define START_CYCLE_CANCEL   100


#endif
