/**************************************************************************
***
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
*** Copyright (c) 2000-2007 Regents of the University of Michigan,
***               Saurabh N. Adya, Jarrod A. Roy, David A. Papa and
***               Igor L. Markov
***
***  Contact author(s): abk@cs.ucsd.edu, imarkov@umich.edu
***  Original Affiliation:   UCLA, Computer Science Department,
***                          Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation
***  the rights to use, copy, modify, merge, publish, distribute, sublicense,
***  and/or sell copies of the Software, and to permit persons to whom the
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/


//created by Andrew Kennings on 06/01/98  akenning@ee.ryerson.ca

// CHANGES
// ilm 980713 --- massive cleanup for release 1.0
// aak 980722 --- added defaultSetting to set options[0] = 0

#ifndef __HMETIS_PARTITIONER_H__
#define __HMETIS_PARTITIONER_H__

#include "HGraph/hgFixed.h"
#include "Partitioning/partitioning.h"
#include "Partitioning/partProb.h"
#include "Partitioners/multiStartPart.h"
#include "PartEvals/partEvals.h"
#include "ABKCommon/abkmessagebuf.h"
#include <sstream>

extern "C"
{
  void HMETIS_PartRecursive( int nvtxs , int nhedges, int *vwgts,
                             int *eptr, int *eind, int *hewgts,
                             int nparts, int ubfactor, int *options,
                             int *hmPart, int *edgecut );
  void HMETIS_PartKway( int nvtxs , int nhedges, int *vwgts,
                        int *eptr, int *eind, int *hewgts,
                        int nparts, int ubfactor, int *options,
                        int *hmPart, int *edgecut );
}

class HMetisLicense
{
public:
  HMetisLicense();
};

class HMetisParams  // : public PartitionerParams
{
public:
  bool    kWay;		// 0 (recursive bisection) or 1 (kway).
  int     defaultSetting;     // options[0] = 0 or 1.
  int     vCycling;		// choices for VCycling.
  int     debugLevel;		// debugging level for hMetis.
  int     balanceFactor;      // balance factor for hMetis.
  double  weightScale;        // scaling factor.
  int     startsPerSolution;  // number of multistarts.
  Verbosity verb;

public:
  HMetisParams():kWay(0), defaultSetting(1), vCycling(1), debugLevel(0),
      balanceFactor(-1), weightScale(1.0), startsPerSolution(10),
      verb("1_1_1") {}
  HMetisParams(int argc, const char *argv[]);
  friend std::ostream& operator<<(std::ostream& os, const HMetisParams&);
};

//*****************************************************************************
//*****************************************************************************

class HMetisPartitioner : public MultiStartPartitioner
{
  // HMetisPartitioner built on top of MultiStartPartitioner

protected:

  static HMetisLicense _license;
  bool	   _kWay;
  double         _scaleFactor;
  unsigned       _numParts;
  uofm::vector<double> _partSizes;
  MaxMem*        _maxMem;

  bool _needDummyNodes; // flag for handing unequal targets.

  // HMetis variables.
  int   nvtxs;
  int   nhedges;
  int   *vwgts;
  int   *eptr;
  int   *eind;
  int   *hewgts;
  int   nparts;
  int   ubfactor;
  int   *options;
  int   *hmPart;
  int   edgecut;

  virtual void doOne(unsigned initSoln)
  {
    doHMetis(initSoln);
  }
  void    doHMetis(unsigned initSoln);
public:
  HMetisPartitioner(PartitioningProblem& problem,
                    MaxMem* maxMem,
                    HMetisParams parms = HMetisParams());
  ~HMetisPartitioner();

  uofm::vector<double>& getPartSizes()
  {
    return _partSizes;
  }
};

std::ostream& operator<<(std::ostream& os, const HMetisParams&);

#endif

