/**************************************************************************
***
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
*** Copyright (c) 2000-2007 Regents of the University of Michigan,
***               Saurabh N. Adya, Jarrod A. Roy, David A. Papa and
***               Igor L. Markov
***
***  Contact author(s): abk@cs.ucsd.edu, imarkov@umich.edu
***  Original Affiliation:   UCLA, Computer Science Department,
***                          Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation
***  the rights to use, copy, modify, merge, publish, distribute, sublicense,
***  and/or sell copies of the Software, and to permit persons to whom the
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/


// Created: Dec 1, 1999 by Igor Markov <imarkov@cs.ucla.edu>


#ifndef _METAPL_H_
#define _METAPL_H_

#include "Placement/placement.h"
#include "Placement/placeWOri.h"
#include "ABKCommon/abkcommon.h"
#include <iostream>
#include <string>
#include <sstream>

struct RowIroningParameters;
class RBPlacement;
class DB;

enum MemSaveMode { Off, Auto, On };

class MetaPlacerParams
{
public:
  Verbosity verb;
  MaxMem maxMem;
  bool save;
  bool savePl;
  bool saving;
  uofm::string saveBookshelfFilename;

  bool capo;
  bool keepOverlaps;
  bool flip;
  bool gSwap;
  bool gMove;
  bool congMoves;
#ifdef USEFLOW
  bool Xflows;
  bool Yflows;
#endif
  bool skipRowIroning;
  bool saveXorder;
  bool doPostProcessing;
  bool snapMacrosX, snapMacrosY;
  unsigned num;

  bool saveAsFloorplan; // save the solution in blocks/nets/pl format
  bool fastRI;
#ifdef USEFLUTE
  bool fluteRI;
#endif
  uofm::stringstream afterMetaMessage;
  bool routing;
  bool printSteiner;
  bool printAllSteiner;

  unsigned legalStrength;

  MemSaveMode memSaveMode;

  MetaPlacerParams():verb("1_1_1"), save(false), savePl(false), saving(false),
      saveBookshelfFilename(""), capo(true), keepOverlaps(false), flip(true),
      gSwap(true), gMove(true), congMoves(false),
#ifdef USEFLOW
      Xflows(true), Yflows(true),
#endif
      skipRowIroning(false), saveXorder(false), doPostProcessing(true),
      snapMacrosX(false), snapMacrosY(false), num(1), saveAsFloorplan(false), fastRI(false),
#ifdef USEFLUTE
      fluteRI(false),
#endif
      routing(false), printSteiner(false), printAllSteiner(false), legalStrength(5),
      memSaveMode(Auto) {};


  MetaPlacerParams(const MetaPlacerParams& srcParams):
      verb(srcParams.verb), save(srcParams.save), savePl(srcParams.savePl),
      saving(srcParams.saving), saveBookshelfFilename(srcParams.saveBookshelfFilename),
      capo(srcParams.capo), keepOverlaps(srcParams.keepOverlaps), gSwap(srcParams.gSwap), gMove(srcParams.gMove), congMoves(srcParams.congMoves),
#ifdef USEFLOW
      Xflows(srcParams.Xflows), Yflows(srcParams.Yflows),
#endif
      skipRowIroning(srcParams.skipRowIroning),
      saveXorder(srcParams.saveXorder), doPostProcessing(srcParams.doPostProcessing),
      snapMacrosX(srcParams.snapMacrosX), snapMacrosY(srcParams.snapMacrosY),
      num(srcParams.num),
      saveAsFloorplan(srcParams.saveAsFloorplan), fastRI(srcParams.fastRI),
#ifdef USEFLUTE
      fluteRI(srcParams.fluteRI),
#endif
      routing(srcParams.routing), printSteiner(false), printAllSteiner(false),
      legalStrength(srcParams.legalStrength), memSaveMode(srcParams.memSaveMode) {};

  MetaPlacerParams(int argc, const char *argv[]);
};

std::ostream& operator<< (std::ostream&, const MetaPlacerParams&);

class MetaPlacer
{
  uofm::vector<double> wls, runTs;

protected:

  MetaPlacerParams _params;
  DB *_db;
  uofm::string _defFileName;

  enum InputType {DB_, RBPlace_, HGraph_};
  InputType inputTyp;
  bool _rbPlaceIsUpdated;

  int  _greetingEtc(int argc, const char *argv[]);
  void _workWithRBPlace(RBPlacement& rbplace,
                        int argc, const char *argv[], unsigned iter, double &bestWL, PlacementWOrient *bestPlacement);
  void _workWithRBPlaceNtimes(RBPlacement& rbplace,
                              int argc, const char *argv[]);

public:

  MetaPlacer(int argc, const char *argv[]);
  MetaPlacer(RBPlacement& rbplace, int argc, const char *argv[]);
  ~MetaPlacer();
  bool isRBPlaceUpdated(void)
  {
    return _rbPlaceIsUpdated;
  }

#ifdef USEFLUTE
  void doMetaPlacerRowIroning(RBPlacement& rbplace, const RowIroningParameters& RIParams, bool fastRI, bool fluteRI);
#else
  void doMetaPlacerRowIroning(RBPlacement& rbplace, const RowIroningParameters& RIParams, bool fastRI);
#endif

};

#endif
