/**************************************************************************
***
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
*** Copyright (c) 2000-2007 Regents of the University of Michigan,
***               Saurabh N. Adya, Jarrod A. Roy, David A. Papa and
***               Igor L. Markov
***
***  Contact author(s): abk@cs.ucsd.edu, imarkov@umich.edu
***  Original Affiliation:   UCLA, Computer Science Department,
***                          Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation
***  the rights to use, copy, modify, merge, publish, distribute, sublicense,
***  and/or sell copies of the Software, and to permit persons to whom the
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/



#ifndef _LIBCELL_H_
#define _LIBCELL_H_

#include <set>
#include <bitset>

// currently, all cells with > 8 inputs does not use turth table and still
// uses blif
class lib_cell
{
public:
  int input_no;  // Number of inputs
  std::bitset<256> ttable; // Truth table, for <= 8 inputs only
  uofm::string blif; // For equivalence checking or too many inputs
};

class compare_lib_cell
{
public:
  int operator()(const lib_cell* s, const lib_cell* t) const
  {
    int no;

    if (s->input_no < t->input_no)
      return 1;
    else if (t->input_no < s->input_no)
      return 0;
    else if (s->input_no <= 8)
    {
      no= (1 << s->input_no) - 1;
      do
      {
        if (s->ttable.test(no) == false && t->ttable.test(no) == true)
          return 1;
        else if (s->ttable.test(no) == true && t->ttable.test(no) == false)
          return 0;
        no--;
      }
      while (no >= 0);
    }
    else
    {
      return s->blif.compare(t->blif) < 0;
    }
    return 0;
  }
};

typedef std::set<lib_cell*, compare_lib_cell> libcell_set;

#endif
