/**************************************************************************
***
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
*** Copyright (c) 2000-2007 Regents of the University of Michigan,
***               Saurabh N. Adya, Jarrod A. Roy, David A. Papa and
***               Igor L. Markov
***
***  Contact author(s): abk@cs.ucsd.edu, imarkov@umich.edu
***  Original Affiliation:   UCLA, Computer Science Department,
***                          Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation
***  the rights to use, copy, modify, merge, publish, distribute, sublicense,
***  and/or sell copies of the Software, and to permit persons to whom the
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/




#ifndef _SWIRE_INCLUDE_
#define _SWIRE_INCLUDE_

#include "ckt.h"
#include "TimeAna/sta.h"
#include <fstream>
#include <vector>
#include <map>

typedef uofm::vector<uofm::string> strvec;
typedef std::map<int, strvec*> strvecmap;

class swire
{
public:
  class Parameters
  {
  public:
    bool outsym_gen; // generate output symmetry
    unsigned int div_bin; // Divide bin
    unsigned int mult_bin; // Multiply bin
    unsigned int pass; // Number of passes
    bool sym_report; // Just report number of symmetries, no rewiring
    unsigned int input_limit; // Number of maximum inputs for rewiring
    bool phase_shift; // if true. use phase-shift symmetry
    int objective; // 0: wirelength, 1: timing
    sta::Parameters sta_parameters;
    Parameters();
  };
  ~swire();
  swire(ckt& inckt, Parameters params= Parameters());
  void rewire();
  void rewire2(unsigned int inst_no, bool bfs);
  void rewire_all();
  void rewire_capo();
  void rewire_bfsdfs(bool bfs);
  void rewire_time();
  unsigned int subckt_no; // Number of subcircuits
  unsigned int outsym_no; // pinswap output symmetry number
  unsigned int neg_outsym_no; // output symmetry involving negation
  unsigned int insym_no; // pinswap input symmetry number
  unsigned int neg_insym_no; // input symmetry involving negation
  unsigned int inoutsym_no; // symmetry involving input and output
private:
  int init();
  double rewire_bin_new(strvec* bin);
  double get_wirelength(uofm::vector<wire*>& wires);
  char* str_get_name(char* buf, uofm::string& token);
  int read_bin(char* bin_name);
  void output_results(uofm::vector<uofm::vector<port*> >& port, uofm::vector<wire*>& wires);
  void output_driver_rewire(uofm::vector<port*>& orig_port, uofm::vector<port*>& new_port);
  bool rewire_output_wires(uofm::vector<wire*>& owires, uofm::vector<wire*>& nowires, bool coerce);
  std::fstream ifile;
  std::fstream opinfile;
  std::fstream outsymfile;
  strvecmap binmap;
  bool debug_new;
  bool equi_check; // if true, check equivalence after every change
  int debug_counter;
  int inst_number; // Number of bins
  ckt& ckt1;
  Parameters params1;
  sta sta1;
};

#endif
