/**************************************************************************
***
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
*** Copyright (c) 2000-2007 Regents of the University of Michigan,
***               Saurabh N. Adya, Jarrod A. Roy, David A. Papa and
***               Igor L. Markov
***
***  Contact author(s): abk@cs.ucsd.edu, imarkov@umich.edu
***  Original Affiliation:   UCLA, Computer Science Department,
***                          Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation
***  the rights to use, copy, modify, merge, publish, distribute, sublicense,
***  and/or sell copies of the Software, and to permit persons to whom the
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/


#ifndef _SYMDETECT_INCLUDE_
#define _SYMDETECT_INCLUDE_


#include "ckt.h"
#include <set>

typedef std::set<int, std::less<int> > intset;
typedef uofm::vector<int> intvector;

class sym_pair
{
public:
  unsigned int v1, v2;
};

class sym_wire_pair
{
public:
  wire *v1, *v2;
};

class one_sym
{
public:
  uofm::vector<sym_pair> inputs, outputs;
};

class buf_sym
{
public:
  uofm::vector<wire*> input_neg;
  uofm::vector<sym_wire_pair> input_perm;
  uofm::vector<wire*> output_neg;
  uofm::vector<sym_wire_pair> output_perm;
};

class one_sym2
{
public:
  uofm::vector<int> input_neg;
  uofm::vector<sym_pair> input_perm;
  uofm::vector<int> output_neg;
  uofm::vector<sym_pair> output_perm;
};

class sym_group
{
public:
  intset ins, outs;
  uofm::vector<one_sym> syms;
  bool marked;
};

// records a permutation, which is also a symmetry generator.
class sym_perm
{
public:
  intvector in, out;
  void init(int in_no, int out_no);
  void init(const sym_perm& in_perm);
  void times(const sym_perm& in_perm);
  void times(one_sym& in_sym);
};

class lt_sym_perm
{
public:
  bool operator()(const sym_perm& sym1, const sym_perm& sym2) const;
};

typedef std::set<sym_perm, lt_sym_perm > sym_perm_set;

// All _new functions are for new algorithm
class symdetect
{
public:
  class Parameters
  {
  public:
    bool naive_sym; // Use naive symmetry detection (no Saucy)
    Parameters();
  };

  symdetect(ckt& inckt, Parameters params= Parameters());
  void init(uofm::vector<inst*>& insts, uofm::vector<wire*>& iwires, uofm::vector<wire*>& owires);
  void run();
  void process();
  void process_rebuf(uofm::vector<buf_sym>& bufsyms); // Process for rebuffering
  void pinswap_orig(uofm::vector<wire*>& wires); // return original place
  void perm(uofm::vector<wire*>& wires);  // return permuted place
  void outsym(uofm::vector<wire*>& wires);  // return outsym
  void redsym(uofm::vector<wire*>& wires); // return redundant
  void run_naive();
  bool last_perm; // if true: it is the last permutation

  int pinswap_no, outsym_no, redundant_no; // Number of symmetries in each type
  void get_sym_numbers(unsigned int& outsym_no, unsigned int& neg_outsym_no, unsigned int& insym_no, unsigned int& neg_insym_no, unsigned int& inoutsym_no);
  void outsym_generate(std::fstream& outfile);
  // new algorithm
  void process_new();
  bool init_group_new();
  bool get_sym_new(uofm::vector<wire*>& in_perm, uofm::vector<wire*>& out_perm, uofm::vector<wire*>& in_neg, uofm::vector<wire*>& out_neg);
  void gen_sym(sym_group& gsym2);
  bool debug_new;
  bool dump_saucy;
  void print_intvec(intvector& inv);

private:
  ckt& ckt1;
  int read_pair(sym_pair& sympair2);
  int read_pair_naive(sym_pair& sympair2);
  bool intset_intersect(intset& set1, intset& set2);
  uofm::vector<sym_pair> naive_syms;
  Parameters params1;

  unsigned int input_size, output_size, vertice_size, minterm_size;
  unsigned int output_base;
  unsigned int symstep;
  unsigned int outsymstep;
  unsigned int redstep;
  unsigned int count;
  unsigned int limit;
  unsigned int naive_count;

  uofm::vector<inst*> sinsts;
  uofm::vector<wire*> siwires, sowires, workwires;
  uofm::vector<one_sym> syms;
  uofm::vector<unsigned int> redundant;
  uofm::vector<one_sym> inputsyms, outputsyms;
  uofm::vector<uofm::vector<unsigned int> > multisyms; // traversed result of multi 2-pin

  // new algorithm
  uofm::vector<sym_group> gsyms; // group of symmetries
  unsigned int count_group_new;
  sym_perm_set symperms;
  sym_perm_set::iterator psymperms;
};

#endif
