/**************************************************************************
***
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
*** Copyright (c) 2000-2007 Regents of the University of Michigan,
***               Saurabh N. Adya, Jarrod A. Roy, David A. Papa and
***               Igor L. Markov
***
***  Contact author(s): abk@cs.ucsd.edu, imarkov@umich.edu
***  Original Affiliation:   UCLA, Computer Science Department,
***                          Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation
***  the rights to use, copy, modify, merge, publish, distribute, sublicense,
***  and/or sell copies of the Software, and to permit persons to whom the
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/


#include "ABKCommon/abkcommon.h"
#ifdef USEFLUTE
#include "Flute/flute.h"
#endif
#include "PhySyn/ckt.h"


class rc_tree
{
public:
  rc_tree();
  // Filled by rc_tree builder
  double dist; // distance to parent
  uofm::vector<rc_tree> child;
  port* leaf_port; // If not null, the port it connects to
  // filled by STA
  double r, c;
  double m1, m2;
  int id;
  void dump(int level);
};

class line_segment
{
public:
  line_segment()
  {
    visited= false;
  };
  Point points[2];
  unsigned int layer;
  int direction; // 0: horizontal, 1: vertical;
  double length();
  void dump();
  bool visited;
};

class via
{
public:
  Point point;
  unsigned int layer1, layer2;
};


class route_to_rc_tree
{
public:
  void run(wire* wire2, const dbRoutedNet& rnet, rc_tree& rc_tree2);
private:
  void add_segment(const dbRoutedNet& rnet);
  void add_seg_break(line_segment& seg2);
  bool intersect(line_segment& seg1, line_segment& seg2, uofm::vector<line_segment>& newsegs);
  void build_map();
  void build_via_map(const dbRoutedNet& rnet);
  void build_tree(Point & orig_point, line_segment& seg , rc_tree& my_tree);
  void get_connected_all(Point& orig, unsigned int layer, uofm::vector<line_segment>& segs, uofm::vector<Point>& new_points);
  void get_connected(Point& orig, unsigned int layer, uofm::vector<line_segment>& segs);
  void bbox_transform(inst* inst2, BBox& box2);
  bool check_port(Point& orig, double dist, rc_tree& my_tree);
  bool check_point_layer(Point& orig, unsigned int layer);
  uofm::vector<line_segment> segments;
  std::map<Point, uofm::vector<int> > seg_map;
  std::map<Point, uofm::vector<via> > via_map;
  unsigned int port_count;
  wire* twire;
};

class rc_tree_flute
{
public:
  bool run(wire* wire2, rc_tree& rc_tree2, int mode=0);
  bool run_super(wire* wire2, rc_tree& rc_tree2,
                 uofm::vector<BBox>& obstacles,
                 uofm::vector<uofm::vector<unsigned> >& inter_obstacles);
  bool run_simple(wire* wire2, rc_tree& rc_tree2,
                  uofm::vector<BBox>& obstacles,
                  uofm::vector<uofm::vector<unsigned> >& inter_obstacles);
private:
  uofm::vector<BBox> obstacles;
  uofm::vector<uofm::vector<unsigned> > inter_obstacles;
#ifdef USEFLUTE
  void add_node(rc_tree& rc_tree2, double px, double py,
                uofm::vector<bool>& node_added, Tree& ftree, int new_node,
                wire* wire2);
#endif
};

